@@ -21,14 +21,8 @@
 
 extern int ottery_valgrind_;
 /** Helper: invoke the RDRAND instruction to get 4 random bytes in the output
- * value. Return 0 on success, and an error on failure. */
-static int
-rdrand(uint32_t *therand) {
- unsigned char status;
- __asm volatile(".byte 0x0F, 0xC7, 0xF0 ; setc %1"
- : "=a" (*therand), "=qm" (status));
- return (status)==1 ? 0 : OTTERY_ERR_INIT_STRONG_RNG;
-}
+ * value. Return 1 on success, and 0 on failure. */
+#define rdrand32(x) ({ unsigned char err = 0; __asm volatile(".byte 0x0f; .byte 0xc7; .byte 0xf0; setc %1":"=a"(x), "=qm"(err) :"a"(0) :"cc"); err; })
 
 /** Generate bytes using the Intel RDRAND instruction. */
 static int
@@ -36,23 +30,23 @@ ottery_get_entropy_rdrand(const struct ottery_entropy_config *cfg,
                           struct ottery_entropy_state *state,
                            uint8_t *out, size_t outlen)
 {
-  int err;
-  uint32_t *up = (uint32_t *) out;
+  uint32_t up;
   (void) cfg;
   (void) state;
   if (! (ottery_get_cpu_capabilities_() & OTTERY_CPUCAP_RAND) || ottery_valgrind_)
     return OTTERY_ERR_INIT_STRONG_RNG;
   while (outlen >= 4) {
-    if ((err = rdrand(up)))
-      return err;
-    up += 1;
+    if (rdrand32(up) != 1)
+      return OTTERY_ERR_INIT_STRONG_RNG;
+    memcpy (out, &up, sizeof (up));
+    out += sizeof (up);
     outlen -= 4;
   }
+
   if (outlen) {
-    uint32_t tmp;
-    if ((err = rdrand(&tmp)))
-      return err;
-    memcpy(up, &tmp, outlen);
+    if (rdrand32(up) != 1)
+      return OTTERY_ERR_INIT_STRONG_RNG;
+    memcpy(out, &up, outlen);
   }
   return 0;
 }
