@@ -332,14 +332,15 @@ rdns_process_timer (void *arg)
 	req->retransmits --;
 	resolver = req->resolver;
 
+	if (req->resolver->ups && req->io->srv->ups_elt) {
+		req->resolver->ups->fail (req->io->srv->ups_elt,
+				req->resolver->ups->data, "timeout waiting reply");
+	}
+	else {
+		UPSTREAM_FAIL (req->io->srv, time (NULL));
+	}
+
 	if (req->retransmits == 0) {
-		if (req->resolver->ups && req->io->srv->ups_elt) {
-			req->resolver->ups->fail (req->io->srv->ups_elt,
-					req->resolver->ups->data);
-		}
-		else {
-			UPSTREAM_FAIL (req->io->srv, time (NULL));
-		}
 
 		rep = rdns_make_reply (req, RDNS_RC_TIMEOUT);
 		rdns_request_unschedule (req);
@@ -371,8 +372,11 @@ rdns_process_timer (void *arg)
 			if (resolver->ups) {
 				struct rdns_upstream_elt *elt;
 
-				elt = resolver->ups->select_retransmit (req->requested_names[0].name,
-						req->requested_names[0].len, resolver->ups->data);
+				elt = resolver->ups->select_retransmit (
+						req->requested_names[0].name,
+						req->requested_names[0].len,
+						req->io->srv->ups_elt,
+						resolver->ups->data);
 
 				if (elt) {
 					serv = elt->server;
@@ -423,7 +427,7 @@ rdns_process_timer (void *arg)
 	else if (r == -1) {
 		if (req->resolver->ups && req->io->srv->ups_elt) {
 			req->resolver->ups->fail (req->io->srv->ups_elt,
-					req->resolver->ups->data);
+					req->resolver->ups->data, "cannot send retransmit after timeout");
 		}
 		else {
 			UPSTREAM_FAIL (req->io->srv, time (NULL));
@@ -533,7 +537,7 @@ rdns_process_retransmit (int fd, void *arg)
 	else if (r == -1) {
 		if (req->resolver->ups && req->io->srv->ups_elt) {
 			req->resolver->ups->fail (req->io->srv->ups_elt,
-					req->resolver->ups->data);
+					req->resolver->ups->data, "retransmit send failed");
 		}
 		else {
 			UPSTREAM_FAIL (req->io->srv, time (NULL));
@@ -551,6 +555,43 @@ rdns_process_retransmit (int fd, void *arg)
 	}
 }
 
+struct rdns_server *
+rdns_select_request_upstream (struct rdns_resolver *resolver,
+							  struct rdns_request *req,
+							  bool is_retransmit,
+							  struct rdns_server *prev_serv)
+{
+	struct rdns_server *serv = NULL;
+
+	if (resolver->ups) {
+		struct rdns_upstream_elt *elt;
+
+		if (is_retransmit && prev_serv) {
+			elt = resolver->ups->select_retransmit (req->requested_names[0].name,
+					req->requested_names[0].len,
+					prev_serv->ups_elt,
+					resolver->ups->data);
+		}
+		else {
+			elt = resolver->ups->select (req->requested_names[0].name,
+					req->requested_names[0].len, resolver->ups->data);
+		}
+
+		if (elt) {
+			serv = elt->server;
+			serv->ups_elt = elt;
+		}
+		else {
+			UPSTREAM_SELECT_ROUND_ROBIN (resolver->servers, serv);
+		}
+	}
+	else {
+		UPSTREAM_SELECT_ROUND_ROBIN (resolver->servers, serv);
+	}
+
+	return serv;
+}
+
 #define align_ptr(p, a)                                                   \
     (guint8 *) (((uintptr_t) (p) + ((uintptr_t) a - 1)) & ~((uintptr_t) a - 1))
 
@@ -717,14 +758,14 @@ rdns_make_request_full (
 			type = req->requested_names[i].type;
 			if (queries > 1) {
 				if (!rdns_add_rr (req, cur_name, clen, type, &comp)) {
-					rdns_err ("cannot add rr", cur_name);
+					rdns_err ("cannot add rr");
 					REF_RELEASE (req);
 					rnds_compression_free (comp);
 					return NULL;
 				}
 			} else {
 				if (!rdns_add_rr (req, cur_name, clen, type, NULL)) {
-					rdns_err ("cannot add rr", cur_name);
+					rdns_err ("cannot add rr");
 					REF_RELEASE (req);
 					rnds_compression_free (comp);
 					return NULL;
@@ -737,30 +778,14 @@ rdns_make_request_full (
 		/* Add EDNS RR */
 		rdns_add_edns0 (req);
 
-		req->retransmits = repeats;
+		req->retransmits = repeats ? repeats : 1;
 		req->timeout = timeout;
 		req->state = RDNS_REQUEST_NEW;
 	}
 
 	req->async = resolver->async;
 
-	if (resolver->ups) {
-		struct rdns_upstream_elt *elt;
-
-		elt = resolver->ups->select (req->requested_names[0].name,
-				req->requested_names[0].len, resolver->ups->data);
-
-		if (elt) {
-			serv = elt->server;
-			serv->ups_elt = elt;
-		}
-		else {
-			UPSTREAM_SELECT_ROUND_ROBIN (resolver->servers, serv);
-		}
-	}
-	else {
-		UPSTREAM_SELECT_ROUND_ROBIN (resolver->servers, serv);
-	}
+	serv = rdns_select_request_upstream (resolver, req, false, NULL);
 
 	if (serv == NULL) {
 		rdns_warn ("cannot find suitable server for request");
@@ -776,24 +801,54 @@ rdns_make_request_full (
 				req->io->sock, req);
 	}
 	else {
-		req->io->uses++;
-
 		/* Now send request to server */
-		r = rdns_send_request (req, req->io->sock, true);
+		do {
+			r = rdns_send_request (req, req->io->sock, true);
 
-		if (r == -1) {
-			rdns_info ("cannot send DNS request: %s", strerror (errno));
-			REF_RELEASE (req);
+			if (r == -1) {
+				req->retransmits --; /* It must be > 0 */
 
-			if (resolver->ups && serv->ups_elt) {
-				resolver->ups->fail (serv->ups_elt, resolver->ups->data);
+				if (req->retransmits > 0) {
+					if (resolver->ups && serv->ups_elt) {
+						resolver->ups->fail (serv->ups_elt, resolver->ups->data,
+								"send IO error");
+					}
+					else {
+						UPSTREAM_FAIL (serv, time (NULL));
+					}
+
+					serv = rdns_select_request_upstream (resolver, req,
+							true, serv);
+
+					if (serv == NULL) {
+						rdns_warn ("cannot find suitable server for request");
+						REF_RELEASE (req);
+						return NULL;
+					}
+
+					req->io = serv->io_channels[ottery_rand_uint32 () % serv->io_cnt];
+				}
+				else {
+					rdns_info ("cannot send DNS request: %s", strerror (errno));
+					REF_RELEASE (req);
+
+					if (resolver->ups && serv->ups_elt) {
+						resolver->ups->fail (serv->ups_elt, resolver->ups->data,
+								"send IO error");
+					}
+					else {
+						UPSTREAM_FAIL (serv, time (NULL));
+					}
+
+					return NULL;
+				}
 			}
 			else {
-				UPSTREAM_FAIL (serv, time (NULL));
+				/* All good */
+				req->io->uses++;
+				break;
 			}
-
-			return NULL;
-		}
+		} while (req->retransmits > 0);
 	}
 
 	REF_RETAIN (req->io);
@@ -994,18 +1049,19 @@ rdns_resolver_free (struct rdns_resolver *resolver)
 
 
 struct rdns_resolver *
-rdns_resolver_new (void)
+rdns_resolver_new (int flags)
 {
-	struct rdns_resolver     *new;
+	struct rdns_resolver     *new_resolver;
 
-	new = calloc (1, sizeof (struct rdns_resolver));
+	new_resolver = calloc (1, sizeof (struct rdns_resolver));
 
-	REF_INIT_RETAIN (new, rdns_resolver_free);
+	REF_INIT_RETAIN (new_resolver, rdns_resolver_free);
 
-	new->logger = rdns_logger_internal;
-	new->log_data = new;
+	new_resolver->logger = rdns_logger_internal;
+	new_resolver->log_data = new_resolver;
+	new_resolver->flags = flags;
 
-	return new;
+	return new_resolver;
 }
 
 void
@@ -1070,4 +1126,4 @@ void rdns_resolver_set_fake_reply (struct rdns_resolver *resolver,
 
 		HASH_ADD (hh, resolver->fake_elts, key, sizeof (*srch) + len, fake_rep);
 	}
-}
\ No newline at end of file
+}
