@@ -38,6 +38,7 @@
 #include "ottery.h"
 #include "util.h"
 #include "logger.h"
+#include "rdns.h"
 
 static int
 rdns_make_socket_nonblocking (int fd)
@@ -495,9 +496,11 @@ rdns_ioc_free (struct rdns_io_channel *ioc)
 	HASH_ITER (hh, ioc->requests, req, rtmp) {
 		REF_RELEASE (req);
 	}
+
 	ioc->resolver->async->del_read (ioc->resolver->async->data,
 			ioc->async_io);
 	close (ioc->sock);
+	free (ioc->saddr);
 	free (ioc);
 }
 
