@@ -15,11 +15,11 @@
  */
 #include "rspamdclient.h"
 #include "libutil/util.h"
-#include "libutil/http_connection.h"
-#include "libutil/http_private.h"
+#include "libserver/http/http_connection.h"
+#include "libserver/http/http_private.h"
+#include "libserver/protocol_internal.h"
 #include "unix-std.h"
 #include "contrib/zstd/zstd.h"
-#include "contrib/zstd/zdict.h"
 
 #ifdef HAVE_FETCH_H
 #include <fetch.h>
@@ -37,8 +37,8 @@ struct rspamd_client_connection {
 	GString *server_name;
 	struct rspamd_cryptobox_pubkey *key;
 	struct rspamd_cryptobox_keypair *keypair;
-	struct event_base *ev_base;
-	struct timeval timeout;
+	struct ev_loop *event_loop;
+	ev_tstamp timeout;
 	struct rspamd_http_connection *http_conn;
 	gboolean req_sent;
 	gdouble start_time;
@@ -96,7 +96,7 @@ rspamd_client_error_handler (struct rspamd_http_connection *conn, GError *err)
 	c = req->conn;
 	req->cb (c, NULL, c->server_name->str, NULL,
 			req->input, req->ud,
-			c->start_time, c->send_time, err);
+			c->start_time, c->send_time, NULL, 0, err);
 }
 
 static gint
@@ -109,6 +109,9 @@ rspamd_client_finish_handler (struct rspamd_http_connection *conn,
 	struct ucl_parser *parser;
 	GError *err;
 	const rspamd_ftok_t *tok;
+	const gchar *start, *body = NULL;
+	guchar *out = NULL;
+	gsize len, bodylen = 0;
 
 	c = req->conn;
 
@@ -118,7 +121,8 @@ rspamd_client_finish_handler (struct rspamd_http_connection *conn,
 		rspamd_http_connection_reset (c->http_conn);
 		rspamd_http_connection_read_message (c->http_conn,
 			c->req,
-			&c->timeout);
+			c->timeout);
+
 		return 0;
 	}
 	else {
@@ -127,13 +131,13 @@ rspamd_client_finish_handler (struct rspamd_http_connection *conn,
 					msg->code,
 					(gint)msg->status->len, msg->status->str);
 			req->cb (c, msg, c->server_name->str, NULL, req->input, req->ud,
-					c->start_time, c->send_time, err);
+					c->start_time, c->send_time, body, bodylen, err);
 			g_error_free (err);
 
 			return 0;
 		}
 
-		tok = rspamd_http_message_find_header (msg, "compression");
+		tok = rspamd_http_message_find_header (msg, COMPRESSION_HEADER);
 
 		if (tok) {
 			/* Need to uncompress */
@@ -146,7 +150,6 @@ rspamd_client_finish_handler (struct rspamd_http_connection *conn,
 				ZSTD_DStream *zstream;
 				ZSTD_inBuffer zin;
 				ZSTD_outBuffer zout;
-				guchar *out;
 				gsize outlen, r;
 
 				zstream = ZSTD_createDStream ();
@@ -174,85 +177,99 @@ rspamd_client_finish_handler (struct rspamd_http_connection *conn,
 								ZSTD_getErrorName (r));
 						req->cb (c, msg, c->server_name->str, NULL,
 								req->input, req->ud, c->start_time,
-								c->send_time, err);
+								c->send_time, body, bodylen, err);
 						g_error_free (err);
 						ZSTD_freeDStream (zstream);
-						g_free (out);
 
-						return 0;
+						goto end;
 					}
 
 					if (zout.pos == zout.size) {
 						/* We need to extend output buffer */
-						zout.size = zout.size * 1.5 + 1.0;
+						zout.size = zout.size * 2;
 						zout.dst = g_realloc (zout.dst, zout.size);
 					}
 				}
 
 				ZSTD_freeDStream (zstream);
 
-				parser = ucl_parser_new (0);
-				if (!ucl_parser_add_chunk (parser, zout.dst, zout.pos)) {
-					err = g_error_new (RCLIENT_ERROR, msg->code, "Cannot parse UCL: %s",
-							ucl_parser_get_error (parser));
-					ucl_parser_free (parser);
-					req->cb (c, msg, c->server_name->str, NULL, req->input,
-							req->ud, c->start_time, c->send_time, err);
-					g_error_free (err);
-					g_free (zout.dst);
-
-					return 0;
-				}
-
-				g_free (zout.dst);
+				start = zout.dst;
+				len = zout.pos;
 			}
 			else {
 				err = g_error_new (RCLIENT_ERROR, 500,
 						"Invalid compression method");
 				req->cb (c, msg, c->server_name->str, NULL,
-						req->input, req->ud, c->start_time, c->send_time, err);
+						req->input, req->ud, c->start_time, c->send_time,
+						body, bodylen, err);
 				g_error_free (err);
 
 				return 0;
 			}
 		}
 		else {
-			parser = ucl_parser_new (0);
-			if (!ucl_parser_add_chunk (parser, msg->body_buf.begin, msg->body_buf.len)) {
-				err = g_error_new (RCLIENT_ERROR, msg->code, "Cannot parse UCL: %s",
-						ucl_parser_get_error (parser));
-				ucl_parser_free (parser);
-				req->cb (c, msg, c->server_name->str, NULL,
-						req->input, req->ud, c->start_time, c->send_time, err);
-				g_error_free (err);
+			start = msg->body_buf.begin;
+			len = msg->body_buf.len;
+		}
 
-				return 0;
+		/* Deal with body */
+		tok = rspamd_http_message_find_header (msg, MESSAGE_OFFSET_HEADER);
+
+		if (tok) {
+			gulong value = 0;
+
+			if (rspamd_strtoul (tok->begin, tok->len, &value) &&
+					value < len) {
+				body = start + value;
+				bodylen = len - value;
+				len = value;
 			}
 		}
 
-		req->cb (c, msg, c->server_name->str, ucl_parser_get_object (
-				parser), req->input, req->ud, c->start_time, c->send_time, NULL);
+		parser = ucl_parser_new (0);
+		if (!ucl_parser_add_chunk (parser, start, len)) {
+			err = g_error_new (RCLIENT_ERROR, msg->code, "Cannot parse UCL: %s",
+					ucl_parser_get_error (parser));
+			ucl_parser_free (parser);
+			req->cb (c, msg, c->server_name->str, NULL,
+					req->input, req->ud,
+					c->start_time, c->send_time, body, bodylen, err);
+			g_error_free (err);
+
+			goto end;
+		}
+
+		req->cb (c, msg, c->server_name->str,
+				ucl_parser_get_object (parser),
+				req->input, req->ud,
+				c->start_time, c->send_time, body, bodylen, NULL);
 		ucl_parser_free (parser);
 	}
 
+end:
+	if (out) {
+		g_free (out);
+	}
+
 	return 0;
 }
 
 struct rspamd_client_connection *
 rspamd_client_init (struct rspamd_http_context *http_ctx,
-					struct event_base *ev_base, const gchar *name,
+					struct ev_loop *ev_base, const gchar *name,
 					guint16 port, gdouble timeout, const gchar *key)
 {
 	struct rspamd_client_connection *conn;
 	gint fd;
 
 	fd = rspamd_socket (name, port, SOCK_STREAM, TRUE, FALSE, TRUE);
+
 	if (fd == -1) {
 		return NULL;
 	}
 
 	conn = g_malloc0 (sizeof (struct rspamd_client_connection));
-	conn->ev_base = ev_base;
+	conn->event_loop = ev_base;
 	conn->fd = fd;
 	conn->req_sent = FALSE;
 	conn->http_conn = rspamd_http_connection_new_client_socket (http_ctx,
@@ -262,12 +279,20 @@ rspamd_client_init (struct rspamd_http_context *http_ctx,
 			0,
 			fd);
 
+	if (!conn->http_conn) {
+		rspamd_client_destroy (conn);
+		return NULL;
+	}
+
+	/* Pass socket ownership */
+	rspamd_http_connection_own_socket (conn->http_conn);
 	conn->server_name = g_string_new (name);
+
 	if (port != 0) {
 		rspamd_printf_gstring (conn->server_name, ":%d", (int)port);
 	}
 
-	double_to_tv (timeout, &conn->timeout);
+	conn->timeout = timeout;
 
 	if (key) {
 		conn->key = rspamd_pubkey_from_base32 (key, 0, RSPAMD_KEYPAIR_KEX,
@@ -307,6 +332,7 @@ rspamd_client_command (struct rspamd_client_connection *conn,
 	gsize dict_len = 0;
 	void *dict = NULL;
 	ZSTD_CCtx *zctx;
+	gboolean ret;
 
 	req = g_malloc0 (sizeof (struct rspamd_client_request));
 	req->conn = conn;
@@ -364,7 +390,7 @@ rspamd_client_command (struct rspamd_client_connection *conn,
 					return FALSE;
 				}
 
-				dict_id = ZDICT_getDictID (comp_dictionary, dict_len);
+				dict_id = -1;
 
 				if (dict_id == 0) {
 					g_set_error (err, RCLIENT_ERROR, errno,
@@ -419,7 +445,7 @@ rspamd_client_command (struct rspamd_client_connection *conn,
 	}
 
 	if (compressed) {
-		rspamd_http_message_add_header (req->msg, "Compression", "zstd");
+		rspamd_http_message_add_header (req->msg, COMPRESSION_HEADER, "zstd");
 
 		if (dict_id != 0) {
 			gchar dict_str[32];
@@ -440,33 +466,38 @@ rspamd_client_command (struct rspamd_client_connection *conn,
 	conn->start_time = rspamd_get_ticks (FALSE);
 
 	if (compressed) {
-		rspamd_http_connection_write_message (conn->http_conn, req->msg, NULL,
-				"application/x-compressed", req,
-				&conn->timeout);
+		ret = rspamd_http_connection_write_message (conn->http_conn, req->msg,
+				NULL,"application/x-compressed", req,
+				conn->timeout);
 	}
 	else {
-		rspamd_http_connection_write_message (conn->http_conn, req->msg, NULL,
-				"text/plain", req, &conn->timeout);
+		ret = rspamd_http_connection_write_message (conn->http_conn, req->msg,
+				NULL,"text/plain", req, conn->timeout);
 	}
 
-	return TRUE;
+	return ret;
 }
 
 void
 rspamd_client_destroy (struct rspamd_client_connection *conn)
 {
 	if (conn != NULL) {
-		rspamd_http_connection_unref (conn->http_conn);
+		if (conn->http_conn) {
+			rspamd_http_connection_unref (conn->http_conn);
+		}
+
 		if (conn->req != NULL) {
 			rspamd_client_request_free (conn->req);
 		}
-		close (conn->fd);
+
 		if (conn->key) {
 			rspamd_pubkey_unref (conn->key);
 		}
+
 		if (conn->keypair) {
 			rspamd_keypair_unref (conn->keypair);
 		}
+
 		g_string_free (conn->server_name, TRUE);
 		g_free (conn);
 	}
