@@ -20,107 +20,130 @@
 #include "contrib/uthash/utlist.h"
 #include "libserver/mempool_vars_internal.h"
 #include "libserver/url.h"
+#include "libserver/cfg_file.h"
+#include "libutil/util.h"
 #include <unicode/utf8.h>
 
+KHASH_INIT (rspamd_mime_headers_htb, gchar *,
+		struct rspamd_mime_header *, 1,
+		rspamd_strcase_hash, rspamd_strcase_equal);
+
+struct rspamd_mime_headers_table {
+	khash_t(rspamd_mime_headers_htb) htb;
+	ref_entry_t ref;
+};
+
+#define RSPAMD_INET_ADDRESS_PARSE_RECEIVED \
+	(RSPAMD_INET_ADDRESS_PARSE_REMOTE|RSPAMD_INET_ADDRESS_PARSE_NO_UNIX)
+
 static void
 rspamd_mime_header_check_special (struct rspamd_task *task,
 		struct rspamd_mime_header *rh)
 {
 	guint64 h;
-	struct received_header *recv;
+	struct rspamd_received_header *recv;
 	const gchar *p, *end;
 	gchar *id;
+	gint max_recipients = -1;
+
+	if (task->cfg) {
+		max_recipients = task->cfg->max_recipients;
+	}
 
 	h = rspamd_icase_hash (rh->name, strlen (rh->name), 0xdeadbabe);
 
 	switch (h) {
 	case 0x88705DC4D9D61ABULL:	/* received */
 		recv = rspamd_mempool_alloc0 (task->task_pool,
-				sizeof (struct received_header));
+				sizeof (struct rspamd_received_header));
 		recv->hdr = rh;
 
 		if (rspamd_smtp_received_parse (task, rh->decoded,
 				strlen (rh->decoded), recv) != -1) {
-			g_ptr_array_add (task->received, recv);
+			DL_APPEND (MESSAGE_FIELD (task, received), recv);
 		}
 
-		rh->type = RSPAMD_HEADER_RECEIVED;
+		rh->flags |= RSPAMD_HEADER_RECEIVED;
 		break;
 	case 0x76F31A09F4352521ULL:	/* to */
-		task->rcpt_mime = rspamd_email_address_from_mime (task->task_pool,
-				rh->decoded, strlen (rh->decoded), task->rcpt_mime);
-		rh->type = RSPAMD_HEADER_TO|RSPAMD_HEADER_RCPT|RSPAMD_HEADER_UNIQUE;
+		MESSAGE_FIELD (task, rcpt_mime) = rspamd_email_address_from_mime (task->task_pool,
+				rh->value, strlen (rh->value),
+				MESSAGE_FIELD (task, rcpt_mime), max_recipients);
+		rh->flags |= RSPAMD_HEADER_TO|RSPAMD_HEADER_RCPT|RSPAMD_HEADER_UNIQUE;
 		break;
 	case 0x7EB117C1480B76ULL:	/* cc */
-		task->rcpt_mime = rspamd_email_address_from_mime (task->task_pool,
-				rh->decoded, strlen (rh->decoded), task->rcpt_mime);
-		rh->type = RSPAMD_HEADER_CC|RSPAMD_HEADER_RCPT|RSPAMD_HEADER_UNIQUE;
+		MESSAGE_FIELD (task, rcpt_mime) = rspamd_email_address_from_mime (task->task_pool,
+				rh->value, strlen (rh->value),
+				MESSAGE_FIELD (task, rcpt_mime), max_recipients);
+		rh->flags |= RSPAMD_HEADER_CC|RSPAMD_HEADER_RCPT|RSPAMD_HEADER_UNIQUE;
 		break;
 	case 0xE4923E11C4989C8DULL:	/* bcc */
-		task->rcpt_mime = rspamd_email_address_from_mime (task->task_pool,
-				rh->decoded, strlen (rh->decoded), task->rcpt_mime);
-		rh->type = RSPAMD_HEADER_BCC|RSPAMD_HEADER_RCPT|RSPAMD_HEADER_UNIQUE;
+		MESSAGE_FIELD (task, rcpt_mime) = rspamd_email_address_from_mime (task->task_pool,
+				rh->value, strlen (rh->value),
+				MESSAGE_FIELD (task, rcpt_mime), max_recipients);
+		rh->flags |= RSPAMD_HEADER_BCC|RSPAMD_HEADER_RCPT|RSPAMD_HEADER_UNIQUE;
 		break;
 	case 0x41E1985EDC1CBDE4ULL:	/* from */
-		task->from_mime = rspamd_email_address_from_mime (task->task_pool,
-				rh->decoded, strlen (rh->decoded), task->from_mime);
-		rh->type = RSPAMD_HEADER_FROM|RSPAMD_HEADER_SENDER|RSPAMD_HEADER_UNIQUE;
+		MESSAGE_FIELD (task, from_mime) = rspamd_email_address_from_mime (task->task_pool,
+				rh->value, strlen (rh->value),
+				MESSAGE_FIELD (task, from_mime), max_recipients);
+		rh->flags |= RSPAMD_HEADER_FROM|RSPAMD_HEADER_SENDER|RSPAMD_HEADER_UNIQUE;
 		break;
 	case 0x43A558FC7C240226ULL:	/* message-id */ {
 
-		rh->type = RSPAMD_HEADER_MESSAGE_ID|RSPAMD_HEADER_UNIQUE;
+		rh->flags = RSPAMD_HEADER_MESSAGE_ID|RSPAMD_HEADER_UNIQUE;
 		p = rh->decoded;
 		end = p + strlen (p);
 
 		if (*p == '<') {
-			p ++;
+			p++;
+		}
 
-			if (end > p) {
-				gchar *d;
+		if (end > p) {
+			gchar *d;
 
-				if (*(end - 1) == '>') {
-					end --;
-				}
-
-				id = rspamd_mempool_alloc (task->task_pool, end - p + 1);
-				d = id;
-
-				while (p < end) {
-					if (g_ascii_isgraph (*p)) {
-						*d++ = *p++;
-					}
-					else {
-						*d++ = '?';
-						p++;
-					}
-				}
-
-				*d = '\0';
-
-				task->message_id = id;
+			if (*(end - 1) == '>') {
+				end --;
 			}
+
+			id = rspamd_mempool_alloc (task->task_pool, end - p + 1);
+			d = id;
+
+			while (p < end) {
+				if (g_ascii_isgraph (*p)) {
+					*d++ = *p++;
+				}
+				else {
+					*d++ = '?';
+					p++;
+				}
+			}
+
+			*d = '\0';
+
+			MESSAGE_FIELD (task, message_id) = id;
 		}
 
 		break;
 	}
 	case 0xB91D3910358E8212ULL:	/* subject */
-		if (task->subject == NULL) {
-			task->subject = rh->decoded;
+		if (MESSAGE_FIELD (task, subject) == NULL) {
+			MESSAGE_FIELD (task, subject) = rh->decoded;
 		}
-		rh->type = RSPAMD_HEADER_SUBJECT|RSPAMD_HEADER_UNIQUE;
+		rh->flags = RSPAMD_HEADER_SUBJECT|RSPAMD_HEADER_UNIQUE;
 		break;
 	case 0xEE4AA2EAAC61D6F4ULL:	/* return-path */
 		if (task->from_envelope == NULL) {
 			task->from_envelope = rspamd_email_address_from_smtp (rh->decoded,
 					strlen (rh->decoded));
 		}
-		rh->type = RSPAMD_HEADER_RETURN_PATH|RSPAMD_HEADER_UNIQUE;
+		rh->flags = RSPAMD_HEADER_RETURN_PATH|RSPAMD_HEADER_UNIQUE;
 		break;
 	case 0xB9EEFAD2E93C2161ULL:	/* delivered-to */
 		if (task->deliver_to == NULL) {
 			task->deliver_to = rh->decoded;
 		}
-		rh->type = RSPAMD_HEADER_DELIVERED_TO;
+		rh->flags = RSPAMD_HEADER_DELIVERED_TO;
 		break;
 	case 0x2EC3BFF3C393FC10ULL: /* date */
 	case 0xAC0DDB1A1D214CAULL: /* sender */
@@ -128,31 +151,37 @@ rspamd_mime_header_check_special (struct rspamd_task *task,
 	case 0x81CD9E9131AB6A9AULL: /* content-type */
 	case 0xC39BD9A75AA25B60ULL: /* content-transfer-encoding */
 	case 0xB3F6704CB3AD6589ULL: /* references */
-		rh->type = RSPAMD_HEADER_UNIQUE;
+		rh->flags = RSPAMD_HEADER_UNIQUE;
 		break;
 	}
 }
 
 static void
 rspamd_mime_header_add (struct rspamd_task *task,
-		GHashTable *target, GQueue *order,
-		struct rspamd_mime_header *rh,
-		gboolean check_special)
+						khash_t(rspamd_mime_headers_htb) *target,
+						struct rspamd_mime_header **order_ptr,
+						struct rspamd_mime_header *rh,
+						gboolean check_special)
 {
-	GPtrArray *ar;
+	khiter_t k;
+	struct rspamd_mime_header *ex;
+	int res;
 
-	if ((ar = g_hash_table_lookup (target, rh->name)) != NULL) {
-		g_ptr_array_add (ar, rh);
+	k = kh_put (rspamd_mime_headers_htb, target, rh->name, &res);
+
+	if (res == 0) {
+		ex = kh_value (target, k);
+		DL_APPEND (ex, rh);
 		msg_debug_task ("append raw header %s: %s", rh->name, rh->value);
 	}
 	else {
-		ar = g_ptr_array_sized_new (2);
-		g_ptr_array_add (ar, rh);
-		g_hash_table_insert (target, rh->name, ar);
+		kh_value (target, k) = rh;
+		rh->prev = rh;
+		rh->next = NULL;
 		msg_debug_task ("add new raw header %s: %s", rh->name, rh->value);
 	}
 
-	g_queue_push_tail (order, rh);
+	LL_PREPEND2 (*order_ptr, rh, ord_next);
 
 	if (check_special) {
 		rspamd_mime_header_check_special (task, rh);
@@ -162,8 +191,9 @@ rspamd_mime_header_add (struct rspamd_task *task,
 
 /* Convert raw headers to a list of struct raw_header * */
 void
-rspamd_mime_headers_process (struct rspamd_task *task, GHashTable *target,
-		GQueue *order,
+rspamd_mime_headers_process (struct rspamd_task *task,
+		struct rspamd_mime_headers_table *target,
+		struct rspamd_mime_header **order_ptr,
 		const gchar *in, gsize len,
 		gboolean check_newlines)
 {
@@ -205,7 +235,7 @@ rspamd_mime_headers_process (struct rspamd_task *task, GHashTable *target,
 				tmp = rspamd_mempool_alloc (task->task_pool, l + 1);
 				rspamd_null_safe_copy (c, l, tmp, l + 1);
 				nh->name = tmp;
-				nh->empty_separator = TRUE;
+				nh->flags |= RSPAMD_HEADER_EMPTY_SEPARATOR;
 				nh->raw_value = c;
 				nh->raw_len = p - c; /* Including trailing ':' */
 				p++;
@@ -225,12 +255,12 @@ rspamd_mime_headers_process (struct rspamd_task *task, GHashTable *target,
 		case 2:
 			/* We got header's name, so skip any \t or spaces */
 			if (*p == '\t') {
-				nh->tab_separated = TRUE;
-				nh->empty_separator = FALSE;
+				nh->flags &= ~RSPAMD_HEADER_EMPTY_SEPARATOR;
+				nh->flags |= RSPAMD_HEADER_TAB_SEPARATED;
 				p++;
 			}
 			else if (*p == ' ') {
-				nh->empty_separator = FALSE;
+				nh->flags &= ~RSPAMD_HEADER_EMPTY_SEPARATOR;
 				p++;
 			}
 			else if (*p == '\n' || *p == '\r') {
@@ -239,7 +269,7 @@ rspamd_mime_headers_process (struct rspamd_task *task, GHashTable *target,
 					if (*p == '\n') {
 						nlines_count[RSPAMD_TASK_NEWLINES_LF] ++;
 					}
-					else if (*(p + 1) == '\n') {
+					else if (p + 1 < end && *(p + 1) == '\n') {
 						nlines_count[RSPAMD_TASK_NEWLINES_CRLF] ++;
 					}
 					else {
@@ -278,7 +308,7 @@ rspamd_mime_headers_process (struct rspamd_task *task, GHashTable *target,
 					if (*p == '\n') {
 						nlines_count[RSPAMD_TASK_NEWLINES_LF] ++;
 					}
-					else if (*(p + 1) == '\n') {
+					else if (p + 1 < end && *(p + 1) == '\n') {
 						nlines_count[RSPAMD_TASK_NEWLINES_CRLF] ++;
 					}
 					else {
@@ -342,7 +372,7 @@ rspamd_mime_headers_process (struct rspamd_task *task, GHashTable *target,
 				}
 			}
 			/* Strip last space that can be added by \r\n parsing */
-			if (*(tp - 1) == ' ') {
+			if (tp > tmp && *(tp - 1) == ' ') {
 				tp--;
 			}
 
@@ -377,7 +407,7 @@ rspamd_mime_headers_process (struct rspamd_task *task, GHashTable *target,
 			/* We also validate utf8 and replace all non-valid utf8 chars */
 			rspamd_mime_charset_utf_enforce (nh->decoded, strlen (nh->decoded));
 			nh->order = norder ++;
-			rspamd_mime_header_add (task, target, order, nh, check_newlines);
+			rspamd_mime_header_add (task, &target->htb, order_ptr, nh, check_newlines);
 			nh = NULL;
 			state = 0;
 			break;
@@ -387,7 +417,7 @@ rspamd_mime_headers_process (struct rspamd_task *task, GHashTable *target,
 			nh->decoded = "";
 			nh->raw_len = p - nh->raw_value;
 			nh->order = norder ++;
-			rspamd_mime_header_add (task, target, order, nh, check_newlines);
+			rspamd_mime_header_add (task, &target->htb, order_ptr, nh, check_newlines);
 			nh = NULL;
 			state = 0;
 			break;
@@ -423,7 +453,7 @@ rspamd_mime_headers_process (struct rspamd_task *task, GHashTable *target,
 			/* Fail state, skip line */
 
 			if (*p == '\r') {
-				if (*(p + 1) == '\n') {
+				if (p + 1 < end && *(p + 1) == '\n') {
 					nlines_count[RSPAMD_TASK_NEWLINES_CRLF] ++;
 					p++;
 				}
@@ -433,7 +463,7 @@ rspamd_mime_headers_process (struct rspamd_task *task, GHashTable *target,
 			else if (*p == '\n') {
 				nlines_count[RSPAMD_TASK_NEWLINES_LF] ++;
 
-				if (*(p + 1) == '\r') {
+				if (p + 1 < end && *(p + 1) == '\r') {
 					p++;
 				}
 				p++;
@@ -450,33 +480,30 @@ rspamd_mime_headers_process (struct rspamd_task *task, GHashTable *target,
 		}
 	}
 
+	/* Since we have prepended headers, we need to reverse the list to get the actual order */
+	LL_REVERSE (*order_ptr);
+
 	if (check_newlines) {
 		guint max_cnt = 0;
 		gint sel = 0;
-		GList *cur;
 		rspamd_cryptobox_hash_state_t hs;
 		guchar hout[rspamd_cryptobox_HASHBYTES], *hexout;
 
-		for (gint i = 0; i < RSPAMD_TASK_NEWLINES_MAX; i ++) {
+		for (gint i = RSPAMD_TASK_NEWLINES_CR; i < RSPAMD_TASK_NEWLINES_MAX; i ++) {
 			if (nlines_count[i] > max_cnt) {
 				max_cnt = nlines_count[i];
 				sel = i;
 			}
 		}
 
-		task->nlines_type = sel;
+		MESSAGE_FIELD (task, nlines_type) = sel;
 
-		cur = order->head;
 		rspamd_cryptobox_hash_init (&hs, NULL, 0);
 
-		while (cur) {
-			nh = cur->data;
-
-			if (nh->name && nh->type != RSPAMD_HEADER_RECEIVED) {
+		LL_FOREACH (*order_ptr, nh) {
+			if (nh->name && nh->flags != RSPAMD_HEADER_RECEIVED) {
 				rspamd_cryptobox_hash_update (&hs, nh->name, strlen (nh->name));
 			}
-
-			cur = g_list_next (cur);
 		}
 
 		rspamd_cryptobox_hash_final (&hs, hout);
@@ -491,9 +518,12 @@ rspamd_mime_headers_process (struct rspamd_task *task, GHashTable *target,
 }
 
 static void
-rspamd_mime_header_maybe_save_token (rspamd_mempool_t *pool, GString *out,
-		GByteArray *token, GByteArray *decoded_token,
-		rspamd_ftok_t *old_charset, rspamd_ftok_t *new_charset)
+rspamd_mime_header_maybe_save_token (rspamd_mempool_t *pool,
+									 GString *out,
+									 GByteArray *token,
+									 GByteArray *decoded_token,
+									 rspamd_ftok_t *old_charset,
+									 rspamd_ftok_t *new_charset)
 {
 	if (new_charset->len == 0) {
 		g_assert_not_reached ();
@@ -517,14 +547,22 @@ rspamd_mime_header_maybe_save_token (rspamd_mempool_t *pool, GString *out,
 	}
 
 	/* We need to flush and decode old token to out string */
-	if (rspamd_mime_to_utf8_byte_array (token, decoded_token,
+	if (rspamd_mime_to_utf8_byte_array (token, decoded_token, pool,
 			rspamd_mime_detect_charset (new_charset, pool))) {
 		g_string_append_len (out, decoded_token->data, decoded_token->len);
 	}
 
 	/* We also reset buffer */
 	g_byte_array_set_size (token, 0);
-	/* Propagate charset */
+	/*
+	 * Propagate charset
+	 *
+	 * Here are dragons: we save the original charset to allow buffers concat
+	 * in the condition at the beginning of the function.
+	 * However, it will likely cause unnecessary calls for
+	 * `rspamd_mime_detect_charset` which could be relatively expensive.
+	 * But we ignore that for now...
+	 */
 	memcpy (old_charset, new_charset, sizeof (*old_charset));
 }
 
@@ -621,9 +659,10 @@ rspamd_mime_header_decode (rspamd_mempool_t *pool, const gchar *in,
 				qmarks = 0;
 			}
 			else {
-				g_string_append_len (out, c, 2);
-				c = p + 1;
+				g_string_append_len (out, c, 1);
+				c = p;
 				state = parse_normal;
+				continue; /* Deal with == case */
 			}
 			p ++;
 			break;
@@ -755,6 +794,7 @@ rspamd_mime_header_decode (rspamd_mempool_t *pool, const gchar *in,
 	g_byte_array_free (token, TRUE);
 	g_byte_array_free (decoded, TRUE);
 	rspamd_mime_header_sanity_check (out);
+	rspamd_mempool_notify_alloc (pool, out->len);
 	ret = g_string_free (out, FALSE);
 	rspamd_mempool_add_destructor (pool, g_free, ret);
 
@@ -867,6 +907,7 @@ enum rspamd_received_part_type {
 	RSPAMD_RECEIVED_PART_BY,
 	RSPAMD_RECEIVED_PART_FOR,
 	RSPAMD_RECEIVED_PART_WITH,
+	RSPAMD_RECEIVED_PART_ID,
 	RSPAMD_RECEIVED_PART_UNKNOWN,
 };
 
@@ -1077,8 +1118,14 @@ rspamd_smtp_received_process_part (struct rspamd_task *task,
 			p ++;
 			break;
 		case all_done:
-			*last = p - (const guchar *)data;
-			return npart;
+			if (p > (const guchar *)data) {
+				*last = p - (const guchar *) data;
+				return npart;
+			}
+			else {
+				/* Empty element */
+				return NULL;
+			}
 			break;
 		}
 	}
@@ -1099,9 +1146,11 @@ rspamd_smtp_received_process_part (struct rspamd_task *task,
 		}
 		break;
 	case skip_spaces:
-		*last = p - (const guchar *)data;
+		if (p > (const guchar *)data) {
+			*last = p - (const guchar *) data;
 
-		return npart;
+			return npart;
+		}
 	default:
 		break;
 	}
@@ -1151,7 +1200,6 @@ rspamd_smtp_received_spill (struct rspamd_task *task,
 	len = end > p ? end - p : 0;
 	DL_APPEND (head, cur_part);
 
-
 	if (len > 2 && (lc_map[p[0]] == 'b' &&
 					lc_map[p[1]] == 'y')) {
 		p += sizeof ("by") - 1;
@@ -1192,6 +1240,12 @@ rspamd_smtp_received_spill (struct rspamd_task *task,
 				cur_part = rspamd_smtp_received_process_part (task, p, end - p,
 						RSPAMD_RECEIVED_PART_FOR, &pos);
 			}
+			else if (len > sizeof ("id") && (lc_map[p[0]] == 'i' &&
+											  lc_map[p[1]] == 'd')) {
+				p += sizeof ("id") - 1;
+				cur_part = rspamd_smtp_received_process_part (task, p, end - p,
+						RSPAMD_RECEIVED_PART_ID, &pos);
+			}
 			else {
 				while (p < end) {
 					if (!(g_ascii_isspace (*p) || *p == '(' || *p == ';')) {
@@ -1216,7 +1270,8 @@ rspamd_smtp_received_spill (struct rspamd_task *task,
 			}
 
 			if (!cur_part) {
-				return NULL;
+				p ++;
+				len = end > p ? end - p : 0;
 			}
 			else {
 				g_assert (pos != 0);
@@ -1287,7 +1342,7 @@ rspamd_smtp_received_process_rdns (struct rspamd_task *task,
 
 static gboolean
 rspamd_smtp_received_process_host_tcpinfo (struct rspamd_task *task,
-										   struct received_header *rh,
+										   struct rspamd_received_header *rh,
 										   const gchar *data,
 										   gsize len)
 {
@@ -1302,7 +1357,8 @@ rspamd_smtp_received_process_host_tcpinfo (struct rspamd_task *task,
 		if (brace_pos) {
 			addr = rspamd_parse_inet_address_pool (data + 1,
 					brace_pos - data - 1,
-					task->task_pool);
+					task->task_pool,
+					RSPAMD_INET_ADDRESS_PARSE_RECEIVED);
 
 			if (addr) {
 				rh->addr = addr;
@@ -1316,7 +1372,7 @@ rspamd_smtp_received_process_host_tcpinfo (struct rspamd_task *task,
 		if (g_ascii_isxdigit (data[0])) {
 			/* Try to parse IP address */
 			addr = rspamd_parse_inet_address_pool (data,
-					len, task->task_pool);
+					len, task->task_pool, RSPAMD_INET_ADDRESS_PARSE_RECEIVED);
 			if (addr) {
 				rh->addr = addr;
 				rh->real_ip = rspamd_mempool_strdup (task->task_pool,
@@ -1337,7 +1393,8 @@ rspamd_smtp_received_process_host_tcpinfo (struct rspamd_task *task,
 				if (ebrace_pos) {
 					addr = rspamd_parse_inet_address_pool (obrace_pos + 1,
 							ebrace_pos - obrace_pos - 1,
-							task->task_pool);
+							task->task_pool,
+							RSPAMD_INET_ADDRESS_PARSE_RECEIVED);
 
 					if (addr) {
 						rh->addr = addr;
@@ -1373,7 +1430,7 @@ rspamd_smtp_received_process_host_tcpinfo (struct rspamd_task *task,
 static void
 rspamd_smtp_received_process_from (struct rspamd_task *task,
 								   struct rspamd_received_part *rpart,
-								   struct received_header *rh)
+								   struct rspamd_received_header *rh)
 {
 	if (rpart->dlen > 0) {
 		/* We have seen multiple cases:
@@ -1384,11 +1441,11 @@ rspamd_smtp_received_process_from (struct rspamd_task *task,
 		 * - hostname ([ip]:port helo=xxx)
 		 * Maybe more...
 		 */
-		gboolean seen_ip_in_data = FALSE, seen_rdns_in_comment = FALSE;
+		gboolean seen_ip_in_data = FALSE;
 
 		if (rpart->head_comment && rpart->head_comment->dlen > 0) {
 			/* We can have info within comment as part of RFC */
-			seen_rdns_in_comment = rspamd_smtp_received_process_host_tcpinfo (
+			rspamd_smtp_received_process_host_tcpinfo (
 					task, rh,
 					rpart->head_comment->data, rpart->head_comment->dlen);
 		}
@@ -1402,7 +1459,8 @@ rspamd_smtp_received_process_from (struct rspamd_task *task,
 				if (brace_pos) {
 					addr = rspamd_parse_inet_address_pool (rpart->data + 1,
 							brace_pos - rpart->data - 1,
-							task->task_pool);
+							task->task_pool,
+							RSPAMD_INET_ADDRESS_PARSE_RECEIVED);
 
 					if (addr) {
 						seen_ip_in_data = TRUE;
@@ -1416,7 +1474,8 @@ rspamd_smtp_received_process_from (struct rspamd_task *task,
 				/* Try to parse IP address */
 				rspamd_inet_addr_t *addr;
 				addr = rspamd_parse_inet_address_pool (rpart->data,
-						rpart->dlen, task->task_pool);
+						rpart->dlen, task->task_pool,
+						RSPAMD_INET_ADDRESS_PARSE_RECEIVED);
 				if (addr) {
 					seen_ip_in_data = TRUE;
 					rh->addr = addr;
@@ -1457,7 +1516,7 @@ int
 rspamd_smtp_received_parse (struct rspamd_task *task,
 							const char *data,
 							size_t len,
-							struct received_header *rh)
+							struct rspamd_received_header *rh)
 {
 	goffset date_pos = -1;
 	struct rspamd_received_part *head, *cur;
@@ -1469,7 +1528,7 @@ rspamd_smtp_received_parse (struct rspamd_task *task,
 		return -1;
 	}
 
-	rh->type = RSPAMD_RECEIVED_UNKNOWN;
+	rh->flags = RSPAMD_RECEIVED_UNKNOWN;
 
 	DL_FOREACH (head, cur) {
 		switch (cur->type) {
@@ -1490,7 +1549,7 @@ rspamd_smtp_received_parse (struct rspamd_task *task,
 				RSPAMD_FTOK_ASSIGN (&t2, "smtp");
 
 				if (rspamd_ftok_cmp (&t1, &t2) == 0) {
-					rh->type = RSPAMD_RECEIVED_SMTP;
+					rh->flags = RSPAMD_RECEIVED_SMTP;
 				}
 
 				RSPAMD_FTOK_ASSIGN (&t2, "esmtp");
@@ -1501,11 +1560,11 @@ rspamd_smtp_received_parse (struct rspamd_task *task,
 					 */
 					if (t1.len == t2.len + 1) {
 						if (t1.begin[t2.len] == 'a') {
-							rh->type = RSPAMD_RECEIVED_ESMTPA;
+							rh->flags = RSPAMD_RECEIVED_ESMTPA;
 							rh->flags |= RSPAMD_RECEIVED_FLAG_AUTHENTICATED;
 						}
 						else if (t1.begin[t2.len] == 's') {
-							rh->type = RSPAMD_RECEIVED_ESMTPS;
+							rh->flags = RSPAMD_RECEIVED_ESMTPS;
 							rh->flags |= RSPAMD_RECEIVED_FLAG_SSL;
 						}
 						continue;
@@ -1513,14 +1572,14 @@ rspamd_smtp_received_parse (struct rspamd_task *task,
 					else if (t1.len == t2.len + 2) {
 						if (t1.begin[t2.len] == 's' &&
 								t1.begin[t2.len + 1] == 'a') {
-							rh->type = RSPAMD_RECEIVED_ESMTPSA;
+							rh->flags = RSPAMD_RECEIVED_ESMTPSA;
 							rh->flags |= RSPAMD_RECEIVED_FLAG_AUTHENTICATED;
 							rh->flags |= RSPAMD_RECEIVED_FLAG_SSL;
 						}
 						continue;
 					}
 					else if (t1.len == t2.len) {
-						rh->type = RSPAMD_RECEIVED_ESMTP;
+						rh->flags = RSPAMD_RECEIVED_ESMTP;
 						continue;
 					}
 				}
@@ -1528,21 +1587,21 @@ rspamd_smtp_received_parse (struct rspamd_task *task,
 				RSPAMD_FTOK_ASSIGN (&t2, "lmtp");
 
 				if (rspamd_ftok_cmp (&t1, &t2) == 0) {
-					rh->type = RSPAMD_RECEIVED_LMTP;
+					rh->flags = RSPAMD_RECEIVED_LMTP;
 					continue;
 				}
 
 				RSPAMD_FTOK_ASSIGN (&t2, "imap");
 
 				if (rspamd_ftok_cmp (&t1, &t2) == 0) {
-					rh->type = RSPAMD_RECEIVED_IMAP;
+					rh->flags = RSPAMD_RECEIVED_IMAP;
 					continue;
 				}
 
 				RSPAMD_FTOK_ASSIGN (&t2, "local");
 
 				if (rspamd_ftok_cmp (&t1, &t2) == 0) {
-					rh->type = RSPAMD_RECEIVED_LOCAL;
+					rh->flags = RSPAMD_RECEIVED_LOCAL;
 					continue;
 				}
 
@@ -1551,18 +1610,34 @@ rspamd_smtp_received_parse (struct rspamd_task *task,
 				if (rspamd_ftok_starts_with (&t1, &t2)) {
 					if (t1.len == t2.len + 1) {
 						if (t1.begin[t2.len] == 's') {
-							rh->type = RSPAMD_RECEIVED_HTTP;
+							rh->flags = RSPAMD_RECEIVED_HTTP;
 							rh->flags |= RSPAMD_RECEIVED_FLAG_SSL;
 						}
 					}
 					else if (t1.len == t2.len) {
-						rh->type = RSPAMD_RECEIVED_HTTP;
+						rh->flags = RSPAMD_RECEIVED_HTTP;
 					}
 
 					continue;
 				}
 			}
 
+			break;
+		case RSPAMD_RECEIVED_PART_FOR:
+			rh->for_addr = rspamd_email_address_from_smtp (cur->data, cur->dlen);
+
+			if (rh->for_addr) {
+				if (rh->for_addr->addr_len > 0) {
+					t1.begin = rh->for_addr->addr;
+					t1.len = rh->for_addr->addr_len;
+					rh->for_mbox = rspamd_mempool_ftokdup (task->task_pool,
+							&t1);
+				}
+
+				rspamd_mempool_add_destructor (task->task_pool,
+						(rspamd_mempool_destruct_t)rspamd_email_address_free,
+						rh->for_addr);
+			}
 			break;
 		default:
 			/* Do nothing */
@@ -1580,8 +1655,73 @@ rspamd_smtp_received_parse (struct rspamd_task *task,
 
 	if (date_pos > 0 && date_pos < len) {
 		rh->timestamp = rspamd_parse_smtp_date (data + date_pos,
-				len - date_pos);
+				len - date_pos, NULL);
 	}
 
 	return 0;
+}
+
+struct rspamd_mime_header *
+rspamd_message_get_header_from_hash (struct rspamd_mime_headers_table *headers,
+									 const gchar *field)
+{
+	khiter_t k;
+	khash_t(rspamd_mime_headers_htb) *htb = &headers->htb;
+
+	if (htb) {
+		k = kh_get (rspamd_mime_headers_htb, htb, (gchar *) field);
+
+		if (k == kh_end (htb)) {
+			return NULL;
+		}
+
+		return kh_value (htb, k);
+	}
+
+	return NULL;
+}
+
+struct rspamd_mime_header *
+rspamd_message_get_header_array (struct rspamd_task *task,
+								 const gchar *field)
+{
+	return rspamd_message_get_header_from_hash (
+			MESSAGE_FIELD_CHECK (task, raw_headers),
+			field);
+}
+
+static void
+rspamd_message_headers_dtor (struct rspamd_mime_headers_table *hdrs)
+{
+	if (hdrs) {
+		kfree (hdrs->htb.keys);
+		kfree (hdrs->htb.vals);
+		kfree (hdrs->htb.flags);
+		g_free (hdrs);
+	}
+}
+
+struct rspamd_mime_headers_table *
+rspamd_message_headers_ref (struct rspamd_mime_headers_table *hdrs)
+{
+	REF_RETAIN (hdrs);
+
+	return hdrs;
+}
+
+void
+rspamd_message_headers_unref (struct rspamd_mime_headers_table *hdrs)
+{
+	REF_RELEASE (hdrs);
+}
+
+struct rspamd_mime_headers_table *
+rspamd_message_headers_new (void)
+{
+	struct rspamd_mime_headers_table *nhdrs;
+
+	nhdrs = g_malloc0 (sizeof (*nhdrs));
+	REF_INIT_RETAIN (nhdrs, rspamd_message_headers_dtor);
+
+	return nhdrs;
 }
\ No newline at end of file
