@@ -14,6 +14,7 @@
  * limitations under the License.
  */
 
+
 #include "config.h"
 #include "task.h"
 #include "mime_parser.h"
@@ -21,6 +22,9 @@
 #include "message.h"
 #include "multipattern.h"
 #include "contrib/libottery/ottery.h"
+#include "contrib/uthash/utlist.h"
+#include <openssl/cms.h>
+#include <openssl/pkcs7.h>
 
 struct rspamd_mime_parser_lib_ctx {
 	struct rspamd_multipattern *mp_boundary;
@@ -75,8 +79,17 @@ static enum rspamd_mime_parse_error
 rspamd_mime_parse_normal_part (struct rspamd_task *task,
 		struct rspamd_mime_part *part,
 		struct rspamd_mime_parser_ctx *st,
+		struct rspamd_content_type *ct,
 		GError **err);
 
+static enum rspamd_mime_parse_error
+rspamd_mime_process_multipart_node (struct rspamd_task *task,
+									struct rspamd_mime_parser_ctx *st,
+									struct rspamd_mime_part *multipart,
+									const gchar *start, const gchar *end,
+									gboolean is_finished,
+									GError **err);
+
 
 #define RSPAMD_MIME_QUARK (rspamd_mime_parser_quark())
 static GQuark
@@ -103,6 +116,9 @@ rspamd_cte_to_string (enum rspamd_cte ct)
 	case RSPAMD_CTE_B64:
 		ret = "base64";
 		break;
+	case RSPAMD_CTE_UUE:
+		ret = "X-uuencode";
+		break;
 	default:
 		break;
 	}
@@ -129,6 +145,15 @@ rspamd_cte_from_string (const gchar *str)
 	else if (strcmp (str, "base64") == 0) {
 		ret = RSPAMD_CTE_B64;
 	}
+	else if (strcmp (str, "X-uuencode") == 0) {
+		ret = RSPAMD_CTE_UUE;
+	}
+	else if (strcmp (str, "uuencode") == 0) {
+		ret = RSPAMD_CTE_UUE;
+	}
+	else if (strcmp (str, "X-uue") == 0) {
+		ret = RSPAMD_CTE_UUE;
+	}
 
 	return ret;
 }
@@ -170,6 +195,11 @@ rspamd_mime_parse_cte (const gchar *in, gsize len)
 	case 0x171029DE1B0423A9ULL: /* base-64 */
 		ret = RSPAMD_CTE_B64;
 		break;
+	case 0x420b54dc00d13cecULL: /* uuencode */
+	case 0x8df6700b8f6c4cf9ULL: /* x-uuencode */
+	case 0x41f725ec544356d3ULL: /* x-uue */
+		ret = RSPAMD_CTE_UUE;
+		break;
 	}
 
 	return ret;
@@ -180,7 +210,8 @@ rspamd_mime_part_get_cte_heuristic (struct rspamd_task *task,
 		struct rspamd_mime_part *part)
 {
 	const guint check_len = 128;
-	guint real_len, nspaces = 0, neqsign = 0, n8bit = 0, nqpencoded = 0;
+	guint real_len, nspaces = 0, neqsign = 0, n8bit = 0, nqpencoded = 0,
+		padeqsign = 0, nupper = 0, nlower = 0;
 	gboolean b64_chars = TRUE;
 	const guchar *p, *end;
 	enum rspamd_cte ret = RSPAMD_CTE_UNKNOWN;
@@ -193,18 +224,51 @@ rspamd_mime_part_get_cte_heuristic (struct rspamd_task *task,
 		p ++;
 	}
 
+	if (end - p > sizeof ("begin-base64 ")) {
+		const guchar *uue_start;
+
+		if (memcmp (p, "begin ", sizeof ("begin ") - 1) == 0) {
+			uue_start = p + sizeof ("begin ") - 1;
+
+			while (uue_start < end && g_ascii_isspace (*uue_start)) {
+				uue_start ++;
+			}
+
+			if (uue_start < end && g_ascii_isdigit (*uue_start)) {
+				return RSPAMD_CTE_UUE;
+			}
+		}
+		else if (memcmp (p, "begin-base64 ", sizeof ("begin-base64 ") - 1) == 0) {
+			uue_start = p + sizeof ("begin ") - 1;
+
+			while (uue_start < end && g_ascii_isspace (*uue_start)) {
+				uue_start ++;
+			}
+
+			if (uue_start < end && g_ascii_isdigit (*uue_start)) {
+				return RSPAMD_CTE_UUE;
+			}
+		}
+	}
+
+	/* Skip trailing spaces */
+	while (end > p && g_ascii_isspace (*(end - 1))) {
+		end --;
+	}
+
 	if (end > p + 2) {
 		if (*(end - 1) == '=') {
-			neqsign ++;
+			padeqsign ++;
 			end --;
 		}
 
 		if (*(end - 1) == '=') {
-			neqsign ++;
+			padeqsign ++;
 			end --;
 		}
 	}
 
+	/* Adjust end to analyse only first characters */
 	if (end - p > real_len) {
 		end = p + real_len;
 	}
@@ -214,6 +278,7 @@ rspamd_mime_part_get_cte_heuristic (struct rspamd_task *task,
 			nspaces ++;
 		}
 		else if (*p == '=') {
+			b64_chars = FALSE; /* Eqsign must not be inside base64 */
 			neqsign ++;
 			p ++;
 
@@ -231,12 +296,74 @@ rspamd_mime_part_get_cte_heuristic (struct rspamd_task *task,
 		else if (!(g_ascii_isalnum (*p) || *p == '/' || *p == '+')) {
 			b64_chars = FALSE;
 		}
+		else if (g_ascii_isupper (*p)) {
+			nupper ++;
+		}
+		else if (g_ascii_islower (*p)) {
+			nlower ++;
+		}
 
 		p ++;
 	}
 
-	if (b64_chars && neqsign < 2 && nspaces == 0) {
-		ret = RSPAMD_CTE_B64;
+	if (b64_chars && neqsign <= 2 && nspaces == 0) {
+		/* Need more thinking */
+
+		if (part->raw_data.len > 80) {
+			if (padeqsign > 0) {
+				ret = RSPAMD_CTE_B64;
+			}
+			else {
+				/* We have a large piece of data with no spaces and base64
+				 * symbols only, no padding is detected as well...
+				 *
+				 * There is a small chance that our first 128 characters
+				 * are either some garbage or it is a base64 with no padding
+				 * (e.g. when it is not needed)
+				 */
+				if (nupper > 1 && nlower > 1) {
+					/*
+					 * We have both uppercase and lowercase letters, so it can be
+					 * base64
+					 */
+					ret = RSPAMD_CTE_B64;
+				}
+				else {
+					ret = RSPAMD_CTE_7BIT;
+				}
+			}
+		}
+		else {
+
+			if (((end - (const guchar *)part->raw_data.begin) + padeqsign) % 4 == 0) {
+				if (padeqsign == 0) {
+					/*
+					 * It can be either base64 or plain text, hard to say
+					 * Let's assume that if we have > 1 uppercase it is
+					 * likely base64
+					 */
+					if (nupper > 1 && nlower > 1) {
+						ret = RSPAMD_CTE_B64;
+					}
+					else {
+						ret = RSPAMD_CTE_7BIT;
+					}
+
+				}
+				else {
+					ret = RSPAMD_CTE_B64;
+				}
+			}
+			else {
+				/* No way */
+				if (padeqsign == 1 || padeqsign == 2) {
+					ret = RSPAMD_CTE_B64;
+				}
+				else {
+					ret = RSPAMD_CTE_7BIT;
+				}
+			}
+		}
 	}
 	else if (n8bit == 0) {
 		if (neqsign > 2 && nqpencoded > 2) {
@@ -251,29 +378,27 @@ rspamd_mime_part_get_cte_heuristic (struct rspamd_task *task,
 	}
 
 	msg_debug_mime ("detected cte: %s", rspamd_cte_to_string (ret));
+
 	return ret;
 }
 
 static void
 rspamd_mime_part_get_cte (struct rspamd_task *task,
-		GHashTable *hdrs,
-		struct rspamd_mime_part *part,
-		gboolean apply_heuristic)
+						  struct rspamd_mime_headers_table *hdrs,
+						  struct rspamd_mime_part *part,
+						  gboolean apply_heuristic)
 {
-	struct rspamd_mime_header *hdr;
-	guint i;
-	GPtrArray *hdrs_cte;
+	struct rspamd_mime_header *hdr, *cur;
 	enum rspamd_cte cte = RSPAMD_CTE_UNKNOWN;
+	gboolean parent_propagated = FALSE;
 
-	hdrs_cte = rspamd_message_get_header_from_hash (hdrs,
-			task->task_pool,
-			"Content-Transfer-Encoding", FALSE);
-
-	if (hdrs_cte == NULL) {
+	hdr = rspamd_message_get_header_from_hash (hdrs, "Content-Transfer-Encoding");
 
+	if (hdr == NULL) {
 		if (part->parent_part && part->parent_part->cte != RSPAMD_CTE_UNKNOWN &&
 				!(part->parent_part->flags & RSPAMD_MIME_PART_MISSING_CTE)) {
 			part->cte = part->parent_part->cte;
+			parent_propagated = TRUE;
 
 			goto check_cte;
 		}
@@ -287,12 +412,11 @@ rspamd_mime_part_get_cte (struct rspamd_task *task,
 		part->flags |= RSPAMD_MIME_PART_MISSING_CTE;
 	}
 	else {
-		for (i = 0; i < hdrs_cte->len; i ++) {
+		DL_FOREACH (hdr, cur) {
 			gsize hlen;
 			gchar lc_buf[128];
 
-			hdr = g_ptr_array_index (hdrs_cte, i);
-			hlen = rspamd_snprintf (lc_buf, sizeof (lc_buf), "%s", hdr->value);
+			hlen = rspamd_snprintf (lc_buf, sizeof (lc_buf), "%s", cur->value);
 			rspamd_str_lc (lc_buf, hlen);
 			cte = rspamd_mime_parse_cte (lc_buf, hlen);
 
@@ -323,6 +447,11 @@ check_cte:
 					part->cte = cte;
 					part->flags |= RSPAMD_MIME_PART_BAD_CTE;
 				}
+				else if (cte != part->cte && parent_propagated) {
+					part->cte = cte;
+					msg_info_task ("detected missing CTE for part as: %s",
+							rspamd_cte_to_string (part->cte));
+				}
 			}
 			else {
 				msg_debug_mime ("processed cte: %s",
@@ -337,19 +466,16 @@ check_cte:
 static void
 rspamd_mime_part_get_cd (struct rspamd_task *task, struct rspamd_mime_part *part)
 {
-	struct rspamd_mime_header *hdr;
-	guint i;
-	GPtrArray *hdrs;
+	struct rspamd_mime_header *hdr, *cur;
 	struct rspamd_content_disposition *cd = NULL;
 	rspamd_ftok_t srch;
 	struct rspamd_content_type_param *found;
 
-	hdrs = rspamd_message_get_header_from_hash (part->raw_headers,
-			task->task_pool,
-			"Content-Disposition", FALSE);
+	hdr = rspamd_message_get_header_from_hash (part->raw_headers,
+			"Content-Disposition");
 
 
-	if (hdrs == NULL) {
+	if (hdr == NULL) {
 		cd = rspamd_mempool_alloc0 (task->task_pool, sizeof (*cd));
 		cd->type = RSPAMD_CT_INLINE;
 
@@ -370,15 +496,13 @@ rspamd_mime_part_get_cd (struct rspamd_task *task, struct rspamd_mime_part *part
 		}
 	}
 	else {
-		for (i = 0; i < hdrs->len; i ++) {
+		DL_FOREACH (hdr, cur) {
 			gsize hlen;
-
-			hdr = g_ptr_array_index (hdrs, i);
 			cd = NULL;
 
-			if (hdr->decoded) {
-				hlen = strlen (hdr->decoded);
-				cd = rspamd_content_disposition_parse (hdr->decoded, hlen,
+			if (cur->decoded) {
+				hlen = strlen (cur->decoded);
+				cd = rspamd_content_disposition_parse (cur->decoded, hlen,
 						task->task_pool);
 			}
 
@@ -406,6 +530,30 @@ rspamd_mime_part_get_cd (struct rspamd_task *task, struct rspamd_mime_part *part
 						cd->lc_data, &cd->filename);
 				break;
 			}
+			else if (part->ct) {
+				/*
+				 * Even in case of malformed Content-Disposition, we can still
+				 * fall back to Content-Type
+				 */
+				cd = rspamd_mempool_alloc0 (task->task_pool, sizeof (*cd));
+				cd->type = RSPAMD_CT_INLINE;
+
+				/* We can also have content dispositon definitions in Content-Type */
+				if (part->ct->attrs) {
+					RSPAMD_FTOK_ASSIGN (&srch, "name");
+					found = g_hash_table_lookup (part->ct->attrs, &srch);
+
+					if (!found) {
+						RSPAMD_FTOK_ASSIGN (&srch, "filename");
+						found = g_hash_table_lookup (part->ct->attrs, &srch);
+					}
+
+					if (found) {
+						cd->type = RSPAMD_CT_ATTACHMENT;
+						memcpy (&cd->filename, &found->value, sizeof (cd->filename));
+					}
+				}
+			}
 		}
 	}
 
@@ -438,6 +586,7 @@ static enum rspamd_mime_parse_error
 rspamd_mime_parse_normal_part (struct rspamd_task *task,
 		struct rspamd_mime_part *part,
 		struct rspamd_mime_parser_ctx *st,
+		struct rspamd_content_type *ct,
 		GError **err)
 {
 	rspamd_fstring_t *parsed;
@@ -445,7 +594,8 @@ rspamd_mime_parse_normal_part (struct rspamd_task *task,
 
 	g_assert (part != NULL);
 
-	rspamd_mime_part_get_cte (task, part->raw_headers, part, TRUE);
+	rspamd_mime_part_get_cte (task, part->raw_headers, part,
+			!(part->ct->flags & RSPAMD_CONTENT_TYPE_MESSAGE));
 	rspamd_mime_part_get_cd (task, part);
 
 	switch (part->cte) {
@@ -465,13 +615,14 @@ rspamd_mime_parse_normal_part (struct rspamd_task *task,
 			}
 		}
 
-		if (IS_CT_TEXT (part->ct)) {
+		if (part->ct && (part->ct->flags & RSPAMD_CONTENT_TYPE_TEXT)) {
 			/* Need to copy text as we have couple of in-place change functions */
 			parsed = rspamd_fstring_sized_new (part->raw_data.len);
 			parsed->len = part->raw_data.len;
 			memcpy (parsed->str, part->raw_data.begin, parsed->len);
 			part->parsed_data.begin = parsed->str;
 			part->parsed_data.len = parsed->len;
+			rspamd_mempool_notify_alloc (task->task_pool, parsed->len);
 			rspamd_mempool_add_destructor (task->task_pool,
 					(rspamd_mempool_destruct_t)rspamd_fstring_free, parsed);
 		}
@@ -488,6 +639,7 @@ rspamd_mime_parse_normal_part (struct rspamd_task *task,
 			parsed->len = r;
 			part->parsed_data.begin = parsed->str;
 			part->parsed_data.len = parsed->len;
+			rspamd_mempool_notify_alloc (task->task_pool, parsed->len);
 			rspamd_mempool_add_destructor (task->task_pool,
 					(rspamd_mempool_destruct_t)rspamd_fstring_free, parsed);
 		}
@@ -499,6 +651,7 @@ rspamd_mime_parse_normal_part (struct rspamd_task *task,
 			parsed->len = part->raw_data.len;
 			part->parsed_data.begin = parsed->str;
 			part->parsed_data.len = parsed->len;
+			rspamd_mempool_notify_alloc (task->task_pool, parsed->len);
 			rspamd_mempool_add_destructor (task->task_pool,
 					(rspamd_mempool_destruct_t)rspamd_fstring_free, parsed);
 		}
@@ -510,20 +663,106 @@ rspamd_mime_parse_normal_part (struct rspamd_task *task,
 				parsed->str, &parsed->len);
 		part->parsed_data.begin = parsed->str;
 		part->parsed_data.len = parsed->len;
+		rspamd_mempool_notify_alloc (task->task_pool, parsed->len);
 		rspamd_mempool_add_destructor (task->task_pool,
 				(rspamd_mempool_destruct_t)rspamd_fstring_free, parsed);
 		break;
+	case RSPAMD_CTE_UUE:
+		parsed = rspamd_fstring_sized_new (part->raw_data.len / 4 * 3 + 12);
+		r = rspamd_decode_uue_buf (part->raw_data.begin, part->raw_data.len,
+				parsed->str, parsed->allocated);
+		rspamd_mempool_notify_alloc (task->task_pool, parsed->len);
+		rspamd_mempool_add_destructor (task->task_pool,
+				(rspamd_mempool_destruct_t)rspamd_fstring_free, parsed);
+		if (r != -1) {
+			parsed->len = r;
+			part->parsed_data.begin = parsed->str;
+			part->parsed_data.len = parsed->len;
+		}
+		else {
+			msg_err_task ("invalid uuencoding in encoded part, assume 8bit");
+			part->ct->flags |= RSPAMD_CONTENT_TYPE_BROKEN;
+			part->cte = RSPAMD_CTE_8BIT;
+			parsed->len = MIN (part->raw_data.len, parsed->allocated);
+			memcpy (parsed->str, part->raw_data.begin, parsed->len);
+			rspamd_mempool_notify_alloc (task->task_pool, parsed->len);
+			part->parsed_data.begin = parsed->str;
+			part->parsed_data.len = parsed->len;
+		}
+		break;
 	default:
 		g_assert_not_reached ();
 	}
 
-	part->id = task->parts->len;
-	g_ptr_array_add (task->parts, part);
+	part->part_number = MESSAGE_FIELD (task, parts)->len;
+	part->urls = g_ptr_array_new ();
+	g_ptr_array_add (MESSAGE_FIELD (task, parts), part);
 	msg_debug_mime ("parsed data part %T/%T of length %z (%z orig), %s cte",
 			&part->ct->type, &part->ct->subtype, part->parsed_data.len,
 			part->raw_data.len, rspamd_cte_to_string (part->cte));
 	rspamd_mime_parser_calc_digest (part);
 
+	if (ct && (ct->flags & RSPAMD_CONTENT_TYPE_SMIME)) {
+		CMS_ContentInfo *cms;
+		const unsigned char *der_beg = part->parsed_data.begin;
+		cms = d2i_CMS_ContentInfo (NULL, &der_beg, part->parsed_data.len);
+
+		if (cms) {
+			const ASN1_OBJECT *asn_ct = CMS_get0_eContentType (cms);
+			int ct_nid = OBJ_obj2nid (asn_ct);
+
+			if (ct_nid == NID_pkcs7_data) {
+				BIO *bio = BIO_new_mem_buf (part->parsed_data.begin,
+						part->parsed_data.len);
+
+				PKCS7 *p7;
+				p7 = d2i_PKCS7_bio (bio, NULL);
+
+				if (p7) {
+					ct_nid = OBJ_obj2nid (p7->type);
+
+					if (ct_nid == NID_pkcs7_signed) {
+						PKCS7 *p7_signed_content = p7->d.sign->contents;
+
+						ct_nid =  OBJ_obj2nid (p7_signed_content->type);
+
+						if (ct_nid == NID_pkcs7_data) {
+							int ret;
+
+							msg_debug_mime ("found an additional part inside of "
+											"smime structure of type %T/%T; length=%d",
+									&ct->type, &ct->subtype, p7_signed_content->d.data->length);
+							/*
+							 * Since ASN.1 structures are freed, we need to copy
+							 * the content
+							 */
+							gchar *cpy = rspamd_mempool_alloc (task->task_pool,
+									p7_signed_content->d.data->length);
+							memcpy (cpy, p7_signed_content->d.data->data,
+									p7_signed_content->d.data->length);
+							ret = rspamd_mime_process_multipart_node (task,
+									st, NULL,
+									cpy,cpy + p7_signed_content->d.data->length,
+									TRUE, err);
+
+							PKCS7_free (p7);
+							BIO_free (bio);
+							CMS_ContentInfo_free (cms);
+
+							return ret;
+						}
+					}
+
+					PKCS7_free (p7);
+				}
+
+				BIO_free (bio);
+			}
+
+			CMS_ContentInfo_free (cms);
+		}
+	}
+
 	return RSPAMD_MIME_PARSE_OK;
 }
 
@@ -546,12 +785,10 @@ rspamd_mime_process_multipart_node (struct rspamd_task *task,
 		GError **err)
 {
 	struct rspamd_content_type *ct, *sel = NULL;
-	struct rspamd_mime_header *hdr;
-	GPtrArray *hdrs = NULL;
+	struct rspamd_mime_header *hdr = NULL, *cur;
 	struct rspamd_mime_part *npart;
 	GString str;
 	goffset hdr_pos, body_pos;
-	guint i;
 	enum rspamd_mime_parse_error ret = RSPAMD_MIME_PARSE_FATAL;
 
 
@@ -592,9 +829,8 @@ rspamd_mime_process_multipart_node (struct rspamd_task *task,
 	npart = rspamd_mempool_alloc0 (task->task_pool,
 			sizeof (struct rspamd_mime_part));
 	npart->parent_part = multipart;
-	npart->raw_headers =  g_hash_table_new_full (rspamd_strcase_hash,
-			rspamd_strcase_equal, NULL, rspamd_ptr_array_free_hard);
-	npart->headers_order = g_queue_new ();
+	npart->raw_headers =  rspamd_message_headers_new ();
+	npart->headers_order = NULL;
 
 	if (multipart) {
 		if (multipart->specific.mp->children == NULL) {
@@ -612,15 +848,19 @@ rspamd_mime_process_multipart_node (struct rspamd_task *task,
 
 		if (npart->raw_headers_len > 0) {
 			rspamd_mime_headers_process (task, npart->raw_headers,
-					npart->headers_order,
+					&npart->headers_order,
 					npart->raw_headers_str,
 					npart->raw_headers_len,
 					FALSE);
+
+			/* Preserve the natural order */
+			if (npart->headers_order) {
+				LL_REVERSE2 (npart->headers_order, ord_next);
+			}
 		}
 
-		hdrs = rspamd_message_get_header_from_hash (npart->raw_headers,
-				task->task_pool,
-				"Content-Type", FALSE);
+		hdr = rspamd_message_get_header_from_hash (npart->raw_headers,
+				"Content-Type");
 
 	}
 	else {
@@ -631,11 +871,10 @@ rspamd_mime_process_multipart_node (struct rspamd_task *task,
 	}
 
 
-	if (hdrs != NULL) {
+	if (hdr != NULL) {
 
-		for (i = 0; i < hdrs->len; i ++) {
-			hdr = g_ptr_array_index (hdrs, i);
-			ct = rspamd_content_type_parse (hdr->decoded, strlen (hdr->decoded),
+		DL_FOREACH (hdr, cur) {
+			ct = rspamd_content_type_parse (cur->decoded, strlen (cur->decoded),
 					task->task_pool);
 
 			/* Here we prefer multipart content-type or any content-type */
@@ -661,6 +900,7 @@ rspamd_mime_process_multipart_node (struct rspamd_task *task,
 	if (sel->flags & RSPAMD_CONTENT_TYPE_MULTIPART) {
 		st->nesting ++;
 		g_ptr_array_add (st->stack, npart);
+		npart->part_type = RSPAMD_MIME_PART_MULTIPART;
 		npart->specific.mp = rspamd_mempool_alloc0 (task->task_pool,
 				sizeof (struct rspamd_mime_multipart));
 		memcpy (&npart->specific.mp->boundary, &sel->orig_boundary,
@@ -670,14 +910,15 @@ rspamd_mime_process_multipart_node (struct rspamd_task *task,
 	else if (sel->flags & RSPAMD_CONTENT_TYPE_MESSAGE) {
 		st->nesting ++;
 		g_ptr_array_add (st->stack, npart);
+		npart->part_type = RSPAMD_MIME_PART_MESSAGE;
 
-		if ((ret = rspamd_mime_parse_normal_part (task, npart, st, err))
+		if ((ret = rspamd_mime_parse_normal_part (task, npart, st, sel, err))
 				== RSPAMD_MIME_PARSE_OK) {
 			ret = rspamd_mime_parse_message (task, npart, st, err);
 		}
 	}
 	else {
-		ret = rspamd_mime_parse_normal_part (task, npart, st, err);
+		ret = rspamd_mime_parse_normal_part (task, npart, st, sel, err);
 	}
 
 	return ret;
@@ -701,10 +942,11 @@ rspamd_mime_parse_multipart_cb (struct rspamd_task *task,
 		st->pos = cb->part_start;
 	}
 	else {
-		/* We have seen the start of the boundary */
-		if (cb->part_start < pos) {
-			/* We should have seen some boundary */
-			g_assert (cb->cur_boundary != NULL);
+		/*
+		 * We have seen the start of the boundary,
+		 * but it might be unsuitable (e.g. in broken headers)
+		 */
+		if (cb->part_start < pos && cb->cur_boundary) {
 
 			if ((ret = rspamd_mime_process_multipart_node (task, cb->st,
 					cb->multipart, cb->part_start, pos, TRUE, cb->err))
@@ -848,8 +1090,9 @@ rspamd_mime_parse_multipart_part (struct rspamd_task *task,
 		return RSPAMD_MIME_PARSE_NESTING;
 	}
 
-	part->id = task->parts->len;
-	g_ptr_array_add (task->parts, part);
+	part->part_number = MESSAGE_FIELD (task, parts)->len;
+	part->urls = g_ptr_array_new ();
+	g_ptr_array_add (MESSAGE_FIELD (task, parts), part);
 	st->nesting ++;
 	rspamd_mime_part_get_cte (task, part->raw_headers, part, FALSE);
 
@@ -940,17 +1183,26 @@ rspamd_mime_preprocess_cb (struct rspamd_multipattern *mp,
 				bend ++;
 			}
 
-			if (*bend == '\r') {
-				bend ++;
+			while (bend < end) {
+				if (*bend == '\r') {
+					bend ++;
 
-				/* \r\n */
-				if (*bend == '\n') {
+					/* \r\n */
+					if (bend < end && *bend == '\n') {
+						bend ++;
+					}
+				}
+				else if (*bend == '\n') {
+					/* \n */
 					bend ++;
 				}
-			}
-			else {
-				/* \n */
-				bend ++;
+				else if (g_ascii_isspace (*bend)){
+					/* Spaces in the same line, skip them */
+					bend ++;
+					continue;
+				}
+
+				break;
 			}
 
 			b.boundary = p - st->start - 2;
@@ -1098,8 +1350,7 @@ rspamd_mime_parse_message (struct rspamd_task *task,
 		GError **err)
 {
 	struct rspamd_content_type *ct, *sel = NULL;
-	struct rspamd_mime_header *hdr;
-	GPtrArray *hdrs = NULL;
+	struct rspamd_mime_header *hdr = NULL, *cur;
 	const gchar *pbegin, *p;
 	gsize plen, len;
 	struct rspamd_mime_part *npart;
@@ -1123,34 +1374,6 @@ rspamd_mime_parse_message (struct rspamd_task *task,
 		/* Top level message */
 		p = task->msg.begin;
 		len = task->msg.len;
-		/* Skip any space characters to avoid some bad messages to be unparsed */
-		while (len > 0 && g_ascii_isspace (*p)) {
-			p ++;
-			len --;
-		}
-		/*
-		 * Exim somehow uses mailbox format for messages being scanned:
-		 * From x@x.com Fri May 13 19:08:48 2016
-		 *
-		 * Need to check that for all inputs due to proxy
-		 */
-		if (len > sizeof ("From ") - 1) {
-			if (memcmp (p, "From ", sizeof ("From ") - 1) == 0) {
-				/* Skip to CRLF */
-				msg_info_task ("mailbox input detected, enable workaround");
-				p += sizeof ("From ") - 1;
-				len -= sizeof ("From ") - 1;
-
-				while (len > 0 && *p != '\n') {
-					p ++;
-					len --;
-				}
-				while (len > 0 && g_ascii_isspace (*p)) {
-					p ++;
-					len --;
-				}
-			}
-		}
 
 		str.str = (gchar *)p;
 		str.len = len;
@@ -1159,42 +1382,59 @@ rspamd_mime_parse_message (struct rspamd_task *task,
 
 		if (hdr_pos > 0 && hdr_pos < str.len) {
 
-			task->raw_headers_content.begin = str.str;
-			task->raw_headers_content.len = hdr_pos;
-			task->raw_headers_content.body_start = str.str + body_pos;
+			MESSAGE_FIELD (task, raw_headers_content).begin = str.str;
+			MESSAGE_FIELD (task, raw_headers_content).len = hdr_pos;
+			MESSAGE_FIELD (task, raw_headers_content).body_start = str.str + body_pos;
 
-			if (task->raw_headers_content.len > 0) {
-				rspamd_mime_headers_process (task, task->raw_headers,
-						task->headers_order,
-						task->raw_headers_content.begin,
-						task->raw_headers_content.len,
+			if (MESSAGE_FIELD (task, raw_headers_content).len > 0) {
+				rspamd_mime_headers_process (task,
+						MESSAGE_FIELD (task, raw_headers),
+						&MESSAGE_FIELD (task, headers_order),
+						MESSAGE_FIELD (task, raw_headers_content).begin,
+						MESSAGE_FIELD (task, raw_headers_content).len,
 						TRUE);
+				npart->raw_headers = rspamd_message_headers_ref (
+						MESSAGE_FIELD (task, raw_headers));
+
+				/* Preserve the natural order */
+				if (MESSAGE_FIELD (task, headers_order)) {
+					LL_REVERSE2 (MESSAGE_FIELD (task, headers_order), ord_next);
+				}
 			}
 
-			hdrs = rspamd_message_get_header_from_hash (task->raw_headers,
-					task->task_pool,
-					"Content-Type", FALSE);
+			hdr = rspamd_message_get_header_from_hash (
+					MESSAGE_FIELD (task, raw_headers),
+					"Content-Type");
 		}
 		else {
 			/* First apply heuristic, maybe we have just headers */
 			hdr_pos = rspamd_mime_parser_headers_heuristic (&str, &body_pos);
 
 			if (hdr_pos > 0 && hdr_pos <= str.len) {
-				task->raw_headers_content.begin = str.str;
-				task->raw_headers_content.len = hdr_pos;
-				task->raw_headers_content.body_start = str.str + body_pos;
+				MESSAGE_FIELD (task, raw_headers_content).begin = str.str;
+				MESSAGE_FIELD (task, raw_headers_content).len = hdr_pos;
+				MESSAGE_FIELD (task, raw_headers_content).body_start = str.str +
+						body_pos;
 
-				if (task->raw_headers_content.len > 0) {
-					rspamd_mime_headers_process (task, task->raw_headers,
-							task->headers_order,
-							task->raw_headers_content.begin,
-							task->raw_headers_content.len,
+				if (MESSAGE_FIELD (task, raw_headers_content).len > 0) {
+					rspamd_mime_headers_process (task,
+							MESSAGE_FIELD (task, raw_headers),
+							&MESSAGE_FIELD (task, headers_order),
+							MESSAGE_FIELD (task, raw_headers_content).begin,
+							MESSAGE_FIELD (task, raw_headers_content).len,
 							TRUE);
+					npart->raw_headers = rspamd_message_headers_ref (
+							MESSAGE_FIELD (task, raw_headers));
+
+					/* Preserve the natural order */
+					if (MESSAGE_FIELD (task, headers_order)) {
+						LL_REVERSE2 (MESSAGE_FIELD (task, headers_order), ord_next);
+					}
 				}
 
-				hdrs = rspamd_message_get_header_from_hash (task->raw_headers,
-						task->task_pool,
-						"Content-Type", FALSE);
+				hdr = rspamd_message_get_header_from_hash (
+						MESSAGE_FIELD (task, raw_headers),
+						"Content-Type");
 				task->flags |= RSPAMD_TASK_FLAG_BROKEN_HEADERS;
 			}
 			else {
@@ -1204,7 +1444,6 @@ rspamd_mime_parse_message (struct rspamd_task *task,
 
 		pbegin = st->start + body_pos;
 		plen = st->end - pbegin;
-		npart->raw_headers = g_hash_table_ref (task->raw_headers);
 		npart->headers_order = NULL;
 	}
 	else {
@@ -1227,9 +1466,8 @@ rspamd_mime_parse_message (struct rspamd_task *task,
 		str.len = part->parsed_data.len;
 
 		hdr_pos = rspamd_string_find_eoh (&str, &body_pos);
-		npart->raw_headers =  g_hash_table_new_full (rspamd_strcase_hash,
-				rspamd_strcase_equal, NULL, rspamd_ptr_array_free_hard);
-		npart->headers_order = g_queue_new ();
+		npart->raw_headers =  rspamd_message_headers_new ();
+		npart->headers_order = NULL;
 
 		if (hdr_pos > 0 && hdr_pos < str.len) {
 			npart->raw_headers_str = str.str;
@@ -1237,16 +1475,21 @@ rspamd_mime_parse_message (struct rspamd_task *task,
 			npart->raw_data.begin = str.str + body_pos;
 
 			if (npart->raw_headers_len > 0) {
-				rspamd_mime_headers_process (task, npart->raw_headers,
-						npart->headers_order,
+				rspamd_mime_headers_process (task,
+						npart->raw_headers,
+						&npart->headers_order,
 						npart->raw_headers_str,
 						npart->raw_headers_len,
 						FALSE);
+
+				/* Preserve the natural order */
+				if (npart->headers_order) {
+					LL_REVERSE2 (npart->headers_order, ord_next);
+				}
 			}
 
-			hdrs = rspamd_message_get_header_from_hash (npart->raw_headers,
-					task->task_pool,
-					"Content-Type", FALSE);
+			hdr = rspamd_message_get_header_from_hash (npart->raw_headers,
+					"Content-Type");
 		}
 		else {
 			body_pos = 0;
@@ -1260,13 +1503,12 @@ rspamd_mime_parse_message (struct rspamd_task *task,
 	npart->raw_data.len = plen;
 	npart->parent_part = part;
 
-	if (hdrs == NULL) {
+	if (hdr == NULL) {
 		sel = NULL;
 	}
 	else {
-		for (i = 0; i < hdrs->len; i ++) {
-			hdr = g_ptr_array_index (hdrs, i);
-			ct = rspamd_content_type_parse (hdr->decoded, strlen (hdr->decoded),
+		DL_FOREACH (hdr, cur) {
+			ct = rspamd_content_type_parse (cur->decoded, strlen (cur->decoded),
 					task->task_pool);
 
 			/* Here we prefer multipart content-type or any content-type */
@@ -1301,6 +1543,7 @@ rspamd_mime_parse_message (struct rspamd_task *task,
 	if (sel->flags & RSPAMD_CONTENT_TYPE_MULTIPART) {
 		g_ptr_array_add (nst->stack, npart);
 		nst->nesting ++;
+		npart->part_type = RSPAMD_MIME_PART_MULTIPART;
 		npart->specific.mp = rspamd_mempool_alloc0 (task->task_pool,
 				sizeof (struct rspamd_mime_multipart));
 		memcpy (&npart->specific.mp->boundary, &sel->orig_boundary,
@@ -1308,15 +1551,21 @@ rspamd_mime_parse_message (struct rspamd_task *task,
 		ret = rspamd_mime_parse_multipart_part (task, npart, nst, err);
 	}
 	else if (sel->flags & RSPAMD_CONTENT_TYPE_MESSAGE) {
-		g_ptr_array_add (nst->stack, npart);
-		nst->nesting ++;
-		ret = rspamd_mime_parse_message (task, npart, nst, err);
+		if ((ret = rspamd_mime_parse_normal_part (task, npart, nst, sel, err))
+			== RSPAMD_MIME_PARSE_OK) {
+			npart->part_type = RSPAMD_MIME_PART_MESSAGE;
+			ret = rspamd_mime_parse_message (task, npart, nst, err);
+		}
 	}
 	else {
-		ret = rspamd_mime_parse_normal_part (task, npart, nst, err);
+		ret = rspamd_mime_parse_normal_part (task, npart, nst, sel, err);
 	}
 
-	if (part) {
+	if (ret != RSPAMD_MIME_PARSE_OK) {
+		return ret;
+	}
+
+	if (part && st->stack->len > 0) {
 		/* Remove message part from the parent stack */
 		g_ptr_array_remove_index_fast (st->stack, st->stack->len - 1);
 		st->nesting --;
@@ -1363,7 +1612,7 @@ rspamd_mime_parse_message (struct rspamd_task *task,
 				start = nst->start + start_boundary->start;
 
 				if (end > start &&
-					(ret = rspamd_mime_process_multipart_node (task, st,
+					(ret = rspamd_mime_process_multipart_node (task, nst,
 						NULL, start, end, FALSE, err)) != RSPAMD_MIME_PARSE_OK) {
 
 					if (nst != st) {
@@ -1408,7 +1657,7 @@ rspamd_mime_parse_task (struct rspamd_task *task, GError **err)
 
 	st = g_malloc0 (sizeof (*st));
 	st->stack = g_ptr_array_sized_new (4);
-	st->pos = task->raw_headers_content.body_start;
+	st->pos = MESSAGE_FIELD (task, raw_headers_content).body_start;
 	st->end = task->msg.begin + task->msg.len;
 	st->boundaries = g_array_sized_new (FALSE, FALSE,
 			sizeof (struct rspamd_mime_boundary), 8);
