@@ -40,7 +40,7 @@ struct rspamd_symbol_result {
 	struct rspamd_symbol_option *opts_head;        /**< head of linked list of options			*/
 	const gchar *name;
 	struct rspamd_symbol *sym;                     /**< symbol configuration					*/
-	gssize opts_len; /**< total size of all options (negative if truncated option is added) */
+	gssize opts_len;                               /**< total size of all options (negative if truncated option is added) */
 	guint nshots;
 	enum rspamd_symbol_result_flags flags;
 };
@@ -53,6 +53,7 @@ struct rspamd_symbol_result {
 
 #define RSPAMD_PASSTHROUGH_LEAST (1u << 0u)
 #define RSPAMD_PASSTHROUGH_NO_SMTP_MESSAGE (1u << 1u)
+#define RSPAMD_PASSTHROUGH_PROCESS_ALL (1u << 2u)
 
 struct rspamd_passthrough_result {
 	struct rspamd_action *action;
@@ -77,14 +78,20 @@ struct rspamd_scan_result {
 	double score;                                    /**< total score							*/
 	double grow_factor;                                /**< current grow factor					*/
 	struct rspamd_passthrough_result *passthrough_result;
-	guint npositive;
-	guint nnegative;
 	double positive_score;
 	double negative_score;
 	struct kh_rspamd_symbols_hash_s *symbols;            /**< symbols of metric						*/
 	struct kh_rspamd_symbols_group_hash_s *sym_groups; /**< groups of symbols						*/
 	struct rspamd_action_result *actions_limits;
+	const gchar *name;                                 /**< for named results, NULL is the default result */
+	struct rspamd_task *task;                          /**< back reference */
+	gint symbol_cbref;                                 /**< lua function that defines if a symbol can be inserted, -1 if unused */
 	guint nactions;
+	guint npositive;
+	guint nnegative;
+	guint nresults; /**< all results: positive, negative, passthrough etc */
+	guint nresults_postfilters; /**< how many results are there before postfilters stage */
+	struct rspamd_scan_result *prev, *next;           /**< double linked list of results */
 };
 
 /**
@@ -92,7 +99,17 @@ struct rspamd_scan_result {
  * @param task task object
  * @return metric result or NULL if metric `name` has not been found
  */
-struct rspamd_scan_result *rspamd_create_metric_result (struct rspamd_task *task);
+struct rspamd_scan_result *rspamd_create_metric_result (struct rspamd_task *task,
+		const gchar *name, gint lua_sym_cbref);
+
+/**
+ * Find result with a specific name (NULL means the default result)
+ * @param task
+ * @param name
+ * @return
+ */
+struct rspamd_scan_result *rspamd_find_metric_result (struct rspamd_task *task,
+													  const gchar *name);
 
 /**
  * Adds a new passthrough result to a task
@@ -104,12 +121,10 @@ struct rspamd_scan_result *rspamd_create_metric_result (struct rspamd_task *task
  * @param module
  */
 void rspamd_add_passthrough_result (struct rspamd_task *task,
-									struct rspamd_action *action,
-									guint priority,
-									double target_score,
-									const gchar *message,
-									const gchar *module,
-									guint flags);
+									struct rspamd_action *action, guint priority,
+									double target_score, const gchar *message,
+									const gchar *module, guint flags,
+									struct rspamd_scan_result *scan_result);
 
 enum rspamd_symbol_insert_flags {
 	RSPAMD_SYMBOL_INSERT_DEFAULT = 0,
@@ -129,14 +144,25 @@ struct rspamd_symbol_result *rspamd_task_insert_result_full (struct rspamd_task
 															 const gchar *symbol,
 															 double weight,
 															 const gchar *opts,
-															 enum rspamd_symbol_insert_flags flags);
-
-#define rspamd_task_insert_result_single(task, symbol, flag, opts) \
-    rspamd_task_insert_result_full (task, symbol, flag, opts, RSPAMD_SYMBOL_INSERT_SINGLE)
-#define rspamd_task_insert_result(task, symbol, flag, opts) \
-    rspamd_task_insert_result_full (task, symbol, flag, opts, RSPAMD_SYMBOL_INSERT_DEFAULT)
+															 enum rspamd_symbol_insert_flags flags,
+															 struct rspamd_scan_result *result);
 
+#define rspamd_task_insert_result_single(task, symbol, weight, opts) \
+    rspamd_task_insert_result_full ((task), (symbol), (weight), (opts), RSPAMD_SYMBOL_INSERT_SINGLE, NULL)
+#define rspamd_task_insert_result(task, symbol, weight, opts) \
+    rspamd_task_insert_result_full ((task), (symbol), (weight), (opts), RSPAMD_SYMBOL_INSERT_DEFAULT, NULL)
 
+/**
+ * Removes a symbol from a specific symbol result
+ * @param task
+ * @param symbol
+ * @param result
+ * @return
+ */
+struct rspamd_symbol_result* rspamd_task_remove_symbol_result (
+		struct rspamd_task *task,
+		const gchar *symbol,
+		struct rspamd_scan_result *result);
 /**
  * Adds new option to symbol
  * @param task
@@ -154,8 +180,9 @@ gboolean rspamd_task_add_result_option (struct rspamd_task *task,
  * @param sym
  * @return
  */
-struct rspamd_symbol_result *rspamd_task_find_symbol_result (
-		struct rspamd_task *task, const char *sym);
+struct rspamd_symbol_result *
+rspamd_task_find_symbol_result (struct rspamd_task *task, const char *sym,
+		struct rspamd_scan_result *result);
 
 /**
  * Compatibility function to iterate on symbols hash
@@ -164,6 +191,7 @@ struct rspamd_symbol_result *rspamd_task_find_symbol_result (
  * @param ud
  */
 void rspamd_task_symbol_result_foreach (struct rspamd_task *task,
+										struct rspamd_scan_result *result,
 										GHFunc func,
 										gpointer ud);
 
@@ -185,7 +213,8 @@ double rspamd_factor_consolidation_func (struct rspamd_task *task,
  * @return
  */
 struct rspamd_action *rspamd_check_action_metric (struct rspamd_task *task,
-												  struct rspamd_passthrough_result **ppr);
+												  struct rspamd_passthrough_result **ppr,
+												  struct rspamd_scan_result *scan_result);
 
 #ifdef  __cplusplus
 }
