@@ -28,6 +28,10 @@ cfg_rcl_error_quark (void)
 	return g_quark_from_static_string ("cfg-rcl-error-quark");
 }
 
+#ifdef  __cplusplus
+extern "C" {
+#endif
+
 struct rspamd_rcl_section;
 struct rspamd_config;
 struct rspamd_rcl_default_handler_data;
@@ -409,6 +413,7 @@ ucl_object_t *rspamd_rcl_add_doc_by_path (struct rspamd_config *cfg,
 		const char *default_value,
 		gboolean required);
 
+
 /**
  * Parses example and adds documentation according to the example:
  *
@@ -491,4 +496,9 @@ gboolean rspamd_config_read (struct rspamd_config *cfg,
 							 GHashTable *vars,
 							 gboolean skip_jinja,
 							 gchar **lua_env);
+
+#ifdef  __cplusplus
+}
+#endif
+
 #endif /* CFG_RCL_H_ */
