@@ -21,7 +21,6 @@
 #include "utlist.h"
 #include "unix-std.h"
 #include "mempool_vars_internal.h"
-#include "libcryptobox/ed25519/ed25519.h"
 
 #include <openssl/evp.h>
 #include <openssl/rsa.h>
@@ -105,15 +104,18 @@ struct rspamd_dkim_common_ctx {
 	rspamd_mempool_t *pool;
 	guint64 sig_hash;
 	gsize len;
-	gint header_canon_type;
-	gint body_canon_type;
-	gboolean is_sign;
 	GPtrArray *hlist;
 	GHashTable *htable; /* header -> count mapping */
 	EVP_MD_CTX *headers_hash;
 	EVP_MD_CTX *body_hash;
 	enum rspamd_dkim_type type;
 	guint idx;
+	gint header_canon_type;
+	gint body_canon_type;
+	guint body_canonicalised;
+	guint headers_canonicalised;
+	gboolean is_sign;
+
 };
 
 enum rspamd_arc_seal_cv {
@@ -128,6 +130,7 @@ enum rspamd_arc_seal_cv {
 struct rspamd_dkim_context_s {
 	struct rspamd_dkim_common_ctx common;
 	rspamd_mempool_t *pool;
+	struct rspamd_dns_resolver *resolver;
 	gsize blen;
 	gsize bhlen;
 	gint sig_alg;
@@ -144,20 +147,23 @@ struct rspamd_dkim_context_s {
 	const gchar *dkim_header;
 };
 
+#define RSPAMD_DKIM_KEY_ID_LEN 16
+
 struct rspamd_dkim_key_s {
 	guint8 *keydata;
 	gsize keylen;
 	gsize decoded_len;
-	guint ttl;
+	gchar key_id[RSPAMD_DKIM_KEY_ID_LEN];
 	union {
 		RSA *key_rsa;
 		EC_KEY *key_ecdsa;
 		guchar *key_eddsa;
 	} key;
-	enum rspamd_dkim_key_type type;
 	BIO *key_bio;
 	EVP_PKEY *key_evp;
 	time_t mtime;
+	guint ttl;
+	enum rspamd_dkim_key_type type;
 	ref_entry_t ref;
 };
 
@@ -168,7 +174,7 @@ struct rspamd_dkim_sign_context_s {
 
 struct rspamd_dkim_header {
 	const gchar *name;
-	guint count;
+	gint count;
 };
 
 /* Parser of dkim params */
@@ -328,8 +334,25 @@ rspamd_dkim_parse_domain (rspamd_dkim_context_t * ctx,
 	gsize len,
 	GError **err)
 {
-	ctx->domain = rspamd_mempool_alloc (ctx->pool, len + 1);
-	rspamd_strlcpy (ctx->domain, param, len + 1);
+	if (!rspamd_str_has_8bit (param, len)) {
+		ctx->domain = rspamd_mempool_alloc (ctx->pool, len + 1);
+		rspamd_strlcpy (ctx->domain, param, len + 1);
+	}
+	else {
+		ctx->domain = rspamd_dns_resolver_idna_convert_utf8 (ctx->resolver,
+				ctx->pool, param, len, NULL);
+
+		if (!ctx->domain) {
+			g_set_error (err,
+					DKIM_ERROR,
+					DKIM_SIGERROR_INVALID_H,
+					"invalid dkim domain tag %*.s: idna failed",
+					(int)len, param);
+
+			return FALSE;
+		}
+	}
+
 	return TRUE;
 }
 
@@ -411,8 +434,26 @@ rspamd_dkim_parse_selector (rspamd_dkim_context_t * ctx,
 	gsize len,
 	GError **err)
 {
-	ctx->selector = rspamd_mempool_alloc (ctx->pool, len + 1);
-	rspamd_strlcpy (ctx->selector, param, len + 1);
+
+	if (!rspamd_str_has_8bit (param, len)) {
+		ctx->selector = rspamd_mempool_alloc (ctx->pool, len + 1);
+		rspamd_strlcpy (ctx->selector, param, len + 1);
+	}
+	else {
+		ctx->selector = rspamd_dns_resolver_idna_convert_utf8 (ctx->resolver,
+				ctx->pool, param, len, NULL);
+
+		if (!ctx->selector) {
+			g_set_error (err,
+					DKIM_ERROR,
+					DKIM_SIGERROR_INVALID_H,
+					"invalid dkim selector tag %*.s: idna failed",
+					(int)len, param);
+
+			return FALSE;
+		}
+	}
+
 	return TRUE;
 }
 
@@ -717,7 +758,7 @@ rspamd_dkim_add_arc_seal_headers (rspamd_mempool_t *pool,
 		struct rspamd_dkim_common_ctx *ctx)
 {
 	struct rspamd_dkim_header *hdr;
-	guint count = ctx->idx, i;
+	gint count = ctx->idx, i;
 
 	ctx->hlist = g_ptr_array_sized_new (count * 3 - 1);
 
@@ -725,23 +766,27 @@ rspamd_dkim_add_arc_seal_headers (rspamd_mempool_t *pool,
 		/* Authentication results */
 		hdr = rspamd_mempool_alloc (pool, sizeof (*hdr));
 		hdr->name = RSPAMD_DKIM_ARC_AUTHHEADER;
-		hdr->count = i;
+		hdr->count = -(i + 1);
 		g_ptr_array_add (ctx->hlist, hdr);
 
 		/* Arc signature */
 		hdr = rspamd_mempool_alloc (pool, sizeof (*hdr));
 		hdr->name = RSPAMD_DKIM_ARC_SIGNHEADER;
-		hdr->count = i;
+		hdr->count = -(i + 1);
 		g_ptr_array_add (ctx->hlist, hdr);
 
 		/* Arc seal (except last one) */
 		if (i != count - 1) {
 			hdr = rspamd_mempool_alloc (pool, sizeof (*hdr));
 			hdr->name = RSPAMD_DKIM_ARC_SEALHEADER;
-			hdr->count = i;
+			hdr->count = -(i + 1);
 			g_ptr_array_add (ctx->hlist, hdr);
 		}
 	}
+
+	rspamd_mempool_add_destructor (ctx->pool,
+			(rspamd_mempool_destruct_t)rspamd_dkim_hlist_free,
+			ctx->hlist);
 }
 
 /**
@@ -753,10 +798,11 @@ rspamd_dkim_add_arc_seal_headers (rspamd_mempool_t *pool,
  */
 rspamd_dkim_context_t *
 rspamd_create_dkim_context (const gchar *sig,
-		rspamd_mempool_t *pool,
-		guint time_jitter,
-		enum rspamd_dkim_type type,
-		GError **err)
+							rspamd_mempool_t *pool,
+							struct rspamd_dns_resolver *resolver,
+							guint time_jitter,
+							enum rspamd_dkim_type type,
+							GError **err)
 {
 	const gchar *p, *c, *tag = NULL, *end;
 	gsize taglen;
@@ -783,6 +829,7 @@ rspamd_create_dkim_context (const gchar *sig,
 
 	ctx = rspamd_mempool_alloc0 (pool, sizeof (rspamd_dkim_context_t));
 	ctx->pool = pool;
+	ctx->resolver = resolver;
 
 	if (type == RSPAMD_DKIM_ARC_SEAL) {
 		ctx->common.header_canon_type = DKIM_CANON_RELAXED;
@@ -806,6 +853,8 @@ rspamd_create_dkim_context (const gchar *sig,
 	ctx->common.sig_hash = rspamd_cryptobox_fast_hash (sig, end - sig,
 			rspamd_hash_seed ());
 
+	msg_debug_dkim ("create dkim context sig = %L", ctx->common.sig_hash);
+
 	while (p <= end) {
 		switch (state) {
 		case DKIM_STATE_TAG:
@@ -1033,6 +1082,17 @@ rspamd_create_dkim_context (const gchar *sig,
 					if (!parser_funcs[param](ctx, c, tlen, err)) {
 						state = DKIM_STATE_ERROR;
 					}
+					if (state == DKIM_STATE_ERROR) {
+						/*
+						 * We need to return from here as state machine won't
+						 * do any more steps after p == end
+						 */
+						if (err) {
+							msg_info_dkim ("dkim parse failed: %e", *err);
+						}
+
+						return NULL;
+					}
 					/* Finish processing */
 					p++;
 				}
@@ -1276,8 +1336,32 @@ rspamd_dkim_make_key (const gchar *keydata,
 	key->keylen = keylen;
 	key->type = type;
 
-	rspamd_cryptobox_base64_decode (keydata, keylen, key->keydata,
-			&key->decoded_len);
+	if (!rspamd_cryptobox_base64_decode (keydata, keylen, key->keydata,
+			&key->decoded_len)) {
+		REF_RELEASE (key);
+		g_set_error (err,
+				DKIM_ERROR,
+				DKIM_SIGERROR_KEYFAIL,
+				"DKIM key is not a valid base64 string");
+
+		return NULL;
+	}
+
+	/* Calculate ID -> md5 */
+	EVP_MD_CTX *mdctx = EVP_MD_CTX_create ();
+
+#ifdef EVP_MD_CTX_FLAG_NON_FIPS_ALLOW
+	EVP_MD_CTX_set_flags (mdctx, EVP_MD_CTX_FLAG_NON_FIPS_ALLOW);
+#endif
+
+	if (EVP_DigestInit_ex (mdctx, EVP_md5 (), NULL) == 1) {
+		guint dlen = sizeof (key->key_id);
+
+		EVP_DigestUpdate (mdctx, key->keydata, key->decoded_len);
+		EVP_DigestFinal_ex (mdctx, key->key_id, &dlen);
+	}
+
+	EVP_MD_CTX_destroy (mdctx);
 
 	if (key->type == RSPAMD_DKIM_KEY_EDDSA) {
 		key->key.key_eddsa = key->keydata;
@@ -1349,6 +1433,16 @@ rspamd_dkim_make_key (const gchar *keydata,
 	return key;
 }
 
+const guchar *
+rspamd_dkim_key_id (rspamd_dkim_key_t *key)
+{
+	if (key) {
+		return key->key_id;
+	}
+
+	return NULL;
+}
+
 /**
  * Free DKIM key
  * @param key
@@ -1499,16 +1593,16 @@ rspamd_dkim_parse_key (const gchar *txt, gsize *keylen, GError **err)
 	}
 
 	if (alglen == 8 && rspamd_lc_cmp (alg, "ecdsa256", alglen) == 0) {
-		return rspamd_dkim_make_key (c, klen,
+		return rspamd_dkim_make_key (key, klen,
 				RSPAMD_DKIM_KEY_ECDSA, err);
 	}
 	else if (alglen == 7 && rspamd_lc_cmp (alg, "ed25519", alglen) == 0) {
-		return rspamd_dkim_make_key (c, klen,
+		return rspamd_dkim_make_key (key, klen,
 				RSPAMD_DKIM_KEY_EDDSA, err);
 	}
 	else {
 		/* We assume RSA default in all cases */
-		return rspamd_dkim_make_key (c, klen,
+		return rspamd_dkim_make_key (key, klen,
 				RSPAMD_DKIM_KEY_RSA, err);
 	}
 
@@ -1600,21 +1694,24 @@ rspamd_get_dkim_key (rspamd_dkim_context_t *ctx,
 static gboolean
 rspamd_dkim_relaxed_body_step (struct rspamd_dkim_common_ctx *ctx, EVP_MD_CTX *ck,
 		const gchar **start, guint size,
-		guint *remain)
+		gssize *remain)
 {
 	const gchar *h;
-	static gchar buf[BUFSIZ];
 	gchar *t;
-	guint len, inlen, added = 0;
+	guint len, inlen;
+	gssize octets_remain;
 	gboolean got_sp;
+	gchar buf[1024];
 
 	len = size;
 	inlen = sizeof (buf) - 1;
 	h = *start;
 	t = buf;
 	got_sp = FALSE;
+	octets_remain = *remain;
+
+	while (len > 0 && inlen > 0 && (octets_remain != 0)) {
 
-	while (len && inlen) {
 		if (*h == '\r' || *h == '\n') {
 			if (got_sp) {
 				/* Ignore spaces at the end of line */
@@ -1622,14 +1719,22 @@ rspamd_dkim_relaxed_body_step (struct rspamd_dkim_common_ctx *ctx, EVP_MD_CTX *c
 			}
 			*t++ = '\r';
 			*t++ = '\n';
+
 			if (len > 1 && (*h == '\r' && h[1] == '\n')) {
 				h += 2;
 				len -= 2;
+				octets_remain -= 2;
 			}
 			else {
 				h ++;
 				len --;
-				added ++;
+				if (octets_remain >= 2) {
+					octets_remain -= 2; /* Input has just \n or \r so we actually add more octets */
+				}
+				else {
+					octets_remain --;
+					break;
+				}
 			}
 			break;
 		}
@@ -1645,6 +1750,7 @@ rspamd_dkim_relaxed_body_step (struct rspamd_dkim_common_ctx *ctx, EVP_MD_CTX *c
 				h++;
 				inlen--;
 				len--;
+				octets_remain --;
 				got_sp = TRUE;
 				continue;
 			}
@@ -1652,77 +1758,97 @@ rspamd_dkim_relaxed_body_step (struct rspamd_dkim_common_ctx *ctx, EVP_MD_CTX *c
 		else {
 			got_sp = FALSE;
 		}
+
 		*t++ = *h++;
 		inlen--;
 		len--;
+		octets_remain --;
 	}
 
 	*start = h;
 
-	if (*remain > 0) {
-		size_t cklen = MIN(t - buf, *remain + added);
+	if (t - buf > 0) {
+		gsize cklen = t - buf;
+
 		EVP_DigestUpdate (ck, buf, cklen);
-		*remain = *remain - (cklen - added);
-#if 0
-		msg_debug_dkim ("update signature with buffer (%ud size, %ud remain, %ud added): %*s",
-				cklen, *remain, added, cklen, buf);
-#else
-		msg_debug_dkim ("update signature with body buffer "
-				"(%ud size, %ud remain, %ud added)",
-						cklen, *remain, added);
-#endif
+		ctx->body_canonicalised += cklen;
+		msg_debug_dkim ("relaxed update signature with body buffer "
+				"(%z size, %z -> %z remain)",
+						cklen, *remain, octets_remain);
+		*remain = octets_remain;
+
 	}
 
-	return (len != 0);
+	return ((len != 0) && (octets_remain != 0));
 }
 
 static gboolean
 rspamd_dkim_simple_body_step (struct rspamd_dkim_common_ctx *ctx,
 		EVP_MD_CTX *ck, const gchar **start, guint size,
-		guint *remain)
+		gssize *remain)
 {
 	const gchar *h;
-	static gchar buf[BUFSIZ];
 	gchar *t;
-	guint len, inlen, added = 0;
+	guint len, inlen;
+	gssize octets_remain;
+	gchar buf[1024];
 
 	len = size;
 	inlen = sizeof (buf) - 1;
 	h = *start;
 	t = &buf[0];
+	octets_remain = *remain;
 
-	while (len && inlen) {
+	while (len > 0 && inlen > 0 && (octets_remain != 0)) {
 		if (*h == '\r' || *h == '\n') {
 			*t++ = '\r';
 			*t++ = '\n';
+
 			if (len > 1 && (*h == '\r' && h[1] == '\n')) {
 				h += 2;
 				len -= 2;
+
+				if (octets_remain >= 2) {
+					octets_remain -= 2; /* Input has just \n or \r so we actually add more octets */
+				}
+				else {
+					octets_remain --;
+				}
 			}
 			else {
 				h ++;
 				len --;
-				added ++;
+
+				if (octets_remain >= 2) {
+					octets_remain -= 2; /* Input has just \n or \r so we actually add more octets */
+				}
+				else {
+					octets_remain --;
+				}
 			}
 			break;
 		}
+
 		*t++ = *h++;
+		octets_remain --;
 		inlen--;
 		len--;
 	}
 
 	*start = h;
 
-	if (*remain > 0) {
-		size_t cklen = MIN(t - buf, *remain + added);
+	if (t - buf > 0) {
+		gsize cklen = t - buf;
+
 		EVP_DigestUpdate (ck, buf, cklen);
-		*remain = *remain - (cklen - added);
-		msg_debug_dkim ("update signature with body buffer "
-				"(%ud size, %ud remain, %ud added)",
-				cklen, *remain, added);
+		ctx->body_canonicalised += cklen;
+		msg_debug_dkim ("simple update signature with body buffer "
+						"(%z size, %z -> %z remain)",
+				cklen, *remain, octets_remain);
+		*remain = octets_remain;
 	}
 
-	return (len != 0);
+	return ((len != 0) && (octets_remain != 0));
 }
 
 static const gchar *
@@ -1901,13 +2027,15 @@ rspamd_dkim_canonize_body (struct rspamd_dkim_common_ctx *ctx,
 	gboolean sign)
 {
 	const gchar *p;
-	guint remain = ctx->len ? ctx->len : (guint)(end - start);
+	gssize remain = ctx->len ? ctx->len : -1;
+	guint total_len = end - start;
 	gboolean need_crlf = FALSE;
 
 	if (start == NULL) {
 		/* Empty body */
 		if (ctx->body_canon_type == DKIM_CANON_SIMPLE) {
 			EVP_DigestUpdate (ctx->body_hash, CRLF, sizeof (CRLF) - 1);
+			ctx->body_canonicalised += sizeof (CRLF) - 1;
 		}
 		else {
 			EVP_DigestUpdate (ctx->body_hash, "", 0);
@@ -1923,6 +2051,7 @@ rspamd_dkim_canonize_body (struct rspamd_dkim_common_ctx *ctx,
 			/* Empty body */
 			if (ctx->body_canon_type == DKIM_CANON_SIMPLE) {
 				EVP_DigestUpdate (ctx->body_hash, CRLF, sizeof (CRLF) - 1);
+				ctx->body_canonicalised += sizeof (CRLF) - 1;
 			}
 			else {
 				EVP_DigestUpdate (ctx->body_hash, "", 0);
@@ -1934,10 +2063,32 @@ rspamd_dkim_canonize_body (struct rspamd_dkim_common_ctx *ctx,
 				while (rspamd_dkim_simple_body_step (ctx, ctx->body_hash,
 						&start, end - start, &remain));
 
+				/*
+				 * If we have l= tag then we cannot add crlf...
+				 */
 				if (need_crlf) {
+					/* l is evil... */
+					if (ctx->len == 0) {
+						remain = 2;
+					}
+					else {
+						if (ctx->len <= total_len) {
+							/* We don't have enough l to add \r\n */
+							remain = 0;
+						}
+						else {
+							if (ctx->len - total_len >= 2) {
+								remain = 2;
+							}
+							else {
+								remain = ctx->len - total_len;
+							}
+						}
+					}
+
 					start = "\r\n";
 					end = start + 2;
-					remain = 2;
+
 					rspamd_dkim_simple_body_step (ctx, ctx->body_hash,
 							&start, end - start, &remain);
 				}
@@ -2017,8 +2168,9 @@ rspamd_dkim_signature_update (struct rspamd_dkim_common_ctx *ctx,
 		if (tag && p[0] == 'b' && p[1] == '=') {
 			/* Add to signature */
 			msg_debug_dkim ("initial update hash with signature part: %*s",
-				p - c + 2,
-				c);
+					(gint)(p - c + 2),
+					c);
+			ctx->headers_canonicalised += p - c + 2;
 			rspamd_dkim_hash_update (ctx->headers_hash, c, p - c + 2);
 			skip = TRUE;
 		}
@@ -2042,7 +2194,9 @@ rspamd_dkim_signature_update (struct rspamd_dkim_common_ctx *ctx,
 	}
 
 	if (p - c + 1 > 0) {
-		msg_debug_dkim ("final update hash with signature part: %*s", p - c + 1, c);
+		msg_debug_dkim ("final update hash with signature part: %*s",
+				(gint)(p - c + 1), c);
+		ctx->headers_canonicalised += p - c + 1;
 		rspamd_dkim_hash_update (ctx->headers_hash, c, p - c + 1);
 	}
 }
@@ -2117,9 +2271,11 @@ rspamd_dkim_canonize_header_relaxed_str (const gchar *hname,
 
 static gboolean
 rspamd_dkim_canonize_header_relaxed (struct rspamd_dkim_common_ctx *ctx,
-	const gchar *header,
-	const gchar *header_name,
-	gboolean is_sign)
+									 const gchar *header,
+									 const gchar *header_name,
+									 gboolean is_sign,
+									 guint count,
+									 bool is_seal)
 {
 	static gchar st_buf[8192];
 	gchar *buf;
@@ -2143,7 +2299,8 @@ rspamd_dkim_canonize_header_relaxed (struct rspamd_dkim_common_ctx *ctx,
 	g_assert (r != -1);
 
 	if (!is_sign) {
-		msg_debug_dkim ("update signature with header: %s", buf);
+		msg_debug_dkim ("update %s with header (idx=%d): %s",
+				is_seal ? "seal" : "signature", count, buf);
 		EVP_DigestUpdate (ctx->headers_hash, buf, r);
 	}
 	else {
@@ -2162,62 +2319,106 @@ static gboolean
 rspamd_dkim_canonize_header (struct rspamd_dkim_common_ctx *ctx,
 	struct rspamd_task *task,
 	const gchar *header_name,
-	guint count,
+	gint count,
 	const gchar *dkim_header,
 	const gchar *dkim_domain)
 {
-	struct rspamd_mime_header *rh;
-	gint rh_num = 0;
-	GPtrArray *ar;
+	struct rspamd_mime_header *rh, *cur, *sel = NULL;
+	gint hdr_cnt = 0;
+	bool use_idx = false;
+
+	if (count < 0) {
+		use_idx = true;
+		count = -(count); /* use i= in header content as it is arc stuff */
+	}
 
 	if (dkim_header == NULL) {
-		ar = g_hash_table_lookup (task->raw_headers, header_name);
+		rh = rspamd_message_get_header_array (task, header_name);
 
-		if (ar) {
-			/* Check uniqueness of the header */
-			rh = g_ptr_array_index (ar, 0);
-			if ((rh->type & RSPAMD_HEADER_UNIQUE) && ar->len > 1) {
-				guint64 random_cookie = ottery_rand_uint64 ();
+		if (rh) {
+			/* Check uniqueness of the header but we count from the bottom to top */
+			if (!use_idx) {
+				for (cur = rh->prev;; cur = cur->prev) {
+					if (hdr_cnt == count) {
+						sel = cur;
+					}
 
-				msg_warn_dkim ("header %s is intended to be unique by"
-						" email standards, but we have %d headers of this"
-						" type, artificially break DKIM check", header_name,
-						ar->len);
-				rspamd_dkim_hash_update (ctx->headers_hash,
-						(const gchar *)&random_cookie,
-						sizeof (random_cookie));
+					hdr_cnt++;
 
-				return FALSE;
-			}
+					if (cur == rh) {
+						/* Cycle */
+						break;
+					}
+				}
 
-			if (ar->len > count) {
-				/* Set skip count */
-				rh_num = ar->len - count - 1;
+				if ((rh->flags & RSPAMD_HEADER_UNIQUE) && hdr_cnt > 1) {
+					guint64 random_cookie = ottery_rand_uint64 ();
+
+					msg_warn_dkim ("header %s is intended to be unique by"
+								   " email standards, but we have %d headers of this"
+								   " type, artificially break DKIM check", header_name,
+							hdr_cnt);
+					rspamd_dkim_hash_update (ctx->headers_hash,
+							(const gchar *)&random_cookie,
+							sizeof (random_cookie));
+					ctx->headers_canonicalised += sizeof (random_cookie);
+
+					return FALSE;
+				}
+
+				if (hdr_cnt <= count) {
+					/*
+					 * If DKIM has less headers requested than there are in a
+					 * message, then it's fine, it allows adding extra headers
+					 */
+					return TRUE;
+				}
 			}
 			else {
-				/*
-				 * If DKIM has less headers requested than there are in a
-				 * message, then it's fine, it allows adding extra headers
-				 */
-				return TRUE;
+				gchar idx_buf[16];
+				gint id_len;
+
+				id_len = rspamd_snprintf (idx_buf, sizeof (idx_buf), "i=%d;",
+						count);
+
+				for (cur = rh->prev; ; cur = cur->prev) {
+					if (cur->decoded &&
+						rspamd_substring_search (cur->decoded, strlen (cur->decoded),
+								idx_buf, id_len) != -1) {
+						sel = cur;
+						break;
+					}
+
+					if (cur == rh) {
+						/* Cycle */
+						break;
+					}
+				}
+
+				if (sel == NULL) {
+					return FALSE;
+				}
 			}
 
-			rh = g_ptr_array_index (ar, rh_num);
+			/* Selected header must be non-null if previous condition is false */
+			g_assert (sel != NULL);
 
 			if (ctx->header_canon_type == DKIM_CANON_SIMPLE) {
-				rspamd_dkim_hash_update (ctx->headers_hash, rh->raw_value,
-						rh->raw_len);
-				msg_debug_dkim ("update signature with header: %*s",
-						(gint)rh->raw_len, rh->raw_value);
+				rspamd_dkim_hash_update (ctx->headers_hash, sel->raw_value,
+						sel->raw_len);
+				ctx->headers_canonicalised += sel->raw_len;
+				msg_debug_dkim ("update %s with header (idx=%d): %*s",
+						(use_idx ? "seal" : "signature"),
+						count, (gint)sel->raw_len, sel->raw_value);
 			}
 			else {
-				if (ctx->is_sign && (rh->type & RSPAMD_HEADER_FROM)) {
+				if (ctx->is_sign && (sel->flags & RSPAMD_HEADER_FROM)) {
 					/* Special handling of the From handling when rewrite is done */
 					gboolean has_rewrite = FALSE;
 					guint i;
 					struct rspamd_email_address *addr;
 
-					PTR_ARRAY_FOREACH (task->from_mime, i, addr) {
+					PTR_ARRAY_FOREACH (MESSAGE_FIELD (task, from_mime), i, addr) {
 						if ((addr->flags & RSPAMD_EMAIL_ADDR_ORIGINAL)
 							&& !(addr->flags & RSPAMD_EMAIL_ADDR_ALIASED)) {
 							has_rewrite = TRUE;
@@ -2225,10 +2426,10 @@ rspamd_dkim_canonize_header (struct rspamd_dkim_common_ctx *ctx,
 					}
 
 					if (has_rewrite) {
-						PTR_ARRAY_FOREACH (task->from_mime, i, addr) {
+						PTR_ARRAY_FOREACH (MESSAGE_FIELD (task, from_mime), i, addr) {
 							if (!(addr->flags & RSPAMD_EMAIL_ADDR_ORIGINAL)) {
 								if (!rspamd_dkim_canonize_header_relaxed (ctx, addr->raw,
-										header_name, FALSE)) {
+										header_name, FALSE, i, use_idx)) {
 									return FALSE;
 								}
 
@@ -2238,8 +2439,8 @@ rspamd_dkim_canonize_header (struct rspamd_dkim_common_ctx *ctx,
 					}
 				}
 
-				if (!rspamd_dkim_canonize_header_relaxed (ctx, rh->value,
-						header_name, FALSE)) {
+				if (!rspamd_dkim_canonize_header_relaxed (ctx, sel->value,
+						header_name, FALSE, count, use_idx)) {
 					return FALSE;
 				}
 			}
@@ -2249,26 +2450,34 @@ rspamd_dkim_canonize_header (struct rspamd_dkim_common_ctx *ctx,
 		/* For signature check just use the saved dkim header */
 		if (ctx->header_canon_type == DKIM_CANON_SIMPLE) {
 			/* We need to find our own signature and use it */
-			guint i;
+			rh = rspamd_message_get_header_array (task, header_name);
 
-			ar = g_hash_table_lookup (task->raw_headers, header_name);
-
-			if (ar) {
+			if (rh) {
 				/* We need to find our own signature */
 				if (!dkim_domain) {
+					msg_err_dkim ("cannot verify dkim as we have no dkim domain!");
 					return FALSE;
 				}
 
-				PTR_ARRAY_FOREACH (ar, i, rh) {
-					guint64 th = rspamd_cryptobox_fast_hash (rh->decoded,
-							strlen (rh->decoded), rspamd_hash_seed ());
+				gboolean found = FALSE;
+
+				DL_FOREACH (rh, cur) {
+					guint64 th = rspamd_cryptobox_fast_hash (cur->decoded,
+							strlen (cur->decoded), rspamd_hash_seed ());
 
 					if (th == ctx->sig_hash) {
-						rspamd_dkim_signature_update (ctx, rh->raw_value,
-								rh->raw_len);
+						rspamd_dkim_signature_update (ctx, cur->raw_value,
+								cur->raw_len);
+						found = TRUE;
 						break;
 					}
 				}
+				if (!found) {
+					msg_err_dkim ("BUGON: cannot verify dkim as we have lost our signature"
+								  " during simple canonicalisation, expected hash=%L",
+								  ctx->sig_hash);
+					return FALSE;
+				}
 			}
 			else {
 				return FALSE;
@@ -2278,7 +2487,7 @@ rspamd_dkim_canonize_header (struct rspamd_dkim_common_ctx *ctx,
 			if (!rspamd_dkim_canonize_header_relaxed (ctx,
 					dkim_header,
 					header_name,
-					TRUE)) {
+					TRUE, 0, use_idx)) {
 				return FALSE;
 			}
 		}
@@ -2321,6 +2530,21 @@ rspamd_dkim_check_bh_cached (struct rspamd_dkim_common_ctx *ctx,
 	return res;
 }
 
+static const char *
+rspamd_dkim_type_to_string (enum rspamd_dkim_type t)
+{
+	switch (t) {
+	case RSPAMD_DKIM_NORMAL:
+		return "dkim";
+	case RSPAMD_DKIM_ARC_SIG:
+		return "arc_sig";
+	case RSPAMD_DKIM_ARC_SEAL:
+	default:
+		return "arc_seal";
+
+	}
+}
+
 /**
  * Check task for dkim context using dkim key
  * @param ctx dkim verify context
@@ -2349,7 +2573,8 @@ rspamd_dkim_check (rspamd_dkim_context_t *ctx,
 
 	/* First of all find place of body */
 	body_end = task->msg.begin + task->msg.len;
-	body_start = task->raw_headers_content.body_start;
+
+	body_start = MESSAGE_FIELD (task, raw_headers_content).body_start;
 
 	res = rspamd_mempool_alloc0 (task->task_pool, sizeof (*res));
 	res->ctx = ctx;
@@ -2417,11 +2642,12 @@ rspamd_dkim_check (rspamd_dkim_context_t *ctx,
 
 		/* Check bh field */
 		if (memcmp (ctx->bh, cached_bh->digest_normal, ctx->bhlen) != 0) {
+			msg_debug_dkim (
+					"bh value mismatch: %*xs versus %*xs, try add LF; try adding CRLF",
+					(gint)dlen, ctx->bh,
+					(gint)dlen, raw_digest);
+
 			if (cpy_ctx) {
-				msg_debug_dkim (
-						"bh value mismatch: %*xs versus %*xs, try add CRLF",
-						dlen, ctx->bh,
-						dlen, cached_bh->digest_normal);
 				/* Try add CRLF */
 #if OPENSSL_VERSION_NUMBER < 0x10100000L || defined(LIBRESSL_VERSION_NUMBER)
 				EVP_MD_CTX_cleanup (cpy_ctx);
@@ -2437,9 +2663,9 @@ rspamd_dkim_check (rspamd_dkim_context_t *ctx,
 
 				if (memcmp (ctx->bh, raw_digest, ctx->bhlen) != 0) {
 					msg_debug_dkim (
-							"bh value mismatch: %*xs versus %*xs, try add LF",
-							dlen, ctx->bh,
-							dlen, raw_digest);
+							"bh value mismatch after added CRLF: %*xs versus %*xs, try add LF",
+							(gint)dlen, ctx->bh,
+							(gint)dlen, raw_digest);
 
 					/* Try add LF */
 #if OPENSSL_VERSION_NUMBER < 0x10100000L || defined(LIBRESSL_VERSION_NUMBER)
@@ -2455,59 +2681,45 @@ rspamd_dkim_check (rspamd_dkim_context_t *ctx,
 					memcpy (cached_bh->digest_cr, raw_digest, sizeof (raw_digest));
 
 					if (memcmp (ctx->bh, raw_digest, ctx->bhlen) != 0) {
-						msg_debug_dkim ("bh value mismatch: %*xs versus %*xs",
-								dlen, ctx->bh,
-								dlen, raw_digest);
-#if OPENSSL_VERSION_NUMBER < 0x10100000L || defined(LIBRESSL_VERSION_NUMBER)
-						EVP_MD_CTX_cleanup (cpy_ctx);
-#else
-						EVP_MD_CTX_reset (cpy_ctx);
-#endif
+						msg_debug_dkim ("bh value mismatch after added LF: %*xs versus %*xs",
+								(gint)dlen, ctx->bh,
+								(gint)dlen, raw_digest);
 						res->fail_reason = "body hash did not verify";
 						res->rcode = DKIM_REJECT;
-						EVP_MD_CTX_destroy (cpy_ctx);
-
-						return res;
 					}
 				}
 			}
 			else if (cached_bh->digest_crlf) {
 				if (memcmp (ctx->bh, cached_bh->digest_crlf, ctx->bhlen) != 0) {
-					msg_debug_dkim ("bh value mismatch: %*xs versus %*xs",
-							dlen, ctx->bh,
-							dlen, cached_bh->digest_crlf);
+					msg_debug_dkim ("bh value mismatch after added CRLF: %*xs versus %*xs",
+							(gint)dlen, ctx->bh,
+							(gint)dlen, cached_bh->digest_crlf);
 
 					if (cached_bh->digest_cr) {
 						if (memcmp (ctx->bh, cached_bh->digest_cr, ctx->bhlen) != 0) {
 							msg_debug_dkim (
-									"bh value mismatch: %*xs versus %*xs",
-									dlen, ctx->bh,
-									dlen, cached_bh->digest_cr);
+									"bh value mismatch after added LF: %*xs versus %*xs",
+									(gint)dlen, ctx->bh,
+									(gint)dlen, cached_bh->digest_cr);
 
 							res->fail_reason = "body hash did not verify";
 							res->rcode = DKIM_REJECT;
-
-							return res;
 						}
 					}
 					else {
 
 						res->fail_reason = "body hash did not verify";
 						res->rcode = DKIM_REJECT;
-
-						return res;
 					}
 				}
 			}
 			else {
 				msg_debug_dkim (
 						"bh value mismatch: %*xs versus %*xs",
-						dlen, ctx->bh,
-						dlen, cached_bh->digest_normal);
+						(gint)dlen, ctx->bh,
+						(gint)dlen, cached_bh->digest_normal);
 				res->fail_reason = "body hash did not verify";
 				res->rcode = DKIM_REJECT;
-
-				return res;
 			}
 		}
 
@@ -2519,6 +2731,19 @@ rspamd_dkim_check (rspamd_dkim_context_t *ctx,
 #endif
 			EVP_MD_CTX_destroy (cpy_ctx);
 		}
+
+		if (res->rcode == DKIM_REJECT) {
+			msg_info_dkim (
+					"%s: bh value mismatch: got %*Bs, expected %*Bs; "
+					"body length %d->%d; d=%s; s=%s",
+					rspamd_dkim_type_to_string (ctx->common.type),
+					(gint)dlen, cached_bh->digest_normal,
+					(gint)dlen, ctx->bh,
+					(gint)(body_end - body_start), ctx->common.body_canonicalised,
+					ctx->domain, ctx->selector);
+
+			return res;
+		}
 	}
 
 	dlen = EVP_MD_CTX_size (ctx->common.headers_hash);
@@ -2546,31 +2771,59 @@ rspamd_dkim_check (rspamd_dkim_context_t *ctx,
 	case RSPAMD_DKIM_KEY_RSA:
 		if (RSA_verify (nid, raw_digest, dlen, ctx->b, ctx->blen,
 				key->key.key_rsa) != 1) {
-			msg_debug_dkim ("rsa verify failed");
+			msg_debug_dkim ("headers rsa verify failed");
 			res->rcode = DKIM_REJECT;
-			res->fail_reason = "rsa verify failed";
+			res->fail_reason = "headers rsa verify failed";
+
+			msg_info_dkim (
+					"%s: headers RSA verification failure; "
+					"body length %d->%d; headers length %d; d=%s; s=%s; key_md5=%*xs; orig header: %s",
+					rspamd_dkim_type_to_string (ctx->common.type),
+					(gint)(body_end - body_start), ctx->common.body_canonicalised,
+					ctx->common.headers_canonicalised,
+					ctx->domain, ctx->selector,
+					RSPAMD_DKIM_KEY_ID_LEN, rspamd_dkim_key_id (key),
+					ctx->dkim_header);
 		}
 		break;
 	case RSPAMD_DKIM_KEY_ECDSA:
 		if (ECDSA_verify (nid, raw_digest, dlen, ctx->b, ctx->blen,
 				key->key.key_ecdsa) != 1) {
-			msg_debug_dkim ("ecdsa verify failed");
+			msg_info_dkim (
+					"%s: headers ECDSA verification failure; "
+					"body length %d->%d; headers length %d; d=%s; s=%s; key_md5=%*xs; orig header: %s",
+					rspamd_dkim_type_to_string (ctx->common.type),
+					(gint)(body_end - body_start), ctx->common.body_canonicalised,
+					ctx->common.headers_canonicalised,
+					ctx->domain, ctx->selector,
+					RSPAMD_DKIM_KEY_ID_LEN, rspamd_dkim_key_id (key),
+					ctx->dkim_header);
+			msg_debug_dkim ("headers ecdsa verify failed");
 			res->rcode = DKIM_REJECT;
-			res->fail_reason = "ecdsa verify failed";
+			res->fail_reason = "headers ecdsa verify failed";
 		}
 		break;
 	case RSPAMD_DKIM_KEY_EDDSA:
 		if (!rspamd_cryptobox_verify (ctx->b, ctx->blen, raw_digest, dlen,
 				key->key.key_eddsa, RSPAMD_CRYPTOBOX_MODE_25519)) {
-			msg_debug_dkim ("eddsa verify failed");
+			msg_info_dkim (
+					"%s: headers EDDSA verification failure; "
+					"body length %d->%d; headers length %d; d=%s; s=%s; key_md5=%*xs; orig header: %s",
+					rspamd_dkim_type_to_string (ctx->common.type),
+					(gint)(body_end - body_start), ctx->common.body_canonicalised,
+					ctx->common.headers_canonicalised,
+					ctx->domain, ctx->selector,
+					RSPAMD_DKIM_KEY_ID_LEN, rspamd_dkim_key_id (key),
+					ctx->dkim_header);
+			msg_debug_dkim ("headers eddsa verify failed");
 			res->rcode = DKIM_REJECT;
-			res->fail_reason = "eddsa verify failed";
+			res->fail_reason = "headers eddsa verify failed";
 		}
 		break;
 	}
 
 
-	if (ctx->common.type == RSPAMD_DKIM_ARC_SEAL && res == DKIM_CONTINUE) {
+	if (ctx->common.type == RSPAMD_DKIM_ARC_SEAL && res->rcode == DKIM_CONTINUE) {
 		switch (ctx->cv) {
 		case RSPAMD_ARC_INVALID:
 			msg_info_dkim ("arc seal is invalid i=%d", ctx->common.idx);
@@ -2697,7 +2950,7 @@ rspamd_dkim_sign_key_load (const gchar *key, gsize len,
 	nkey = g_malloc0 (sizeof (*nkey));
 	nkey->mtime = mtime;
 
-	msg_debug_dkim_taskless ("got public key with length %d and type %d",
+	msg_debug_dkim_taskless ("got public key with length %z and type %d",
 			len, type);
 
 	/* Load key file if needed */
@@ -2763,7 +3016,7 @@ rspamd_dkim_sign_key_load (const gchar *key, gsize len,
 			nkey->type = RSPAMD_DKIM_KEY_EDDSA;
 			nkey->key.key_eddsa = g_malloc (
 					rspamd_cryptobox_sk_sig_bytes (RSPAMD_CRYPTOBOX_MODE_25519));
-			ed25519_seed_keypair (pk, nkey->key.key_eddsa, (char *) key);
+			crypto_sign_ed25519_seed_keypair (pk, nkey->key.key_eddsa, key);
 			nkey->keylen = rspamd_cryptobox_sk_sig_bytes (RSPAMD_CRYPTOBOX_MODE_25519);
 		}
 		else {
@@ -2949,7 +3202,7 @@ rspamd_dkim_sign (struct rspamd_task *task, const gchar *selector,
 
 	/* First of all find place of body */
 	body_end = task->msg.begin + task->msg.len;
-	body_start = task->raw_headers_content.body_start;
+	body_start = MESSAGE_FIELD (task, raw_headers_content).body_start;
 
 	if (len > 0) {
 		ctx->common.len = len;
@@ -2999,12 +3252,13 @@ rspamd_dkim_sign (struct rspamd_task *task, const gchar *selector,
 	}
 	else {
 		g_assert (arc_cv != NULL);
-		rspamd_printf_gstring (hdr, "i=%d; a=%s; c=%s/%s; d=%s; s=%s; cv=%s; ",
-				arc_cv,
+		rspamd_printf_gstring (hdr, "i=%d; a=%s; d=%s; s=%s; cv=%s; ",
+				idx,
 				ctx->key->type == RSPAMD_DKIM_KEY_RSA ?
 						"rsa-sha256" : "ed25519-sha256",
-				idx,
-				domain, selector);
+				domain,
+				selector,
+				arc_cv);
 	}
 
 	if (expire > 0) {
@@ -3021,6 +3275,8 @@ rspamd_dkim_sign (struct rspamd_task *task, const gchar *selector,
 
 	/* Now canonize headers */
 	for (i = 0; i < ctx->common.hlist->len; i++) {
+		struct rspamd_mime_header *rh, *cur;
+
 		dh = g_ptr_array_index (ctx->common.hlist, i);
 
 		/* We allow oversigning if dh->count > number of headers with this name */
@@ -3028,25 +3284,25 @@ rspamd_dkim_sign (struct rspamd_task *task, const gchar *selector,
 
 		if (hstat.s.flags & RSPAMD_DKIM_FLAG_OVERSIGN) {
 			/* Do oversigning */
-			GPtrArray *ar;
 			guint count = 0;
 
-			ar = g_hash_table_lookup (task->raw_headers, dh->name);
+			rh = rspamd_message_get_header_array (task, dh->name);
 
-			if (ar) {
-				count = ar->len;
-			}
-
-			for (j = 0; j < count; j ++) {
-				/* Sign all existing headers */
-				rspamd_dkim_canonize_header (&ctx->common, task, dh->name, j,
-						NULL, NULL);
+			if (rh) {
+				DL_FOREACH (rh, cur) {
+					/* Sign all existing headers */
+					rspamd_dkim_canonize_header (&ctx->common, task, dh->name,
+							count,
+							NULL, NULL);
+					count++;
+				}
 			}
 
 			/* Now add one more entry to oversign */
 			if (count > 0 || !(hstat.s.flags & RSPAMD_DKIM_FLAG_OVERSIGN_EXISTING)) {
 				cur_len = (strlen (dh->name) + 1) * (count + 1);
 				headers_len += cur_len;
+
 				if (headers_len > 70 && i > 0 && i < ctx->common.hlist->len - 1) {
 					rspamd_printf_gstring (hdr, "  ");
 					headers_len = cur_len;
@@ -3058,7 +3314,9 @@ rspamd_dkim_sign (struct rspamd_task *task, const gchar *selector,
 			}
 		}
 		else {
-			if (g_hash_table_lookup (task->raw_headers, dh->name)) {
+			rh = rspamd_message_get_header_array (task, dh->name);
+
+			if (rh) {
 				if (hstat.s.count > 0) {
 
 					cur_len = (strlen (dh->name) + 1) * (hstat.s.count);
@@ -3120,7 +3378,9 @@ rspamd_dkim_sign (struct rspamd_task *task, const gchar *selector,
 		if (!rspamd_dkim_canonize_header_relaxed (&ctx->common,
 				hdr->str,
 				hname,
-				TRUE)) {
+				TRUE,
+				0,
+				ctx->common.type == RSPAMD_DKIM_ARC_SEAL)) {
 
 			g_string_free (hdr, TRUE);
 			return NULL;
@@ -3130,6 +3390,7 @@ rspamd_dkim_sign (struct rspamd_task *task, const gchar *selector,
 		/* Will likely have issues with folding */
 		rspamd_dkim_hash_update (ctx->common.headers_hash, hdr->str,
 				hdr->len);
+		ctx->common.headers_canonicalised += hdr->len;
 		msg_debug_task ("update signature with header: %*s",
 				(gint)hdr->len, hdr->str);
 	}
@@ -3161,13 +3422,13 @@ rspamd_dkim_sign (struct rspamd_task *task, const gchar *selector,
 		return NULL;
 	}
 
-	if (task->flags & RSPAMD_TASK_FLAG_MILTER) {
+	if (task->protocol_flags & RSPAMD_TASK_PROTOCOL_FLAG_MILTER) {
 		b64_data = rspamd_encode_base64_fold (sig_buf, sig_len, 70, NULL,
 				RSPAMD_TASK_NEWLINES_LF);
 	}
 	else {
 		b64_data = rspamd_encode_base64_fold (sig_buf, sig_len, 70, NULL,
-				task->nlines_type);
+				MESSAGE_FIELD (task, nlines_type));
 	}
 
 	rspamd_printf_gstring (hdr, "%s", b64_data);
