@@ -131,7 +131,7 @@ rspamd_emergency_logger_dtor (gpointer d)
 }
 
 rspamd_logger_t *
-rspamd_log_open_emergency (rspamd_mempool_t *pool)
+rspamd_log_open_emergency (rspamd_mempool_t *pool, gint flags)
 {
 	rspamd_logger_t *logger;
 	GError *err = NULL;
@@ -147,7 +147,7 @@ rspamd_log_open_emergency (rspamd_mempool_t *pool)
 		logger = g_malloc0 (sizeof (rspamd_logger_t));
 	}
 
-
+	logger->flags = flags;
 	logger->pool = pool;
 	logger->process_type = "main";
 
@@ -258,7 +258,7 @@ rspamd_log_open_specific (rspamd_mempool_t *pool,
 					"IP addresses for which debug logs are enabled",
 					&logger->debug_ip,
 					NULL,
-					NULL);
+					NULL, "debug ip");
 		}
 
 		if (cfg->log_encryption_key) {
@@ -415,19 +415,36 @@ rspamd_common_logv (rspamd_logger_t *rspamd_log, gint level_flags,
 		const gchar *module, const gchar *id, const gchar *function,
 		const gchar *fmt, va_list args)
 {
-	gchar logbuf[RSPAMD_LOGBUF_SIZE], *end;
+	gchar *end;
 	gint level = level_flags & (RSPAMD_LOG_LEVEL_MASK & G_LOG_LEVEL_MASK), mod_id;
 	bool ret = false;
+	gchar logbuf[RSPAMD_LOGBUF_SIZE], *log_line;
+	gsize nescaped;
 
 	if (G_UNLIKELY (rspamd_log == NULL)) {
 		rspamd_log = default_logger;
 	}
 
+	log_line = logbuf;
+
 	if (G_UNLIKELY (rspamd_log == NULL)) {
 		/* Just fprintf message to stderr */
 		if (level >= G_LOG_LEVEL_INFO) {
 			end = rspamd_vsnprintf (logbuf, sizeof (logbuf), fmt, args);
-			rspamd_fprintf (stderr, "%*s\n", (gint)(end - logbuf), logbuf);
+
+			if (!(rspamd_log->flags & RSPAMD_LOG_FLAG_RSPAMADM)) {
+				if ((nescaped = rspamd_log_line_need_escape (logbuf, end - logbuf)) != 0) {
+					gsize unsecaped_len = end - logbuf;
+					gchar *logbuf_escaped = g_alloca (unsecaped_len + nescaped * 4);
+					log_line = logbuf_escaped;
+
+					end = rspamd_log_line_hex_escape (logbuf, unsecaped_len,
+							logbuf_escaped, unsecaped_len + nescaped * 4);
+				}
+			}
+
+			rspamd_fprintf (stderr, "%*s\n", (gint)(end - log_line),
+					log_line);
 		}
 	}
 	else {
@@ -441,11 +458,22 @@ rspamd_common_logv (rspamd_logger_t *rspamd_log, gint level_flags,
 		if (rspamd_logger_need_log (rspamd_log, level_flags, mod_id)) {
 			end = rspamd_vsnprintf (logbuf, sizeof (logbuf), fmt, args);
 
+			if (!(rspamd_log->flags & RSPAMD_LOG_FLAG_RSPAMADM)) {
+				if ((nescaped = rspamd_log_line_need_escape (logbuf, end - logbuf)) != 0) {
+					gsize unsecaped_len = end - logbuf;
+					gchar *logbuf_escaped = g_alloca (unsecaped_len + nescaped * 4);
+					log_line = logbuf_escaped;
+
+					end = rspamd_log_line_hex_escape (logbuf, unsecaped_len,
+							logbuf_escaped, unsecaped_len + nescaped * 4);
+				}
+			}
+
 			if ((level_flags & RSPAMD_LOG_ENCRYPTED) && rspamd_log->pk) {
 				gchar *encrypted;
 				gsize enc_len;
 
-				encrypted = rspamd_log_encrypt_message (logbuf, end, &enc_len,
+				encrypted = rspamd_log_encrypt_message (log_line, end, &enc_len,
 						rspamd_log);
 				ret = rspamd_log->ops.log (module, id,
 						function,
@@ -460,8 +488,8 @@ rspamd_common_logv (rspamd_logger_t *rspamd_log, gint level_flags,
 				ret = rspamd_log->ops.log (module, id,
 						function,
 						level_flags,
-						logbuf,
-						end - logbuf,
+						log_line,
+						end - log_line,
 						rspamd_log,
 						rspamd_log->ops.specific);
 			}
@@ -469,8 +497,8 @@ rspamd_common_logv (rspamd_logger_t *rspamd_log, gint level_flags,
 			switch (level) {
 			case G_LOG_LEVEL_CRITICAL:
 				rspamd_log->log_cnt[0] ++;
-				rspamd_log_write_ringbuffer (rspamd_log, module, id, logbuf,
-						end - logbuf);
+				rspamd_log_write_ringbuffer (rspamd_log, module, id, log_line,
+						end - log_line);
 				break;
 			case G_LOG_LEVEL_WARNING:
 				rspamd_log->log_cnt[1]++;
@@ -853,6 +881,8 @@ rspamd_logger_configure_modules (GHashTable *mods_enabled)
 	gpointer k, v;
 	guint id;
 
+	/* Clear all in bitset_allocated -> this are bytes not bits */
+	memset (log_modules->bitset, 0, log_modules->bitset_allocated);
 	/* On first iteration, we go through all modules enabled and add missing ones */
 	g_hash_table_iter_init (&it, mods_enabled);
 
@@ -880,4 +910,92 @@ rspamd_logger_set_log_function (rspamd_logger_t *logger,
 	/* TODO: write this */
 
 	return NULL;
+}
+
+
+
+gchar *
+rspamd_log_line_hex_escape (const guchar *src, gsize srclen,
+								  gchar *dst, gsize dstlen)
+{
+	static const gchar hexdigests[16] = "0123456789ABCDEF";
+	gchar *d = dst;
+
+	static guint32 escape[] = {
+			0xffffffff, /* 1111 1111 1111 1111  1111 1111 1111 1111 */
+
+			/* ?>=< ;:98 7654 3210  /.-, +*)( '&%$ #"!  */
+			0x00000000, /* 0000 0000 0000 0000  0000 0000 0000 0100 */
+
+			/* _^]\ [ZYX WVUT SRQP  ONML KJIH GFED CBA@ */
+			0x00000000, /* 0001 0000 0000 0000  0000 0000 0000 0000 */
+
+			/*  ~}| {zyx wvut srqp  onml kjih gfed cba` */
+			0x80000000, /* 1000 0000 0000 0000  0000 0000 0000 0000 */
+
+			/* Allow all 8bit characters (assuming they are valid utf8) */
+			0x00000000,
+			0x00000000,
+			0x00000000,
+			0x00000000,
+	};
+
+	while (srclen && dstlen) {
+		if (escape[*src >> 5] & (1U << (*src & 0x1f))) {
+			if (dstlen >= 4) {
+				*d++ = '\\';
+				*d++ = 'x';
+				*d++ = hexdigests[*src >> 4];
+				*d++ = hexdigests[*src & 0xf];
+				src++;
+				dstlen -= 4;
+			}
+			else {
+				/* Overflow */
+				break;
+			}
+		} else {
+			*d++ = *src++;
+			dstlen --;
+		}
+
+		srclen--;
+	}
+
+	return d;
+}
+
+gsize
+rspamd_log_line_need_escape (const guchar *src, gsize srclen)
+{
+	static guint32 escape[] = {
+			0xffffffff, /* 1111 1111 1111 1111  1111 1111 1111 1111 */
+
+			/* ?>=< ;:98 7654 3210  /.-, +*)( '&%$ #"!  */
+			0x00000000, /* 0000 0000 0000 0000  0000 0000 0000 0100 */
+
+			/* _^]\ [ZYX WVUT SRQP  ONML KJIH GFED CBA@ */
+			0x00000000, /* 0001 0000 0000 0000  0000 0000 0000 0000 */
+
+			/*  ~}| {zyx wvut srqp  onml kjih gfed cba` */
+			0x80000000, /* 1000 0000 0000 0000  0000 0000 0000 0000 */
+
+			/* Allow all 8bit characters (assuming they are valid utf8) */
+			0x00000000,
+			0x00000000,
+			0x00000000,
+			0x00000000,
+	};
+	gsize n = 0;
+
+	while (srclen) {
+		if (escape[*src >> 5] & (1U << (*src & 0x1f))) {
+			n++;
+		}
+
+		src ++;
+		srclen --;
+	}
+
+	return n;
 }
\ No newline at end of file
