@@ -20,6 +20,7 @@
 #include "radix.h"
 #include "rspamd.h"
 #include "cryptobox.h"
+#include "mempool_vars_internal.h"
 #include "contrib/fastutf8/fastutf8.h"
 #include "contrib/cdb/cdb.h"
 
@@ -34,7 +35,7 @@
 
 
 static const guint64 map_hash_seed = 0xdeadbabeULL;
-static const gchar *hash_fill = "1";
+static const gchar * const hash_fill = "1";
 
 struct rspamd_map_helper_value {
 	gsize hits;
@@ -42,36 +43,43 @@ struct rspamd_map_helper_value {
 	gchar value[]; /* Null terminated */
 };
 
-KHASH_INIT (rspamd_map_hash, const gchar *,
+#define rspamd_map_ftok_hash(t) (rspamd_icase_hash((t).begin, (t).len, rspamd_hash_seed ()))
+#define rspamd_map_ftok_equal(a, b) ((a).len == (b).len && rspamd_lc_cmp((a).begin, (b).begin, (a).len) == 0)
+
+KHASH_INIT (rspamd_map_hash, rspamd_ftok_t,
 		struct rspamd_map_helper_value *, true,
-		rspamd_strcase_hash, rspamd_strcase_equal);
+		rspamd_map_ftok_hash, rspamd_map_ftok_equal);
 
 struct rspamd_radix_map_helper {
 	rspamd_mempool_t *pool;
 	khash_t(rspamd_map_hash) *htb;
 	radix_compressed_t *trie;
+	struct rspamd_map *map;
 	rspamd_cryptobox_fast_hash_state_t hst;
 };
 
 struct rspamd_hash_map_helper {
 	rspamd_mempool_t *pool;
 	khash_t(rspamd_map_hash) *htb;
+	struct rspamd_map *map;
 	rspamd_cryptobox_fast_hash_state_t hst;
 };
 
 struct rspamd_cdb_map_helper {
 	GQueue cdbs;
+	struct rspamd_map *map;
 	rspamd_cryptobox_fast_hash_state_t hst;
 	gsize total_size;
 };
 
 struct rspamd_regexp_map_helper {
+	rspamd_cryptobox_hash_state_t hst;
+	guchar re_digest[rspamd_cryptobox_HASHBYTES];
 	rspamd_mempool_t *pool;
 	struct rspamd_map *map;
 	GPtrArray *regexps;
 	GPtrArray *values;
 	khash_t(rspamd_map_hash) *htb;
-	rspamd_cryptobox_fast_hash_state_t hst;
 	enum rspamd_regexp_map_flags map_flags;
 #ifdef WITH_HYPERSCAN
 	hs_database_t *hs_db;
@@ -105,7 +113,7 @@ rspamd_parse_kv_list (
 		gchar * chunk,
 		gint len,
 		struct map_cb_data *data,
-		insert_func func,
+		rspamd_map_insert_func func,
 		const gchar *default_value,
 		gboolean final)
 {
@@ -452,25 +460,51 @@ rspamd_map_helper_insert_radix (gpointer st, gconstpointer key, gconstpointer va
 	gsize vlen;
 	khiter_t k;
 	gconstpointer nk;
+	rspamd_ftok_t tok;
 	gint res;
+	struct rspamd_map *map;
 
-	vlen = strlen (value);
-	val = rspamd_mempool_alloc0 (r->pool, sizeof (*val) +
-			vlen + 1);
-	memcpy (val->value, value, vlen);
+	map = r->map;
+	tok.begin = key;
+	tok.len = strlen (key);
 
-	k = kh_get (rspamd_map_hash, r->htb, key);
+	k = kh_get (rspamd_map_hash, r->htb, tok);
 
 	if (k == kh_end (r->htb)) {
 		nk = rspamd_mempool_strdup (r->pool, key);
-		k = kh_put (rspamd_map_hash, r->htb, nk, &res);
+		tok.begin = nk;
+		k = kh_put (rspamd_map_hash, r->htb, tok, &res);
+	}
+	else {
+		val = kh_value (r->htb, k);
+
+		if (strcmp (value, val->value) == 0) {
+			/* Same element, skip */
+			return;
+		}
+		else {
+			msg_warn_map ("duplicate radix entry found for map %s: %s (old value: '%s', new: '%s')",
+					map->name, key, val->value, value);
+		}
+
+		nk = kh_key (r->htb, k).begin;
+		val->key = nk;
+		kh_value (r->htb, k) = val;
+
+		return; /* do not touch radix in case of exact duplicate */
 	}
 
-	nk = kh_key (r->htb, k);
+	vlen = strlen (value);
+	val = rspamd_mempool_alloc0 (r->pool, sizeof (*val) +
+										  vlen + 1);
+	memcpy (val->value, value, vlen);
+
+	nk = kh_key (r->htb, k).begin;
 	val->key = nk;
 	kh_value (r->htb, k) = val;
-	rspamd_radix_add_iplist (key, ",", r->trie, val, FALSE);
-	rspamd_cryptobox_fast_hash_update (&r->hst, nk, strlen (nk));
+	rspamd_radix_add_iplist (key, ",", r->trie, val, FALSE,
+			r->map->name);
+	rspamd_cryptobox_fast_hash_update (&r->hst, nk, tok.len);
 }
 
 void
@@ -481,25 +515,50 @@ rspamd_map_helper_insert_radix_resolve (gpointer st, gconstpointer key, gconstpo
 	gsize vlen;
 	khiter_t k;
 	gconstpointer nk;
+	rspamd_ftok_t tok;
 	gint res;
+	struct rspamd_map *map;
 
-	vlen = strlen (value);
-	val = rspamd_mempool_alloc0 (r->pool, sizeof (*val) +
-			vlen + 1);
-	memcpy (val->value, value, vlen);
+	map = r->map;
+	tok.begin = key;
+	tok.len = strlen (key);
 
-	k = kh_get (rspamd_map_hash, r->htb, key);
+	k = kh_get (rspamd_map_hash, r->htb, tok);
 
 	if (k == kh_end (r->htb)) {
 		nk = rspamd_mempool_strdup (r->pool, key);
-		k = kh_put (rspamd_map_hash, r->htb, nk, &res);
+		tok.begin = nk;
+		k = kh_put (rspamd_map_hash, r->htb, tok, &res);
+	}
+	else {
+		val = kh_value (r->htb, k);
+
+		if (strcmp (value, val->value) == 0) {
+			/* Same element, skip */
+			return;
+		}
+		else {
+			msg_warn_map ("duplicate radix entry found for map %s: %s (old value: '%s', new: '%s')",
+					map->name, key, val->value, value);
+		}
+
+		nk = kh_key (r->htb, k).begin;
+		val->key = nk;
+		kh_value (r->htb, k) = val;
+
+		return; /* do not touch radix in case of exact duplicate */
 	}
 
-	nk = kh_key (r->htb, k);
+	vlen = strlen (value);
+	val = rspamd_mempool_alloc0 (r->pool, sizeof (*val) +
+										  vlen + 1);
+	memcpy (val->value, value, vlen);
+	nk = kh_key (r->htb, k).begin;
 	val->key = nk;
 	kh_value (r->htb, k) = val;
-	rspamd_radix_add_iplist (key, ",", r->trie, val, TRUE);
-	rspamd_cryptobox_fast_hash_update (&r->hst, nk, strlen (nk));
+	rspamd_radix_add_iplist (key, ",", r->trie, val, TRUE,
+			r->map->name);
+	rspamd_cryptobox_fast_hash_update (&r->hst, nk, tok.len);
 }
 
 void
@@ -511,13 +570,19 @@ rspamd_map_helper_insert_hash (gpointer st, gconstpointer key, gconstpointer val
 	gconstpointer nk;
 	gsize vlen;
 	gint r;
+	rspamd_ftok_t tok;
+	struct rspamd_map *map;
 
-	k = kh_get (rspamd_map_hash, ht->htb, key);
-	vlen = strlen (value);
+	tok.begin = key;
+	tok.len = strlen (key);
+	map = ht->map;
+
+	k = kh_get (rspamd_map_hash, ht->htb, tok);
 
 	if (k == kh_end (ht->htb)) {
 		nk = rspamd_mempool_strdup (ht->pool, key);
-		k = kh_put (rspamd_map_hash, ht->htb, nk, &r);
+		tok.begin = nk;
+		k = kh_put (rspamd_map_hash, ht->htb, tok, &r);
 	}
 	else {
 		val = kh_value (ht->htb, k);
@@ -526,16 +591,23 @@ rspamd_map_helper_insert_hash (gpointer st, gconstpointer key, gconstpointer val
 			/* Same element, skip */
 			return;
 		}
+		else {
+			msg_warn_map ("duplicate hash entry found for map %s: %s (old value: '%s', new: '%s')",
+					map->name, key, val->value, value);
+		}
 	}
 
 	/* Null termination due to alloc0 */
+	vlen = strlen (value);
 	val = rspamd_mempool_alloc0 (ht->pool, sizeof (*val) + vlen + 1);
 	memcpy (val->value, value, vlen);
 
-	nk = kh_key (ht->htb, k);
+	tok = kh_key (ht->htb, k);
+	nk = tok.begin;
 	val->key = nk;
 	kh_value (ht->htb, k) = val;
-	rspamd_cryptobox_fast_hash_update (&ht->hst, nk, strlen (nk));
+
+	rspamd_cryptobox_fast_hash_update (&ht->hst, nk, tok.len);
 }
 
 void
@@ -550,12 +622,44 @@ rspamd_map_helper_insert_re (gpointer st, gconstpointer key, gconstpointer value
 	gsize escaped_len;
 	struct rspamd_map_helper_value *val;
 	khiter_t k;
+	rspamd_ftok_t tok;
 	gconstpointer nk;
 	gsize vlen;
 	gint r;
 
 	map = re_map->map;
 
+	tok.begin = key;
+	tok.len = strlen (key);
+
+	k = kh_get (rspamd_map_hash, re_map->htb, tok);
+
+	if (k == kh_end (re_map->htb)) {
+		nk = rspamd_mempool_strdup (re_map->pool, key);
+		tok.begin = nk;
+		k = kh_put (rspamd_map_hash, re_map->htb, tok, &r);
+	}
+	else {
+		val = kh_value (re_map->htb, k);
+
+		/* Always warn about regexp duplicate as it's likely a bad mistake */
+		msg_warn_map ("duplicate re entry found for map %s: %s (old value: '%s', new: '%s')",
+				map->name, key, val->value, value);
+
+		if (strcmp (val->value, value) == 0) {
+			/* Same value, skip */
+			return;
+		}
+
+		/* Replace value but do not touch regexp */
+		nk = kh_key (re_map->htb, k).begin;
+		val->key = nk;
+		kh_value (re_map->htb, k) = val;
+
+		return;
+	}
+
+	/* Check regexp stuff */
 	if (re_map->map_flags & RSPAMD_REGEXP_MAP_FLAG_GLOB) {
 		escaped = rspamd_str_regexp_escape (key, strlen (key), &escaped_len,
 				RSPAMD_REGEXP_ESCAPE_GLOB|RSPAMD_REGEXP_ESCAPE_UTF);
@@ -578,20 +682,12 @@ rspamd_map_helper_insert_re (gpointer st, gconstpointer key, gconstpointer value
 
 	vlen = strlen (value);
 	val = rspamd_mempool_alloc0 (re_map->pool, sizeof (*val) +
-			vlen + 1);
-	memcpy (val->value, value, vlen);
-
-	k = kh_get (rspamd_map_hash, re_map->htb, key);
-
-	if (k == kh_end (re_map->htb)) {
-		nk = rspamd_mempool_strdup (re_map->pool, key);
-		k = kh_put (rspamd_map_hash, re_map->htb, nk, &r);
-	}
-
-	nk = kh_key (re_map->htb, k);
+											   vlen + 1);
+	memcpy (val->value, value, vlen); /* Null terminated due to alloc0 previously */
+	nk = kh_key (re_map->htb, k).begin;
 	val->key = nk;
 	kh_value (re_map->htb, k) = val;
-	rspamd_cryptobox_fast_hash_update (&re_map->hst, nk, strlen (nk));
+	rspamd_cryptobox_hash_update (&re_map->hst, nk, tok.len);
 
 	pcre_flags = rspamd_regexp_get_pcre_flags (re);
 
@@ -615,12 +711,12 @@ rspamd_map_helper_traverse_regexp (void *data,
 		gpointer cbdata,
 		gboolean reset_hits)
 {
-	gconstpointer k;
+	rspamd_ftok_t tok;
 	struct rspamd_map_helper_value *val;
 	struct rspamd_regexp_map_helper *re_map = data;
 
-	kh_foreach (re_map->htb, k, val, {
-		if (!cb (k, val->value, val->hits, cbdata)) {
+	kh_foreach (re_map->htb, tok, val, {
+		if (!cb (tok.begin, val->value, val->hits, cbdata)) {
 			break;
 		}
 
@@ -648,6 +744,7 @@ rspamd_map_helper_new_hash (struct rspamd_map *map)
 	htb = rspamd_mempool_alloc0 (pool, sizeof (*htb));
 	htb->htb = kh_init (rspamd_map_hash);
 	htb->pool = pool;
+	htb->map = map;
 	rspamd_cryptobox_fast_hash_init (&htb->hst, map_hash_seed);
 
 	return htb;
@@ -672,12 +769,12 @@ rspamd_map_helper_traverse_hash (void *data,
 		gpointer cbdata,
 		gboolean reset_hits)
 {
-	gconstpointer k;
+	rspamd_ftok_t tok;
 	struct rspamd_map_helper_value *val;
 	struct rspamd_hash_map_helper *ht = data;
 
-	kh_foreach (ht->htb, k, val, {
-		if (!cb (k, val->value, val->hits, cbdata)) {
+	kh_foreach (ht->htb, tok, val, {
+		if (!cb (tok.begin, val->value, val->hits, cbdata)) {
 			break;
 		}
 
@@ -692,10 +789,12 @@ rspamd_map_helper_new_radix (struct rspamd_map *map)
 {
 	struct rspamd_radix_map_helper *r;
 	rspamd_mempool_t *pool;
+	const gchar *name = "unnamed";
 
 	if (map) {
 		pool = rspamd_mempool_new (rspamd_mempool_suggest_size (),
 				map->tag, 0);
+		name = map->name;
 	}
 	else {
 		pool = rspamd_mempool_new (rspamd_mempool_suggest_size (),
@@ -703,9 +802,10 @@ rspamd_map_helper_new_radix (struct rspamd_map *map)
 	}
 
 	r = rspamd_mempool_alloc0 (pool, sizeof (*r));
-	r->trie = radix_create_compressed_with_pool (pool);
+	r->trie = radix_create_compressed_with_pool (pool, name);
 	r->htb = kh_init (rspamd_map_hash);
 	r->pool = pool;
+	r->map = map;
 	rspamd_cryptobox_fast_hash_init (&r->hst, map_hash_seed);
 
 	return r;
@@ -730,12 +830,12 @@ rspamd_map_helper_traverse_radix (void *data,
 		gpointer cbdata,
 		gboolean reset_hits)
 {
-	gconstpointer k;
+	rspamd_ftok_t tok;
 	struct rspamd_map_helper_value *val;
 	struct rspamd_radix_map_helper *r = data;
 
-	kh_foreach (r->htb, k, val, {
-		if (!cb (k, val->value, val->hits, cbdata)) {
+	kh_foreach (r->htb, tok, val, {
+		if (!cb (tok.begin, val->value, val->hits, cbdata)) {
 			break;
 		}
 
@@ -762,7 +862,7 @@ rspamd_map_helper_new_regexp (struct rspamd_map *map,
 	re_map->map = map;
 	re_map->map_flags = flags;
 	re_map->htb = kh_init (rspamd_map_hash);
-	rspamd_cryptobox_fast_hash_init (&re_map->hst, map_hash_seed);
+	rspamd_cryptobox_hash_init (&re_map->hst, NULL, 0);
 
 	return re_map;
 }
@@ -842,7 +942,8 @@ rspamd_kv_list_fin (struct map_cb_data *data, void **target)
 
 	if (data->cur_data) {
 		htb = (struct rspamd_hash_map_helper *)data->cur_data;
-		msg_info_map ("read hash of %d elements", kh_size (htb->htb));
+		msg_info_map ("read hash of %d elements from %s", kh_size (htb->htb),
+				map->name);
 		data->map->traverse_function = rspamd_map_helper_traverse_hash;
 		data->map->nelts = kh_size (htb->htb);
 		data->map->digest = rspamd_cryptobox_fast_hash_final (&htb->hst);
@@ -929,6 +1030,243 @@ rspamd_radix_dtor (struct map_cb_data *data)
 	}
 }
 
+#ifdef WITH_HYPERSCAN
+struct rspamd_re_maps_cache_dtor_cbdata {
+	struct rspamd_config *cfg;
+	GHashTable *valid_re_hashes;
+	gchar *dirname;
+};
+
+static void
+rspamd_re_maps_cache_cleanup_dtor (gpointer ud)
+{
+	struct rspamd_re_maps_cache_dtor_cbdata *cbd =
+			(struct rspamd_re_maps_cache_dtor_cbdata *)ud;
+	GPtrArray *cache_files;
+	GError *err = NULL;
+	struct rspamd_config *cfg;
+
+	cfg = cbd->cfg;
+
+	if (cfg->cur_worker != NULL) {
+		/* Skip dtor, limit it to main process only */
+		return;
+	}
+
+	cache_files = rspamd_glob_path (cbd->dirname, "*.hsmc", FALSE, &err);
+
+	if (!cache_files) {
+		msg_err_config ("cannot glob files in %s: %e", cbd->dirname, err);
+		g_error_free (err);
+	}
+	else {
+		const gchar *fname;
+		guint i;
+
+		PTR_ARRAY_FOREACH (cache_files, i, fname) {
+			gchar *basename = g_path_get_basename (fname);
+
+			if (g_hash_table_lookup (cbd->valid_re_hashes, basename) == NULL) {
+				gchar *dir;
+
+				dir = g_path_get_dirname (fname);
+
+				/* Sanity check to avoid removal of something bad */
+				if (strcmp (dir, cbd->dirname) != 0) {
+					msg_err_config ("bogus file found: %s in %s, skip deleting",
+							fname, dir);
+				}
+				else {
+					if (unlink (fname) == -1) {
+						msg_err_config ("cannot delete obsolete file %s in %s: %s",
+								fname, dir, strerror (errno));
+					}
+					else {
+						msg_info_config ("deleted obsolete file %s in %s",
+								fname, dir);
+					}
+				}
+
+				g_free (dir);
+			}
+			else {
+				msg_debug_config ("valid re cache file %s", fname);
+			}
+
+			g_free (basename);
+		}
+	}
+
+	g_hash_table_unref (cbd->valid_re_hashes);
+	g_free (cbd->dirname);
+}
+
+static void
+rspamd_re_map_cache_update (const gchar *fname, struct rspamd_config *cfg)
+{
+	GHashTable *valid_re_hashes;
+
+	valid_re_hashes = rspamd_mempool_get_variable (cfg->cfg_pool,
+			RSPAMD_MEMPOOL_RE_MAPS_CACHE);
+
+	if (!valid_re_hashes) {
+		valid_re_hashes = g_hash_table_new_full (g_str_hash, g_str_equal,
+				g_free, NULL);
+		rspamd_mempool_set_variable (cfg->cfg_pool,
+				RSPAMD_MEMPOOL_RE_MAPS_CACHE,
+				valid_re_hashes, (rspamd_mempool_destruct_t)g_hash_table_unref);
+
+		/* We also add a cleanup dtor for all hashes */
+		static struct rspamd_re_maps_cache_dtor_cbdata cbd;
+
+		cbd.valid_re_hashes = g_hash_table_ref (valid_re_hashes);
+		cbd.cfg = cfg;
+		cbd.dirname = g_path_get_dirname (fname);
+		rspamd_mempool_add_destructor (cfg->cfg_pool,
+				rspamd_re_maps_cache_cleanup_dtor, &cbd);
+	}
+
+	g_hash_table_insert (valid_re_hashes, g_path_get_basename (fname), "1");
+}
+
+static gboolean
+rspamd_try_load_re_map_cache (struct rspamd_regexp_map_helper *re_map)
+{
+	gchar fp[PATH_MAX];
+	gpointer data;
+	gsize len;
+	struct rspamd_map *map;
+
+	map = re_map->map;
+
+	if (!map->cfg->hs_cache_dir) {
+		return FALSE;
+	}
+
+	rspamd_snprintf (fp, sizeof (fp), "%s/%*xs.hsmc",
+			map->cfg->hs_cache_dir,
+			(gint)rspamd_cryptobox_HASHBYTES / 2, re_map->re_digest);
+
+	if ((data = rspamd_file_xmap (fp, PROT_READ, &len, TRUE)) != NULL) {
+		if (hs_deserialize_database (data, len, &re_map->hs_db) == HS_SUCCESS) {
+			rspamd_re_map_cache_update (fp, map->cfg);
+			munmap (data, len);
+
+			msg_info_map ("loaded hypersan cache from %s (%Hz length) for %s",
+					fp, len, map->name);
+
+			return TRUE;
+		}
+
+		msg_info_map ("invalid hypersan cache in %s (%Hz length) for %s, removing file",
+				fp, len, map->name);
+		munmap (data, len);
+		/* Remove stale file */
+		(void)unlink (fp);
+	}
+
+	return FALSE;
+}
+
+static gboolean
+rspamd_try_save_re_map_cache (struct rspamd_regexp_map_helper *re_map)
+{
+	gchar fp[PATH_MAX], np[PATH_MAX];
+	gsize len;
+	gint fd;
+	char *bytes = NULL;
+	struct rspamd_map *map;
+
+	map = re_map->map;
+
+	if (!map->cfg->hs_cache_dir) {
+		return FALSE;
+	}
+
+	rspamd_snprintf (fp, sizeof (fp), "%s/%*xs.hsmc.tmp",
+			re_map->map->cfg->hs_cache_dir,
+			(gint)rspamd_cryptobox_HASHBYTES / 2, re_map->re_digest);
+
+	if ((fd = rspamd_file_xopen (fp, O_WRONLY | O_CREAT | O_EXCL, 00644, 0)) != -1) {
+		if (hs_serialize_database (re_map->hs_db, &bytes, &len) == HS_SUCCESS) {
+			if (write (fd, bytes, len) == -1) {
+				msg_warn_map ("cannot write hyperscan cache to %s: %s",
+						fp, strerror (errno));
+				unlink (fp);
+				free (bytes);
+			}
+			else {
+				free (bytes);
+				fsync (fd);
+
+				rspamd_snprintf (np, sizeof (np), "%s/%*xs.hsmc",
+						re_map->map->cfg->hs_cache_dir,
+						(gint)rspamd_cryptobox_HASHBYTES / 2, re_map->re_digest);
+
+				if (rename (fp, np) == -1) {
+					msg_warn_map ("cannot rename hyperscan cache from %s to %s: %s",
+							fp, np, strerror (errno));
+					unlink (fp);
+				}
+				else {
+					msg_info_map ("written cached hyperscan data for %s to %s (%Hz length)",
+							map->name, np, len);
+
+					rspamd_re_map_cache_update (np, map->cfg);
+				}
+			}
+		}
+		else {
+			msg_warn_map ("cannot serialize hyperscan cache to %s: %s",
+					fp, strerror (errno));
+			unlink (fp);
+		}
+
+
+		close (fd);
+	}
+
+	return FALSE;
+}
+
+static gboolean
+rspamd_re_map_cache_cleanup_old (struct rspamd_regexp_map_helper *old_re_map)
+{
+	gchar fp[PATH_MAX];
+	struct rspamd_map *map;
+
+	map = old_re_map->map;
+
+	if (!map->cfg->hs_cache_dir) {
+		return FALSE;
+	}
+
+	rspamd_snprintf (fp, sizeof (fp), "%s/%*xs.hsmc",
+			map->cfg->hs_cache_dir,
+			(gint)rspamd_cryptobox_HASHBYTES / 2, old_re_map->re_digest);
+
+	msg_info_map ("unlink stale cache file for %s: %s", map->name, fp);
+
+	if (unlink (fp) == -1) {
+		msg_warn_map ("cannot unlink stale cache file for %s (%s): %s",
+				map->name, fp, strerror (errno));
+		return FALSE;
+	}
+
+	GHashTable *valid_re_hashes;
+
+	valid_re_hashes = rspamd_mempool_get_variable (map->cfg->cfg_pool,
+			RSPAMD_MEMPOOL_RE_MAPS_CACHE);
+
+	if (valid_re_hashes) {
+		g_hash_table_remove (valid_re_hashes, fp);
+	}
+
+	return TRUE;
+}
+
+#endif
+
 static void
 rspamd_re_map_finalize (struct rspamd_regexp_map_helper *re_map)
 {
@@ -942,11 +1280,13 @@ rspamd_re_map_finalize (struct rspamd_regexp_map_helper *re_map)
 
 	map = re_map->map;
 
+#ifndef __aarch64__
 	if (!(map->cfg->libs_ctx->crypto_ctx->cpu_config & CPUID_SSSE3)) {
 		msg_info_map ("disable hyperscan for map %s, ssse3 instructons are not supported by CPU",
 				map->name);
 		return;
 	}
+#endif
 
 	if (hs_populate_platform (&plt) != HS_SUCCESS) {
 		msg_err_map ("cannot populate hyperscan platform");
@@ -1006,23 +1346,37 @@ rspamd_re_map_finalize (struct rspamd_regexp_map_helper *re_map)
 	}
 
 	if (re_map->regexps->len > 0 && re_map->patterns) {
-		if (hs_compile_multi ((const gchar **)re_map->patterns,
-				re_map->flags,
-				re_map->ids,
-				re_map->regexps->len,
-				HS_MODE_BLOCK,
-				&plt,
-				&re_map->hs_db,
-				&err) != HS_SUCCESS) {
 
-			msg_err_map ("cannot create tree of regexp when processing '%s': %s",
-					err->expression >= 0 ?
-							re_map->patterns[err->expression] :
-							"unknown regexp", err->message);
-			re_map->hs_db = NULL;
-			hs_free_compile_error (err);
+		if (!rspamd_try_load_re_map_cache (re_map)) {
+			gdouble ts1 = rspamd_get_ticks (FALSE);
 
-			return;
+			if (hs_compile_multi ((const gchar **) re_map->patterns,
+					re_map->flags,
+					re_map->ids,
+					re_map->regexps->len,
+					HS_MODE_BLOCK,
+					&plt,
+					&re_map->hs_db,
+					&err) != HS_SUCCESS) {
+
+				msg_err_map ("cannot create tree of regexp when processing '%s': %s",
+						err->expression >= 0 ?
+						re_map->patterns[err->expression] :
+						"unknown regexp", err->message);
+				re_map->hs_db = NULL;
+				hs_free_compile_error (err);
+
+				return;
+			}
+
+			ts1 = (rspamd_get_ticks (FALSE) - ts1) * 1000.0;
+			msg_info_map ("hyperscan compiled %d regular expressions from %s in %.1f ms",
+					re_map->regexps->len, re_map->map->name, ts1);
+			rspamd_try_save_re_map_cache (re_map);
+		}
+		else {
+			msg_info_map ("hyperscan read %d cached regular expressions from %s",
+					re_map->regexps->len, re_map->map->name);
 		}
 
 		if (hs_alloc_scratch (re_map->hs_db, &re_map->hs_scratch) != HS_SUCCESS) {
@@ -1135,17 +1489,18 @@ rspamd_glob_list_read_multiple (
 void
 rspamd_regexp_list_fin (struct map_cb_data *data, void **target)
 {
-	struct rspamd_regexp_map_helper *re_map;
+	struct rspamd_regexp_map_helper *re_map, *old_re_map;
 	struct rspamd_map *map = data->map;
 
 	if (data->cur_data) {
 		re_map = data->cur_data;
+		rspamd_cryptobox_hash_final (&re_map->hst, re_map->re_digest);
+		memcpy (&data->map->digest, re_map->re_digest, sizeof (data->map->digest));
 		rspamd_re_map_finalize (re_map);
 		msg_info_map ("read regexp list of %ud elements",
 				re_map->regexps->len);
 		data->map->traverse_function = rspamd_map_helper_traverse_regexp;
 		data->map->nelts = kh_size (re_map->htb);
-		data->map->digest = rspamd_cryptobox_fast_hash_final (&re_map->hst);
 	}
 
 	if (target) {
@@ -1153,7 +1508,17 @@ rspamd_regexp_list_fin (struct map_cb_data *data, void **target)
 	}
 
 	if (data->prev_data) {
-		rspamd_map_helper_destroy_regexp (data->prev_data);
+		old_re_map = data->prev_data;
+
+#ifdef WITH_HYPERSCAN
+		if (memcmp (re_map->re_digest, old_re_map->re_digest,
+				sizeof (re_map->re_digest)) != 0) {
+			/* Cleanup old stuff */
+			rspamd_re_map_cache_cleanup_old (old_re_map);
+		}
+#endif
+
+		rspamd_map_helper_destroy_regexp (old_re_map);
 	}
 }
 void
@@ -1344,12 +1709,16 @@ rspamd_match_hash_map (struct rspamd_hash_map_helper *map, const gchar *in,
 {
 	khiter_t k;
 	struct rspamd_map_helper_value *val;
+	rspamd_ftok_t tok;
 
 	if (map == NULL || map->htb == NULL) {
 		return NULL;
 	}
 
-	k = kh_get (rspamd_map_hash, map->htb, in);
+	tok.begin = in;
+	tok.len = len;
+
+	k = kh_get (rspamd_map_hash, map->htb, tok);
 
 	if (k != kh_end (map->htb)) {
 		val = kh_value (map->htb, k);
@@ -1416,6 +1785,7 @@ rspamd_map_helper_new_cdb (struct rspamd_map *map)
 
 	n = g_malloc0 (sizeof (*n));
 	n->cdbs = (GQueue)G_QUEUE_INIT;
+	n->map = map;
 
 	rspamd_cryptobox_fast_hash_init (&n->hst, map_hash_seed);
 
@@ -1572,4 +1942,4 @@ rspamd_match_cdb_map (struct rspamd_cdb_map_helper *map,
 	}
 
 	return NULL;
-}
\ No newline at end of file
+}
