@@ -108,6 +108,7 @@ enum rspamd_re_cache_elt_match_type {
 
 struct rspamd_re_cache_elt {
 	rspamd_regexp_t *re;
+	gint lua_cbref;
 	enum rspamd_re_cache_elt_match_type match_type;
 };
 
@@ -212,6 +213,15 @@ rspamd_re_cache_destroy (struct rspamd_re_cache *cache)
 			luaL_unref (cache->L, LUA_REGISTRYINDEX, sref);
 			g_free (skey);
 		});
+
+		struct rspamd_re_cache_elt *elt;
+		guint i;
+
+		PTR_ARRAY_FOREACH (cache->re, i, elt) {
+			if (elt->lua_cbref != -1) {
+				luaL_unref (cache->L, LUA_REGISTRYINDEX, elt->lua_cbref);
+			}
+		}
 	}
 
 	kh_destroy (lua_selectors_hash, cache->selectors);
@@ -261,8 +271,11 @@ rspamd_re_cache_is_hs_loaded (struct rspamd_re_cache *cache)
 }
 
 rspamd_regexp_t *
-rspamd_re_cache_add (struct rspamd_re_cache *cache, rspamd_regexp_t *re,
-		enum rspamd_re_type type, gconstpointer type_data, gsize datalen)
+rspamd_re_cache_add (struct rspamd_re_cache *cache,
+					 rspamd_regexp_t *re,
+					 enum rspamd_re_type type,
+					 gconstpointer type_data, gsize datalen,
+					 gint lua_cbref)
 {
 	guint64 class_id;
 	struct rspamd_re_class *re_class;
@@ -304,6 +317,8 @@ rspamd_re_cache_add (struct rspamd_re_cache *cache, rspamd_regexp_t *re,
 		elt->re = rspamd_regexp_ref (re);
 		g_ptr_array_add (cache->re, elt);
 		rspamd_regexp_set_class (re, re_class);
+		elt->lua_cbref = lua_cbref;
+
 		g_hash_table_insert (re_class->re, rspamd_regexp_get_id (nre), nre);
 	}
 
@@ -391,7 +406,7 @@ rspamd_re_cache_init (struct rspamd_re_cache *cache, struct rspamd_config *cfg)
 		rspamd_regexp_set_cache_id (re, i);
 
 		if (re_class->st == NULL) {
-			posix_memalign ((void **)&re_class->st, _Alignof (rspamd_cryptobox_hash_state_t),
+			(void) !posix_memalign ((void **)&re_class->st, _Alignof (rspamd_cryptobox_hash_state_t),
 			 		sizeof (*re_class->st));
 			g_assert (re_class->st != NULL);
 			rspamd_cryptobox_hash_init (re_class->st, NULL, 0);
@@ -529,11 +544,49 @@ rspamd_re_cache_get_stat (struct rspamd_re_runtime *rt)
 	return &rt->stat;
 }
 
+static gboolean
+rspamd_re_cache_check_lua_condition (struct rspamd_task *task,
+									 rspamd_regexp_t *re,
+									 const guchar *in, gsize len,
+									 goffset start, goffset end,
+									 gint lua_cbref)
+{
+	lua_State *L = (lua_State *)task->cfg->lua_state;
+	GError *err = NULL;
+	struct rspamd_lua_text __attribute__ ((unused)) *t;
+	gint text_pos;
+
+	if (G_LIKELY (lua_cbref == -1)) {
+		return TRUE;
+	}
+
+	t = lua_new_text (L, in, len, FALSE);
+	text_pos = lua_gettop (L);
+
+	if (!rspamd_lua_universal_pcall (L, lua_cbref,
+			G_STRLOC, 1, "utii", &err,
+			"rspamd{task}", task,
+			text_pos, start, end)) {
+		msg_warn_task ("cannot call for re_cache_check_lua_condition for re %s: %e",
+				rspamd_regexp_get_pattern (re), err);
+		g_error_free (err);
+
+		return TRUE;
+	}
+
+	gboolean res = lua_toboolean (L, -1);
+
+	lua_settop (L, text_pos - 1);
+
+	return res;
+}
+
 static guint
 rspamd_re_cache_process_pcre (struct rspamd_re_runtime *rt,
 		rspamd_regexp_t *re, struct rspamd_task *task,
 		const guchar *in, gsize len,
-		gboolean is_raw)
+		gboolean is_raw,
+		gint lua_cbref)
 {
 	guint r = 0;
 	const gchar *start = NULL, *end = NULL;
@@ -570,9 +623,12 @@ rspamd_re_cache_process_pcre (struct rspamd_re_runtime *rt,
 				&end,
 				is_raw,
 				NULL)) {
-			r++;
-			msg_debug_re_task ("found regexp /%s/, total hits: %d",
-					rspamd_regexp_get_pattern (re), r);
+			if (rspamd_re_cache_check_lua_condition (task, re, in, len,
+					start - (const gchar *)in, end - (const gchar *)in, lua_cbref)) {
+				r++;
+				msg_debug_re_task ("found regexp /%s/, total hits: %d",
+						rspamd_regexp_get_pattern (re), r);
+			}
 
 			if (max_hits > 0 && r >= max_hits) {
 				break;
@@ -621,25 +677,28 @@ rspamd_re_cache_hyperscan_cb (unsigned int id,
 {
 	struct rspamd_re_hyperscan_cbdata *cbdata = ud;
 	struct rspamd_re_runtime *rt;
-	struct rspamd_re_cache_elt *pcre_elt;
+	struct rspamd_re_cache_elt *cache_elt;
 	guint ret, maxhits, i, processed;
 	struct rspamd_task *task;
 
 	rt = cbdata->rt;
 	task = cbdata->task;
-	pcre_elt = g_ptr_array_index (rt->cache->re, id);
-	maxhits = rspamd_regexp_get_maxhits (pcre_elt->re);
+	cache_elt = g_ptr_array_index (rt->cache->re, id);
+	maxhits = rspamd_regexp_get_maxhits (cache_elt->re);
 
-	if (pcre_elt->match_type == RSPAMD_RE_CACHE_HYPERSCAN) {
-		ret = 1;
-		setbit (rt->checked, id);
+	if (cache_elt->match_type == RSPAMD_RE_CACHE_HYPERSCAN) {
+		if (rspamd_re_cache_check_lua_condition (task, cache_elt->re,
+				cbdata->ins[0], cbdata->lens[0], from, to, cache_elt->lua_cbref)) {
+			ret = 1;
+			setbit (rt->checked, id);
 
-		if (maxhits == 0 || rt->results[id] < maxhits) {
-			rt->results[id] += ret;
-			rt->stat.regexp_matched++;
+			if (maxhits == 0 || rt->results[id] < maxhits) {
+				rt->results[id] += ret;
+				rt->stat.regexp_matched++;
+			}
+			msg_debug_re_task ("found regexp /%s/ using hyperscan only, total hits: %d",
+					rspamd_regexp_get_pattern (cache_elt->re), rt->results[id]);
 		}
-		msg_debug_re_task ("found regexp /%s/ using hyperscan only, total hits: %d",
-				rspamd_regexp_get_pattern (pcre_elt->re), rt->results[id]);
 	}
 	else {
 		if (!isset (rt->checked, id)) {
@@ -648,11 +707,12 @@ rspamd_re_cache_hyperscan_cb (unsigned int id,
 
 			for (i = 0; i < cbdata->count; i ++) {
 				rspamd_re_cache_process_pcre (rt,
-						pcre_elt->re,
+						cache_elt->re,
 						cbdata->task,
 						cbdata->ins[i],
 						cbdata->lens[i],
-						FALSE);
+						FALSE,
+						cache_elt->lua_cbref);
 				setbit (rt->checked, id);
 
 				processed += cbdata->lens[i];
@@ -680,6 +740,7 @@ rspamd_re_cache_process_regexp_data (struct rspamd_re_runtime *rt,
 	guint64 re_id;
 	guint ret = 0;
 	guint i;
+	struct rspamd_re_cache_elt *cache_elt;
 
 	re_id = rspamd_regexp_get_cache_id (re);
 
@@ -690,6 +751,8 @@ rspamd_re_cache_process_regexp_data (struct rspamd_re_runtime *rt,
 		return ret;
 	}
 
+	cache_elt = (struct rspamd_re_cache_elt *)g_ptr_array_index (rt->cache->re, re_id);
+
 #ifndef WITH_HYPERSCAN
 	for (i = 0; i < count; i++) {
 		ret = rspamd_re_cache_process_pcre (rt,
@@ -697,20 +760,20 @@ rspamd_re_cache_process_regexp_data (struct rspamd_re_runtime *rt,
 				task,
 				in[i],
 				lens[i],
-				is_raw);
+				is_raw,
+				cache_elt->lua_cbref);
 		rt->results[re_id] = ret;
 	}
 
 	setbit (rt->checked, re_id);
 #else
-	struct rspamd_re_cache_elt *elt;
 	struct rspamd_re_class *re_class;
 	struct rspamd_re_hyperscan_cbdata cbdata;
 
-	elt = g_ptr_array_index (rt->cache->re, re_id);
+	cache_elt = g_ptr_array_index (rt->cache->re, re_id);
 	re_class = rspamd_regexp_get_class (re);
 
-	if (rt->cache->disable_hyperscan || elt->match_type == RSPAMD_RE_CACHE_PCRE ||
+	if (rt->cache->disable_hyperscan || cache_elt->match_type == RSPAMD_RE_CACHE_PCRE ||
 			!rt->has_hs || (is_raw && re_class->has_utf8)) {
 		for (i = 0; i < count; i++) {
 			ret = rspamd_re_cache_process_pcre (rt,
@@ -718,17 +781,20 @@ rspamd_re_cache_process_regexp_data (struct rspamd_re_runtime *rt,
 					task,
 					in[i],
 					lens[i],
-					is_raw);
+					is_raw,
+					cache_elt->lua_cbref);
 		}
 
 		setbit (rt->checked, re_id);
 	}
 	else {
 		for (i = 0; i < count; i ++) {
+			/* For Hyperscan we can probably safely disable all those limits */
+#if 0
 			if (rt->cache->max_re_data > 0 && lens[i] > rt->cache->max_re_data) {
 				lens[i] = rt->cache->max_re_data;
 			}
-
+#endif
 			rt->stat.bytes_scanned += lens[i];
 		}
 
@@ -871,19 +937,24 @@ rspamd_re_cache_process_selector (struct rspamd_task *task,
 						lua_tostring (L, -1));
 	}
 	else {
+		struct rspamd_lua_text *txt;
 		gsize slen;
 		const gchar *sel_data;
 
-		if (lua_type (L, -1) == LUA_TSTRING) {
-			sel_data = lua_tolstring (L, -1, &slen);
-			*n = 1;
-			*svec = g_malloc (sizeof (guchar *));
-			*lenvec = g_malloc (sizeof (guint));
-			(*svec)[0] = g_malloc (slen);
-			memcpy ((*svec)[0], sel_data, slen);
-			(*lenvec)[0] = slen;
+		if (lua_type (L, -1) != LUA_TTABLE) {
+			txt = lua_check_text_or_string (L, -1);
 
-			result = TRUE;
+			if (txt) {
+				sel_data = txt->start;
+				slen = txt->len;
+				*n = 1;
+				*svec = g_malloc (sizeof (guchar *));
+				*lenvec = g_malloc (sizeof (guint));
+				(*svec)[0] = g_malloc (slen);
+				memcpy ((*svec)[0], sel_data, slen);
+				(*lenvec)[0] = slen;
+				result = TRUE;
+			}
 		}
 		else {
 			*n = rspamd_lua_table_size (L, -1);
@@ -895,7 +966,16 @@ rspamd_re_cache_process_selector (struct rspamd_task *task,
 				for (guint i = 0; i < *n; i ++) {
 					lua_rawgeti (L, -1, i + 1);
 
-					sel_data = lua_tolstring (L, -1, &slen);
+					txt = lua_check_text_or_string (L, -1);
+					if (txt) {
+						sel_data = txt->start;
+						slen = txt->len;
+					}
+					else {
+						sel_data = "";
+						slen = 0;
+					}
+
 					(*svec)[i] = g_malloc (slen);
 					memcpy ((*svec)[i], sel_data, slen);
 					(*lenvec)[i] = slen;
@@ -1134,13 +1214,13 @@ rspamd_re_cache_exec_re (struct rspamd_task *task,
 				}
 				else {
 					/* Skip empty parts */
-					if (IS_PART_EMPTY (text_part)) {
+					if (IS_TEXT_PART_EMPTY (text_part)) {
 						len = 0;
 						in = "";
 					}
 					else {
 						/* Check raw flags */
-						if (!IS_PART_UTF (text_part)) {
+						if (!IS_TEXT_PART_UTF (text_part)) {
 							raw = TRUE;
 						}
 
@@ -1281,7 +1361,7 @@ rspamd_re_cache_exec_re (struct rspamd_task *task,
 				scvec[i + 1] = (guchar *)text_part->utf_stripped_content->data;
 				lenvec[i + 1] = text_part->utf_stripped_content->len;
 
-				if (!IS_PART_UTF (text_part)) {
+				if (!IS_TEXT_PART_UTF (text_part)) {
 					raw = TRUE;
 				}
 			}
@@ -1318,7 +1398,7 @@ rspamd_re_cache_exec_re (struct rspamd_task *task,
 					scvec[i] = (guchar *)text_part->parsed.begin;
 					lenvec[i] = text_part->parsed.len;
 
-					if (!IS_PART_UTF (text_part)) {
+					if (!IS_TEXT_PART_UTF (text_part)) {
 						raw = TRUE;
 					}
 				}
@@ -1703,7 +1783,6 @@ rspamd_re_cache_is_finite (struct rspamd_re_cache *cache,
 		/* Try to compile pattern */
 
 		gchar *pat = rspamd_re_cache_hs_pattern_from_pcre (re);
-		/* Memory leak here but ok since we do exit */
 
 		if (hs_compile (pat,
 				flags | HS_FLAG_PREFILTER,
@@ -1711,9 +1790,18 @@ rspamd_re_cache_is_finite (struct rspamd_re_cache *cache,
 				&cache->plt,
 				&test_db,
 				&hs_errors) != HS_SUCCESS) {
+
+			msg_info_re_cache ("cannot compile (prefilter mode) '%s' to hyperscan: '%s'",
+					pat,
+					hs_errors != NULL ? hs_errors->message : "unknown error");
+
+			hs_free_compile_error (hs_errors);
+			g_free (pat);
+
 			exit (EXIT_FAILURE);
 		}
 
+		g_free (pat);
 		exit (EXIT_SUCCESS);
 	}
 	else if (cld > 0) {
@@ -1911,7 +1999,12 @@ rspamd_re_cache_compile_timer_cb (EV_P_ ev_timer *w, int revents )
 		if (pcre_flags & PCRE_FLAG(DOTALL)) {
 			hs_flags[i] |= HS_FLAG_DOTALL;
 		}
-		if (rspamd_regexp_get_maxhits (re) == 1) {
+
+
+		if (re_flags & RSPAMD_REGEXP_FLAG_LEFTMOST) {
+			hs_flags[i] |= HS_FLAG_SOM_LEFTMOST;
+		}
+		else if (rspamd_regexp_get_maxhits (re) == 1) {
 			hs_flags[i] |= HS_FLAG_SINGLEMATCH;
 		}
 
@@ -1923,8 +2016,9 @@ rspamd_re_cache_compile_timer_cb (EV_P_ ev_timer *w, int revents )
 				&cache->plt,
 				&test_db,
 				&hs_errors) != HS_SUCCESS) {
-			msg_info_re_cache ("cannot compile %s to hyperscan, try prefilter match",
-					pat);
+			msg_info_re_cache ("cannot compile '%s' to hyperscan: '%s', try prefilter match",
+					pat,
+					hs_errors != NULL ? hs_errors->message : "unknown error");
 			hs_free_compile_error (hs_errors);
 
 			/* The approximation operation might take a significant
