@@ -5,11 +5,15 @@
 #include "ref.h"
 #include "addr.h"
 
+#ifdef  __cplusplus
+extern "C" {
+#endif
+
 struct rspamd_task;
 struct spf_resolved;
 
-typedef void (*spf_cb_t)(struct spf_resolved *record,
-		struct rspamd_task *task, gpointer cbdata);
+typedef void (*spf_cb_t) (struct spf_resolved *record,
+						  struct rspamd_task *task, gpointer cbdata);
 
 typedef enum spf_mech_e {
 	SPF_FAIL,
@@ -18,6 +22,21 @@ typedef enum spf_mech_e {
 	SPF_NEUTRAL
 } spf_mech_t;
 
+static inline gchar spf_mech_char (spf_mech_t mech)
+{
+	switch (mech) {
+	case SPF_FAIL:
+		return '-';
+	case SPF_SOFT_FAIL:
+		return '~';
+	case SPF_PASS:
+		return '+';
+	case SPF_NEUTRAL:
+	default:
+		return '?';
+	}
+}
+
 typedef enum spf_action_e {
 	SPF_RESOLVE_MX,
 	SPF_RESOLVE_A,
@@ -41,6 +60,12 @@ typedef enum spf_action_e {
 #define RSPAMD_SPF_FLAG_NA (1u << 9u)
 #define RSPAMD_SPF_FLAG_PERMFAIL (1u << 10u)
 #define RSPAMD_SPF_FLAG_RESOLVED (1u << 11u)
+#define RSPAMD_SPF_FLAG_CACHED (1u << 12u)
+
+/** Default SPF limits for avoiding abuse **/
+#define SPF_MAX_NESTING 10
+#define SPF_MAX_DNS_REQUESTS 30
+#define SPF_MIN_CACHE_TTL (60 * 5) /* 5 minutes */
 
 struct spf_addr {
 	guchar addr6[sizeof (struct in6_addr)];
@@ -58,39 +83,55 @@ struct spf_addr {
 	struct spf_addr *prev, *next;
 };
 
+enum rspamd_spf_resolved_flags {
+	RSPAMD_SPF_RESOLVED_NORMAL = 0,
+	RSPAMD_SPF_RESOLVED_TEMP_FAILED = (1u << 0u),
+	RSPAMD_SPF_RESOLVED_PERM_FAILED = (1u << 1u),
+	RSPAMD_SPF_RESOLVED_NA = (1u << 2u),
+};
+
 struct spf_resolved {
 	gchar *domain;
 	guint ttl;
-	gboolean temp_failed;
-	gboolean na;
-	gboolean perm_failed;
+	gint flags;
+	gdouble timestamp;
 	guint64 digest;
 	GArray *elts; /* Flat list of struct spf_addr */
 	ref_entry_t ref; /* Refcounting */
 };
 
+struct rspamd_spf_cred {
+	gchar *local_part;
+	gchar *domain;
+	gchar *sender;
+};
 
 /*
  * Resolve spf record for specified task and call a callback after resolution fails/succeed
  */
-gboolean rspamd_spf_resolve (struct rspamd_task *task, spf_cb_t callback,
-		gpointer cbdata);
+gboolean rspamd_spf_resolve (struct rspamd_task *task,
+							 spf_cb_t callback,
+							 gpointer cbdata,
+							 struct rspamd_spf_cred *cred);
 
 /*
  * Get a domain for spf for specified task
  */
-const gchar * rspamd_spf_get_domain (struct rspamd_task *task);
-
+const gchar *rspamd_spf_get_domain (struct rspamd_task *task);
 
+struct rspamd_spf_cred *rspamd_spf_get_cred (struct rspamd_task *task);
 /*
  * Increase refcount
  */
-struct spf_resolved * spf_record_ref (struct spf_resolved *rec);
-
+struct spf_resolved *_spf_record_ref (struct spf_resolved *rec, const gchar *loc);
+#define spf_record_ref(rec) \
+    _spf_record_ref ((rec), G_STRLOC)
 /*
  * Decrease refcount
  */
-void spf_record_unref (struct spf_resolved *rec);
+void _spf_record_unref (struct spf_resolved *rec, const gchar *loc);
+#define spf_record_unref(rec) \
+    _spf_record_unref((rec), G_STRLOC)
 
 /**
  * Prints address + mask in a freshly allocated string (must be freed)
@@ -105,6 +146,13 @@ gchar *spf_addr_mask_to_string (struct spf_addr *addr);
  * @param rec
  * @return
  */
-struct spf_addr * spf_addr_match_task (struct rspamd_task *task,
-									   struct spf_resolved *rec);
+struct spf_addr *spf_addr_match_task (struct rspamd_task *task,
+									  struct spf_resolved *rec);
+
+void spf_library_config (const ucl_object_t *obj);
+
+#ifdef  __cplusplus
+}
+#endif
+
 #endif
