@@ -111,7 +111,8 @@ rspamd_task_new (struct rspamd_worker *worker,
 	new_task->request_headers = kh_init (rspamd_req_headers_hash);
 	new_task->sock = -1;
 	new_task->flags |= (RSPAMD_TASK_FLAG_MIME);
-	new_task->result = rspamd_create_metric_result (new_task);
+	/* Default results chain */
+	rspamd_create_metric_result (new_task, NULL, -1);
 
 	new_task->queue_id = "undef";
 	new_task->messages = ucl_object_typed_new (UCL_OBJECT);
@@ -203,6 +204,10 @@ rspamd_task_free (struct rspamd_task *task)
 			rspamd_email_address_free (task->from_envelope);
 		}
 
+		if (task->from_envelope_orig) {
+			rspamd_email_address_free (task->from_envelope_orig);
+		}
+
 		if (task->meta_words) {
 			g_array_free (task->meta_words, TRUE);
 		}
@@ -716,24 +721,27 @@ rspamd_task_process (struct rspamd_task *task, guint stages)
 	st = rspamd_task_select_processing_stage (task, stages);
 
 	switch (st) {
+	case RSPAMD_TASK_STAGE_CONNFILTERS:
+		all_done = rspamd_symcache_process_symbols (task, task->cfg->cache, st);
+		break;
+
 	case RSPAMD_TASK_STAGE_READ_MESSAGE:
 		if (!rspamd_message_parse (task)) {
 			ret = FALSE;
 		}
 		break;
 
-	case RSPAMD_TASK_STAGE_PRE_FILTERS_EMPTY:
-	case RSPAMD_TASK_STAGE_PRE_FILTERS:
-	case RSPAMD_TASK_STAGE_FILTERS:
-		all_done = rspamd_symcache_process_symbols (task, task->cfg->cache, st);
-		break;
-
 	case RSPAMD_TASK_STAGE_PROCESS_MESSAGE:
 		if (!(task->flags & RSPAMD_TASK_FLAG_SKIP_PROCESS)) {
 			rspamd_message_process (task);
 		}
 		break;
 
+	case RSPAMD_TASK_STAGE_PRE_FILTERS:
+	case RSPAMD_TASK_STAGE_FILTERS:
+		all_done = rspamd_symcache_process_symbols (task, task->cfg->cache, st);
+		break;
+
 	case RSPAMD_TASK_STAGE_CLASSIFIERS:
 	case RSPAMD_TASK_STAGE_CLASSIFIERS_PRE:
 	case RSPAMD_TASK_STAGE_CLASSIFIERS_POST:
@@ -747,7 +755,8 @@ rspamd_task_process (struct rspamd_task *task, guint stages)
 		break;
 
 	case RSPAMD_TASK_STAGE_COMPOSITES:
-		rspamd_make_composites (task);
+		rspamd_composites_process_task (task);
+		task->result->nresults_postfilters = task->result->nresults;
 		break;
 
 	case RSPAMD_TASK_STAGE_POST_FILTERS:
@@ -805,8 +814,13 @@ rspamd_task_process (struct rspamd_task *task, guint stages)
 		}
 		break;
 	case RSPAMD_TASK_STAGE_COMPOSITES_POST:
-		/* Second run of composites processing before idempotent filters */
-		rspamd_make_composites (task);
+		/* Second run of composites processing before idempotent filters (if needed) */
+		if (task->result->nresults_postfilters != task->result->nresults) {
+			rspamd_composites_process_task (task);
+		}
+		else {
+			msg_debug_task ("skip second run of composites as the result has not been changed");
+		}
 		break;
 
 	case RSPAMD_TASK_STAGE_IDEMPOTENT:
@@ -1076,7 +1090,7 @@ rspamd_task_log_metric_res (struct rspamd_task *task,
 	khiter_t k;
 
 	mres = task->result;
-	act = rspamd_check_action_metric (task, NULL);
+	act = rspamd_check_action_metric (task, NULL, NULL);
 
 	if (mres != NULL) {
 		switch (lf->type) {
@@ -1802,8 +1816,8 @@ rspamd_task_stage_name (enum rspamd_task_stage stg)
 	case RSPAMD_TASK_STAGE_CONNECT:
 		ret = "connect";
 		break;
-	case RSPAMD_TASK_STAGE_ENVELOPE:
-		ret = "envelope";
+	case RSPAMD_TASK_STAGE_CONNFILTERS:
+		ret = "connection_filter";
 		break;
 	case RSPAMD_TASK_STAGE_READ_MESSAGE:
 		ret = "read_message";
@@ -1875,7 +1889,7 @@ rspamd_task_timeout (EV_P_ ev_timer *w, int revents)
 		if (task->cfg->soft_reject_on_timeout) {
 			struct rspamd_action *action, *soft_reject;
 
-			action = rspamd_check_action_metric (task, NULL);
+			action = rspamd_check_action_metric (task, NULL, NULL);
 
 			if (action->action_type != METRIC_ACTION_REJECT) {
 				soft_reject = rspamd_config_get_action_by_type (task->cfg,
@@ -1886,7 +1900,7 @@ rspamd_task_timeout (EV_P_ ev_timer *w, int revents)
 						NAN,
 						"timeout processing message",
 						"task timeout",
-						0);
+						0, NULL);
 			}
 		}
 
@@ -1904,7 +1918,7 @@ rspamd_task_timeout (EV_P_ ev_timer *w, int revents)
 		if (task->cfg->soft_reject_on_timeout) {
 			struct rspamd_action *action, *soft_reject;
 
-			action = rspamd_check_action_metric (task, NULL);
+			action = rspamd_check_action_metric (task, NULL, NULL);
 
 			if (action->action_type != METRIC_ACTION_REJECT) {
 				soft_reject = rspamd_config_get_action_by_type (task->cfg,
@@ -1915,7 +1929,7 @@ rspamd_task_timeout (EV_P_ ev_timer *w, int revents)
 						NAN,
 						"timeout post-processing message",
 						"task timeout",
-						0);
+						0, NULL);
 			}
 		}
 
