@@ -18,13 +18,19 @@
 
 #include "config.h"
 #include "util.h"
-#include "http_connection.h"
+#include "libserver/http/http_connection.h"
 #include "rspamd.h"
 
+#ifdef  __cplusplus
+extern "C" {
+#endif
+
 #ifndef HAVE_SA_SIGINFO
 typedef void (*rspamd_sig_handler_t) (gint);
 #else
+
 typedef void (*rspamd_sig_handler_t) (gint, siginfo_t *, void *);
+
 #endif
 
 struct rspamd_worker;
@@ -33,9 +39,12 @@ struct rspamd_worker_signal_handler;
 /**
  * Init basic signals for a worker
  * @param worker
- * @param base
+ * @param event_loop
  */
-void rspamd_worker_init_signals (struct rspamd_worker *worker, struct event_base *base);
+void rspamd_worker_init_signals (struct rspamd_worker *worker, struct ev_loop *event_loop);
+
+typedef void (*rspamd_accept_handler) (struct ev_loop *loop, ev_io *w, int revents);
+
 /**
  * Prepare worker's startup
  * @param worker worker structure
@@ -44,18 +53,26 @@ void rspamd_worker_init_signals (struct rspamd_worker *worker, struct event_base
  * @param accept_handler handler of accept event for listen sockets
  * @return event base suitable for a worker
  */
-struct event_base *
+struct ev_loop *
 rspamd_prepare_worker (struct rspamd_worker *worker, const char *name,
-	void (*accept_handler)(int, short, void *));
+					   rspamd_accept_handler hdl);
+
+/**
+ * Should be used to validate context for a worker as in assert like invocation
+ * @param ctx
+ * @param magic
+ * @return
+ */
+gboolean rspamd_worker_check_context (gpointer ctx, guint64 magic);
 
 /**
  * Set special signal handler for a worker
  */
 void rspamd_worker_set_signal_handler (int signo,
-		struct rspamd_worker *worker,
-		struct event_base *base,
-		rspamd_worker_signal_handler handler,
-		void *handler_data);
+									   struct rspamd_worker *worker,
+									   struct ev_loop *event_loop,
+									   rspamd_worker_signal_cb_t handler,
+									   void *handler_data);
 
 /**
  * Stop accepting new connections for a worker
@@ -64,9 +81,9 @@ void rspamd_worker_set_signal_handler (int signo,
 void rspamd_worker_stop_accept (struct rspamd_worker *worker);
 
 typedef gint (*rspamd_controller_func_t) (
-	struct rspamd_http_connection_entry *conn_ent,
-	struct rspamd_http_message *msg,
-	struct module_ctx *ctx);
+		struct rspamd_http_connection_entry *conn_ent,
+		struct rspamd_http_message *msg,
+		struct module_ctx *ctx);
 
 struct rspamd_custom_controller_command {
 	const gchar *command;
@@ -99,7 +116,7 @@ struct rspamd_controller_session {
  * @param error_msg error message
  */
 void rspamd_controller_send_error (struct rspamd_http_connection_entry *entry,
-	gint code, const gchar *error_msg, ...);
+								   gint code, const gchar *error_msg, ...);
 
 /**
  * Send a custom string using HTTP
@@ -107,7 +124,7 @@ void rspamd_controller_send_error (struct rspamd_http_connection_entry *entry,
  * @param str string to send
  */
 void rspamd_controller_send_string (struct rspamd_http_connection_entry *entry,
-	const gchar *str);
+									const gchar *str);
 
 /**
  * Send UCL using HTTP and JSON serialization
@@ -115,16 +132,14 @@ void rspamd_controller_send_string (struct rspamd_http_connection_entry *entry,
  * @param obj object to send
  */
 void rspamd_controller_send_ucl (struct rspamd_http_connection_entry *entry,
-	ucl_object_t *obj);
+								 ucl_object_t *obj);
 
 /**
  * Return worker's control structure by its type
  * @param type
  * @return worker's control structure or NULL
  */
-worker_t * rspamd_get_worker_by_type (struct rspamd_config *cfg, GQuark type);
-
-void rspamd_worker_stop_accept (struct rspamd_worker *worker);
+worker_t *rspamd_get_worker_by_type (struct rspamd_config *cfg, GQuark type);
 
 /**
  * Block signals before terminations
@@ -161,8 +176,8 @@ gboolean rspamd_worker_is_primary_controller (struct rspamd_worker *w);
  * @param w
  * @return
  */
-void * rspamd_worker_session_cache_new (struct rspamd_worker *w,
-		struct event_base *ev_base);
+void *rspamd_worker_session_cache_new (struct rspamd_worker *w,
+									   struct ev_loop *ev_base);
 
 /**
  * Adds a new session identified by pointer
@@ -172,7 +187,7 @@ void * rspamd_worker_session_cache_new (struct rspamd_worker *w,
  * @param ptr
  */
 void rspamd_worker_session_cache_add (void *cache, const gchar *tag,
-		guint *pref, void *ptr);
+									  guint *pref, void *ptr);
 
 /**
  * Removes session from cache
@@ -185,12 +200,15 @@ void rspamd_worker_session_cache_remove (void *cache, void *ptr);
  * Fork new worker with the specified configuration
  */
 struct rspamd_worker *rspamd_fork_worker (struct rspamd_main *,
-		struct rspamd_worker_conf *, guint idx, struct event_base *ev_base);
+										  struct rspamd_worker_conf *, guint idx,
+										  struct ev_loop *ev_base,
+										  rspamd_worker_term_cb term_handler,
+										  GHashTable *listen_sockets);
 
 /**
  * Sets crash signals handlers if compiled with libunwind
  */
-void rspamd_set_crash_handler (struct rspamd_main *);
+RSPAMD_NO_SANITIZE void rspamd_set_crash_handler (struct rspamd_main *);
 
 /**
  * Initialise the main monitoring worker
@@ -199,8 +217,69 @@ void rspamd_set_crash_handler (struct rspamd_main *);
  * @param resolver
  */
 void rspamd_worker_init_monitored (struct rspamd_worker *worker,
-		struct event_base *ev_base,
-		struct rspamd_dns_resolver *resolver);
+								   struct ev_loop *ev_base,
+								   struct rspamd_dns_resolver *resolver);
+
+/**
+ * Performs throttling for accept events
+ * @param sock
+ * @param data struct rspamd_worker_accept_event * list
+ */
+void rspamd_worker_throttle_accept_events (gint sock, void *data);
+
+/**
+ * Checks (and logs) the worker's termination status. Returns TRUE if a worker
+ * should be restarted.
+ * @param rspamd_main
+ * @param wrk
+ * @param status waitpid res
+ * @return TRUE if refork is desired
+ */
+gboolean rspamd_check_termination_clause (struct rspamd_main *rspamd_main,
+										  struct rspamd_worker *wrk, int status);
+
+/**
+ * Call for final scripts for a worker
+ * @param worker
+ * @return
+ */
+gboolean rspamd_worker_call_finish_handlers (struct rspamd_worker *worker);
+
+struct rspamd_rrd_file;
+/**
+ * Terminate controller worker
+ * @param worker
+ */
+void rspamd_controller_on_terminate (struct rspamd_worker *worker,
+		struct rspamd_rrd_file *rrd);
+
+/**
+ * Inits controller worker
+ * @param worker
+ * @param ev_base
+ * @param prrd
+ */
+void rspamd_worker_init_controller (struct rspamd_worker *worker,
+								   struct rspamd_rrd_file **prrd);
+
+/**
+ * Saves stats
+ * @param rspamd_main
+ * @param cfg
+ */
+void rspamd_controller_store_saved_stats (struct rspamd_main *rspamd_main,
+									 struct rspamd_config *cfg);
+
+#ifdef WITH_HYPERSCAN
+struct rspamd_control_command;
+
+gboolean rspamd_worker_hyperscan_ready (struct rspamd_main *rspamd_main,
+										struct rspamd_worker *worker, gint fd,
+										gint attached_fd,
+										struct rspamd_control_command *cmd,
+										gpointer ud);
+
+#endif
 
 #define msg_err_main(...) rspamd_default_log_function (G_LOG_LEVEL_CRITICAL, \
         rspamd_main->server_pool->tag.tagname, rspamd_main->server_pool->tag.uid, \
@@ -215,4 +294,8 @@ void rspamd_worker_init_monitored (struct rspamd_worker *worker,
         G_STRFUNC, \
         __VA_ARGS__)
 
+#ifdef  __cplusplus
+}
+#endif
+
 #endif /* WORKER_UTIL_H_ */
