@@ -22,7 +22,7 @@
 
 #ifdef WITH_HIREDIS
 #include "hiredis.h"
-#include "adapters/libevent.h"
+#include "adapters/libev.h"
 #include "ref.h"
 
 #define msg_debug_stat_redis(...)  rspamd_conditional_debug_fast (NULL, NULL, \
@@ -70,8 +70,9 @@ struct redis_stat_runtime {
 	struct redis_stat_ctx *ctx;
 	struct rspamd_task *task;
 	struct upstream *selected;
-	struct event timeout_event;
+	ev_timer timeout_event;
 	GArray *results;
+	GPtrArray *tokens;
 	struct rspamd_statfile_config *stcf;
 	gchar *redis_object_expanded;
 	redisAsyncContext *redis;
@@ -87,7 +88,7 @@ struct rspamd_redis_stat_cbdata;
 struct rspamd_redis_stat_elt {
 	struct redis_stat_ctx *ctx;
 	struct rspamd_stat_async_elt *async;
-	struct event_base *ev_base;
+	struct ev_loop *event_loop;
 	ucl_object_t *stat;
 	struct rspamd_redis_stat_cbdata *cbdata;
 };
@@ -148,7 +149,6 @@ rspamd_redis_expand_object (const gchar *pattern,
 	struct rspamd_statfile_config *stcf;
 	lua_State *L = NULL;
 	struct rspamd_task **ptask;
-	GString *tb;
 	const gchar *rcpt = NULL;
 	gint err_idx;
 
@@ -172,16 +172,15 @@ rspamd_redis_expand_object (const gchar *pattern,
 			rspamd_lua_setclass (L, "rspamd{task}", -1);
 
 			if (lua_pcall (L, 1, 1, err_idx) != 0) {
-				tb = lua_touserdata (L, -1);
-				msg_err_task ("call to user extraction script failed: %v", tb);
-				g_string_free (tb, TRUE);
+				msg_err_task ("call to user extraction script failed: %s",
+						lua_tostring (L, -1));
 			}
 			else {
 				rcpt = rspamd_mempool_strdup (task->task_pool, lua_tostring (L, -1));
 			}
 
 			/* Result + error function */
-			lua_pop (L, 2);
+			lua_settop (L, err_idx - 1);
 		}
 
 		if (rcpt) {
@@ -373,6 +372,12 @@ rspamd_redis_maybe_auth (struct redis_stat_ctx *ctx, redisAsyncContext *redis)
 	}
 }
 
+// the `b` conversion type character is unknown to gcc
+#ifdef __GNUC__
+#pragma GCC diagnostic push
+#pragma GCC diagnostic ignored "-Wformat"
+#pragma GCC diagnostic ignored "-Wformat-extra-args"
+#endif
 static rspamd_fstring_t *
 rspamd_redis_tokens_to_query (struct rspamd_task *task,
 		struct redis_stat_runtime *rt,
@@ -527,7 +532,8 @@ rspamd_redis_tokens_to_query (struct rspamd_task *task,
 								"HSET %b_tokens %b %b",
 								prefix, (size_t) prefix_len,
 								n0, (size_t) l0,
-								tok->t1->stemmed.begin, tok->t1->stemmed.len);
+								tok->t1->stemmed.begin,
+								tok->t1->stemmed.len);
 					}
 				}
 				else {
@@ -625,6 +631,9 @@ rspamd_redis_tokens_to_query (struct rspamd_task *task,
 
 	return out;
 }
+#ifdef __GNUC__
+#pragma GCC diagnostic pop
+#endif
 
 static void
 rspamd_redis_store_stat_signature (struct rspamd_task *task,
@@ -637,7 +646,6 @@ rspamd_redis_store_stat_signature (struct rspamd_task *task,
 	guint i, blen, klen;
 	rspamd_fstring_t *out;
 
-	out = rspamd_fstring_sized_new (1024);
 	sig = rspamd_mempool_get_variable (task->task_pool,
 			RSPAMD_MEMPOOL_STAT_SIGNATURE);
 
@@ -646,11 +654,10 @@ rspamd_redis_store_stat_signature (struct rspamd_task *task,
 		return;
 	}
 
+	out = rspamd_fstring_sized_new (1024);
 	klen = rspamd_snprintf (keybuf, sizeof (keybuf), "%s_%s_%s",
 			prefix, sig, rt->stcf->is_spam ? "S" : "H");
 
-	out->len = 0;
-
 	/* Cleanup key */
 	rspamd_printf_fstring (&out, ""
 					"*2\r\n"
@@ -751,12 +758,15 @@ rspamd_redis_async_cbdata_cleanup (struct rspamd_redis_stat_cbdata *cbdata)
 static void
 rspamd_redis_stat_learns (redisAsyncContext *c, gpointer r, gpointer priv)
 {
-	struct rspamd_redis_stat_cbdata *cbdata = priv;
+	struct rspamd_redis_stat_elt *redis_elt = (struct rspamd_redis_stat_elt *)priv;
+	struct rspamd_redis_stat_cbdata *cbdata;
 	redisReply *reply = r;
 	ucl_object_t *obj;
 	gulong num = 0;
 
-	if (cbdata->wanna_die) {
+	cbdata = redis_elt->cbdata;
+
+	if (cbdata == NULL || cbdata->wanna_die) {
 		return;
 	}
 
@@ -778,6 +788,7 @@ rspamd_redis_stat_learns (redisAsyncContext *c, gpointer r, gpointer priv)
 
 	if (cbdata->inflight == 0) {
 		rspamd_redis_async_cbdata_cleanup (cbdata);
+		redis_elt->cbdata = NULL;
 	}
 }
 
@@ -785,12 +796,15 @@ rspamd_redis_stat_learns (redisAsyncContext *c, gpointer r, gpointer priv)
 static void
 rspamd_redis_stat_key (redisAsyncContext *c, gpointer r, gpointer priv)
 {
-	struct rspamd_redis_stat_cbdata *cbdata = priv;
+	struct rspamd_redis_stat_elt *redis_elt = (struct rspamd_redis_stat_elt *)priv;
+	struct rspamd_redis_stat_cbdata *cbdata;
 	redisReply *reply = r;
 	ucl_object_t *obj;
 	glong num = 0;
 
-	if (cbdata->wanna_die) {
+	cbdata = redis_elt->cbdata;
+
+	if (cbdata == NULL || cbdata->wanna_die) {
 		return;
 	}
 
@@ -829,6 +843,7 @@ rspamd_redis_stat_key (redisAsyncContext *c, gpointer r, gpointer priv)
 
 	if (cbdata->inflight == 0) {
 		rspamd_redis_async_cbdata_cleanup (cbdata);
+		redis_elt->cbdata = NULL;
 	}
 }
 
@@ -836,13 +851,16 @@ rspamd_redis_stat_key (redisAsyncContext *c, gpointer r, gpointer priv)
 static void
 rspamd_redis_stat_keys (redisAsyncContext *c, gpointer r, gpointer priv)
 {
-	struct rspamd_redis_stat_cbdata *cbdata = priv;
-	redisReply *reply = r, *elt;
+	struct rspamd_redis_stat_elt *redis_elt = (struct rspamd_redis_stat_elt *)priv;
+	struct rspamd_redis_stat_cbdata *cbdata;
+	redisReply *reply = r, *more_elt, *elts, *elt;
 	gchar **pk, *k;
 	guint i, processed = 0;
+	gboolean more = false;
 
+	cbdata = redis_elt->cbdata;
 
-	if (cbdata->wanna_die) {
+	if (cbdata == NULL || cbdata->wanna_die) {
 		return;
 	}
 
@@ -850,10 +868,24 @@ rspamd_redis_stat_keys (redisAsyncContext *c, gpointer r, gpointer priv)
 
 	if (c->err == 0 && r != NULL) {
 		if (reply->type == REDIS_REPLY_ARRAY) {
-			g_ptr_array_set_size (cbdata->cur_keys, reply->elements);
+			more_elt = reply->element[0];
+			elts = reply->element[1];
 
-			for (i = 0; i < reply->elements; i ++) {
-				elt = reply->element[i];
+			if (more_elt != NULL && more_elt->str != NULL && strcmp (more_elt->str, "0") != 0) {
+				more = true;
+			}
+
+			/* Clear the existing stuff */
+			PTR_ARRAY_FOREACH (cbdata->cur_keys, i, k) {
+				if (k) {
+					g_free (k);
+				}
+			}
+
+			g_ptr_array_set_size (cbdata->cur_keys, elts->elements);
+
+			for (i = 0; i < elts->elements; i ++) {
+				elt = elts->element[i];
 
 				if (elt->type == REDIS_REPLY_STRING) {
 					pk = (gchar **)&g_ptr_array_index (cbdata->cur_keys, i);
@@ -861,12 +893,14 @@ rspamd_redis_stat_keys (redisAsyncContext *c, gpointer r, gpointer priv)
 					rspamd_strlcpy (*pk, elt->str, elt->len + 1);
 					processed ++;
 				}
+				else {
+					pk = (gchar **)&g_ptr_array_index (cbdata->cur_keys, i);
+					*pk = NULL;
+				}
 			}
 
 			if (processed) {
-				for (i = 0; i < cbdata->cur_keys->len; i ++) {
-					k = (gchar *)g_ptr_array_index (cbdata->cur_keys, i);
-
+				PTR_ARRAY_FOREACH (cbdata->cur_keys, i, k) {
 					if (k) {
 						const gchar *learned_key = "learns";
 
@@ -879,7 +913,7 @@ rspamd_redis_stat_keys (redisAsyncContext *c, gpointer r, gpointer priv)
 							}
 							redisAsyncCommand (cbdata->redis,
 									rspamd_redis_stat_learns,
-									cbdata,
+									redis_elt,
 									"HGET %s %s",
 									k, learned_key);
 							cbdata->inflight += 1;
@@ -887,12 +921,12 @@ rspamd_redis_stat_keys (redisAsyncContext *c, gpointer r, gpointer priv)
 						else {
 							redisAsyncCommand (cbdata->redis,
 									rspamd_redis_stat_key,
-									cbdata,
+									redis_elt,
 									"HLEN %s",
 									k);
 							redisAsyncCommand (cbdata->redis,
 									rspamd_redis_stat_learns,
-									cbdata,
+									redis_elt,
 									"HGET %s %s",
 									k, learned_key);
 							cbdata->inflight += 2;
@@ -902,29 +936,40 @@ rspamd_redis_stat_keys (redisAsyncContext *c, gpointer r, gpointer priv)
 			}
 		}
 
-		/* Set up the required keys */
-		ucl_object_insert_key (cbdata->cur,
-				ucl_object_typed_new (UCL_INT), "revision", 0, false);
-		ucl_object_insert_key (cbdata->cur,
-				ucl_object_typed_new (UCL_INT), "used", 0, false);
-		ucl_object_insert_key (cbdata->cur,
-				ucl_object_typed_new (UCL_INT), "total", 0, false);
-		ucl_object_insert_key (cbdata->cur,
-				ucl_object_typed_new (UCL_INT), "size", 0, false);
-		ucl_object_insert_key (cbdata->cur,
-				ucl_object_fromstring (cbdata->elt->ctx->stcf->symbol),
-				"symbol", 0, false);
-		ucl_object_insert_key (cbdata->cur, ucl_object_fromstring ("redis"),
-				"type", 0, false);
-		ucl_object_insert_key (cbdata->cur, ucl_object_fromint (0),
-				"languages", 0, false);
-		ucl_object_insert_key (cbdata->cur, ucl_object_fromint (processed),
-				"users", 0, false);
+		if (more) {
+			/* Get more stat keys */
+			redisAsyncCommand (cbdata->redis, rspamd_redis_stat_keys, redis_elt,
+					"SSCAN %s_keys %s COUNT 1000",
+					cbdata->elt->ctx->stcf->symbol, more_elt->str);
 
-		rspamd_upstream_ok (cbdata->selected);
+			cbdata->inflight += 1;
+		}
+		else {
+			/* Set up the required keys */
+			ucl_object_insert_key (cbdata->cur,
+					ucl_object_typed_new (UCL_INT), "revision", 0, false);
+			ucl_object_insert_key (cbdata->cur,
+					ucl_object_typed_new (UCL_INT), "used", 0, false);
+			ucl_object_insert_key (cbdata->cur,
+					ucl_object_typed_new (UCL_INT), "total", 0, false);
+			ucl_object_insert_key (cbdata->cur,
+					ucl_object_typed_new (UCL_INT), "size", 0, false);
+			ucl_object_insert_key (cbdata->cur,
+					ucl_object_fromstring (cbdata->elt->ctx->stcf->symbol),
+					"symbol", 0, false);
+			ucl_object_insert_key (cbdata->cur, ucl_object_fromstring ("redis"),
+					"type", 0, false);
+			ucl_object_insert_key (cbdata->cur, ucl_object_fromint (0),
+					"languages", 0, false);
+			ucl_object_insert_key (cbdata->cur, ucl_object_fromint (processed),
+					"users", 0, false);
 
-		if (cbdata->inflight == 0) {
-			rspamd_redis_async_cbdata_cleanup (cbdata);
+			rspamd_upstream_ok (cbdata->selected);
+
+			if (cbdata->inflight == 0) {
+				rspamd_redis_async_cbdata_cleanup (cbdata);
+				redis_elt->cbdata = NULL;
+			}
 		}
 	}
 	else {
@@ -935,8 +980,9 @@ rspamd_redis_stat_keys (redisAsyncContext *c, gpointer r, gpointer priv)
 			msg_err ("cannot get keys to gather stat: unknown error");
 		}
 
-		rspamd_upstream_fail (cbdata->selected, FALSE);
+		rspamd_upstream_fail (cbdata->selected, FALSE, c->errstr);
 		rspamd_redis_async_cbdata_cleanup (cbdata);
+		redis_elt->cbdata = NULL;
 	}
 }
 
@@ -948,6 +994,8 @@ rspamd_redis_async_stat_cb (struct rspamd_stat_async_elt *elt, gpointer d)
 	struct rspamd_redis_stat_cbdata *cbdata;
 	rspamd_inet_addr_t *addr;
 	struct upstream_list *ups;
+	redisAsyncContext *redis_ctx;
+	struct upstream *selected;
 
 	g_assert (redis_elt != NULL);
 
@@ -956,6 +1004,7 @@ rspamd_redis_async_stat_cb (struct rspamd_stat_async_elt *elt, gpointer d)
 	if (redis_elt->cbdata) {
 		/* We have some other process pending */
 		rspamd_redis_async_cbdata_cleanup (redis_elt->cbdata);
+		redis_elt->cbdata = NULL;
 	}
 
 	/* Disable further events unless needed */
@@ -967,40 +1016,54 @@ rspamd_redis_async_stat_cb (struct rspamd_stat_async_elt *elt, gpointer d)
 		return;
 	}
 
-	cbdata = g_malloc0 (sizeof (*cbdata));
-
-	cbdata->selected = rspamd_upstream_get (ups,
+	selected = rspamd_upstream_get (ups,
 					RSPAMD_UPSTREAM_ROUND_ROBIN,
 					NULL,
 					0);
 
-	g_assert (cbdata->selected != NULL);
-	addr = rspamd_upstream_addr_next (cbdata->selected);
+	g_assert (selected != NULL);
+	addr = rspamd_upstream_addr_next (selected);
 	g_assert (addr != NULL);
 
 	if (rspamd_inet_address_get_af (addr) == AF_UNIX) {
-		cbdata->redis = redisAsyncConnectUnix (rspamd_inet_address_to_string (addr));
+		redis_ctx = redisAsyncConnectUnix (rspamd_inet_address_to_string (addr));
 	}
 	else {
-		cbdata->redis = redisAsyncConnect (rspamd_inet_address_to_string (addr),
+		redis_ctx = redisAsyncConnect (rspamd_inet_address_to_string (addr),
 				rspamd_inet_address_get_port (addr));
 	}
 
-	g_assert (cbdata->redis != NULL);
+	if (redis_ctx == NULL) {
+		msg_warn ("cannot connect to redis server %s: %s",
+				rspamd_inet_address_to_string_pretty (addr),
+				strerror (errno));
 
-	redisLibeventAttach (cbdata->redis, redis_elt->ev_base);
+		return;
+	}
+	else if (redis_ctx->err != REDIS_OK) {
+		msg_warn ("cannot connect to redis server %s: %s",
+				rspamd_inet_address_to_string_pretty (addr),
+				redis_ctx->errstr);
+		redisAsyncFree (redis_ctx);
 
+		return;
+	}
+
+	redisLibevAttach (redis_elt->event_loop, redis_ctx);
+	cbdata = g_malloc0 (sizeof (*cbdata));
+	cbdata->redis = redis_ctx;
+	cbdata->selected = selected;
 	cbdata->inflight = 1;
 	cbdata->cur = ucl_object_typed_new (UCL_OBJECT);
 	cbdata->elt = redis_elt;
-	cbdata->cur_keys = g_ptr_array_new ();
+	cbdata->cur_keys = g_ptr_array_sized_new (1000);
 	redis_elt->cbdata = cbdata;
 
 	/* XXX: deal with timeouts maybe */
 	/* Get keys in redis that match our symbol */
 	rspamd_redis_maybe_auth (ctx, cbdata->redis);
-	redisAsyncCommand (cbdata->redis, rspamd_redis_stat_keys, cbdata,
-			"SMEMBERS %s_keys",
+	redisAsyncCommand (cbdata->redis, rspamd_redis_stat_keys, redis_elt,
+			"SSCAN %s_keys 0 COUNT 1000",
 			ctx->stcf->symbol);
 }
 
@@ -1009,7 +1072,10 @@ rspamd_redis_async_stat_fin (struct rspamd_stat_async_elt *elt, gpointer d)
 {
 	struct rspamd_redis_stat_elt *redis_elt = elt->ud;
 
-	rspamd_redis_async_cbdata_cleanup (redis_elt->cbdata);
+	if (redis_elt->cbdata != NULL) {
+		rspamd_redis_async_cbdata_cleanup (redis_elt->cbdata);
+		redis_elt->cbdata = NULL;
+	}
 }
 
 /* Called on connection termination */
@@ -1019,10 +1085,20 @@ rspamd_redis_fin (gpointer data)
 	struct redis_stat_runtime *rt = REDIS_RUNTIME (data);
 	redisAsyncContext *redis;
 
-	rt->has_event = FALSE;
+	if (rt->has_event) {
+		/* Should not happen ! */
+		msg_err ("FIXME: this code path should not be reached!");
+		rspamd_session_remove_event (rt->task->s, NULL, rt);
+		rt->has_event = FALSE;
+	}
 	/* Stop timeout */
-	if (rspamd_event_pending (&rt->timeout_event, EV_TIMEOUT)) {
-		event_del (&rt->timeout_event);
+	if (ev_can_stop (&rt->timeout_event)) {
+		ev_timer_stop (rt->task->event_loop, &rt->timeout_event);
+	}
+
+	if (rt->tokens) {
+		g_ptr_array_unref (rt->tokens);
+		rt->tokens = NULL;
 	}
 
 	if (rt->redis) {
@@ -1031,32 +1107,16 @@ rspamd_redis_fin (gpointer data)
 		/* This calls for all callbacks pending */
 		redisAsyncFree (redis);
 	}
-}
 
-static void
-rspamd_redis_fin_learn (gpointer data)
-{
-	struct redis_stat_runtime *rt = REDIS_RUNTIME (data);
-	redisAsyncContext *redis;
-
-	rt->has_event = FALSE;
-	/* Stop timeout */
-	if (rspamd_event_pending (&rt->timeout_event, EV_TIMEOUT)) {
-		event_del (&rt->timeout_event);
-	}
-
-	if (rt->redis) {
-		redis = rt->redis;
-		rt->redis = NULL;
-		/* This calls for all callbacks pending */
-		redisAsyncFree (redis);
+	if (rt->err) {
+		g_error_free (rt->err);
 	}
 }
 
 static void
-rspamd_redis_timeout (gint fd, short what, gpointer d)
+rspamd_redis_timeout (EV_P_ ev_timer *w, int revents)
 {
-	struct redis_stat_runtime *rt = REDIS_RUNTIME (d);
+	struct redis_stat_runtime *rt = REDIS_RUNTIME (w->data);
 	struct rspamd_task *task;
 	redisAsyncContext *redis;
 
@@ -1065,7 +1125,7 @@ rspamd_redis_timeout (gint fd, short what, gpointer d)
 	msg_err_task_check ("connection to redis server %s timed out",
 			rspamd_upstream_name (rt->selected));
 
-	rspamd_upstream_fail (rt->selected, FALSE);
+	rspamd_upstream_fail (rt->selected, FALSE, "timeout");
 
 	if (rt->redis) {
 		redis = rt->redis;
@@ -1074,11 +1134,112 @@ rspamd_redis_timeout (gint fd, short what, gpointer d)
 		redisAsyncFree (redis);
 	}
 
+	if (rt->tokens) {
+		g_ptr_array_unref (rt->tokens);
+		rt->tokens = NULL;
+	}
+
 	if (!rt->err) {
 		g_set_error (&rt->err, rspamd_redis_stat_quark (), ETIMEDOUT,
 				"error getting reply from redis server %s: timeout",
 				rspamd_upstream_name (rt->selected));
 	}
+	if (rt->has_event) {
+		rt->has_event = FALSE;
+		rspamd_session_remove_event (task->s, NULL, rt);
+	}
+}
+
+/* Called when we have received tokens values from redis */
+static void
+rspamd_redis_processed (redisAsyncContext *c, gpointer r, gpointer priv)
+{
+	struct redis_stat_runtime *rt = REDIS_RUNTIME (priv);
+	redisReply *reply = r, *elt;
+	struct rspamd_task *task;
+	rspamd_token_t *tok;
+	guint i, processed = 0, found = 0;
+	gulong val;
+	gdouble float_val;
+
+	task = rt->task;
+
+	if (c->err == 0 && rt->has_event) {
+		if (r != NULL) {
+			if (reply->type == REDIS_REPLY_ARRAY) {
+
+				if (reply->elements == task->tokens->len) {
+					for (i = 0; i < reply->elements; i ++) {
+						tok = g_ptr_array_index (task->tokens, i);
+						elt = reply->element[i];
+
+						if (G_UNLIKELY (elt->type == REDIS_REPLY_INTEGER)) {
+							tok->values[rt->id] = elt->integer;
+							found ++;
+						}
+						else if (elt->type == REDIS_REPLY_STRING) {
+							if (rt->stcf->clcf->flags &
+								RSPAMD_FLAG_CLASSIFIER_INTEGER) {
+								rspamd_strtoul (elt->str, elt->len, &val);
+								tok->values[rt->id] = val;
+							}
+							else {
+								float_val = strtof (elt->str, NULL);
+								tok->values[rt->id] = float_val;
+							}
+
+							found ++;
+						}
+						else {
+							tok->values[rt->id] = 0;
+						}
+
+						processed ++;
+					}
+
+					if (rt->stcf->is_spam) {
+						task->flags |= RSPAMD_TASK_FLAG_HAS_SPAM_TOKENS;
+					}
+					else {
+						task->flags |= RSPAMD_TASK_FLAG_HAS_HAM_TOKENS;
+					}
+				}
+				else {
+					msg_err_task_check ("got invalid length of reply vector from redis: "
+										"%d, expected: %d",
+							(gint)reply->elements,
+							(gint)task->tokens->len);
+				}
+			}
+			else {
+				msg_err_task_check ("got invalid reply from redis: %s, array expected",
+						rspamd_redis_type_to_string (reply->type));
+			}
+
+			msg_debug_stat_redis ("received tokens for %s: %d processed, %d found",
+					rt->redis_object_expanded, processed, found);
+			rspamd_upstream_ok (rt->selected);
+		}
+	}
+	else {
+		msg_err_task ("error getting reply from redis server %s: %s",
+				rspamd_upstream_name (rt->selected), c->errstr);
+
+		if (rt->redis) {
+			rspamd_upstream_fail (rt->selected, FALSE, c->errstr);
+		}
+
+		if (!rt->err) {
+			g_set_error (&rt->err, rspamd_redis_stat_quark (), c->err,
+					"cannot get values: error getting reply from redis server %s: %s",
+					rspamd_upstream_name (rt->selected), c->errstr);
+		}
+	}
+
+	if (rt->has_event) {
+		rt->has_event = FALSE;
+		rspamd_session_remove_event (task->s, NULL, rt);
+	}
 }
 
 /* Called when we have connected to the redis server and got stats */
@@ -1089,10 +1250,11 @@ rspamd_redis_connected (redisAsyncContext *c, gpointer r, gpointer priv)
 	redisReply *reply = r;
 	struct rspamd_task *task;
 	glong val = 0;
+	gboolean final = TRUE;
 
 	task = rt->task;
 
-	if (c->err == 0) {
+	if (c->err == 0 && rt->has_event) {
 		if (r != NULL) {
 			if (G_UNLIKELY (reply->type == REDIS_REPLY_INTEGER)) {
 				val = reply->integer;
@@ -1120,12 +1282,80 @@ rspamd_redis_connected (redisAsyncContext *c, gpointer r, gpointer priv)
 			msg_debug_stat_redis ("connected to redis server, tokens learned for %s: %uL",
 					rt->redis_object_expanded, rt->learned);
 			rspamd_upstream_ok (rt->selected);
+
+			/* Save learn count in mempool variable */
+			gint64 *learns_cnt;
+			const gchar *var_name;
+
+			if (rt->stcf->is_spam) {
+				var_name = RSPAMD_MEMPOOL_SPAM_LEARNS;
+			}
+			else {
+				var_name = RSPAMD_MEMPOOL_HAM_LEARNS;
+			}
+
+			learns_cnt = rspamd_mempool_get_variable (task->task_pool,
+					var_name);
+
+			if (learns_cnt) {
+				(*learns_cnt) += rt->learned;
+			}
+			else {
+				learns_cnt = rspamd_mempool_alloc (task->task_pool,
+						sizeof (*learns_cnt));
+				*learns_cnt = rt->learned;
+				rspamd_mempool_set_variable (task->task_pool,
+						var_name,
+						learns_cnt, NULL);
+			}
+
+			if (rt->learned >= rt->stcf->clcf->min_learns && rt->learned > 0) {
+				rspamd_fstring_t *query = rspamd_redis_tokens_to_query (
+						task,
+						rt,
+						rt->tokens,
+						rt->ctx->new_schema ? "HGET" : "HMGET",
+						rt->redis_object_expanded, FALSE, -1,
+						rt->stcf->clcf->flags & RSPAMD_FLAG_CLASSIFIER_INTEGER);
+				g_assert (query != NULL);
+				rspamd_mempool_add_destructor (task->task_pool,
+						(rspamd_mempool_destruct_t)rspamd_fstring_free, query);
+
+				int ret = redisAsyncFormattedCommand (rt->redis,
+						rspamd_redis_processed, rt,
+						query->str, query->len);
+
+				if (ret != REDIS_OK) {
+					msg_err_task ("call to redis failed: %s", rt->redis->errstr);
+				}
+				else {
+					/* Further is handled by rspamd_redis_processed */
+					final = FALSE;
+					/* Restart timeout */
+					if (ev_can_stop (&rt->timeout_event)) {
+						rt->timeout_event.repeat = rt->ctx->timeout;
+						ev_timer_again (task->event_loop, &rt->timeout_event);
+					}
+					else {
+						rt->timeout_event.data = rt;
+						ev_timer_init (&rt->timeout_event, rspamd_redis_timeout,
+								rt->ctx->timeout, 0.);
+						ev_timer_start (task->event_loop, &rt->timeout_event);
+					}
+				}
+			}
+			else {
+				msg_warn_task ("skip obtaining bayes tokens for %s of classifier "
+							   "%s: not enough learns %d; %d required",
+						rt->stcf->symbol, rt->stcf->clcf->name,
+						(int)rt->learned, rt->stcf->clcf->min_learns);
+			}
 		}
 	}
-	else {
+	else if (rt->has_event) {
 		msg_err_task ("error getting reply from redis server %s: %s",
 				rspamd_upstream_name (rt->selected), c->errstr);
-		rspamd_upstream_fail (rt->selected, FALSE);
+		rspamd_upstream_fail (rt->selected, FALSE,  c->errstr);
 
 		if (!rt->err) {
 			g_set_error (&rt->err, rspamd_redis_stat_quark (), c->err,
@@ -1134,96 +1364,9 @@ rspamd_redis_connected (redisAsyncContext *c, gpointer r, gpointer priv)
 		}
 	}
 
-}
-
-/* Called when we have received tokens values from redis */
-static void
-rspamd_redis_processed (redisAsyncContext *c, gpointer r, gpointer priv)
-{
-	struct redis_stat_runtime *rt = REDIS_RUNTIME (priv);
-	redisReply *reply = r, *elt;
-	struct rspamd_task *task;
-	rspamd_token_t *tok;
-	guint i, processed = 0, found = 0;
-	gulong val;
-	gdouble float_val;
-
-	task = rt->task;
-
-	if (c->err == 0) {
-		if (r != NULL) {
-			if (reply->type == REDIS_REPLY_ARRAY) {
-
-				if (reply->elements == task->tokens->len) {
-					for (i = 0; i < reply->elements; i ++) {
-						tok = g_ptr_array_index (task->tokens, i);
-						elt = reply->element[i];
-
-						if (G_UNLIKELY (elt->type == REDIS_REPLY_INTEGER)) {
-							tok->values[rt->id] = elt->integer;
-							found ++;
-						}
-						else if (elt->type == REDIS_REPLY_STRING) {
-							if (rt->stcf->clcf->flags &
-									RSPAMD_FLAG_CLASSIFIER_INTEGER) {
-								rspamd_strtoul (elt->str, elt->len, &val);
-								tok->values[rt->id] = val;
-							}
-							else {
-								float_val = strtod (elt->str, NULL);
-								tok->values[rt->id] = float_val;
-							}
-
-							found ++;
-						}
-						else {
-							tok->values[rt->id] = 0;
-						}
-
-						processed ++;
-					}
-
-					if (rt->stcf->is_spam) {
-						task->flags |= RSPAMD_TASK_FLAG_HAS_SPAM_TOKENS;
-					}
-					else {
-						task->flags |= RSPAMD_TASK_FLAG_HAS_HAM_TOKENS;
-					}
-				}
-				else {
-					msg_err_task_check ("got invalid length of reply vector from redis: "
-							"%d, expected: %d",
-							(gint)reply->elements,
-							(gint)task->tokens->len);
-				}
-			}
-			else {
-				msg_err_task_check ("got invalid reply from redis: %s, array expected",
-						rspamd_redis_type_to_string (reply->type));
-			}
-
-			msg_debug_stat_redis ("received tokens for %s: %d processed, %d found",
-					rt->redis_object_expanded, processed, found);
-			rspamd_upstream_ok (rt->selected);
-		}
-	}
-	else {
-		msg_err_task ("error getting reply from redis server %s: %s",
-				rspamd_upstream_name (rt->selected), c->errstr);
-
-		if (rt->redis) {
-			rspamd_upstream_fail (rt->selected, FALSE);
-		}
-
-		if (!rt->err) {
-			g_set_error (&rt->err, rspamd_redis_stat_quark (), c->err,
-					"cannot get values: error getting reply from redis server %s: %s",
-					rspamd_upstream_name (rt->selected), c->errstr);
-		}
-	}
-
-	if (rt->has_event) {
-		rspamd_session_remove_event (task->s, rspamd_redis_fin, rt);
+	if (final && rt->has_event) {
+		rt->has_event = FALSE;
+		rspamd_session_remove_event (task->s, NULL, rt);
 	}
 }
 
@@ -1244,7 +1387,7 @@ rspamd_redis_learned (redisAsyncContext *c, gpointer r, gpointer priv)
 				rspamd_upstream_name (rt->selected), c->errstr);
 
 		if (rt->redis) {
-			rspamd_upstream_fail (rt->selected, FALSE);
+			rspamd_upstream_fail (rt->selected, FALSE, c->errstr);
 		}
 
 		if (!rt->err) {
@@ -1255,7 +1398,8 @@ rspamd_redis_learned (redisAsyncContext *c, gpointer r, gpointer priv)
 	}
 
 	if (rt->has_event) {
-		rspamd_session_remove_event (task->s, rspamd_redis_fin_learn, rt);
+		rt->has_event = FALSE;
+		rspamd_session_remove_event (task->s, NULL, rt);
 	}
 }
 static void
@@ -1446,7 +1590,7 @@ rspamd_redis_init (struct rspamd_stat_ctx *ctx,
 	backend->stcf = stf;
 
 	st_elt = g_malloc0 (sizeof (*st_elt));
-	st_elt->ev_base = ctx->ev_base;
+	st_elt->event_loop = ctx->event_loop;
 	st_elt->ctx = backend;
 	backend->stat_elt = rspamd_stat_ctx_register_async (
 			rspamd_redis_async_stat_cb,
@@ -1507,15 +1651,14 @@ rspamd_redis_runtime (struct rspamd_task *task,
 
 	if (rspamd_redis_expand_object (ctx->redis_object, ctx, task,
 			&object_expanded) == 0) {
-		msg_err_task ("expansion for learning failed for symbol %s "
+		msg_err_task ("expansion for %s failed for symbol %s "
 				 "(maybe learning per user classifier with no user or recipient)",
+				 learn ? "learning" : "classifying",
 				 stcf->symbol);
 		return NULL;
 	}
 
 	rt = rspamd_mempool_alloc0 (task->task_pool, sizeof (*rt));
-	rspamd_mempool_add_destructor (task->task_pool,
-			rspamd_gerror_free_maybe, &rt->err);
 	rt->selected = up;
 	rt->task = task;
 	rt->ctx = ctx;
@@ -1534,13 +1677,26 @@ rspamd_redis_runtime (struct rspamd_task *task,
 	}
 
 	if (rt->redis == NULL) {
-		msg_err_task ("cannot connect redis");
+		msg_warn_task ("cannot connect to redis server %s: %s",
+				rspamd_inet_address_to_string_pretty (addr),
+				strerror (errno));
+		return NULL;
+	}
+	else if (rt->redis->err != REDIS_OK) {
+		msg_warn_task ("cannot connect to redis server %s: %s",
+				rspamd_inet_address_to_string_pretty (addr),
+				rt->redis->errstr);
+		redisAsyncFree (rt->redis);
+		rt->redis = NULL;
+
 		return NULL;
 	}
 
-	redisLibeventAttach (rt->redis, task->ev_base);
+	redisLibevAttach (task->event_loop, rt->redis);
 	rspamd_redis_maybe_auth (ctx, rt->redis);
 
+	rspamd_mempool_add_destructor (task->task_pool, rspamd_redis_fin, rt);
+
 	return rt;
 }
 
@@ -1563,9 +1719,6 @@ rspamd_redis_process_tokens (struct rspamd_task *task,
 		gint id, gpointer p)
 {
 	struct redis_stat_runtime *rt = REDIS_RUNTIME (p);
-	rspamd_fstring_t *query;
-	struct timeval tv;
-	gint ret;
 	const gchar *learned_key = "learns";
 
 	if (rspamd_session_blocked (task->s)) {
@@ -1590,33 +1743,19 @@ rspamd_redis_process_tokens (struct rspamd_task *task,
 	if (redisAsyncCommand (rt->redis, rspamd_redis_connected, rt, "HGET %s %s",
 			rt->redis_object_expanded, learned_key) == REDIS_OK) {
 
-		rspamd_session_add_event (task->s, rspamd_redis_fin, rt, M);
+		rspamd_session_add_event (task->s, NULL, rt, M);
 		rt->has_event = TRUE;
+		rt->tokens = g_ptr_array_ref (tokens);
 
-		if (rspamd_event_pending (&rt->timeout_event, EV_TIMEOUT)) {
-			event_del (&rt->timeout_event);
-		}
-		event_set (&rt->timeout_event, -1, EV_TIMEOUT, rspamd_redis_timeout, rt);
-		event_base_set (task->ev_base, &rt->timeout_event);
-		double_to_tv (rt->ctx->timeout, &tv);
-		event_add (&rt->timeout_event, &tv);
-
-		query = rspamd_redis_tokens_to_query (task, rt, tokens,
-				rt->ctx->new_schema ? "HGET" : "HMGET",
-				rt->redis_object_expanded, FALSE, -1,
-				rt->stcf->clcf->flags & RSPAMD_FLAG_CLASSIFIER_INTEGER);
-		g_assert (query != NULL);
-		rspamd_mempool_add_destructor (task->task_pool,
-				(rspamd_mempool_destruct_t)rspamd_fstring_free, query);
-
-		ret = redisAsyncFormattedCommand (rt->redis, rspamd_redis_processed, rt,
-				query->str, query->len);
-
-		if (ret == REDIS_OK) {
-			return TRUE;
+		if (ev_can_stop (&rt->timeout_event)) {
+			rt->timeout_event.repeat = rt->ctx->timeout;
+			ev_timer_again (task->event_loop, &rt->timeout_event);
 		}
 		else {
-			msg_err_task ("call to redis failed: %s", rt->redis->errstr);
+			rt->timeout_event.data = rt;
+			ev_timer_init (&rt->timeout_event, rspamd_redis_timeout,
+					rt->ctx->timeout, 0.);
+			ev_timer_start (task->event_loop, &rt->timeout_event);
 		}
 	}
 
@@ -1628,22 +1767,18 @@ rspamd_redis_finalize_process (struct rspamd_task *task, gpointer runtime,
 		gpointer ctx)
 {
 	struct redis_stat_runtime *rt = REDIS_RUNTIME (runtime);
-	redisAsyncContext *redis;
-
-	if (rspamd_event_pending (&rt->timeout_event, EV_TIMEOUT)) {
-		event_del (&rt->timeout_event);
-	}
-
-	if (rt->redis) {
-		redis = rt->redis;
-		rt->redis = NULL;
-		redisAsyncFree (redis);
-	}
 
 	if (rt->err) {
+		msg_info_task ("cannot retrieve stat tokens from Redis: %e", rt->err);
+		g_error_free (rt->err);
+		rt->err = NULL;
+		rspamd_redis_fin (rt);
+
 		return FALSE;
 	}
 
+	rspamd_redis_fin (rt);
+
 	return TRUE;
 }
 
@@ -1652,10 +1787,6 @@ rspamd_redis_learn_tokens (struct rspamd_task *task, GPtrArray *tokens,
 		gint id, gpointer p)
 {
 	struct redis_stat_runtime *rt = REDIS_RUNTIME (p);
-	struct upstream *up;
-	struct upstream_list *ups;
-	rspamd_inet_addr_t *addr;
-	struct timeval tv;
 	rspamd_fstring_t *query;
 	const gchar *redis_cmd;
 	rspamd_token_t *tok;
@@ -1667,23 +1798,6 @@ rspamd_redis_learn_tokens (struct rspamd_task *task, GPtrArray *tokens,
 		return FALSE;
 	}
 
-	ups = rspamd_redis_get_servers (rt->ctx, "write_servers");
-
-	if (!ups) {
-		return FALSE;
-	}
-	up = rspamd_upstream_get (ups,
-			RSPAMD_UPSTREAM_MASTER_SLAVE,
-			NULL,
-			0);
-
-	if (up == NULL) {
-		msg_err_task ("no upstreams reachable");
-		return FALSE;
-	}
-
-	rt->selected = up;
-
 	if (rt->ctx->new_schema) {
 		if (rt->ctx->stcf->is_spam) {
 			learned_key = "learns_spam";
@@ -1693,22 +1807,6 @@ rspamd_redis_learn_tokens (struct rspamd_task *task, GPtrArray *tokens,
 		}
 	}
 
-	addr = rspamd_upstream_addr_next (up);
-	g_assert (addr != NULL);
-
-	if (rspamd_inet_address_get_af (addr) == AF_UNIX) {
-		rt->redis = redisAsyncConnectUnix (rspamd_inet_address_to_string (addr));
-	}
-	else {
-		rt->redis = redisAsyncConnect (rspamd_inet_address_to_string (addr),
-				rspamd_inet_address_get_port (addr));
-	}
-
-	g_assert (rt->redis != NULL);
-
-	redisLibeventAttach (rt->redis, task->ev_base);
-	rspamd_redis_maybe_auth (rt->ctx, rt->redis);
-
 	/*
 	 * Add the current key to the set of learned keys
 	 */
@@ -1800,17 +1898,20 @@ rspamd_redis_learn_tokens (struct rspamd_task *task, GPtrArray *tokens,
 					"RSIG");
 		}
 
-		rspamd_session_add_event (task->s, rspamd_redis_fin_learn, rt, M);
+		rspamd_session_add_event (task->s, NULL, rt, M);
 		rt->has_event = TRUE;
 
 		/* Set timeout */
-		if (rspamd_event_pending (&rt->timeout_event, EV_TIMEOUT)) {
-			event_del (&rt->timeout_event);
+		if (ev_can_stop (&rt->timeout_event)) {
+			rt->timeout_event.repeat = rt->ctx->timeout;
+			ev_timer_again (task->event_loop, &rt->timeout_event);
+		}
+		else {
+			rt->timeout_event.data = rt;
+			ev_timer_init (&rt->timeout_event, rspamd_redis_timeout,
+					rt->ctx->timeout, 0.);
+			ev_timer_start (task->event_loop, &rt->timeout_event);
 		}
-		event_set (&rt->timeout_event, -1, EV_TIMEOUT, rspamd_redis_timeout, rt);
-		event_base_set (task->ev_base, &rt->timeout_event);
-		double_to_tv (rt->ctx->timeout, &tv);
-		event_add (&rt->timeout_event, &tv);
 
 		return TRUE;
 	}
@@ -1827,25 +1928,17 @@ rspamd_redis_finalize_learn (struct rspamd_task *task, gpointer runtime,
 		gpointer ctx, GError **err)
 {
 	struct redis_stat_runtime *rt = REDIS_RUNTIME (runtime);
-	redisAsyncContext *redis;
-
-	if (rspamd_event_pending (&rt->timeout_event, EV_TIMEOUT)) {
-		event_del (&rt->timeout_event);
-	}
-
-	if (rt->redis) {
-		redis = rt->redis;
-		rt->redis = NULL;
-		redisAsyncFree (redis);
-	}
 
 	if (rt->err) {
 		g_propagate_error (err, rt->err);
 		rt->err = NULL;
+		rspamd_redis_fin (rt);
 
 		return FALSE;
 	}
 
+	rspamd_redis_fin (rt);
+
 	return TRUE;
 }
 
