@@ -20,14 +20,20 @@
 #include "mem_pool.h"
 #include "fstring.h"
 
+#ifdef  __cplusplus
+extern "C" {
+#endif
+
 #define RSPAMD_EXPRESSION_MAX_PRIORITY 1024
 
 #define RSPAMD_EXPRESSION_FLAG_NOOPT (1 << 0)
 
 enum rspamd_expression_op {
 	OP_INVALID = 0,
-	OP_PLUS, /* || or + */
-	OP_MULT, /* && or * */
+	OP_PLUS, /* + */
+	OP_MULT, /* * */
+	OP_MINUS, /* - */
+	OP_DIVIDE, /* / */
 	OP_OR, /* || or | */
 	OP_AND, /* && or & */
 	OP_NOT, /* ! */
@@ -35,6 +41,8 @@ enum rspamd_expression_op {
 	OP_GT, /* > */
 	OP_LE, /* <= */
 	OP_GE, /* >= */
+	OP_EQ, /* == */
+	OP_NE, /* != */
 	OP_OBRACE, /* ( */
 	OP_CBRACE /* ) */
 };
@@ -56,26 +64,20 @@ typedef struct rspamd_expression_atom_s {
 	gint priority;
 } rspamd_expression_atom_t;
 
-struct rspamd_expr_process_data {
-	/* Current Lua state to run atom processing */
-	struct lua_State *L;
-	/* Parameter of lua-function processing atom*/
-	gint stack_item;
-	gint flags;
-	/* != NULL if trace is collected */
-	GPtrArray *trace;
-	struct composites_data *cd;
-	struct rspamd_task *task;
-};
+typedef gdouble (*rspamd_expression_process_cb) (gpointer runtime_data,
+												 rspamd_expression_atom_t *atom);
 
 struct rspamd_atom_subr {
 	/* Parses atom from string and returns atom structure */
-	rspamd_expression_atom_t * (*parse)(const gchar *line, gsize len,
-			rspamd_mempool_t *pool, gpointer ud, GError **err);
+	rspamd_expression_atom_t *(*parse) (const gchar *line, gsize len,
+										rspamd_mempool_t *pool, gpointer ud, GError **err);
+
 	/* Process atom via the opaque pointer (e.g. struct rspamd_task *) */
-	gdouble (*process) (struct rspamd_expr_process_data *process_data, rspamd_expression_atom_t *atom);
+	rspamd_expression_process_cb process;
+
 	/* Calculates the relative priority of the expression */
 	gint (*priority) (rspamd_expression_atom_t *atom);
+
 	void (*destroy) (rspamd_expression_atom_t *atom);
 };
 
@@ -94,9 +96,9 @@ struct rspamd_expression;
  * @return TRUE if an expression have been parsed
  */
 gboolean rspamd_parse_expression (const gchar *line, gsize len,
-		const struct rspamd_atom_subr *subr, gpointer subr_data,
-		rspamd_mempool_t *pool, GError **err,
-		struct rspamd_expression **target);
+								  const struct rspamd_atom_subr *subr, gpointer subr_data,
+								  rspamd_mempool_t *pool, GError **err,
+								  struct rspamd_expression **target);
 
 /**
  * Process the expression and return its value using atom 'process' functions with the specified data pointer
@@ -105,7 +107,8 @@ gboolean rspamd_parse_expression (const gchar *line, gsize len,
  * @return the value of expression
  */
 gdouble rspamd_process_expression (struct rspamd_expression *expr,
-		struct rspamd_expr_process_data *process_data);
+								   gint flags,
+								   gpointer runtime_ud);
 
 /**
  * Process the expression and return its value using atom 'process' functions with the specified data pointer.
@@ -116,7 +119,22 @@ gdouble rspamd_process_expression (struct rspamd_expression *expr,
  * @return the value of expression
  */
 gdouble rspamd_process_expression_track (struct rspamd_expression *expr,
-		struct rspamd_expr_process_data *process_data);
+										 gint flags,
+										 gpointer runtime_ud,
+										 GPtrArray **track);
+
+/**
+ * Process the expression with the custom processor
+ * @param expr
+ * @param cb
+ * @param process_data
+ * @return
+ */
+gdouble rspamd_process_expression_closure (struct rspamd_expression *expr,
+										   rspamd_expression_process_cb cb,
+										   gint flags,
+										   gpointer runtime_ud,
+										   GPtrArray **track);
 
 /**
  * Shows string representation of an expression
@@ -130,7 +148,7 @@ GString *rspamd_expression_tostring (struct rspamd_expression *expr);
  * and should not be modified within callback
  */
 typedef void (*rspamd_expression_atom_foreach_cb) (const rspamd_ftok_t *atom,
-		gpointer ud);
+												   gpointer ud);
 
 /**
  * Traverse over all atoms in the expression
@@ -139,7 +157,7 @@ typedef void (*rspamd_expression_atom_foreach_cb) (const rspamd_ftok_t *atom,
  * @param ud opaque data passed to `cb`
  */
 void rspamd_expression_atom_foreach (struct rspamd_expression *expr,
-		rspamd_expression_atom_foreach_cb cb, gpointer cbdata);
+									 rspamd_expression_atom_foreach_cb cb, gpointer cbdata);
 
 /**
  * Checks if a specified node in AST is the specified operation
@@ -149,4 +167,8 @@ void rspamd_expression_atom_foreach (struct rspamd_expression *expr,
  */
 gboolean rspamd_expression_node_is_op (GNode *node, enum rspamd_expression_op op);
 
+#ifdef  __cplusplus
+}
+#endif
+
 #endif /* SRC_LIBUTIL_EXPRESSION_H_ */
