@@ -1,3 +1,19 @@
+/*-
+ * Copyright 2019 Vsevolod Stakhov
+ *
+ * Licensed under the Apache License, Version 2.0 (the "License");
+ * you may not use this file except in compliance with the License.
+ * You may obtain a copy of the License at
+ *
+ *   http://www.apache.org/licenses/LICENSE-2.0
+ *
+ * Unless required by applicable law or agreed to in writing, software
+ * distributed under the License is distributed on an "AS IS" BASIS,
+ * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+ * See the License for the specific language governing permissions and
+ * limitations under the License.
+ */
+
 /**
  * @file mem_pool.h
  * \brief Memory pools library.
@@ -13,14 +29,18 @@
 #define RSPAMD_MEM_POOL_H
 
 #include "config.h"
+
+
 #if defined(HAVE_PTHREAD_PROCESS_SHARED) && !defined(DISABLE_PTHREAD_MUTEX)
 #include <pthread.h>
 #endif
 
+#ifdef  __cplusplus
+extern "C" {
+#endif
+
 struct f_str_s;
 
-
-
 #ifdef __has_attribute
 #  if __has_attribute(alloc_size)
 #    define RSPAMD_ATTR_ALLOC_SIZE(pos) __attribute__((alloc_size(pos)))
@@ -46,21 +66,12 @@ struct f_str_s;
 
 #define MEMPOOL_TAG_LEN 20
 #define MEMPOOL_UID_LEN 20
-#define MEM_ALIGNMENT   sizeof (guint64)
-#define align_ptr(p, a)                                                   \
-    (guint8 *) (((uintptr_t) (p) + ((uintptr_t) a - 1)) & ~((uintptr_t) a - 1))
-
-enum rspamd_mempool_chain_type {
-	RSPAMD_MEMPOOL_NORMAL = 0,
-	RSPAMD_MEMPOOL_TMP,
-	RSPAMD_MEMPOOL_SHARED,
-	RSPAMD_MEMPOOL_MAX
-};
-
+/* All pointers are aligned as this variable */
+#define MIN_MEM_ALIGNMENT   sizeof (guint64)
 /**
  * Destructor type definition
  */
-typedef void (*rspamd_mempool_destruct_t)(void *ptr);
+typedef void (*rspamd_mempool_destruct_t) (void *ptr);
 
 /**
  * Pool mutex structure
@@ -83,26 +94,6 @@ typedef pthread_mutex_t rspamd_mempool_mutex_t;
 typedef pthread_rwlock_t rspamd_mempool_rwlock_t;
 #endif
 
-/**
- * Pool page structure
- */
-struct _pool_chain {
-	guint8 *begin;                  /**< begin of pool chain block              */
-	guint8 *pos;                    /**< current start of free space in block   */
-	gsize len;                      /**< length of block                        */
-	rspamd_mempool_mutex_t *lock;
-};
-
-/**
- * Destructors list item structure
- */
-struct _pool_destructors {
-	rspamd_mempool_destruct_t func;         /**< pointer to destructor					*/
-	void *data;                             /**< data to free							*/
-	const gchar *function;                  /**< function from which this destructor was added */
-	const gchar *loc;                       /**< line number                            */
-};
-
 /**
  * Tag to use for logging purposes
  */
@@ -111,18 +102,18 @@ struct rspamd_mempool_tag {
 	gchar uid[MEMPOOL_UID_LEN];             /**< unique id								*/
 };
 
+enum rspamd_mempool_flags {
+	RSPAMD_MEMPOOL_DEBUG = (1u << 0u),
+};
+
 /**
  * Memory pool type
  */
 struct rspamd_mempool_entry_point;
 struct rspamd_mutex_s;
+struct rspamd_mempool_specific;
 typedef struct memory_pool_s {
-	GPtrArray *pools[RSPAMD_MEMPOOL_MAX];
-	GArray *destructors;
-	GPtrArray *trash_stack;
-	GHashTable *variables;                  /**< private memory pool variables			*/
-	gsize elt_len;							/**< size of an element						*/
-	struct rspamd_mempool_entry_point *entry;
+	struct rspamd_mempool_specific *priv;
 	struct rspamd_mempool_tag tag;          /**< memory pool tag						*/
 } rspamd_mempool_t;
 
@@ -141,14 +132,16 @@ typedef struct memory_pool_stat_s {
 } rspamd_mempool_stat_t;
 
 
-
 /**
  * Allocate new memory poll
  * @param size size of pool's page
  * @return new memory pool object
  */
-rspamd_mempool_t * rspamd_mempool_new_ (gsize size, const gchar *tag, const gchar *loc);
-#define rspamd_mempool_new(size, tag) rspamd_mempool_new_((size), (tag), G_STRLOC)
+rspamd_mempool_t *rspamd_mempool_new_ (gsize size, const gchar *tag, gint flags,
+		const gchar *loc);
+
+#define rspamd_mempool_new(size, tag, flags) \
+	rspamd_mempool_new_((size), (tag), (flags), G_STRLOC)
 
 /**
  * Get memory from pool
@@ -156,16 +149,20 @@ rspamd_mempool_t * rspamd_mempool_new_ (gsize size, const gchar *tag, const gcha
  * @param size bytes to allocate
  * @return pointer to allocated object
  */
-void * rspamd_mempool_alloc (rspamd_mempool_t * pool, gsize size)
-		RSPAMD_ATTR_ALLOC_SIZE(2) RSPAMD_ATTR_ALLOC_ALIGN(MEM_ALIGNMENT) RSPAMD_ATTR_RETURNS_NONNUL;
+void *rspamd_mempool_alloc_ (rspamd_mempool_t *pool, gsize size, const gchar *loc)
+RSPAMD_ATTR_ALLOC_SIZE(2) RSPAMD_ATTR_ALLOC_ALIGN(MIN_MEM_ALIGNMENT) RSPAMD_ATTR_RETURNS_NONNUL;
+#define rspamd_mempool_alloc(pool, size) \
+	rspamd_mempool_alloc_((pool), (size), G_STRLOC)
 
 /**
- * Get memory from temporary pool
- * @param pool memory pool object
- * @param size bytes to allocate
- * @return pointer to allocated object
+ * Notify external memory usage for memory pool
+ * @param pool
+ * @param size
+ * @param loc
  */
-void * rspamd_mempool_alloc_tmp (rspamd_mempool_t * pool, gsize size) RSPAMD_ATTR_RETURNS_NONNUL;
+void rspamd_mempool_notify_alloc_ (rspamd_mempool_t *pool, gsize size, const gchar *loc);
+#define rspamd_mempool_notify_alloc(pool, size) \
+	rspamd_mempool_notify_alloc_((pool), (size), G_STRLOC)
 
 /**
  * Get memory and set it to zero
@@ -173,21 +170,10 @@ void * rspamd_mempool_alloc_tmp (rspamd_mempool_t * pool, gsize size) RSPAMD_ATT
  * @param size bytes to allocate
  * @return pointer to allocated object
  */
-void * rspamd_mempool_alloc0 (rspamd_mempool_t * pool, gsize size)
-	RSPAMD_ATTR_ALLOC_SIZE(2) RSPAMD_ATTR_ALLOC_ALIGN(MEM_ALIGNMENT) RSPAMD_ATTR_RETURNS_NONNUL;
-
-/**
- * Get memory and set it to zero
- * @param pool memory pool object
- * @param size bytes to allocate
- * @return pointer to allocated object
- */
-void * rspamd_mempool_alloc0_tmp (rspamd_mempool_t * pool, gsize size) RSPAMD_ATTR_RETURNS_NONNUL;
-
-/**
- * Cleanup temporary data in pool
- */
-void rspamd_mempool_cleanup_tmp (rspamd_mempool_t * pool);
+void *rspamd_mempool_alloc0_ (rspamd_mempool_t *pool, gsize size, const gchar *loc)
+RSPAMD_ATTR_ALLOC_SIZE(2) RSPAMD_ATTR_ALLOC_ALIGN(MIN_MEM_ALIGNMENT) RSPAMD_ATTR_RETURNS_NONNUL;
+#define rspamd_mempool_alloc0(pool, size) \
+	rspamd_mempool_alloc0_((pool), (size), G_STRLOC)
 
 /**
  * Make a copy of string in pool
@@ -195,8 +181,10 @@ void rspamd_mempool_cleanup_tmp (rspamd_mempool_t * pool);
  * @param src source string
  * @return pointer to newly created string that is copy of src
  */
-gchar * rspamd_mempool_strdup (rspamd_mempool_t * pool, const gchar *src)
-	RSPAMD_ATTR_ALLOC_ALIGN(MEM_ALIGNMENT);
+gchar *rspamd_mempool_strdup_ (rspamd_mempool_t *pool, const gchar *src, const gchar *loc)
+RSPAMD_ATTR_ALLOC_ALIGN(MIN_MEM_ALIGNMENT);
+#define rspamd_mempool_strdup(pool, src) \
+	rspamd_mempool_strdup_ ((pool), (src), G_STRLOC)
 
 /**
  * Make a copy of fixed string in pool as null terminated string
@@ -204,8 +192,12 @@ gchar * rspamd_mempool_strdup (rspamd_mempool_t * pool, const gchar *src)
  * @param src source string
  * @return pointer to newly created string that is copy of src
  */
-gchar * rspamd_mempool_fstrdup (rspamd_mempool_t * pool,
-	const struct f_str_s *src) RSPAMD_ATTR_ALLOC_ALIGN(MEM_ALIGNMENT);
+gchar *rspamd_mempool_fstrdup_ (rspamd_mempool_t *pool,
+								const struct f_str_s *src,
+								const gchar *loc)
+RSPAMD_ATTR_ALLOC_ALIGN(MIN_MEM_ALIGNMENT);
+#define rspamd_mempool_fstrdup(pool, src) \
+	rspamd_mempool_fstrdup_ ((pool), (src), G_STRLOC)
 
 struct f_str_tok;
 
@@ -215,18 +207,28 @@ struct f_str_tok;
  * @param src source string
  * @return pointer to newly created string that is copy of src
  */
-gchar * rspamd_mempool_ftokdup (rspamd_mempool_t *pool,
-		const struct f_str_tok *src) RSPAMD_ATTR_ALLOC_ALIGN(MEM_ALIGNMENT);
+gchar *rspamd_mempool_ftokdup_ (rspamd_mempool_t *pool,
+								const struct f_str_tok *src,
+								const gchar *loc)
+RSPAMD_ATTR_ALLOC_ALIGN(MIN_MEM_ALIGNMENT);
+#define rspamd_mempool_ftokdup(pool, src) \
+	rspamd_mempool_ftokdup_ ((pool), (src), G_STRLOC)
 
 /**
  * Allocate piece of shared memory
  * @param pool memory pool object
  * @param size bytes to allocate
  */
-void * rspamd_mempool_alloc_shared (rspamd_mempool_t * pool, gsize size)
-	RSPAMD_ATTR_ALLOC_SIZE(2) RSPAMD_ATTR_ALLOC_ALIGN(MEM_ALIGNMENT) RSPAMD_ATTR_RETURNS_NONNUL;
-void * rspamd_mempool_alloc0_shared (rspamd_mempool_t *pool, gsize size)
-	RSPAMD_ATTR_ALLOC_SIZE(2) RSPAMD_ATTR_ALLOC_ALIGN(MEM_ALIGNMENT) RSPAMD_ATTR_RETURNS_NONNUL;
+void *rspamd_mempool_alloc_shared_ (rspamd_mempool_t *pool, gsize size, const gchar *loc)
+RSPAMD_ATTR_ALLOC_SIZE(2) RSPAMD_ATTR_ALLOC_ALIGN(MIN_MEM_ALIGNMENT) RSPAMD_ATTR_RETURNS_NONNUL;
+#define rspamd_mempool_alloc_shared(pool, size) \
+	rspamd_mempool_alloc_shared_((pool), (size), G_STRLOC)
+
+void *rspamd_mempool_alloc0_shared_ (rspamd_mempool_t *pool, gsize size, const gchar *loc)
+RSPAMD_ATTR_ALLOC_SIZE(2) RSPAMD_ATTR_ALLOC_ALIGN(MIN_MEM_ALIGNMENT) RSPAMD_ATTR_RETURNS_NONNUL;
+#define rspamd_mempool_alloc0_shared(pool, size) \
+	rspamd_mempool_alloc0_shared_((pool), (size), G_STRLOC)
+
 /**
  * Add destructor callback to pool
  * @param pool memory pool object
@@ -234,14 +236,14 @@ void * rspamd_mempool_alloc0_shared (rspamd_mempool_t *pool, gsize size)
  * @param data pointer to data that would be passed to destructor
  */
 void rspamd_mempool_add_destructor_full (rspamd_mempool_t *pool,
-	rspamd_mempool_destruct_t func,
-	void *data,
-	const gchar *function,
-	const gchar *line);
+										 rspamd_mempool_destruct_t func,
+										 void *data,
+										 const gchar *function,
+										 const gchar *line);
 
 /* Macros for common usage */
 #define rspamd_mempool_add_destructor(pool, func, data) \
-	rspamd_mempool_add_destructor_full (pool, func, data, G_STRFUNC, G_STRLOC)
+    rspamd_mempool_add_destructor_full (pool, func, data, G_STRFUNC, G_STRLOC)
 
 /**
  * Replace destructor callback to pool for specified pointer
@@ -251,7 +253,7 @@ void rspamd_mempool_add_destructor_full (rspamd_mempool_t *pool,
  * @param new_data pointer to data that would be passed to destructor
  */
 void rspamd_mempool_replace_destructor (rspamd_mempool_t *pool,
-	rspamd_mempool_destruct_t func, void *old_data, void *new_data);
+										rspamd_mempool_destruct_t func, void *old_data, void *new_data);
 
 /**
  * Calls all destructors associated with the specific memory pool without removing
@@ -271,7 +273,7 @@ void rspamd_mempool_delete (rspamd_mempool_t *pool);
  * @param pool memory pool object
  * @return mutex object
  */
-rspamd_mempool_mutex_t * rspamd_mempool_get_mutex (rspamd_mempool_t *pool);
+rspamd_mempool_mutex_t *rspamd_mempool_get_mutex (rspamd_mempool_t *pool);
 
 /**
  * Lock mutex
@@ -290,7 +292,7 @@ void rspamd_mempool_unlock_mutex (rspamd_mempool_mutex_t *mutex);
  * @param pool memory pool object
  * @return rwlock object
  */
-rspamd_mempool_rwlock_t * rspamd_mempool_get_rwlock (rspamd_mempool_t *pool);
+rspamd_mempool_rwlock_t *rspamd_mempool_get_rwlock (rspamd_mempool_t *pool);
 
 /**
  * Acquire read lock
@@ -332,8 +334,12 @@ void rspamd_mempool_stat_reset (void);
  * @return size of memory page in system
  */
 #define rspamd_mempool_suggest_size() rspamd_mempool_suggest_size_(G_STRLOC)
+
 gsize rspamd_mempool_suggest_size_ (const char *loc);
 
+gsize rspamd_mempool_get_used_size (rspamd_mempool_t *pool);
+gsize rspamd_mempool_get_wasted_size (rspamd_mempool_t *pool);
+
 /**
  * Set memory pool variable
  * @param pool memory pool object
@@ -341,8 +347,10 @@ gsize rspamd_mempool_suggest_size_ (const char *loc);
  * @param gpointer value value of variable
  * @param destructor pointer to function-destructor
  */
-void rspamd_mempool_set_variable (rspamd_mempool_t *pool, const gchar *name,
-	gpointer value, rspamd_mempool_destruct_t destructor);
+void rspamd_mempool_set_variable (rspamd_mempool_t *pool,
+								  const gchar *name,
+								  gpointer value,
+								  rspamd_mempool_destruct_t destructor);
 
 /**
  * Get memory pool variable
@@ -351,7 +359,7 @@ void rspamd_mempool_set_variable (rspamd_mempool_t *pool, const gchar *name,
  * @return NULL or pointer to variable data
  */
 gpointer rspamd_mempool_get_variable (rspamd_mempool_t *pool,
-	const gchar *name);
+									  const gchar *name);
 
 /**
  * Removes variable from memory pool
@@ -359,7 +367,8 @@ gpointer rspamd_mempool_get_variable (rspamd_mempool_t *pool,
  * @param name name of variable
  */
 void rspamd_mempool_remove_variable (rspamd_mempool_t *pool,
-		const gchar *name);
+									 const gchar *name);
+
 /**
  * Prepend element to a list creating it in the memory pool
  * @param l
@@ -367,7 +376,8 @@ void rspamd_mempool_remove_variable (rspamd_mempool_t *pool,
  * @return
  */
 GList *rspamd_mempool_glist_prepend (rspamd_mempool_t *pool,
-		GList *l, gpointer p) G_GNUC_WARN_UNUSED_RESULT;
+									 GList *l, gpointer p) G_GNUC_WARN_UNUSED_RESULT;
+
 /**
  * Append element to a list creating it in the memory pool
  * @param l
@@ -375,6 +385,10 @@ GList *rspamd_mempool_glist_prepend (rspamd_mempool_t *pool,
  * @return
  */
 GList *rspamd_mempool_glist_append (rspamd_mempool_t *pool,
-		GList *l, gpointer p) G_GNUC_WARN_UNUSED_RESULT;
+									GList *l, gpointer p) G_GNUC_WARN_UNUSED_RESULT;
+
+#ifdef  __cplusplus
+}
+#endif
 
 #endif
