@@ -1,5 +1,20 @@
-/* Copyright (c) 2010-2015, Vsevolod Stakhov
- * Copyright (C) 2002-2015 Igor Sysoev
+/*-
+ * Copyright 2016 Vsevolod Stakhov
+ *
+ * Licensed under the Apache License, Version 2.0 (the "License");
+ * you may not use this file except in compliance with the License.
+ * You may obtain a copy of the License at
+ *
+ *   http://www.apache.org/licenses/LICENSE-2.0
+ *
+ * Unless required by applicable law or agreed to in writing, software
+ * distributed under the License is distributed on an "AS IS" BASIS,
+ * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+ * See the License for the specific language governing permissions and
+ * limitations under the License.
+ */
+
+/* Copyright (C) 2002-2015 Igor Sysoev
  * Copyright (C) 2011-2015 Nginx, Inc.
  * All rights reserved.
  *
@@ -24,8 +39,8 @@
  */
 
 #include "printf.h"
-#include "fstring.h"
-#include "main.h"
+#include "str_util.h"
+#include "contrib/fpconv/fpconv.h"
 
 /**
  * From FreeBSD libutil code
@@ -40,36 +55,27 @@ rspamd_humanize_number (gchar *buf, gchar *last, gint64 num, gboolean bytes)
 	const gchar *prefixes;
 	int i, r, remainder, sign;
 	gint64 divisor;
-	gsize baselen, len = last - buf;
+	gsize len = last - buf;
 
 	remainder = 0;
 
-	baselen = 1;
 	if (!bytes) {
 		divisor = 1000;
-		prefixes = "\0\0\0k\0\0M\0\0G\0\0T\0\0P\0\0E";
+		prefixes = "\0\0\0\0k\0\0\0M\0\0\0G\0\0\0T\0\0\0P\0\0\0E";
 	}
 	else {
 		divisor = 1024;
-		prefixes = "B\0\0k\0\0M\0\0G\0\0T\0\0P\0\0E";
+		prefixes = "B\0\0\0KiB\0MiB\0GiB\0TiB\0PiB\0EiB";
 	}
 
-
-#define SCALE2PREFIX(scale)     (&prefixes[(scale) * 3])
+#define SCALE2PREFIX(scale)     (&prefixes[(scale) * 4])
 
 	if (num < 0) {
 		sign = -1;
 		num = -num;
-		baselen += 2; /* sign, digit */
 	}
 	else {
 		sign = 1;
-		baselen += 1; /* digit */
-	}
-
-	/* Check if enough room for `x y' + suffix + `\0' */
-	if (len < baselen + 1) {
-		return buf;
 	}
 
 	/*
@@ -82,9 +88,17 @@ rspamd_humanize_number (gchar *buf, gchar *last, gint64 num, gboolean bytes)
 		num /= divisor;
 	}
 
-	r = rspamd_snprintf (buf, len, "%L%s",
-			sign * (num + (remainder + 50) / 1000),
-			SCALE2PREFIX (i));
+	if (remainder == 0 || num > divisor / 2) {
+		r = rspamd_snprintf (buf, len, "%L%s",
+				sign * (num + (remainder + 50) / divisor),
+				SCALE2PREFIX (i));
+	}
+	else {
+		/* Floating point version */
+		r = rspamd_snprintf (buf, len, "%.2f%s",
+				sign * (num + remainder / (gdouble)divisor),
+				SCALE2PREFIX (i));
+	}
 
 #undef SCALE2PREFIX
 
@@ -92,78 +106,295 @@ rspamd_humanize_number (gchar *buf, gchar *last, gint64 num, gboolean bytes)
 }
 
 
+static inline unsigned
+rspamd_decimal_digits32 (guint32 val)
+{
+	static const guint32 powers_of_10[] = {
+			0,
+			10,
+			100,
+			1000,
+			10000,
+			100000,
+			1000000,
+			10000000,
+			100000000,
+			1000000000
+	};
+	unsigned tmp;
+
+#if defined(_MSC_VER)
+	unsigned long r = 0;
+	_BitScanReverse (&r, val | 1);
+	tmp = (r + 1) * 1233 >> 12;
+#elif defined(__GNUC__) && (__GNUC__ >= 3)
+	tmp = (32 - __builtin_clz (val | 1U)) * 1233 >> 12;
+
+#else /* Software version */
+	static const unsigned debruijn_tbl[32] = { 0,  9,  1, 10, 13, 21,  2, 29,
+											   11, 14, 16, 18, 22, 25,  3, 30,
+											   8, 12, 20, 28, 15, 17, 24,  7,
+											   19, 27, 23,  6, 26,  5,  4, 31 };
+	guint32 v = val | 1;
+
+	v |= v >> 1;
+	v |= v >> 2;
+	v |= v >> 4;
+	v |= v >> 8;
+	v |= v >> 16;
+	tmp = (1 + debruijn_tbl[(v * 0x07C4ACDDU) >> 27]) * 1233 >> 12;
+#endif
+	return tmp - (val < powers_of_10[tmp]) + 1;
+}
+
+static inline unsigned
+rspamd_decimal_digits64 (guint64 val)
+{
+	static const guint64 powers_of_10[] = {
+			0,
+			10ULL,
+			100ULL,
+			1000ULL,
+			10000ULL,
+			100000ULL,
+			1000000ULL,
+			10000000ULL,
+			100000000ULL,
+			1000000000ULL,
+			10000000000ULL,
+			100000000000ULL,
+			1000000000000ULL,
+			10000000000000ULL,
+			100000000000000ULL,
+			1000000000000000ULL,
+			10000000000000000ULL,
+			100000000000000000ULL,
+			1000000000000000000ULL,
+			10000000000000000000ULL
+	};
+	unsigned tmp;
+
+#if defined(_MSC_VER)
+#if _M_IX86
+	unsigned long r = 0;
+	guint64 m = val | 1;
+	if (_BitScanReverse (&r, m >> 32)) {
+		r += 32;
+	}
+	else {
+		_BitScanReverse (&r, m & 0xFFFFFFFF);
+	}
+	tmp = (r + 1) * 1233 >> 12;
+#else
+	unsigned long r = 0;
+	_BitScanReverse64 (&r, val | 1);
+	tmp = (r + 1) * 1233 >> 12;
+#endif
+#elif defined(__GNUC__) && (__GNUC__ >= 3)
+	tmp = (64 - __builtin_clzll (val | 1ULL)) * 1233 >> 12;
+#else /* Software version */
+	static const unsigned debruijn_tbl[32] = { 0,  9,  1, 10, 13, 21,  2, 29,
+											   11, 14, 16, 18, 22, 25,  3, 30,
+											   8, 12, 20, 28, 15, 17, 24,  7,
+											   19, 27, 23,  6, 26,  5,  4, 31 };
+	guint32 v = val >> 32;
+
+	if (v) {
+		v |= 1;
+		v |= v >> 1;
+		v |= v >> 2;
+		v |= v >> 4;
+		v |= v >> 8;
+		v |= v >> 16;
+		tmp = 32 + debruijn_tbl[(v * 0x07C4ACDDU) >> 27];
+	}
+	else {
+		v = val & 0xFFFFFFFF;
+		v |= 1;
+		v |= v >> 1;
+		v |= v >> 2;
+		v |= v >> 4;
+		v |= v >> 8;
+		v |= v >> 16;
+
+		tmp = debruijn_tbl[(v * 0x07C4ACDDU) >> 27];
+	}
+
+
+	tmp = (tmp + 1) * 1233 >> 12;
+#endif
+
+	return tmp - (val < powers_of_10[tmp]) + 1;
+}
+
+/*
+ * Idea from https://github.com/miloyip/itoa-benchmark:
+ * Uses lookup table (LUT) of digit pairs for division/modulo of 100.
+ *
+ * Mentioned in:
+ * https://www.slideshare.net/andreialexandrescu1/three-optimization-tips-for-c-15708507
+ */
+
+static const char int_lookup_table[200] = {
+		'0','0','0','1','0','2','0','3','0','4',
+		'0','5','0','6','0','7','0','8','0','9',
+		'1','0','1','1','1','2','1','3','1','4',
+		'1','5','1','6','1','7','1','8','1','9',
+		'2','0','2','1','2','2','2','3','2','4',
+		'2','5','2','6','2','7','2','8','2','9',
+		'3','0','3','1','3','2','3','3','3','4',
+		'3','5','3','6','3','7','3','8','3','9',
+		'4','0','4','1','4','2','4','3','4','4',
+		'4','5','4','6','4','7','4','8','4','9',
+		'5','0','5','1','5','2','5','3','5','4',
+		'5','5','5','6','5','7','5','8','5','9',
+		'6','0','6','1','6','2','6','3','6','4',
+		'6','5','6','6','6','7','6','8','6','9',
+		'7','0','7','1','7','2','7','3','7','4',
+		'7','5','7','6','7','7','7','8','7','9',
+		'8','0','8','1','8','2','8','3','8','4',
+		'8','5','8','6','8','7','8','8','8','9',
+		'9','0','9','1','9','2','9','3','9','4',
+		'9','5','9','6','9','7','9','8','9','9'
+};
+
+static inline guint
+rspamd_uint32_print (guint32 in, gchar *out)
+{
+	guint ndigits = rspamd_decimal_digits32 (in);
+	gchar *p;
+
+	p = out + ndigits - 1;
+
+	while (in >= 100) {
+		unsigned idx = (in % 100) * 2;
+
+		/* Do two digits at once */
+		*p-- = int_lookup_table[idx + 1];
+		*p-- = int_lookup_table[idx];
+
+		in /= 100;
+	}
+
+	if (in < 10) {
+		*p = ((char)in) + '0';
+	}
+	else {
+		unsigned idx = in * 2;
+
+		*p-- = int_lookup_table[idx + 1];
+		*p = int_lookup_table[idx];
+	}
+
+	return ndigits;
+}
+
+static inline guint
+rspamd_uint64_print (guint64 in, gchar *out)
+{
+	guint ndigits = rspamd_decimal_digits64 (in);
+	guint32 v32;
+	gchar *p;
+
+	p = out + ndigits - 1;
+
+	while (in >= 100000000) {
+		v32 = (guint32)(in % 100000000);
+		guint32 a, b, a1, a2, b1, b2;
+
+		/* Initial spill */
+		a = v32 / 10000;
+		b = v32 % 10000;
+		a1 = (a / 100) * 2;
+		a2 = (a % 100) * 2;
+		b1 = (b / 100) * 2;
+		b2 = (b % 100) * 2;
+
+		/* Fill 8 digits at once */
+		*p-- = int_lookup_table[b2 + 1];
+		*p-- = int_lookup_table[b2];
+		*p-- = int_lookup_table[b1 + 1];
+		*p-- = int_lookup_table[b1];
+		*p-- = int_lookup_table[a2 + 1];
+		*p-- = int_lookup_table[a2];
+		*p-- = int_lookup_table[a1 + 1];
+		*p-- = int_lookup_table[a1];
+
+		in /= 100000000;
+	}
+
+	/* Remaining 32 bit */
+	v32 = (guint32)in;
+
+	while (v32 >= 100) {
+		unsigned idx = (v32 % 100) << 1;
+
+		/* Do 2 digits at once */
+		*p-- = int_lookup_table[idx + 1];
+		*p-- = int_lookup_table[idx];
+
+		v32 /= 100;
+	}
+
+	if (v32 < 10) {
+		*p = ((char)v32) + '0';
+	}
+	else {
+		unsigned idx = v32 * 2;
+
+		*p-- = int_lookup_table[idx + 1];
+		*p = int_lookup_table[idx];
+	}
+
+	return ndigits;
+}
+
 static gchar *
 rspamd_sprintf_num (gchar *buf, gchar *last, guint64 ui64, gchar zero,
-	guint hexadecimal, guint width)
+					  guint hexadecimal, guint width)
 {
 	gchar *p, temp[sizeof ("18446744073709551615")];
 	size_t len;
-	guint32 ui32;
-
-	p = temp + sizeof(temp);
 
 	if (hexadecimal == 0) {
+		p = temp;
 
-		if (ui64 <= G_MAXUINT32) {
-
-			/*
-			 * To divide 64-bit numbers and to find remainders
-			 * on the x86 platform gcc and icc call the libc functions
-			 * [u]divdi3() and [u]moddi3(), they call another function
-			 * in its turn.  On FreeBSD it is the qdivrem() function,
-			 * its source code is about 170 lines of the code.
-			 * The glibc counterpart is about 150 lines of the code.
-			 *
-			 * For 32-bit numbers and some divisors gcc and icc use
-			 * a inlined multiplication and shifts.  For example,
-			 * guint "i32 / 10" is compiled to
-			 *
-			 *	 (i32 * 0xCCCCCCCD) >> 35
-			 */
-
-			ui32 = (guint32) ui64;
-
-			do {
-				*--p = (gchar) (ui32 % 10 + '0');
-			} while (ui32 /= 10);
-
-		} else {
-			do {
-				*--p = (gchar) (ui64 % 10 + '0');
-			} while (ui64 /= 10);
+		if (ui64 < G_MAXUINT32) {
+			len = rspamd_uint32_print ((guint32)ui64, temp);
 		}
-
-	} else if (hexadecimal == 1) {
-
+		else {
+			len = rspamd_uint64_print (ui64, temp);
+		}
+	}
+	else if (hexadecimal == 1) {
+		p = temp + sizeof(temp);
 		do {
-
-			/* the "(guint32)" cast disables the BCC's warning */
 			*--p = _hex[(guint32) (ui64 & 0xf)];
-
 		} while (ui64 >>= 4);
 
-	} else { /* hexadecimal == 2 */
-
+		len = (temp + sizeof (temp)) - p;
+	}
+	else { /* hexadecimal == 2 */
+		p = temp + sizeof(temp);
 		do {
-
-			/* the "(guint32)" cast disables the BCC's warning */
 			*--p = _HEX[(guint32) (ui64 & 0xf)];
-
 		} while (ui64 >>= 4);
+
+		len = (temp + sizeof (temp)) - p;
 	}
 
 	/* zero or space padding */
 
-	len = (temp + sizeof (temp)) - p;
+	if (len < width) {
+		width -= len;
 
-	while (len++ < width && buf < last) {
-		*buf++ = zero;
+		while (width-- > 0 && buf < last) {
+			*buf++ = zero;
+		}
 	}
 
 	/* number safe copy */
 
-	len = (temp + sizeof (temp)) - p;
-
 	if (buf + len > last) {
 		len = last - buf;
 	}
@@ -219,6 +450,18 @@ rspamd_printf_append_gstring (const gchar *buf, glong buflen, gpointer ud)
 	return buflen;
 }
 
+static glong
+rspamd_printf_append_fstring (const gchar *buf, glong buflen, gpointer ud)
+{
+	rspamd_fstring_t **dst = ud;
+
+	if (buflen > 0) {
+		*dst = rspamd_fstring_append (*dst, buf, buflen);
+	}
+
+	return buflen;
+}
+
 glong
 rspamd_fprintf (FILE *f, const gchar *fmt, ...)
 {
@@ -307,6 +550,25 @@ rspamd_vprintf_gstring (GString *s, const gchar *fmt, va_list args)
 	return rspamd_vprintf_common (rspamd_printf_append_gstring, s, fmt, args);
 }
 
+glong
+rspamd_printf_fstring (rspamd_fstring_t **s, const gchar *fmt, ...)
+{
+	va_list args;
+	glong r;
+
+	va_start (args, fmt);
+	r = rspamd_vprintf_fstring (s, fmt, args);
+	va_end (args);
+
+	return r;
+}
+
+glong
+rspamd_vprintf_fstring (rspamd_fstring_t **s, const gchar *fmt, va_list args)
+{
+	return rspamd_vprintf_common (rspamd_printf_append_fstring, s, fmt, args);
+}
+
 #define RSPAMD_PRINTF_APPEND(buf, len)                                         \
 	do {                                                                       \
 		RSPAMD_PRINTF_APPEND_BUF(buf, len);                                    \
@@ -329,18 +591,19 @@ rspamd_vprintf_common (rspamd_printf_append_func func,
 	const gchar *fmt,
 	va_list args)
 {
-	gchar zero, numbuf[G_ASCII_DTOSTR_BUF_SIZE], *p, *last, c;
+	gchar zero, numbuf[G_ASCII_DTOSTR_BUF_SIZE], dtoabuf[32], *p, *last;
+	guchar c;
 	const gchar *buf_start = fmt;
 	gint d;
-	long double f, scale;
+	gdouble f;
 	glong written = 0, wr, slen;
 	gint64 i64;
 	guint64 ui64;
-	guint width, sign, hex, humanize, bytes, frac_width, i;
+	guint width, sign, hex, humanize, bytes, frac_width, b32, b64;
 	rspamd_fstring_t *v;
+	rspamd_ftok_t *tok;
 	GString *gs;
 	GError *err;
-	gboolean bv;
 
 	while (*fmt) {
 
@@ -367,6 +630,8 @@ rspamd_vprintf_common (rspamd_printf_append_func func,
 			width = 0;
 			sign = 1;
 			hex = 0;
+			b32 = 0;
+			b64 = 0;
 			bytes = 0;
 			humanize = 0;
 			frac_width = 0;
@@ -400,6 +665,16 @@ rspamd_vprintf_common (rspamd_printf_append_func func,
 					sign = 0;
 					fmt++;
 					continue;
+				case 'b':
+					b32 = 1;
+					sign = 0;
+					fmt++;
+					continue;
+				case 'B':
+					b64 = 1;
+					sign = 0;
+					fmt++;
+					continue;
 				case 'H':
 					humanize = 1;
 					bytes = 1;
@@ -417,7 +692,6 @@ rspamd_vprintf_common (rspamd_printf_append_func func,
 					if (*fmt == '*') {
 						d = (gint)va_arg (args, gint);
 						if (G_UNLIKELY (d < 0)) {
-							msg_err ("critical error: fraction width is less than 0");
 							return 0;
 						}
 						frac_width = (guint)d;
@@ -434,7 +708,6 @@ rspamd_vprintf_common (rspamd_printf_append_func func,
 				case '*':
 					d = (gint)va_arg (args, gint);
 					if (G_UNLIKELY (d < 0)) {
-						msg_err ("critical error: size is less than 0");
 						return 0;
 					}
 					slen = (glong)d;
@@ -453,13 +726,53 @@ rspamd_vprintf_common (rspamd_printf_append_func func,
 
 			case 'V':
 				v = va_arg (args, rspamd_fstring_t *);
-				RSPAMD_PRINTF_APPEND (v->begin, v->len);
 
+				if (v) {
+					slen = v->len;
+
+					if (G_UNLIKELY (width != 0)) {
+						slen = MIN (v->len, width);
+					}
+
+					RSPAMD_PRINTF_APPEND (v->str, slen);
+				}
+				else {
+					RSPAMD_PRINTF_APPEND ("(NULL)", 6);
+				}
+
+				continue;
+
+			case 'T':
+				tok = va_arg (args, rspamd_ftok_t *);
+
+				if (tok) {
+					slen = tok->len;
+
+					if (G_UNLIKELY (width != 0)) {
+						slen = MIN (tok->len, width);
+					}
+					RSPAMD_PRINTF_APPEND (tok->begin, slen);
+				}
+				else {
+					RSPAMD_PRINTF_APPEND ("(NULL)", 6);
+				}
 				continue;
 
 			case 'v':
 				gs = va_arg (args, GString *);
-				RSPAMD_PRINTF_APPEND (gs->str, gs->len);
+
+				if (gs) {
+					slen = gs->len;
+
+					if (G_UNLIKELY (width != 0)) {
+						slen = MIN (gs->len, width);
+					}
+
+					RSPAMD_PRINTF_APPEND (gs->str, slen);
+				}
+				else {
+					RSPAMD_PRINTF_APPEND ("(NULL)", 6);
+				}
 
 				continue;
 
@@ -486,30 +799,94 @@ rspamd_vprintf_common (rspamd_printf_append_func func,
 				p = va_arg (args, gchar *);
 				if (p == NULL) {
 					p = "(NULL)";
+					slen = sizeof ("(NULL)") - 1;
 				}
 
-				if (slen == -1) {
-					/* NULL terminated string */
-					slen = strlen (p);
-				}
+				if (G_UNLIKELY (b32)) {
+					gchar *b32buf;
 
-				if (G_LIKELY(hex == 0)) {
-					RSPAMD_PRINTF_APPEND (p, slen);
-				}
-				else {
-					gchar hexbuf[2];
-					while (slen) {
-						hexbuf[0] = hex == 2 ? _HEX[*p & 0xf] : _hex[*p & 0xf];
-						hexbuf[1] = hex == 2 ? _HEX[(*p >> 8) & 0xf] :
-								_hex[(*p >> 8) & 0xf];
-						RSPAMD_PRINTF_APPEND_BUF (hexbuf, 2);
-						p ++;
-						slen --;
+					if (G_UNLIKELY (slen == -1)) {
+						if (G_LIKELY (width != 0)) {
+							slen = width;
+						}
+						else {
+							/* NULL terminated string */
+							slen = strlen (p);
+						}
 					}
-					fmt ++;
+
+					b32buf = rspamd_encode_base32 (p, slen, RSPAMD_BASE32_DEFAULT);
+
+					if (b32buf) {
+						RSPAMD_PRINTF_APPEND (b32buf, strlen (b32buf));
+						g_free (b32buf);
+					}
+					else {
+						RSPAMD_PRINTF_APPEND ("(NULL)", sizeof ("(NULL)") - 1);
+					}
+				}
+				else if (G_UNLIKELY (hex)) {
+					gchar hexbuf[2];
+
+					if (G_UNLIKELY (slen == -1)) {
+						if (G_LIKELY (width != 0)) {
+							slen = width;
+						}
+						else {
+							/* NULL terminated string */
+							slen = strlen (p);
+						}
+					}
+
+					while (slen) {
+						hexbuf[0] = hex == 2 ? _HEX[(*p >> 4u) & 0xfu] :
+								_hex[(*p >> 4u) & 0xfu];
+						hexbuf[1] = hex == 2 ? _HEX[*p & 0xfu] : _hex[*p & 0xfu];
+						RSPAMD_PRINTF_APPEND_BUF (hexbuf, 2);
+						p++;
+						slen--;
+					}
+
+					fmt++;
 					buf_start = fmt;
 
 				}
+				else if (G_UNLIKELY (b64)) {
+					gchar *b64buf;
+					gsize olen = 0;
+
+					if (G_UNLIKELY (slen == -1)) {
+						if (G_LIKELY (width != 0)) {
+							slen = width;
+						}
+						else {
+							/* NULL terminated string */
+							slen = strlen (p);
+						}
+					}
+
+					b64buf = rspamd_encode_base64 (p, slen, 0, &olen);
+
+					if (b64buf) {
+						RSPAMD_PRINTF_APPEND (b64buf, olen);
+						g_free (b64buf);
+					}
+					else {
+						RSPAMD_PRINTF_APPEND ("(NULL)", sizeof ("(NULL)") - 1);
+					}
+				}
+				else {
+					if (slen == -1) {
+						/* NULL terminated string */
+						slen = strlen (p);
+					}
+
+					if (G_UNLIKELY (width != 0)) {
+						slen = MIN (slen, width);
+					}
+
+					RSPAMD_PRINTF_APPEND (p, slen);
+				}
 
 				continue;
 
@@ -523,7 +900,7 @@ rspamd_vprintf_common (rspamd_printf_append_func func,
 				sign = 1;
 				break;
 
-			case 'T':
+			case 't':
 				i64 = (gint64) va_arg (args, time_t);
 				sign = 1;
 				break;
@@ -570,71 +947,32 @@ rspamd_vprintf_common (rspamd_printf_append_func func,
 
 
 			case 'f':
-			case 'F':
-				if (*fmt == 'f') {
-					f = (long double) va_arg (args, double);
-				}
-				else {
-					f = (long double) va_arg (args, long double);
-				}
-				p = numbuf;
-				last = p + sizeof (numbuf);
-				if (f < 0) {
-					*p++ = '-';
-					f = -f;
-				}
-				if (frac_width == 0) {
-					frac_width = 6;
-				}
+				f = (gdouble) va_arg (args, double);
+				slen = fpconv_dtoa (f, dtoabuf, frac_width, false);
 
-				ui64 = (gint64) f;
-
-				p = rspamd_sprintf_num (p, last, ui64, zero, 0, width);
-
-				if (frac_width) {
-
-					if (p < last) {
-						*p++ = '.';
-					}
-
-					scale = 1.0;
-
-					for (i = 0; i < frac_width; i++) {
-						scale *= 10.0;
-					}
-
-					/*
-					 * (gint64) cast is required for msvc6:
-					 * it can not convert guint64 to double
-					 */
-					ui64 = (guint64) ((f - (gint64) ui64) * scale);
-
-					p = rspamd_sprintf_num (p, last, ui64, '0', 0, frac_width);
-				}
-
-				slen = p - numbuf;
-				RSPAMD_PRINTF_APPEND (numbuf, slen);
+				RSPAMD_PRINTF_APPEND (dtoabuf, slen);
 
 				continue;
 
 			case 'g':
-			case 'G':
-				if (*fmt == 'g') {
-					f = (long double) va_arg (args, double);
-				}
-				else {
-					f = (long double) va_arg (args, long double);
-				}
-
-				g_ascii_formatd (numbuf, sizeof (numbuf), "%g", (double)f);
-				slen = strlen (numbuf);
-				RSPAMD_PRINTF_APPEND (numbuf, slen);
+				f = (gdouble) va_arg (args, double);
+				slen = fpconv_dtoa (f, dtoabuf, 0, true);
+				RSPAMD_PRINTF_APPEND (dtoabuf, slen);
 
 				continue;
 
-			case 'b':
-				bv = (gboolean) va_arg (args, double);
-				RSPAMD_PRINTF_APPEND (bv ? "true" : "false", bv ? 4 : 5);
+			case 'F':
+				f = (gdouble) va_arg (args, long double);
+				slen = fpconv_dtoa (f, dtoabuf, frac_width, false);
+
+				RSPAMD_PRINTF_APPEND (dtoabuf, slen);
+
+				continue;
+
+			case 'G':
+				f = (gdouble) va_arg (args, long double);
+				slen = fpconv_dtoa (f, dtoabuf, 0, true);
+				RSPAMD_PRINTF_APPEND (dtoabuf, slen);
 
 				continue;
 
@@ -648,8 +986,18 @@ rspamd_vprintf_common (rspamd_printf_append_func func,
 
 			case 'c':
 				c = va_arg (args, gint);
-				c &= 0xff;
-				RSPAMD_PRINTF_APPEND (&c, 1);
+				c &= 0xffu;
+				if (G_UNLIKELY (hex)) {
+					gchar hexbuf[2];
+					hexbuf[0] = hex == 2 ? _HEX[(c >> 4u) & 0xfu] :
+								_hex[(c >> 4u) & 0xfu];
+					hexbuf[1] = hex == 2 ? _HEX[c & 0xfu] : _hex[c & 0xfu];
+
+					RSPAMD_PRINTF_APPEND (hexbuf, 2);
+				}
+				else {
+					RSPAMD_PRINTF_APPEND (&c, 1);
+				}
 
 				continue;
 
@@ -660,7 +1008,7 @@ rspamd_vprintf_common (rspamd_printf_append_func func,
 				continue;
 
 			case 'N':
-				c = LF;
+				c = '\n';
 				RSPAMD_PRINTF_APPEND (&c, 1);
 
 				continue;
