@@ -16,6 +16,7 @@
 #include "config.h"
 #include "rrd.h"
 #include "util.h"
+#include "cfg_file.h"
 #include "logger.h"
 #include "unix-std.h"
 #include "cryptobox.h"
@@ -345,7 +346,7 @@ rspamd_rrd_calculate_checksum (struct rspamd_rrd_file *file)
 
 		rspamd_cryptobox_hash_final (&st, sigbuf);
 
-		file->id = rspamd_encode_base32 (sigbuf, sizeof (sigbuf));
+		file->id = rspamd_encode_base32 (sigbuf, sizeof (sigbuf), RSPAMD_BASE32_DEFAULT);
 	}
 }
 
