@@ -66,7 +66,7 @@ gint rspamd_socket (const gchar *credits, guint16 port, gint type,
 /*
  * Create socketpair
  */
-gboolean rspamd_socketpair (gint pair[2], gboolean is_stream);
+gboolean rspamd_socketpair (gint pair[2], gint af);
 
 /*
  * Make specified socket non-blocking
@@ -478,6 +478,12 @@ double rspamd_set_counter_ema (struct rspamd_counter_data *cd,
 double rspamd_set_counter (struct rspamd_counter_data *cd,
 						   gdouble value);
 
+/**
+ * Shuffle elements in an array inplace
+ * @param ar
+ */
+void rspamd_ptr_array_shuffle (GPtrArray *ar);
+
 enum rspamd_pbkdf_version_id {
 	RSPAMD_PBKDF_ID_V1 = 1,
 	RSPAMD_PBKDF_ID_V2 = 2,
