@@ -50,6 +50,7 @@ lua_cdb_create (lua_State *L)
 	struct cdb *cdb, **pcdb;
 	const gchar *filename;
 	gint fd;
+	struct ev_loop *ev_base = lua_check_ev_base (L, 2);
 
 	filename = luaL_checkstring (L, 1);
 	/* If file begins with cdb://, just skip it */
@@ -64,13 +65,14 @@ lua_cdb_create (lua_State *L)
 	else {
 		cdb = g_malloc (sizeof (struct cdb));
 		cdb->filename = g_strdup (filename);
-		cdb->check_timer_ev = NULL;
-		cdb->check_timer_tv = NULL;
 		if (cdb_init (cdb, fd) == -1) {
+			g_free (cdb->filename);
+			g_free (cdb);
 			msg_warn ("cannot open cdb: %s, %s", filename, strerror (errno));
 			lua_pushnil (L);
 		}
 		else {
+			cdb_add_timer (cdb, ev_base, CDB_REFRESH_TIME);
 			pcdb = lua_newuserdata (L, sizeof (struct cdb *));
 			rspamd_lua_setclass (L, "rspamd{cdb}", -1);
 			*pcdb = cdb;
@@ -106,13 +108,6 @@ lua_cdb_lookup (lua_State *L)
 		lua_error (L);
 		return 1;
 	}
-	/*
-	 * XXX: this code is placed here because event_loop is called inside workers, so start
-	 * monitoring on first check, not on creation
-	 */
-	if (cdb->check_timer_ev == NULL) {
-		cdb_add_timer (cdb, CDB_REFRESH_TIME);
-	}
 
 	what = luaL_checkstring (L, 2);
 	if (cdb_find (cdb, what, strlen (what)) > 0) {
@@ -158,17 +153,7 @@ lua_load_cdb (lua_State *L)
 void
 luaopen_cdb (lua_State * L)
 {
-	luaL_newmetatable (L, "rspamd{cdb}");
-	lua_pushstring (L, "__index");
-	lua_pushvalue (L, -2);
-	lua_settable (L, -3);
-
-	lua_pushstring (L, "class");
-	lua_pushstring (L, "rspamd{cdb}");
-	lua_rawset (L, -3);
-
-	luaL_register (L, NULL,	 cdblib_m);
-	lua_pop (L, 1);                      /* remove metatable from stack */
-
+	rspamd_lua_new_class (L, "rspamd{cdb}", cdblib_m);
+	lua_pop (L, 1);
 	rspamd_lua_add_preload (L, "rspamd_cdb", lua_load_cdb);
 }
