@@ -89,6 +89,7 @@ rspamd_lua_new_class (lua_State * L,
 	void *class_ptr;
 	khiter_t k;
 	gint r, nmethods = 0;
+	gboolean seen_index = false;
 
 	k = kh_put (lua_class_set, lua_classes, classname, &r);
 	class_ptr = RSPAMD_LIGHTUSERDATA_MASK (kh_key (lua_classes, k));
@@ -96,6 +97,9 @@ rspamd_lua_new_class (lua_State * L,
 	if (methods) {
 		for (;;) {
 			if (methods[nmethods].name != NULL) {
+				if (strcmp (methods[nmethods].name, "__index") == 0) {
+					seen_index = true;
+				}
 				nmethods ++;
 			}
 			else {
@@ -105,9 +109,12 @@ rspamd_lua_new_class (lua_State * L,
 	}
 
 	lua_createtable (L, 0, 3 + nmethods);
-	lua_pushstring (L, "__index");
-	lua_pushvalue (L, -2);      /* pushes the metatable */
-	lua_settable (L, -3);       /* metatable.__index = metatable */
+
+	if (!seen_index) {
+		lua_pushstring (L, "__index");
+		lua_pushvalue (L, -2);      /* pushes the metatable */
+		lua_settable (L, -3);       /* metatable.__index = metatable */
+	}
 
 	lua_pushstring (L, "class");
 	lua_pushstring (L, classname);
@@ -197,6 +204,18 @@ rspamd_lua_setclass (lua_State * L, const gchar *classname, gint objidx)
 	lua_setmetatable (L, objidx);
 }
 
+void
+rspamd_lua_class_metatable (lua_State *L, const gchar *classname)
+{
+	khiter_t k;
+
+	k = kh_get (lua_class_set, lua_classes, classname);
+
+	g_assert (k != kh_end (lua_classes));
+	lua_rawgetp (L, LUA_REGISTRYINDEX,
+			RSPAMD_LIGHTUSERDATA_MASK (kh_key (lua_classes, k)));
+}
+
 void
 rspamd_lua_add_metamethod (lua_State *L, const gchar *classname,
 								luaL_Reg *meth)
@@ -960,14 +979,13 @@ rspamd_lua_init (bool wipe_mem)
 	luaopen_worker (L);
 	luaopen_kann (L);
 	luaopen_spf (L);
+	luaopen_tensor (L);
+	luaopen_parsers (L);
 #ifndef WITH_LUAJIT
 	rspamd_lua_add_preload (L, "bit", luaopen_bit);
 	lua_settop (L, 0);
 #endif
 
-	rspamd_lua_new_class (L, "rspamd{ev_base}", NULL);
-	lua_pop (L, 1);
-
 	rspamd_lua_new_class (L, "rspamd{session}", NULL);
 	lua_pop (L, 1);
 
@@ -1328,7 +1346,7 @@ rspamd_lua_parse_table_arguments (lua_State *L, gint pos,
 		read_semicolon
 	} state = read_key;
 	gsize keylen = 0, *valuelen, clslen;
-	gint idx = 0, t;
+	gint idx = 0, t, direct_userdata = 0;
 
 	g_assert (extraction_pattern != NULL);
 
@@ -1383,7 +1401,7 @@ rspamd_lua_parse_table_arguments (lua_State *L, gint pos,
 
 			t = lua_type (L, idx);
 
-			switch (g_ascii_toupper (*p)) {
+			switch (*p) {
 			case 'S':
 				if (t == LUA_TSTRING) {
 					*(va_arg (ap, const gchar **)) = lua_tostring (L, idx);
@@ -1449,6 +1467,38 @@ rspamd_lua_parse_table_arguments (lua_State *L, gint pos,
 				}
 				break;
 
+			case 'i':
+				if (t == LUA_TNUMBER) {
+					*(va_arg (ap, gint32 *)) = lua_tointeger (L, idx);
+				}
+				else if (t == LUA_TNIL || t == LUA_TNONE) {
+					failed = TRUE;
+					if (how != RSPAMD_LUA_PARSE_ARGUMENTS_IGNORE_MISSING) {
+						*(va_arg (ap, gint32 *)) = 0;
+					}
+					else {
+						(void)va_arg (ap, gint32 *);
+					}
+				}
+				else {
+					g_set_error (err,
+							lua_error_quark (),
+							1,
+							"bad type for key:"
+							" %.*s: '%s', '%s' is expected",
+							(gint) keylen,
+							key,
+							lua_typename (L, lua_type (L, idx)),
+							"int64");
+					va_end (ap);
+
+					return FALSE;
+				}
+				if (is_table) {
+					lua_pop (L, 1);
+				}
+				break;
+
 			case 'F':
 				if (t == LUA_TFUNCTION) {
 					if (!is_table) {
@@ -1676,6 +1726,39 @@ rspamd_lua_parse_table_arguments (lua_State *L, gint pos,
 
 				state = read_class_start;
 				clslen = 0;
+				direct_userdata = 0;
+				cls = NULL;
+				p ++;
+				continue;
+			case 'u':
+				if (t == LUA_TNIL || t == LUA_TNONE) {
+					failed = TRUE;
+
+					if (how != RSPAMD_LUA_PARSE_ARGUMENTS_IGNORE_MISSING) {
+						*(va_arg (ap, void **)) = NULL;
+					}
+					else {
+						(void)va_arg (ap, void **);
+					}
+				}
+				else if (t != LUA_TUSERDATA) {
+					g_set_error (err,
+							lua_error_quark (),
+							1,
+							"bad type for key:"
+							" %.*s: '%s', '%s' is expected",
+							(gint) keylen,
+							key,
+							lua_typename (L, lua_type (L, idx)),
+							"int64");
+					va_end (ap);
+
+					return FALSE;
+				}
+
+				state = read_class_start;
+				clslen = 0;
+				direct_userdata = 1;
 				cls = NULL;
 				p ++;
 				continue;
@@ -1753,7 +1836,14 @@ rspamd_lua_parse_table_arguments (lua_State *L, gint pos,
 				 */
 				if (!failed && (!is_table ||
 						rspamd_lua_check_class (L, idx, classbuf))) {
-					*(va_arg (ap, void **)) = *(void **)lua_touserdata (L, idx);
+					if (direct_userdata) {
+						void **arg_p = (va_arg (ap, void **));
+						*arg_p = lua_touserdata (L, idx);
+					}
+					else {
+						*(va_arg (ap,
+						void **)) = *(void **) lua_touserdata (L, idx);
+					}
 				}
 				else {
 					if (!failed) {
@@ -2069,8 +2159,8 @@ rspamd_lua_run_config_post_init (lua_State *L, struct rspamd_config *cfg)
 		rspamd_lua_setclass (L, "rspamd{config}", -1);
 
 		if (lua_pcall (L, 1, 0, err_idx) != 0) {
-			msg_err_config ("cannot run config post init script: %s",
-					lua_tostring (L, -1));
+			msg_err_config ("cannot run config post init script: %s; priority = %d",
+					lua_tostring (L, -1), sc->priority);
 		}
 
 		lua_settop (L, err_idx - 1);
@@ -2407,4 +2497,106 @@ rspamd_lua_get_module_name (lua_State *L)
 	}
 
 	return NULL;
-}
\ No newline at end of file
+}
+
+bool
+rspamd_lua_universal_pcall (lua_State *L, gint cbref, const gchar* strloc,
+								 gint nret, const gchar *args, GError **err, ...)
+{
+	va_list ap;
+	const gchar *argp = args, *classname;
+	gint err_idx, nargs = 0;
+	gpointer *cls_ptr;
+	gsize sz;
+
+	/* Error function */
+	lua_pushcfunction (L, &rspamd_lua_traceback);
+	err_idx = lua_gettop (L);
+
+	va_start (ap, err);
+	/* Called function */
+	lua_rawgeti (L, LUA_REGISTRYINDEX, cbref);
+	/*
+	 * Possible arguments
+	 * - i - lua_integer, argument - gint64
+	 * - n - lua_number, argument - gdouble
+	 * - s - lua_string, argument - const gchar * (zero terminated)
+	 * - l - lua_lstring, argument - (size_t + const gchar *) pair
+	 * - u - lua_userdata, argument - (const char * + void *) - classname + pointer
+	 * - b - lua_boolean, argument - gboolean (not bool due to varargs promotion)
+	 * - f - lua_function, argument - int - position of the function on stack (not lua_registry)
+	 * - t - lua_text, argument - int - position of the lua_text on stack (not lua_registry)
+	 */
+	while (*argp) {
+		switch (*argp) {
+		case 'i':
+			lua_pushinteger (L, va_arg (ap, gint64));
+			nargs ++;
+			break;
+		case 'n':
+			lua_pushnumber (L, va_arg (ap, gdouble));
+			nargs ++;
+			break;
+		case 's':
+			lua_pushstring (L, va_arg (ap, const gchar *));
+			nargs ++;
+			break;
+		case 'l':
+			sz = va_arg (ap, gsize);
+			lua_pushlstring (L, va_arg (ap, const gchar *), sz);
+			nargs ++;
+			break;
+		case 'b':
+			lua_pushboolean (L, va_arg (ap, gboolean));
+			nargs ++;
+			break;
+		case 'u':
+			classname = va_arg (ap, const gchar *);
+			cls_ptr = (gpointer *)lua_newuserdata (L, sizeof (gpointer));
+			*cls_ptr = va_arg (ap, gpointer);
+			rspamd_lua_setclass (L, classname, -1);
+			nargs ++;
+			break;
+		case 'f':
+		case 't':
+			lua_pushvalue (L, va_arg (ap, gint));
+			nargs ++;
+			break;
+		default:
+			lua_settop (L, err_idx - 1);
+			g_set_error (err, lua_error_quark (), EINVAL,
+					"invalid argument character: %c at %s",
+					*argp, argp);
+
+			return false;
+		}
+
+		argp ++;
+	}
+
+	if (lua_pcall (L, nargs, nret, err_idx) != 0) {
+		g_set_error (err, lua_error_quark (), EBADF,
+				"error when calling lua function from %s: %s",
+				strloc, lua_tostring (L, -1));
+		lua_settop (L, err_idx - 1);
+
+		return false;
+	}
+
+	lua_remove (L, err_idx);
+	va_end (ap);
+
+	return true;
+}
+
+#if defined( LUA_VERSION_NUM ) && LUA_VERSION_NUM <= 502
+gint
+rspamd_lua_geti (lua_State *L, int pos, int i)
+{
+	pos = lua_absindex (L, pos);
+	lua_pushinteger (L, i);
+	lua_gettable (L, pos);
+
+	return lua_type (L, -1);
+}
+#endif
\ No newline at end of file
