@@ -51,6 +51,7 @@ LUA_FUNCTION_DEF (dns_resolver, resolve_txt);
 LUA_FUNCTION_DEF (dns_resolver, resolve_mx);
 LUA_FUNCTION_DEF (dns_resolver, resolve_ns);
 LUA_FUNCTION_DEF (dns_resolver, resolve);
+LUA_FUNCTION_DEF (dns_resolver, idna_convert_utf8);
 
 void lua_push_dns_reply (lua_State *L, const struct rdns_reply *reply);
 
@@ -66,6 +67,7 @@ static const struct luaL_reg dns_resolverlib_m[] = {
 	LUA_INTERFACE_DEF (dns_resolver, resolve_mx),
 	LUA_INTERFACE_DEF (dns_resolver, resolve_ns),
 	LUA_INTERFACE_DEF (dns_resolver, resolve),
+	LUA_INTERFACE_DEF (dns_resolver, idna_convert_utf8),
 	{"__tostring", rspamd_lua_class_tostring},
 	{NULL, NULL}
 };
@@ -143,7 +145,6 @@ lua_dns_resolver_callback (struct rdns_reply *reply, gpointer arg)
 	struct lua_callback_state cbs;
 	rspamd_mempool_t *pool;
 	gint err_idx;
-	GString *tb = NULL;
 
 	pool = cd->pool;
 	lua_thread_pool_prepare_callback (cd->resolver->cfg->lua_thread_pool, &cbs);
@@ -199,12 +200,8 @@ lua_dns_resolver_callback (struct rdns_reply *reply, gpointer arg)
 	}
 
 	if (lua_pcall (L, 7, 0, err_idx) != 0) {
-		tb = lua_touserdata (L, -1);
-
-		if (tb) {
-			msg_err_pool_check ("call to dns callback failed: %s", tb->str);
-			g_string_free (tb, TRUE);
-		}
+		msg_err_pool_check ("call to dns callback failed: %s",
+				lua_tostring (L, -1));
 	}
 
 	lua_settop (L, err_idx - 1);
@@ -318,7 +315,7 @@ lua_dns_resolver_init (lua_State *L)
 {
 	struct rspamd_dns_resolver *resolver, **presolver;
 	struct rspamd_config *cfg, **pcfg;
-	struct event_base *base, **pbase;
+	struct ev_loop *base, **pbase;
 
 	/* Check args */
 	pbase = rspamd_lua_check_udata (L, 1, "rspamd{ev_base}");
@@ -365,6 +362,7 @@ lua_dns_resolver_resolve_common (lua_State *L,
 
 	/* Check arguments */
 	if (!rspamd_lua_parse_table_arguments (L, first, &err,
+			RSPAMD_LUA_PARSE_ARGUMENTS_DEFAULT,
 			"session=U{session};mempool=U{mempool};*name=S;*callback=F;"
 			"option=S;task=U{task};forced=B",
 			&session, &pool, &to_resolve, &cbref, &user_str, &task, &forced)) {
@@ -481,7 +479,6 @@ lua_dns_resolver_resolve_common (lua_State *L,
 			if (ret) {
 				cbdata->s = session;
 
-
 				if (item) {
 					cbdata->item = item;
 					rspamd_symcache_item_async_inc (task, item, M);
@@ -490,7 +487,11 @@ lua_dns_resolver_resolve_common (lua_State *L,
 				lua_pushboolean (L, TRUE);
 			}
 			else {
-				lua_pushnil (L);
+				if (item) {
+					rspamd_symcache_item_async_dec_check (task, item, M);
+				}
+
+				goto err;
 			}
 
 			if (item) {
@@ -510,6 +511,11 @@ err:
 		g_free (cbdata->user_str);
 	}
 
+	/* Callback is not called in this case */
+	if (cbdata->cbref != -1) {
+		luaL_unref (L, LUA_REGISTRYINDEX, cbdata->cbref);
+	}
+
 	lua_pushnil (L);
 
 	return 1;
@@ -684,6 +690,50 @@ lua_dns_resolver_resolve (lua_State *L)
 	return 1;
 }
 
+/***
+ * @method resolver:idna_convert_utf8(hostname[, pool])
+ * Converts domain name from IDN (in utf8 format) to punycode
+ * @return {string} new name converted
+ */
+static int
+lua_dns_resolver_idna_convert_utf8 (lua_State *L)
+{
+	struct rspamd_dns_resolver *dns_resolver = lua_check_dns_resolver (L, 1);
+	gsize hlen;
+	guint conv_len = 0;
+	const gchar *hname = luaL_checklstring (L, 2, &hlen);
+	gchar *converted;
+	rspamd_mempool_t *pool = rspamd_lua_check_udata_maybe (L, 3, "rspamd{mempool}");
+
+
+	if (dns_resolver && hname) {
+		if (!rspamd_str_has_8bit (hname, hlen)) {
+			/* No 8 bit, no reasons to call idna */
+			lua_pushlstring (L, hname, hlen);
+		}
+		else {
+			converted = rspamd_dns_resolver_idna_convert_utf8 (dns_resolver, pool,
+					hname, hlen, &conv_len);
+
+			if (converted == NULL) {
+				lua_pushnil (L);
+			}
+			else {
+				lua_pushlstring (L, converted, conv_len);
+
+				if (pool == NULL) {
+					g_free (converted);
+				}
+			}
+		}
+	}
+	else {
+		return luaL_error (L, "invalid arguments");
+	}
+
+	return 1;
+}
+
 static gint
 lua_load_dns_resolver (lua_State *L)
 {
@@ -697,15 +747,7 @@ void
 luaopen_dns_resolver (lua_State * L)
 {
 
-	luaL_newmetatable (L, "rspamd{resolver}");
-	lua_pushstring (L, "__index");
-	lua_pushvalue (L, -2);
-	lua_settable (L, -3);
-
-	lua_pushstring (L, "class");
-	lua_pushstring (L, "rspamd{resolver}");
-	lua_rawset (L, -3);
-
+	rspamd_lua_new_class (L, "rspamd{resolver}", dns_resolverlib_m);
 	{
 		LUA_ENUM (L, DNS_A,	 RDNS_REQUEST_A);
 		LUA_ENUM (L, DNS_PTR, RDNS_REQUEST_PTR);
@@ -717,8 +759,7 @@ luaopen_dns_resolver (lua_State * L)
 		LUA_ENUM (L, DNS_SOA, RDNS_REQUEST_SOA);
 	}
 
-	luaL_register (L, NULL, dns_resolverlib_m);
-	rspamd_lua_add_preload (L, "rspamd_resolver", lua_load_dns_resolver);
+	lua_pop (L, 1);
 
-	lua_pop (L, 1);                      /* remove metatable from stack */
+	rspamd_lua_add_preload (L, "rspamd_resolver", lua_load_dns_resolver);
 }
