@@ -15,6 +15,7 @@
  */
 
 #include "lua_common.h"
+#include "lua_tensor.h"
 #include "contrib/kann/kann.h"
 
 /***
@@ -1022,6 +1023,7 @@ static int
 lua_kann_train1 (lua_State *L)
 {
 	kann_t *k = lua_check_kann (L, 1);
+	struct rspamd_lua_tensor *pca = NULL;
 
 	/* Default train params */
 	double lr = 0.001;
@@ -1054,8 +1056,8 @@ lua_kann_train1 (lua_State *L)
 
 			if (!rspamd_lua_parse_table_arguments (L, 4, &err,
 					RSPAMD_LUA_PARSE_ARGUMENTS_IGNORE_MISSING,
-					"lr=N;mini_size=I;max_epoch=I;max_drop_streak=I;frac_val=N;cb=F",
-					&lr, &mini_size, &max_epoch, &max_drop_streak, &frac_val, &cbref)) {
+					"lr=N;mini_size=I;max_epoch=I;max_drop_streak=I;frac_val=N;cb=F;pca=u{tensor}",
+					&lr, &mini_size, &max_epoch, &max_drop_streak, &frac_val, &cbref, &pca)) {
 				n = luaL_error (L, "invalid params: %s",
 						err ? err->message : "unknown error");
 				g_error_free (err);
@@ -1064,35 +1066,83 @@ lua_kann_train1 (lua_State *L)
 			}
 		}
 
-		float **x, **y;
+		if (pca) {
+			/* Check pca matrix validity */
+			if (pca->ndims != 2) {
+				return luaL_error (L, "invalid pca tensor: matrix expected, got a row");
+			}
 
-		/* Fill vectors */
+			if (pca->dim[0] != n_in) {
+				return luaL_error (L, "invalid pca tensor: "
+						  "matrix must have %d rows and it has %d rows instead",
+						  n_in, pca->dim[0]);
+			}
+		}
+
+		float **x, **y, *tmp_row = NULL;
+
+		/* Fill vectors row by row */
 		x = (float **)g_malloc0 (sizeof (float *) * n);
 		y = (float **)g_malloc0 (sizeof (float *) * n);
 
+		if (pca) {
+			tmp_row = g_malloc (sizeof (float) * pca->dim[1]);
+		}
+
 		for (int s = 0; s < n; s ++) {
 			/* Inputs */
 			lua_rawgeti (L, 2, s + 1);
 			x[s] = (float *)g_malloc (sizeof (float) * n_in);
 
-			if (rspamd_lua_table_size (L, -1) != n_in) {
-				FREE_VEC (x, n);
-				FREE_VEC (y, n);
+			if (pca == NULL) {
+				if (rspamd_lua_table_size (L, -1) != n_in) {
+					FREE_VEC (x, n);
+					FREE_VEC (y, n);
 
-				n = luaL_error (L, "invalid params at pos %d: "
-					   "bad input dimension %d; %d expected",
-						s + 1,
-						(int)rspamd_lua_table_size (L, -1),
-						n_in);
+					n = luaL_error (L, "invalid params at pos %d: "
+									   "bad input dimension %d; %d expected",
+							s + 1,
+							(int) rspamd_lua_table_size (L, -1),
+							n_in);
+					lua_pop (L, 1);
 
-				return n;
+					return n;
+				}
+
+				for (int i = 0; i < n_in; i++) {
+					lua_rawgeti (L, -1, i + 1);
+					x[s][i] = lua_tonumber (L, -1);
+
+					lua_pop (L, 1);
+				}
 			}
+			else {
+				if (rspamd_lua_table_size (L, -1) != pca->dim[1]) {
+					FREE_VEC (x, n);
+					FREE_VEC (y, n);
+					g_free (tmp_row);
 
-			for (int i = 0; i < n_in; i ++) {
-				lua_rawgeti (L, -1, i + 1);
-				x[s][i] = lua_tonumber (L, -1);
+					n = luaL_error (L, "(pca on) invalid params at pos %d: "
+									   "bad input dimension %d; %d expected",
+							s + 1,
+							(int) rspamd_lua_table_size (L, -1),
+							pca->dim[1]);
+					lua_pop (L, 1);
 
-				lua_pop (L, 1);
+					return n;
+				}
+
+
+				for (int i = 0; i < pca->dim[1]; i++) {
+					lua_rawgeti (L, -1, i + 1);
+					tmp_row[i] = lua_tonumber (L, -1);
+
+					lua_pop (L, 1);
+				}
+
+				kad_sgemm_simple (0, 1, 1, n_in,
+						pca->dim[1], tmp_row, pca->data,
+						x[s]);
 			}
 
 			lua_pop (L, 1);
@@ -1104,6 +1154,7 @@ lua_kann_train1 (lua_State *L)
 			if (rspamd_lua_table_size (L, -1) != n_out) {
 				FREE_VEC (x, n);
 				FREE_VEC (y, n);
+				g_free (tmp_row);
 
 				n = luaL_error (L, "invalid params at pos %d: "
 					   "bad output dimension %d; "
@@ -1111,6 +1162,7 @@ lua_kann_train1 (lua_State *L)
 						s + 1,
 						(int)rspamd_lua_table_size (L, -1),
 						n_out);
+				lua_pop (L, 1);
 
 				return n;
 			}
@@ -1139,6 +1191,7 @@ lua_kann_train1 (lua_State *L)
 
 		FREE_VEC (x, n);
 		FREE_VEC (y, n);
+		g_free (tmp_row);
 	}
 	else {
 		return luaL_error (L, "invalid arguments: kann, inputs, outputs and"
@@ -1152,49 +1205,128 @@ static int
 lua_kann_apply1 (lua_State *L)
 {
 	kann_t *k = lua_check_kann (L, 1);
+	struct rspamd_lua_tensor *pca = NULL;
 
-	if (k && lua_istable (L, 2)) {
-		gsize vec_len = rspamd_lua_table_size (L, 2);
-		float *vec = (float *)g_malloc (sizeof (float) * vec_len);
-		int i_out;
-		int n_in = kann_dim_in (k);
+	if (k) {
+		if (lua_istable (L, 2)) {
+			gsize vec_len = rspamd_lua_table_size (L, 2);
+			float *vec = (float *) g_malloc (sizeof (float) * vec_len),
+				*pca_out = NULL;
+			int i_out;
+			int n_in = kann_dim_in (k);
 
-		if (n_in <= 0) {
-			return luaL_error (L, "invalid inputs count: %d", n_in);
-		}
+			if (n_in <= 0) {
+				g_free (vec);
+				return luaL_error (L, "invalid inputs count: %d", n_in);
+			}
 
-		if (n_in != vec_len) {
-			return luaL_error (L, "invalid params: bad input dimension %d; %d expected",
-					(int)vec_len, n_in);
-		}
+			if (lua_isuserdata (L, 3)) {
+				pca = lua_check_tensor (L, 3);
 
-		for (gsize i = 0; i < vec_len; i ++) {
-			lua_rawgeti (L, 2, i + 1);
-			vec[i] = lua_tonumber (L, -1);
-			lua_pop (L, 1);
-		}
+				if (pca) {
+					if (pca->ndims != 2) {
+						g_free (vec);
+						return luaL_error (L, "invalid pca tensor: matrix expected, got a row");
+					}
 
-		i_out = kann_find (k, KANN_F_OUT, 0);
+					if (pca->dim[0] != n_in) {
+						g_free (vec);
+						return luaL_error (L, "invalid pca tensor: "
+											  "matrix must have %d rows and it has %d rows instead",
+								n_in, pca->dim[0]);
+					}
+				}
+				else {
+					g_free (vec);
+					return luaL_error (L, "invalid params: pca matrix expected");
+				}
+			}
+			else {
+				if (n_in != vec_len) {
+					g_free (vec);
+					return luaL_error (L, "invalid params: bad input dimension %d; %d expected",
+							(int) vec_len, n_in);
+				}
+			}
+
+			for (gsize i = 0; i < vec_len; i++) {
+				lua_rawgeti (L, 2, i + 1);
+				vec[i] = lua_tonumber (L, -1);
+				lua_pop (L, 1);
+			}
+
+			i_out = kann_find (k, KANN_F_OUT, 0);
+
+			if (i_out <= 0) {
+				g_free (vec);
+				return luaL_error (L, "invalid ANN: output layer is missing or is "
+									  "at the input pos");
+			}
+
+			kann_set_batch_size (k, 1);
+			if (pca) {
+				pca_out = g_malloc (sizeof (float) * n_in);
+
+				kad_sgemm_simple (0, 1, 1, n_in,
+						vec_len, vec, pca->data,
+						pca_out);
+
+				kann_feed_bind (k, KANN_F_IN, 0, &pca_out);
+			}
+			else {
+				kann_feed_bind (k, KANN_F_IN, 0, &vec);
+			}
+
+			kad_eval_at (k->n, k->v, i_out);
+
+			gsize outlen = kad_len (k->v[i_out]);
+			lua_createtable (L, outlen, 0);
+
+			for (gsize i = 0; i < outlen; i++) {
+				lua_pushnumber (L, k->v[i_out]->x[i]);
+				lua_rawseti (L, -2, i + 1);
+			}
 
-		if (i_out <= 0) {
 			g_free (vec);
-			return luaL_error (L, "invalid ANN: output layer is missing or is "
-						 "at the input pos");
+			g_free (pca_out);
 		}
+		else if (lua_isuserdata (L, 2)) {
+			struct rspamd_lua_tensor *t = lua_check_tensor (L, 2);
 
-		kann_set_batch_size (k, 1);
-		kann_feed_bind (k, KANN_F_IN, 0, &vec);
-		kad_eval_at (k->n, k->v, i_out);
+			if (t && t->ndims == 1) {
+				int i_out;
+				int n_in = kann_dim_in (k);
 
-		gsize outlen = kad_len (k->v[i_out]);
-		lua_createtable (L, outlen, 0);
+				if (n_in != t->dim[0]) {
+					return luaL_error (L, "invalid params: bad input dimension %d; %d expected",
+							(int) t->dim[0], n_in);
+				}
 
-		for (gsize i = 0; i < outlen; i ++) {
-			lua_pushnumber (L, k->v[i_out]->x[i]);
-			lua_rawseti (L, -2, i + 1);
+				i_out = kann_find (k, KANN_F_OUT, 0);
+
+				if (i_out <= 0) {
+					return luaL_error (L, "invalid ANN: output layer is missing or is "
+										  "at the input pos");
+				}
+
+				kann_set_batch_size (k, 1);
+				kann_feed_bind (k, KANN_F_IN, 0, &t->data);
+				kad_eval_at (k->n, k->v, i_out);
+
+				gint outlen = kad_len (k->v[i_out]);
+				struct rspamd_lua_tensor *out;
+				out = lua_newtensor (L, 1, &outlen, false, false);
+				/* Ensure that kann and tensor have the same understanding of floats */
+				G_STATIC_ASSERT (sizeof (float) == sizeof (rspamd_tensor_num_t));
+				memcpy (out->data, k->v[i_out]->x, outlen * sizeof (float));
+			}
+			else {
+				return luaL_error (L, "invalid arguments: 1D rspamd{tensor} expected");
+			}
+		}
+		else {
+			return luaL_error (L, "invalid arguments: 1D rspamd{tensor} expected");
 		}
-
-		g_free (vec);
 	}
 	else {
 		return luaL_error (L, "invalid arguments: rspamd{kann} expected");
