@@ -15,9 +15,9 @@
  */
 
 #include "lua_common.h"
-#include "libutil/map.h"
-#include "libutil/map_helpers.h"
-#include "libutil/map_private.h"
+#include "libserver/maps/map.h"
+#include "libserver/maps/map_helpers.h"
+#include "libserver/maps/map_private.h"
 #include "contrib/libucl/lua_ucl.h"
 
 /***
@@ -129,6 +129,7 @@ static const struct luaL_reg maplib_m[] = {
 struct lua_map_callback_data {
 	lua_State *L;
 	gint ref;
+	gboolean opaque;
 	rspamd_fstring_t *data;
 	struct rspamd_lua_map *lua_map;
 };
@@ -161,7 +162,8 @@ lua_config_add_radix_map (lua_State *L)
 				rspamd_radix_read,
 				rspamd_radix_fin,
 				rspamd_radix_dtor,
-				(void **)&map->data.radix)) == NULL) {
+				(void **)&map->data.radix,
+				NULL, RSPAMD_MAP_DEFAULT)) == NULL) {
 			msg_warn_config ("invalid radix map %s", map_line);
 			lua_pushnil (L);
 
@@ -218,7 +220,8 @@ lua_config_radix_from_config (lua_State *L)
 					rspamd_radix_read,
 					rspamd_radix_fin,
 					rspamd_radix_dtor,
-					(void **)&map->data.radix)) == NULL) {
+					(void **)&map->data.radix,
+					NULL, RSPAMD_MAP_DEFAULT)) == NULL) {
 				msg_err_config ("invalid radix map static");
 				lua_pushnil (L);
 				ucl_object_unref (fake_obj);
@@ -279,7 +282,8 @@ lua_config_radix_from_ucl (lua_State *L)
 				rspamd_radix_read,
 				rspamd_radix_fin,
 				rspamd_radix_dtor,
-				(void **)&map->data.radix)) == NULL) {
+				(void **)&map->data.radix,
+				NULL, RSPAMD_MAP_DEFAULT)) == NULL) {
 			msg_err_config ("invalid radix map static");
 			lua_pushnil (L);
 			ucl_object_unref (fake_obj);
@@ -324,7 +328,8 @@ lua_config_add_hash_map (lua_State *L)
 				rspamd_kv_list_read,
 				rspamd_kv_list_fin,
 				rspamd_kv_list_dtor,
-				(void **)&map->data.hash)) == NULL) {
+				(void **)&map->data.hash,
+				NULL, RSPAMD_MAP_DEFAULT)) == NULL) {
 			msg_warn_config ("invalid set map %s", map_line);
 			lua_pushnil (L);
 			return 1;
@@ -364,7 +369,8 @@ lua_config_add_kv_map (lua_State *L)
 				rspamd_kv_list_read,
 				rspamd_kv_list_fin,
 				rspamd_kv_list_dtor,
-				(void **)&map->data.hash)) == NULL) {
+				(void **)&map->data.hash,
+				NULL, RSPAMD_MAP_DEFAULT)) == NULL) {
 			msg_warn_config ("invalid hash map %s", map_line);
 			lua_pushnil (L);
 
@@ -436,17 +442,38 @@ lua_map_fin (struct map_cb_data *data, void **target)
 		msg_err_map ("map has no callback set");
 	}
 	else if (cbdata->data != NULL && cbdata->data->len != 0) {
+
+		lua_pushcfunction (cbdata->L, &rspamd_lua_traceback);
+		int err_idx = lua_gettop (cbdata->L);
+
 		lua_rawgeti (cbdata->L, LUA_REGISTRYINDEX, cbdata->ref);
-		lua_pushlstring (cbdata->L, cbdata->data->str, cbdata->data->len);
+
+		if (!cbdata->opaque) {
+			lua_pushlstring (cbdata->L, cbdata->data->str, cbdata->data->len);
+		}
+		else {
+			struct rspamd_lua_text *t;
+
+			t = lua_newuserdata (cbdata->L, sizeof (*t));
+			rspamd_lua_setclass (cbdata->L, "rspamd{text}", -1);
+			t->flags = 0;
+			t->len = cbdata->data->len;
+			t->start = cbdata->data->str;
+		}
+
 		pmap = lua_newuserdata (cbdata->L, sizeof (void *));
 		*pmap = cbdata->lua_map;
 		rspamd_lua_setclass (cbdata->L, "rspamd{map}", -1);
 
-		if (lua_pcall (cbdata->L, 2, 0, 0) != 0) {
-			msg_info_map ("call to %s failed: %s", "local function",
+		gint ret = lua_pcall (cbdata->L, 2, 0, err_idx);
+
+		if (ret != 0) {
+			msg_info_map ("call to %s failed (%d): %s", "map fin function",
+				ret,
 				lua_tostring (cbdata->L, -1));
-			lua_pop (cbdata->L, 1);
 		}
+
+		lua_settop (cbdata->L, err_idx - 1);
 	}
 
 	cbdata->data = rspamd_fstring_assign (cbdata->data, "", 0);
@@ -488,13 +515,15 @@ lua_config_add_map (lua_State *L)
 	struct lua_map_callback_data *cbdata;
 	struct rspamd_lua_map *map, **pmap;
 	struct rspamd_map *m;
+	gboolean opaque_data = FALSE;
 	int cbidx = -1, ret;
 	GError *err = NULL;
 
 	if (cfg) {
 		if (!rspamd_lua_parse_table_arguments (L, 2, &err,
-				"*url=O;description=S;callback=F;type=S",
-				&map_obj, &description, &cbidx, &type)) {
+				RSPAMD_LUA_PARSE_ARGUMENTS_DEFAULT,
+				"*url=O;description=S;callback=F;type=S;opaque_data=B",
+				&map_obj, &description, &cbidx, &type, &opaque_data)) {
 			ret = luaL_error (L, "invalid table arguments: %s", err->message);
 			g_error_free (err);
 			if (map_obj) {
@@ -523,12 +552,14 @@ lua_config_add_map (lua_State *L)
 			cbdata->data = NULL;
 			cbdata->lua_map = map;
 			cbdata->ref = cbidx;
+			cbdata->opaque = opaque_data;
 
 			if ((m = rspamd_map_add_from_ucl (cfg, map_obj, description,
 					lua_map_read,
 					lua_map_fin,
 					lua_map_dtor,
-					(void **)&map->data.cbdata)) == NULL) {
+					(void **)&map->data.cbdata,
+					NULL, RSPAMD_MAP_DEFAULT)) == NULL) {
 
 				if (cbidx != -1) {
 					luaL_unref (L, LUA_REGISTRYINDEX, cbidx);
@@ -553,7 +584,8 @@ lua_config_add_map (lua_State *L)
 					rspamd_kv_list_read,
 					rspamd_kv_list_fin,
 					rspamd_kv_list_dtor,
-					(void **)&map->data.hash)) == NULL) {
+					(void **)&map->data.hash,
+					NULL, RSPAMD_MAP_DEFAULT)) == NULL) {
 				lua_pushnil (L);
 				ucl_object_unref (map_obj);
 
@@ -570,7 +602,8 @@ lua_config_add_map (lua_State *L)
 					rspamd_kv_list_read,
 					rspamd_kv_list_fin,
 					rspamd_kv_list_dtor,
-					(void **)&map->data.hash)) == NULL) {
+					(void **)&map->data.hash,
+					NULL, RSPAMD_MAP_DEFAULT)) == NULL) {
 				lua_pushnil (L);
 				ucl_object_unref (map_obj);
 
@@ -587,7 +620,8 @@ lua_config_add_map (lua_State *L)
 					rspamd_radix_read,
 					rspamd_radix_fin,
 					rspamd_radix_dtor,
-					(void **)&map->data.radix)) == NULL) {
+					(void **)&map->data.radix,
+					NULL, RSPAMD_MAP_DEFAULT)) == NULL) {
 				lua_pushnil (L);
 				ucl_object_unref (map_obj);
 
@@ -604,7 +638,8 @@ lua_config_add_map (lua_State *L)
 					rspamd_regexp_list_read_single,
 					rspamd_regexp_list_fin,
 					rspamd_regexp_list_dtor,
-					(void **) &map->data.re_map)) == NULL) {
+					(void **) &map->data.re_map,
+					NULL, RSPAMD_MAP_DEFAULT)) == NULL) {
 				lua_pushnil (L);
 				ucl_object_unref (map_obj);
 
@@ -621,7 +656,8 @@ lua_config_add_map (lua_State *L)
 					rspamd_regexp_list_read_multiple,
 					rspamd_regexp_list_fin,
 					rspamd_regexp_list_dtor,
-					(void **) &map->data.re_map)) == NULL) {
+					(void **) &map->data.re_map,
+					NULL, RSPAMD_MAP_DEFAULT)) == NULL) {
 				lua_pushnil (L);
 				ucl_object_unref (map_obj);
 
@@ -638,7 +674,8 @@ lua_config_add_map (lua_State *L)
 					rspamd_glob_list_read_single,
 					rspamd_regexp_list_fin,
 					rspamd_regexp_list_dtor,
-					(void **) &map->data.re_map)) == NULL) {
+					(void **) &map->data.re_map,
+					NULL, RSPAMD_MAP_DEFAULT)) == NULL) {
 				lua_pushnil (L);
 				ucl_object_unref (map_obj);
 
@@ -655,7 +692,26 @@ lua_config_add_map (lua_State *L)
 					rspamd_glob_list_read_multiple,
 					rspamd_regexp_list_fin,
 					rspamd_regexp_list_dtor,
-					(void **) &map->data.re_map)) == NULL) {
+					(void **) &map->data.re_map,
+					NULL, RSPAMD_MAP_DEFAULT)) == NULL) {
+				lua_pushnil (L);
+				ucl_object_unref (map_obj);
+
+				return 1;
+			}
+			m->lua_map = map;
+		}
+		else if (strcmp (type, "cdb") == 0) {
+			map = rspamd_mempool_alloc0 (cfg->cfg_pool, sizeof (*map));
+			map->data.cdb_map = NULL;
+			map->type = RSPAMD_LUA_MAP_CDB;
+
+			if ((m = rspamd_map_add_from_ucl (cfg, map_obj, description,
+					rspamd_cdb_list_read,
+					rspamd_cdb_list_fin,
+					rspamd_cdb_list_dtor,
+					(void **) &map->data.cdb_map,
+					NULL, RSPAMD_MAP_FILE_ONLY|RSPAMD_MAP_FILE_NO_READ)) == NULL) {
 				lua_pushnil (L);
 				ucl_object_unref (map_obj);
 
@@ -780,7 +836,6 @@ lua_map_get_key (lua_State * L)
 
 			if (lua_type (L, 2) == LUA_TSTRING) {
 				const gchar *addr_str;
-				gsize len;
 
 				addr_str = luaL_checklstring (L, 2, &len);
 				addr = g_alloca (sizeof (*addr));
@@ -788,7 +843,6 @@ lua_map_get_key (lua_State * L)
 
 				if (!rspamd_parse_inet_address_ip (addr_str, len, addr->addr)) {
 					addr = NULL;
-					msg_err ("invalid ip address: %*s", (gint)len, addr_str);
 				}
 			}
 			else if (lua_type (L, 2) == LUA_TUSERDATA) {
@@ -805,7 +859,7 @@ lua_map_get_key (lua_State * L)
 				}
 			}
 			else if (lua_type (L, 2) == LUA_TNUMBER) {
-				key_num = luaL_checknumber (L, 2);
+				key_num = luaL_checkinteger (L, 2);
 				key_num = htonl (key_num);
 			}
 
@@ -843,7 +897,7 @@ lua_map_get_key (lua_State * L)
 			key = lua_map_process_string_key (L, 2, &len);
 
 			if (key && map->data.hash) {
-				ret = rspamd_match_hash_map (map->data.hash, key) != NULL;
+				ret = rspamd_match_hash_map (map->data.hash, key, len) != NULL;
 			}
 		}
 		else if (map->type == RSPAMD_LUA_MAP_REGEXP) {
@@ -889,7 +943,7 @@ lua_map_get_key (lua_State * L)
 			key = lua_map_process_string_key (L, 2, &len);
 
 			if (key && map->data.hash) {
-				value = rspamd_match_hash_map (map->data.hash, key);
+				value = rspamd_match_hash_map (map->data.hash, key, len);
 			}
 
 			if (value) {
@@ -897,6 +951,21 @@ lua_map_get_key (lua_State * L)
 				return 1;
 			}
 		}
+		else if (map->type == RSPAMD_LUA_MAP_CDB) {
+			/* cdb map */
+			const rspamd_ftok_t *tok = NULL;
+
+			key = lua_map_process_string_key (L, 2, &len);
+
+			if (key && map->data.cdb_map) {
+				tok = rspamd_match_cdb_map (map->data.cdb_map, key, len);
+			}
+
+			if (tok) {
+				lua_pushlstring (L, tok->begin, tok->len);
+				return 1;
+			}
+		}
 		else {
 			/* callback map or unknown type map */
 			lua_pushnil (L);
@@ -996,7 +1065,7 @@ lua_map_is_signed (lua_State *L)
 		if (map->map) {
 			for (i = 0; i < map->map->backends->len; i ++) {
 				bk = g_ptr_array_index (map->map->backends, i);
-				if (bk->is_signed) {
+				if (bk->is_signed && bk->protocol == MAP_PROTO_FILE) {
 					ret = TRUE;
 					break;
 				}
@@ -1151,15 +1220,13 @@ lua_map_get_uri (lua_State *L)
 {
 	LUA_TRACE_POINT;
 	struct rspamd_lua_map *map = lua_check_map (L, 1);
-	const gchar *ret = "undefined";
 	struct rspamd_map_backend *bk;
-		guint i;
+	guint i;
 
 	if (map != NULL) {
 		for (i = 0; i < map->map->backends->len; i ++) {
 			bk = g_ptr_array_index (map->map->backends, i);
-			ret = bk->uri;
-			lua_pushstring (L, ret);
+			lua_pushstring (L, bk->uri);
 		}
 	}
 	else {
