@@ -15,12 +15,15 @@
  */
 
 #include "lua_common.h"
+#include "lua_url.h"
 #include "libmime/message.h"
 #include "libmime/lang_detection.h"
 #include "libstat/stat_api.h"
 #include "libcryptobox/cryptobox.h"
 #include "libutil/shingles.h"
 
+#include "contrib/uthash/utlist.h"
+
 /* Textpart methods */
 /***
  * @module rspamd_textpart
@@ -146,6 +149,24 @@ LUA_FUNCTION_DEF (textpart, get_words_count);
  */
 LUA_FUNCTION_DEF (textpart, get_words);
 
+/***
+ * @method mime_part:filter_words(regexp, [how][, max]])
+ * Filter words using some regexp:
+ * - `stem`: stemmed words (default)
+ * - `norm`: normalised words (utf normalised + lowercased)
+ * - `raw`: raw words in utf (if possible)
+ * - `full`: list of tables, each table has the following fields:
+ *   - [1] - stemmed word
+ *   - [2] - normalised word
+ *   - [3] - raw word
+ *   - [4] - flags (table of strings)
+ * @param {rspamd_regexp} regexp regexp to match
+ * @param {string} how what words to extract
+ * @param {number} max maximum number of hits returned (all hits if <= 0 or nil)
+ * @return {table/strings} words matching regexp
+ */
+LUA_FUNCTION_DEF (textpart, filter_words);
+
 /***
  * @method text_part:is_empty()
  * Returns `true` if the specified part is empty
@@ -216,6 +237,7 @@ static const struct luaL_reg textpartlib_m[] = {
 	LUA_INTERFACE_DEF (textpart, get_lines_count),
 	LUA_INTERFACE_DEF (textpart, get_words_count),
 	LUA_INTERFACE_DEF (textpart, get_words),
+	LUA_INTERFACE_DEF (textpart, filter_words),
 	LUA_INTERFACE_DEF (textpart, is_empty),
 	LUA_INTERFACE_DEF (textpart, is_html),
 	LUA_INTERFACE_DEF (textpart, get_html),
@@ -344,18 +366,25 @@ LUA_FUNCTION_DEF (mimepart, get_type_full);
 
 /***
  * @method mime_part:get_detected_type()
- * Extract content-type string of the mime part. Use libmagic detection
+ * Extract content-type string of the mime part. Use lua_magic detection
  * @return {string,string} content type in form 'type','subtype'
  */
 LUA_FUNCTION_DEF (mimepart, get_detected_type);
 
 /***
  * @method mime_part:get_detected_type_full()
- * Extract content-type string of the mime part with all attributes. Use libmagic detection
+ * Extract content-type string of the mime part with all attributes. Use lua_magic detection
  * @return {string,string,table} content type in form 'type','subtype', {attrs}
  */
 LUA_FUNCTION_DEF (mimepart, get_detected_type_full);
 
+/***
+ * @method mime_part:get_detected_ext()
+ * Returns a msdos extension name according to lua_magic detection
+ * @return {string} detected extension (see lua_magic.types)
+ */
+LUA_FUNCTION_DEF (mimepart, get_detected_ext);
+
 /***
  * @method mime_part:get_cte()
  * Extract content-transfer-encoding for a part
@@ -494,6 +523,35 @@ LUA_FUNCTION_DEF (mimepart, headers_foreach);
  */
 LUA_FUNCTION_DEF (mimepart, get_parent);
 
+/***
+ * @method mime_part:get_specific()
+ * Returns specific lua content for this part
+ * @return {any} specific lua content
+ */
+LUA_FUNCTION_DEF (mimepart, get_specific);
+
+/***
+ * @method mime_part:set_specific(<any>)
+ * Sets a specific content for this part
+ * @return {any} previous specific lua content (or nil)
+ */
+LUA_FUNCTION_DEF (mimepart, set_specific);
+
+/***
+ * @method mime_part:is_specific(<any>)
+ * Returns true if part has specific lua content
+ * @return {boolean} flag
+ */
+LUA_FUNCTION_DEF (mimepart, is_specific);
+
+/***
+ * @method mime_part:get_urls([need_emails|list_protos][, need_images])
+ * Get all URLs found in a mime part. Telephone urls and emails are not included unless explicitly asked in `list_protos`
+ * @param {boolean} need_emails if `true` then reutrn also email urls, this can be a comma separated string of protocols desired or a table (e.g. `mailto` or `telephone`)
+ * @param {boolean} need_images return urls from images (<img src=...>) as well
+ * @return {table rspamd_url} list of all urls found
+ */
+LUA_FUNCTION_DEF (mimepart, get_urls);
 
 static const struct luaL_reg mimepartlib_m[] = {
 	LUA_INTERFACE_DEF (mimepart, get_content),
@@ -502,6 +560,7 @@ static const struct luaL_reg mimepartlib_m[] = {
 	LUA_INTERFACE_DEF (mimepart, get_type),
 	LUA_INTERFACE_DEF (mimepart, get_type_full),
 	LUA_INTERFACE_DEF (mimepart, get_detected_type),
+	LUA_INTERFACE_DEF (mimepart, get_detected_ext),
 	LUA_INTERFACE_DEF (mimepart, get_detected_type_full),
 	LUA_INTERFACE_DEF (mimepart, get_cte),
 	LUA_INTERFACE_DEF (mimepart, get_filename),
@@ -519,6 +578,7 @@ static const struct luaL_reg mimepartlib_m[] = {
 	LUA_INTERFACE_DEF (mimepart, is_message),
 	LUA_INTERFACE_DEF (mimepart, get_children),
 	LUA_INTERFACE_DEF (mimepart, get_parent),
+	LUA_INTERFACE_DEF (mimepart, get_urls),
 	LUA_INTERFACE_DEF (mimepart, is_text),
 	LUA_INTERFACE_DEF (mimepart, is_broken),
 	LUA_INTERFACE_DEF (mimepart, is_attachment),
@@ -526,6 +586,9 @@ static const struct luaL_reg mimepartlib_m[] = {
 	LUA_INTERFACE_DEF (mimepart, get_digest),
 	LUA_INTERFACE_DEF (mimepart, get_id),
 	LUA_INTERFACE_DEF (mimepart, headers_foreach),
+	LUA_INTERFACE_DEF (mimepart, get_specific),
+	LUA_INTERFACE_DEF (mimepart, set_specific),
+	LUA_INTERFACE_DEF (mimepart, is_specific),
 	{"__tostring", rspamd_lua_class_tostring},
 	{NULL, NULL}
 };
@@ -554,12 +617,12 @@ lua_textpart_is_utf (lua_State * L)
 	LUA_TRACE_POINT;
 	struct rspamd_mime_text_part *part = lua_check_textpart (L);
 
-	if (part == NULL || IS_PART_EMPTY (part)) {
+	if (part == NULL || IS_TEXT_PART_EMPTY (part)) {
 		lua_pushboolean (L, FALSE);
 		return 1;
 	}
 
-	lua_pushboolean (L, IS_PART_UTF (part));
+	lua_pushboolean (L, IS_TEXT_PART_UTF (part));
 
 	return 1;
 }
@@ -572,7 +635,7 @@ lua_textpart_has_8bit_raw (lua_State * L)
 	struct rspamd_mime_text_part *part = lua_check_textpart (L);
 
 	if (part) {
-		if (part->flags & RSPAMD_MIME_TEXT_PART_FLAG_8BIT) {
+		if (part->flags & RSPAMD_MIME_TEXT_PART_FLAG_8BIT_RAW) {
 			lua_pushboolean (L, TRUE);
 		}
 		else {
@@ -627,7 +690,7 @@ lua_textpart_get_content (lua_State * L)
 	}
 
 	if (!type) {
-		if (IS_PART_EMPTY (part)) {
+		if (IS_TEXT_PART_EMPTY (part)) {
 			lua_pushnil (L);
 			return 1;
 		}
@@ -635,7 +698,7 @@ lua_textpart_get_content (lua_State * L)
 		len = part->utf_content->len;
 	}
 	else if (strcmp (type, "content") == 0) {
-		if (IS_PART_EMPTY (part)) {
+		if (IS_TEXT_PART_EMPTY (part)) {
 			lua_pushnil (L);
 			return 1;
 		}
@@ -644,7 +707,7 @@ lua_textpart_get_content (lua_State * L)
 		len = part->utf_content->len;
 	}
 	else if (strcmp (type, "content_oneline") == 0) {
-		if (IS_PART_EMPTY (part)) {
+		if (IS_TEXT_PART_EMPTY (part)) {
 			lua_pushnil (L);
 			return 1;
 		}
@@ -700,7 +763,7 @@ lua_textpart_get_raw_content (lua_State * L)
 	struct rspamd_mime_text_part *part = lua_check_textpart (L);
 	struct rspamd_lua_text *t;
 
-	if (part == NULL || IS_PART_EMPTY (part)) {
+	if (part == NULL || IS_TEXT_PART_EMPTY (part)) {
 		lua_pushnil (L);
 		return 1;
 	}
@@ -721,7 +784,7 @@ lua_textpart_get_content_oneline (lua_State * L)
 	struct rspamd_mime_text_part *part = lua_check_textpart (L);
 	struct rspamd_lua_text *t;
 
-	if (part == NULL || IS_PART_EMPTY (part)) {
+	if (part == NULL || IS_TEXT_PART_EMPTY (part)) {
 		lua_pushnil (L);
 		return 1;
 	}
@@ -746,7 +809,7 @@ lua_textpart_get_length (lua_State * L)
 		return 1;
 	}
 
-	if (IS_PART_EMPTY (part) || part->utf_content == NULL) {
+	if (IS_TEXT_PART_EMPTY (part) || part->utf_content == NULL) {
 		lua_pushinteger (L, 0);
 	}
 	else {
@@ -810,7 +873,7 @@ lua_textpart_get_lines_count (lua_State * L)
 		return 1;
 	}
 
-	if (IS_PART_EMPTY (part)) {
+	if (IS_TEXT_PART_EMPTY (part)) {
 		lua_pushinteger (L, 0);
 	}
 	else {
@@ -831,7 +894,7 @@ lua_textpart_get_words_count (lua_State *L)
 		return 1;
 	}
 
-	if (IS_PART_EMPTY (part) || part->utf_words == NULL) {
+	if (IS_TEXT_PART_EMPTY (part) || part->utf_words == NULL) {
 		lua_pushinteger (L, 0);
 	}
 	else {
@@ -841,6 +904,27 @@ lua_textpart_get_words_count (lua_State *L)
 	return 1;
 }
 
+static inline enum rspamd_lua_words_type
+word_extract_type_from_string (const gchar *how_str)
+{
+	enum rspamd_lua_words_type how = RSPAMD_LUA_WORDS_MAX;
+
+	if (strcmp (how_str, "stem") == 0) {
+		how = RSPAMD_LUA_WORDS_STEM;
+	}
+	else if (strcmp (how_str, "norm") == 0) {
+		how = RSPAMD_LUA_WORDS_NORM;
+	}
+	else if (strcmp (how_str, "raw") == 0) {
+		how = RSPAMD_LUA_WORDS_RAW;
+	}
+	else if (strcmp (how_str, "full") == 0) {
+		how = RSPAMD_LUA_WORDS_FULL;
+	}
+
+	return how;
+}
+
 static gint
 lua_textpart_get_words (lua_State *L)
 {
@@ -852,27 +936,17 @@ lua_textpart_get_words (lua_State *L)
 		return luaL_error (L, "invalid arguments");
 	}
 
-	if (IS_PART_EMPTY (part) || part->utf_words == NULL) {
+	if (IS_TEXT_PART_EMPTY (part) || part->utf_words == NULL) {
 		lua_createtable (L, 0, 0);
 	}
 	else {
 		if (lua_type (L, 2) == LUA_TSTRING) {
 			const gchar *how_str = lua_tostring (L, 2);
 
-			if (strcmp (how_str, "stem") == 0) {
-				how = RSPAMD_LUA_WORDS_STEM;
-			}
-			else if (strcmp (how_str, "norm") == 0) {
-				how = RSPAMD_LUA_WORDS_NORM;
-			}
-			else if (strcmp (how_str, "raw") == 0) {
-				how = RSPAMD_LUA_WORDS_RAW;
-			}
-			else if (strcmp (how_str, "full") == 0) {
-				how = RSPAMD_LUA_WORDS_FULL;
-			}
-			else {
-				return luaL_error (L, "unknown words type: %s", how_str);
+			how = word_extract_type_from_string (how_str);
+
+			if (how == RSPAMD_LUA_WORDS_MAX) {
+				return luaL_error (L, "invalid extraction type: %s", how_str);
 			}
 		}
 
@@ -882,6 +956,94 @@ lua_textpart_get_words (lua_State *L)
 	return 1;
 }
 
+static gint
+lua_textpart_filter_words (lua_State *L)
+{
+	LUA_TRACE_POINT;
+	struct rspamd_mime_text_part *part = lua_check_textpart (L);
+	struct rspamd_lua_regexp *re = lua_check_regexp (L, 2);
+	gint lim = -1;
+	enum rspamd_lua_words_type how = RSPAMD_LUA_WORDS_STEM;
+
+	if (part == NULL || re == NULL) {
+		return luaL_error (L, "invalid arguments");
+	}
+
+	if (IS_TEXT_PART_EMPTY (part) || part->utf_words == NULL) {
+		lua_createtable (L, 0, 0);
+	}
+	else {
+		if (lua_type (L, 3) == LUA_TSTRING) {
+			const gchar *how_str = lua_tostring (L, 3);
+
+			how = word_extract_type_from_string (how_str);
+
+			if (how == RSPAMD_LUA_WORDS_MAX) {
+				return luaL_error (L, "invalid extraction type: %s", how_str);
+			}
+		}
+
+		if (lua_type (L, 4) == LUA_TNUMBER) {
+			lim = lua_tointeger (L, 4);
+		}
+
+		guint cnt, i;
+
+		lua_createtable (L, 8, 0);
+
+		for (i = 0, cnt = 1; i < part->utf_words->len; i ++) {
+			rspamd_stat_token_t *w = &g_array_index (part->utf_words,
+					rspamd_stat_token_t, i);
+
+			switch (how) {
+			case RSPAMD_LUA_WORDS_STEM:
+				if (w->stemmed.len > 0) {
+					if (rspamd_regexp_match (re->re, w->stemmed.begin,
+							w->stemmed.len, FALSE)) {
+						lua_pushlstring (L, w->stemmed.begin, w->stemmed.len);
+						lua_rawseti (L, -2, cnt++);
+					}
+				}
+				break;
+			case RSPAMD_LUA_WORDS_NORM:
+				if (w->normalized.len > 0) {
+					if (rspamd_regexp_match (re->re, w->normalized.begin,
+							w->normalized.len, FALSE)) {
+						lua_pushlstring (L, w->normalized.begin, w->normalized.len);
+						lua_rawseti (L, -2, cnt++);
+					}
+				}
+				break;
+			case RSPAMD_LUA_WORDS_RAW:
+				if (w->original.len > 0) {
+					if (rspamd_regexp_match (re->re, w->original.begin,
+							w->original.len, TRUE)) {
+						lua_pushlstring (L, w->original.begin, w->original.len);
+						lua_rawseti (L, -2, cnt++);
+					}
+				}
+				break;
+			case RSPAMD_LUA_WORDS_FULL:
+				if (rspamd_regexp_match (re->re, w->normalized.begin,
+						w->normalized.len, FALSE)) {
+					rspamd_lua_push_full_word (L, w);
+					/* Push to the resulting vector */
+					lua_rawseti (L, -2, cnt++);
+				}
+				break;
+			default:
+				break;
+			}
+
+			if (lim > 0 && cnt >= lim) {
+				break;
+			}
+		}
+	}
+
+	return 1;
+}
+
 static gint
 lua_textpart_is_empty (lua_State * L)
 {
@@ -893,7 +1055,7 @@ lua_textpart_is_empty (lua_State * L)
 		return 1;
 	}
 
-	lua_pushboolean (L, IS_PART_EMPTY (part));
+	lua_pushboolean (L, IS_TEXT_PART_EMPTY (part));
 
 	return 1;
 }
@@ -909,7 +1071,7 @@ lua_textpart_is_html (lua_State * L)
 		return 1;
 	}
 
-	lua_pushboolean (L, IS_PART_HTML (part));
+	lua_pushboolean (L, IS_TEXT_PART_HTML (part));
 
 	return 1;
 }
@@ -1374,6 +1536,26 @@ lua_mimepart_get_detected_type_full (lua_State * L)
 	return lua_mimepart_get_type_common (L, part->detected_ct, TRUE);
 }
 
+static gint
+lua_mimepart_get_detected_ext (lua_State * L)
+{
+	LUA_TRACE_POINT;
+	struct rspamd_mime_part *part = lua_check_mimepart (L);
+
+	if (part == NULL) {
+		return luaL_error (L, "invalid arguments");
+	}
+
+	if (part->detected_ext) {
+		lua_pushstring (L, part->detected_ext);
+	}
+	else {
+		lua_pushnil (L);
+	}
+
+	return 1;
+}
+
 static gint
 lua_mimepart_get_cte (lua_State * L)
 {
@@ -1416,14 +1598,14 @@ lua_mimepart_get_boundary (lua_State * L)
 		return luaL_error (L, "invalid arguments");
 	}
 
-	if (IS_CT_MULTIPART (part->ct)) {
+	if (IS_PART_MULTIPART (part)) {
 		lua_pushlstring (L, part->specific.mp->boundary.begin,
 				part->specific.mp->boundary.len);
 	}
 	else {
 		parent = part->parent_part;
 
-		if (!parent || !IS_CT_MULTIPART (parent->ct)) {
+		if (!parent || !IS_PART_MULTIPART (parent)) {
 			lua_pushnil (L);
 		}
 		else {
@@ -1440,16 +1622,21 @@ lua_mimepart_get_header_common (lua_State *L, enum rspamd_lua_task_header_type h
 {
 	struct rspamd_mime_part *part = lua_check_mimepart (L);
 	const gchar *name;
-	GPtrArray *ar;
+	gboolean strong = FALSE;
 
 	name = luaL_checkstring (L, 2);
 
 	if (name && part) {
 
-		ar = rspamd_message_get_header_from_hash (part->raw_headers, NULL,
-				name, FALSE);
+		if (lua_isboolean (L, 3)) {
+			strong = lua_toboolean (L, 3);
+		}
 
-		return rspamd_lua_push_header_array (L, ar, how);
+		return rspamd_lua_push_header_array (L,
+				name,
+				rspamd_message_get_header_from_hash (part->raw_headers, name),
+				how,
+				strong);
 	}
 
 	lua_pushnil (L);
@@ -1517,7 +1704,7 @@ lua_mimepart_is_image (lua_State * L)
 		return luaL_error (L, "invalid arguments");
 	}
 
-	lua_pushboolean (L, (part->flags & RSPAMD_MIME_PART_IMAGE) ? true : false);
+	lua_pushboolean (L, part->part_type == RSPAMD_MIME_PART_IMAGE);
 
 	return 1;
 }
@@ -1532,7 +1719,7 @@ lua_mimepart_is_archive (lua_State * L)
 		return luaL_error (L, "invalid arguments");
 	}
 
-	lua_pushboolean (L, (part->flags & RSPAMD_MIME_PART_ARCHIVE) ? true : false);
+	lua_pushboolean (L, part->part_type == RSPAMD_MIME_PART_ARCHIVE);
 
 	return 1;
 }
@@ -1547,7 +1734,7 @@ lua_mimepart_is_multipart (lua_State * L)
 		return luaL_error (L, "invalid arguments");
 	}
 
-	lua_pushboolean (L, IS_CT_MULTIPART (part->ct) ? true : false);
+	lua_pushboolean (L, IS_PART_MULTIPART (part) ? true : false);
 
 	return 1;
 }
@@ -1562,7 +1749,7 @@ lua_mimepart_is_message (lua_State * L)
 		return luaL_error (L, "invalid arguments");
 	}
 
-	lua_pushboolean (L, IS_CT_MESSAGE (part->ct) ? true : false);
+	lua_pushboolean (L, IS_PART_MESSAGE (part) ? true : false);
 
 	return 1;
 }
@@ -1577,22 +1764,27 @@ lua_mimepart_is_attachment (lua_State * L)
 		return luaL_error (L, "invalid arguments");
 	}
 
-	if (!(part->flags & (RSPAMD_MIME_PART_IMAGE|RSPAMD_MIME_PART_TEXT))) {
-		if (part->cd && part->cd->type == RSPAMD_CT_ATTACHMENT) {
-			lua_pushboolean (L, true);
-		}
-		else {
-			if (part->cd && part->cd->filename.len > 0) {
-				/* We still have filename and it is not an image */
+	if (part->cd && part->cd->type == RSPAMD_CT_ATTACHMENT) {
+		lua_pushboolean (L, true);
+	}
+	else {
+		/* if has_name and not (image and Content-ID_header_present) */
+		if (part->cd && part->cd->filename.len > 0) {
+			if (part->part_type != RSPAMD_MIME_PART_IMAGE &&
+				rspamd_message_get_header_from_hash (part->raw_headers,
+						"Content-Id") == NULL) {
+				/* Filename is presented but no content id and not image */
 				lua_pushboolean (L, true);
 			}
 			else {
+				/* Image or an embeded object */
 				lua_pushboolean (L, false);
 			}
 		}
-	}
-	else {
-		lua_pushboolean (L, false);
+		else {
+			/* No filename */
+			lua_pushboolean (L, false);
+		}
 	}
 
 	return 1;
@@ -1608,7 +1800,7 @@ lua_mimepart_is_text (lua_State * L)
 		return luaL_error (L, "invalid arguments");
 	}
 
-	lua_pushboolean (L, (part->flags & RSPAMD_MIME_PART_TEXT) ? true : false);
+	lua_pushboolean (L, part->part_type == RSPAMD_MIME_PART_TEXT);
 
 	return 1;
 }
@@ -1645,7 +1837,7 @@ lua_mimepart_get_image (lua_State * L)
 		return luaL_error (L, "invalid arguments");
 	}
 
-	if (!(part->flags & RSPAMD_MIME_PART_IMAGE) || part->specific.img == NULL) {
+	if (part->part_type != RSPAMD_MIME_PART_IMAGE || part->specific.img == NULL) {
 		lua_pushnil (L);
 	}
 	else {
@@ -1668,7 +1860,7 @@ lua_mimepart_get_archive (lua_State * L)
 		return luaL_error (L, "invalid arguments");
 	}
 
-	if (!(part->flags & RSPAMD_MIME_PART_ARCHIVE) || part->specific.arch == NULL) {
+	if (part->part_type != RSPAMD_MIME_PART_ARCHIVE || part->specific.arch == NULL) {
 		lua_pushnil (L);
 	}
 	else {
@@ -1692,7 +1884,7 @@ lua_mimepart_get_children (lua_State * L)
 		return luaL_error (L, "invalid arguments");
 	}
 
-	if (!IS_CT_MULTIPART (part->ct) || part->specific.mp->children == NULL) {
+	if (!IS_PART_MULTIPART (part) || part->specific.mp->children == NULL) {
 		lua_pushnil (L);
 	}
 	else {
@@ -1744,7 +1936,7 @@ lua_mimepart_get_text (lua_State * L)
 		return luaL_error (L, "invalid arguments");
 	}
 
-	if (!(part->flags & RSPAMD_MIME_PART_TEXT) || part->specific.txt == NULL) {
+	if (part->part_type != RSPAMD_MIME_PART_TEXT || part->specific.txt == NULL) {
 		lua_pushnil (L);
 	}
 	else {
@@ -1785,7 +1977,7 @@ lua_mimepart_get_id (lua_State * L)
 		return luaL_error (L, "invalid arguments");
 	}
 
-	lua_pushinteger (L, part->id);
+	lua_pushinteger (L, part->part_number);
 
 	return 1;
 }
@@ -1797,8 +1989,7 @@ lua_mimepart_headers_foreach (lua_State *L)
 	struct rspamd_mime_part *part = lua_check_mimepart (L);
 	enum rspamd_lua_task_header_type how = RSPAMD_TASK_HEADER_PUSH_SIMPLE;
 	struct rspamd_lua_regexp *re = NULL;
-	GList *cur;
-	struct rspamd_mime_header *hdr;
+	struct rspamd_mime_header *hdr, *cur;
 	gint old_top;
 
 	if (part && lua_isfunction (L, 2)) {
@@ -1833,23 +2024,20 @@ lua_mimepart_headers_foreach (lua_State *L)
 		}
 
 		if (part->headers_order) {
-			cur = part->headers_order->head;
-
-			while (cur) {
-				hdr = cur->data;
+			hdr = part->headers_order;
 
+			LL_FOREACH2 (hdr, cur, ord_next) {
 				if (re && re->re) {
-					if (!rspamd_regexp_match (re->re, hdr->name,
-							strlen (hdr->name),FALSE)) {
-						cur = g_list_next (cur);
+					if (!rspamd_regexp_match (re->re, cur->name,
+							strlen (cur->name),FALSE)) {
 						continue;
 					}
 				}
 
 				old_top = lua_gettop (L);
 				lua_pushvalue (L, 2);
-				lua_pushstring (L, hdr->name);
-				rspamd_lua_push_header (L, hdr, how);
+				lua_pushstring (L, cur->name);
+				rspamd_lua_push_header (L, cur, how);
 
 				if (lua_pcall (L, 2, LUA_MULTRET, 0) != 0) {
 					msg_err ("call to header_foreach failed: %s",
@@ -1869,7 +2057,6 @@ lua_mimepart_headers_foreach (lua_State *L)
 				}
 
 				lua_settop (L, old_top);
-				cur = g_list_next (cur);
 			}
 		}
 	}
@@ -1877,6 +2064,141 @@ lua_mimepart_headers_foreach (lua_State *L)
 	return 0;
 }
 
+static gint
+lua_mimepart_get_specific (lua_State * L)
+{
+	LUA_TRACE_POINT;
+	struct rspamd_mime_part *part = lua_check_mimepart (L);
+
+	if (part == NULL) {
+		return luaL_error (L, "invalid arguments");
+	}
+
+	if (part->part_type != RSPAMD_MIME_PART_CUSTOM_LUA) {
+		lua_pushnil (L);
+	}
+	else {
+		lua_rawgeti (L, LUA_REGISTRYINDEX, part->specific.lua_specific.cbref);
+	}
+
+	return 1;
+}
+
+static gint
+lua_mimepart_get_urls (lua_State * L)
+{
+	LUA_TRACE_POINT;
+	struct rspamd_mime_part *part = lua_check_mimepart (L);
+
+	if (part == NULL) {
+		return luaL_error (L, "invalid arguments");
+	}
+
+	struct lua_tree_cb_data cb;
+	struct rspamd_url *u;
+	static const gint default_protocols_mask = PROTOCOL_HTTP|PROTOCOL_HTTPS|
+											   PROTOCOL_FILE|PROTOCOL_FTP;
+	gsize sz, max_urls = 0, i;
+
+	if (part->urls == NULL) {
+		lua_newtable (L);
+
+		return 1;
+	}
+
+	if (!lua_url_cbdata_fill (L, 2, &cb, default_protocols_mask,
+			~(0), max_urls)) {
+		return luaL_error (L, "invalid arguments");
+	}
+
+	sz = part->urls->len;
+
+	lua_createtable (L, sz, 0);
+
+	PTR_ARRAY_FOREACH (part->urls, i, u) {
+		lua_tree_url_callback (u, u, &cb);
+	}
+
+	lua_url_cbdata_dtor (&cb);
+
+	return 1;
+}
+
+static gint
+lua_mimepart_is_specific (lua_State * L)
+{
+	LUA_TRACE_POINT;
+	struct rspamd_mime_part *part = lua_check_mimepart (L);
+
+	if (part == NULL) {
+		return luaL_error (L, "invalid arguments");
+	}
+
+	lua_pushboolean (L, part->part_type == RSPAMD_MIME_PART_CUSTOM_LUA);
+
+	return 1;
+}
+
+static gint
+lua_mimepart_set_specific (lua_State * L)
+{
+	LUA_TRACE_POINT;
+	struct rspamd_mime_part *part = lua_check_mimepart (L);
+
+	if (part == NULL || lua_isnil (L, 2)) {
+		return luaL_error (L, "invalid arguments");
+	}
+
+	if (part->part_type != RSPAMD_MIME_PART_UNDEFINED &&
+			part->part_type != RSPAMD_MIME_PART_CUSTOM_LUA) {
+		return luaL_error (L,
+				"internal error: trying to set specific lua content on part of type %d",
+				part->part_type);
+	}
+
+	if (part->part_type == RSPAMD_MIME_PART_CUSTOM_LUA) {
+		/* Push old specific data */
+		lua_rawgeti (L, LUA_REGISTRYINDEX, part->specific.lua_specific.cbref);
+		luaL_unref (L, LUA_REGISTRYINDEX, part->specific.lua_specific.cbref);
+	}
+	else {
+		part->part_type = RSPAMD_MIME_PART_CUSTOM_LUA;
+		lua_pushnil (L);
+	}
+
+	/* Now, we push argument on the position 2 and save its reference */
+	lua_pushvalue (L, 2);
+	part->specific.lua_specific.cbref = luaL_ref (L, LUA_REGISTRYINDEX);
+	/* Now stack has just a return value as luaL_ref removes value from stack */
+
+	gint ltype = lua_type (L, 2);
+
+	switch (ltype) {
+	case LUA_TTABLE:
+		part->specific.lua_specific.type = RSPAMD_LUA_PART_TABLE;
+		break;
+	case LUA_TSTRING:
+		part->specific.lua_specific.type = RSPAMD_LUA_PART_STRING;
+		break;
+	case LUA_TUSERDATA:
+		if (rspamd_lua_check_udata_maybe (L, 2, "rspamd{text}")) {
+			part->specific.lua_specific.type = RSPAMD_LUA_PART_TEXT;
+		}
+		else {
+			part->specific.lua_specific.type = RSPAMD_LUA_PART_UNKNOWN;
+		}
+		break;
+	case LUA_TFUNCTION:
+		part->specific.lua_specific.type = RSPAMD_LUA_PART_FUNCTION;
+		break;
+	default:
+		part->specific.lua_specific.type = RSPAMD_LUA_PART_UNKNOWN;
+		break;
+	}
+
+	return 1;
+}
+
 void
 luaopen_textpart (lua_State * L)
 {
