@@ -27,7 +27,7 @@
  * You can convert rspamd_text into string but it will copy data.
  */
 
-/**
+/***
  * @function rspamd_text.fromstring(str)
  * Creates rspamd_text from Lua string (copied to the text)
  * @param {string} str string to use
@@ -35,15 +35,22 @@
  */
 LUA_FUNCTION_DEF (text, fromstring);
 
-/**
+/***
  * @function rspamd_text.null()
  * Creates rspamd_text with NULL pointer for testing purposes
  * @param {string} str string to use
  * @return {rspamd_text} resulting text
  */
 LUA_FUNCTION_DEF (text, null);
+/***
+ * @function rspamd_text.randombytes(nbytes)
+ * Creates rspamd_text with random bytes inside (raw bytes)
+ * @param {number} nbytes number of random bytes generated
+ * @return {rspamd_text} random bytes text
+ */
+LUA_FUNCTION_DEF (text, randombytes);
 
-/**
+/***
  * @function rspamd_text.fromtable(tbl[, delim])
  * Same as `table.concat` but generates rspamd_text instead of the Lua string
  * @param {table} tbl table to use
@@ -51,6 +58,14 @@ LUA_FUNCTION_DEF (text, null);
  * @return {rspamd_text} resulting text
  */
 LUA_FUNCTION_DEF (text, fromtable);
+/***
+ * @method rspamd_text:byte(pos[, pos2])
+ * Returns a byte at the position `pos` or bytes from `pos` to `pos2` if specified
+ * @param {integer} pos index
+ * @param {integer} pos2 index
+ * @return {integer} byte at the position `pos` or varargs of bytes
+ */
+LUA_FUNCTION_DEF (text, byte);
 /***
  * @method rspamd_text:len()
  * Returns length of a string
@@ -84,6 +99,12 @@ LUA_FUNCTION_DEF (text, save_in_file);
  * @return {rspamd_text} new rspamd_text with span (must be careful when using with owned texts...)
  */
 LUA_FUNCTION_DEF (text, span);
+/***
+ * @method rspamd_text:sub(start[, len])
+ * Returns a substrin for lua_text similar to string.sub from Lua
+ * @return {rspamd_text} new rspamd_text with span (must be careful when using with owned texts...)
+ */
+LUA_FUNCTION_DEF (text, sub);
 /***
  * @method rspamd_text:lines([stringify])
  * Returns an iter over all lines as rspamd_text objects or as strings if `stringify` is true
@@ -106,12 +127,29 @@ LUA_FUNCTION_DEF (text, split);
  * @return {integer} byte at the position `pos` or nil if pos out of bound
  */
 LUA_FUNCTION_DEF (text, at);
+/***
+ * @method rspamd_text:memchr(chr, [reverse])
+ * Returns the first or the last position of the character `chr` in the text or
+ * -1 in case if a character has not been found. Indexes start from `1`
+ * @param {string/number} chr character or a character code to find
+ * @param {boolean} reverse last character if `true`
+ * @return {integer} position of the character or `-1`
+ */
+LUA_FUNCTION_DEF (text, memchr);
 /***
  * @method rspamd_text:bytes()
  * Converts text to an array of bytes
  * @return {table|integer} bytes in the array (as unsigned char)
  */
 LUA_FUNCTION_DEF (text, bytes);
+/***
+ * @method rspamd_text:lower([is_utf, [inplace]])
+ * Return a new text with lowercased characters, if is_utf is true then Rspamd applies utf8 lowercase
+ * @param {boolean} is_utf apply utf8 lowercase
+ * @param {boolean} inplace lowercase the original text
+ * @return {rspamd_text} new rspamd_text (or the original text if inplace) with lowercased letters
+ */
+LUA_FUNCTION_DEF (text, lower);
 LUA_FUNCTION_DEF (text, take_ownership);
 /***
  * @method rspamd_text:exclude_chars(set_to_exclude, [always_copy])
@@ -127,7 +165,7 @@ LUA_FUNCTION_DEF (text, take_ownership);
  *
  * @param {string} set_to_exclude characters to exclude
  * @param {boolean} always_copy always copy the source text
- * @return {tspamd_text} modified or copied text
+ * @return {rspamd_text} modified or copied text
  */
 LUA_FUNCTION_DEF (text, exclude_chars);
 /***
@@ -141,11 +179,52 @@ LUA_FUNCTION_DEF (text, exclude_chars);
  * - UTF8 sequences are normalised
  *
  * @param {boolean} always_copy always copy the source text
- * @return {tspamd_text} modified or copied text
+ * @return {rspamd_text} modified or copied text
  */
 LUA_FUNCTION_DEF (text, oneline);
+/***
+ * @method rspamd_text:base32([b32type])
+ * Returns a text encoded in base32 (new rspamd_text is allocated)
+ *
+ * @param {string} b32type base32 type (default, bleach, rfc)
+ * @return {rspamd_text} new text encoded in base32
+ */
+LUA_FUNCTION_DEF (text, base32);
+/***
+ * @method rspamd_text:base64([line_length, [nline, [fold]]])
+ * Returns a text encoded in base64 (new rspamd_text is allocated)
+ *
+ * @param {number} line_length return text splited with newlines up to this attribute
+ * @param {string} nline newline type: `cr`, `lf`, `crlf`
+ * @param {boolean} fold use folding when splitting into lines (false by default)
+ * @return {rspamd_text} new text encoded in base64
+ */
+LUA_FUNCTION_DEF (text, base64);
+/***
+ * @method rspamd_text:hex()
+ * Returns a text encoded in hex (new rspamd_text is allocated)
+ *
+ * @return {rspamd_text} new text encoded in hex
+ */
+LUA_FUNCTION_DEF (text, hex);
+/***
+ * @method rspamd_text:find(pattern [, init])
+ * Looks for the first match of pattern in the string s.
+ * If it finds a match, then find returns the indices of s where this occurrence
+ * starts and ends; otherwise, it returns nil. A third,
+ * optional numerical argument init specifies where to start the search;
+ * its default value is 1 and can be negative.
+ * This method currently supports merely a plain search, no patterns.
+ *
+ * @param {string} pattern pattern to find
+ * @param {number} init specifies where to start the search (1 default)
+ * @return {number,number/nil} If it finds a match, then find returns the indices of s where this occurrence starts and ends; otherwise, it returns nil
+ */
+LUA_FUNCTION_DEF (text, find);
 LUA_FUNCTION_DEF (text, gc);
 LUA_FUNCTION_DEF (text, eq);
+LUA_FUNCTION_DEF (text, lt);
+LUA_FUNCTION_DEF (text, concat);
 
 static const struct luaL_reg textlib_f[] = {
 		LUA_INTERFACE_DEF (text, fromstring),
@@ -153,6 +232,7 @@ static const struct luaL_reg textlib_f[] = {
 		LUA_INTERFACE_DEF (text, fromtable),
 		{"from_table", lua_text_fromtable},
 		LUA_INTERFACE_DEF (text, null),
+		LUA_INTERFACE_DEF (text, randombytes),
 		{NULL, NULL}
 };
 
@@ -163,17 +243,27 @@ static const struct luaL_reg textlib_m[] = {
 		LUA_INTERFACE_DEF (text, take_ownership),
 		LUA_INTERFACE_DEF (text, save_in_file),
 		LUA_INTERFACE_DEF (text, span),
+		LUA_INTERFACE_DEF (text, sub),
 		LUA_INTERFACE_DEF (text, lines),
 		LUA_INTERFACE_DEF (text, split),
 		LUA_INTERFACE_DEF (text, at),
+		LUA_INTERFACE_DEF (text, memchr),
+		LUA_INTERFACE_DEF (text, byte),
 		LUA_INTERFACE_DEF (text, bytes),
+		LUA_INTERFACE_DEF (text, lower),
 		LUA_INTERFACE_DEF (text, exclude_chars),
 		LUA_INTERFACE_DEF (text, oneline),
+		LUA_INTERFACE_DEF (text, base32),
+		LUA_INTERFACE_DEF (text, base64),
+		LUA_INTERFACE_DEF (text, hex),
+		LUA_INTERFACE_DEF (text, find),
 		{"write", lua_text_save_in_file},
 		{"__len", lua_text_len},
 		{"__tostring", lua_text_str},
 		{"__gc", lua_text_gc},
 		{"__eq", lua_text_eq},
+		{"__lt", lua_text_lt},
+		{"__concat", lua_text_concat},
 		{NULL, NULL}
 };
 
@@ -185,6 +275,35 @@ lua_check_text (lua_State * L, gint pos)
 	return ud ? (struct rspamd_lua_text *)ud : NULL;
 }
 
+struct rspamd_lua_text *
+lua_check_text_or_string (lua_State * L, gint pos)
+{
+	gint pos_type = lua_type (L, pos);
+
+	if (pos_type == LUA_TUSERDATA) {
+		void *ud = rspamd_lua_check_udata (L, pos, "rspamd{text}");
+		luaL_argcheck (L, ud != NULL, pos, "'text' expected");
+		return ud ? (struct rspamd_lua_text *) ud : NULL;
+	}
+	else if (pos_type == LUA_TSTRING) {
+		/* Fake static lua_text */
+		static struct rspamd_lua_text fake_text;
+		gsize len;
+
+		fake_text.start = lua_tolstring (L, pos, &len);
+		if (len >= G_MAXUINT) {
+			return NULL;
+		}
+
+		fake_text.len = len;
+		fake_text.flags = RSPAMD_TEXT_FLAG_FAKE;
+
+		return &fake_text;
+	}
+
+	return NULL;
+}
+
 struct rspamd_lua_text *
 lua_new_text (lua_State *L, const gchar *start, gsize len, gboolean own)
 {
@@ -198,7 +317,11 @@ lua_new_text (lua_State *L, const gchar *start, gsize len, gboolean own)
 
 		if (len > 0) {
 			storage = g_malloc (len);
-			memcpy (storage, start, len);
+
+			if (start != NULL) {
+				memcpy (storage, start, len);
+			}
+
 			t->start = storage;
 			t->flags = RSPAMD_TEXT_FLAG_OWN;
 		}
@@ -252,6 +375,20 @@ lua_text_null (lua_State *L)
 	return 1;
 }
 
+static gint
+lua_text_randombytes (lua_State *L)
+{
+	LUA_TRACE_POINT;
+	guint nbytes = luaL_checkinteger (L, 1);
+	struct rspamd_lua_text *out;
+
+	out = lua_new_text (L, NULL, nbytes, TRUE);
+	randombytes_buf ((char *)out->start, nbytes);
+	out->len = nbytes;
+
+	return 1;
+}
+
 #define MAX_REC 10
 
 static void
@@ -502,6 +639,68 @@ lua_text_span (lua_State *L)
 	return 1;
 }
 
+/* Helpers to behave exactly as Lua does */
+static inline gsize
+relative_pos_start (gint pos, gsize len)
+{
+	if (pos > 0) {
+		return pos;
+	}
+	else if (pos == 0) {
+		return 1;
+	}
+	else if (pos < -((gint) len)) {
+		return 1;
+	}
+
+	/* Negative pos inside str */
+	return len + ((gsize)pos) + 1;
+}
+
+static inline gsize
+relative_pos_end (gint pos, gsize len)
+{
+	if (pos > (gint)len) {
+		return len;
+	}
+	else if (pos >= 0) {
+		return (size_t) pos;
+	}
+	else if (pos < -((gint)len)) {
+		return 0;
+	}
+
+	return len + ((gsize)pos) + 1;
+}
+
+static gint
+lua_text_sub (lua_State *L)
+{
+	LUA_TRACE_POINT;
+	struct rspamd_lua_text *t = lua_check_text (L, 1);
+
+	if (t) {
+		size_t start = relative_pos_start (luaL_checkinteger (L, 2),
+				t->len);
+		size_t end = relative_pos_end (luaL_optinteger (L, 3, -1),
+				t->len);
+
+
+		if (start <= end) {
+			lua_new_text (L, t->start + (start - 1),
+					(end - start) + 1, FALSE);
+		}
+		else {
+			lua_new_text (L, "", 0, TRUE);
+		}
+	}
+	else {
+		return luaL_error (L, "invalid arguments");
+	}
+
+	return 1;
+}
+
 static gint64
 rspamd_lua_text_push_line (lua_State *L,
 						   struct rspamd_lua_text *t,
@@ -771,17 +970,79 @@ lua_text_split (lua_State *L)
 
 static gint
 lua_text_at (lua_State *L)
+{
+	return lua_text_byte(L);
+}
+
+static gint
+lua_text_byte (lua_State *L)
 {
 	LUA_TRACE_POINT;
 	struct rspamd_lua_text *t = lua_check_text (L, 1);
-	gint pos = lua_tointeger (L, 2);
+	if (!t) {
+		return luaL_error (L, "invalid arguments");
+	}
 
-	if (t) {
-		if (pos > 0 && pos <= t->len) {
-			lua_pushinteger (L, t->start[pos - 1]);
+	gsize start = relative_pos_start (luaL_optinteger (L, 2, 1), t->len);
+	gsize end = relative_pos_end (luaL_optinteger (L, 3, start), t->len);
+	start--;
+
+	if (start >= end) {
+		return 0;
+	}
+
+	for (gsize i = start; i < end; i++) {
+		lua_pushinteger (L, t->start[i]);
+	}
+	return end - start;
+}
+
+static gint
+lua_text_memchr (lua_State *L)
+{
+	LUA_TRACE_POINT;
+	struct rspamd_lua_text *t = lua_check_text (L, 1);
+	int c;
+	bool reverse = false;
+
+	if (lua_isnumber (L, 2)) {
+		c = lua_tonumber (L, 2);
+	}
+	else {
+		gsize l;
+		const gchar *str = lua_tolstring (L, 2, &l);
+
+		if (str) {
+			c = str[0];
+
+			if (l != 1) {
+				return luaL_error (L, "need exactly one character to search");
+			}
 		}
 		else {
-			lua_pushnil (L);
+			return luaL_error (L, "invalid arguments");
+		}
+	}
+
+	if (t) {
+		void *f;
+
+		if (lua_isboolean (L, 3)) {
+			reverse = lua_toboolean (L, 3);
+		}
+
+		if (reverse) {
+			f = rspamd_memrchr (t->start, c, t->len);
+		}
+		else {
+			f = memchr (t->start, c, t->len);
+		}
+
+		if (f) {
+			lua_pushinteger (L, ((const char *)f) - t->start + 1);
+		}
+		else {
+			lua_pushinteger (L, -1);
 		}
 	}
 	else {
@@ -883,6 +1144,8 @@ lua_text_gc (lua_State *L)
 	struct rspamd_lua_text *t = lua_check_text (L, 1);
 
 	if (t != NULL) {
+		g_assert (!(t->flags & RSPAMD_TEXT_FLAG_FAKE));
+
 		if (t->flags & RSPAMD_TEXT_FLAG_OWN) {
 			if (t->flags & RSPAMD_TEXT_FLAG_WIPE) {
 				rspamd_explicit_memzero ((guchar *)t->start, t->len);
@@ -910,8 +1173,8 @@ static gint
 lua_text_eq (lua_State *L)
 {
 	LUA_TRACE_POINT;
-	struct rspamd_lua_text *t1 = lua_check_text (L, 1),
-			*t2 = lua_check_text (L, 2);
+	struct rspamd_lua_text *t1 = lua_check_text_or_string (L, 1),
+			*t2 = lua_check_text_or_string (L, 2);
 
 	if (t1->len == t2->len) {
 		lua_pushboolean (L, memcmp (t1->start, t2->start, t1->len) == 0);
@@ -923,6 +1186,43 @@ lua_text_eq (lua_State *L)
 	return 1;
 }
 
+static gint
+lua_text_lt (lua_State *L)
+{
+	LUA_TRACE_POINT;
+	struct rspamd_lua_text *t1 = lua_check_text_or_string (L, 1),
+			*t2 = lua_check_text_or_string (L, 2);
+
+	if (t1 && t2) {
+		if (t1->len == t2->len) {
+			lua_pushboolean (L, memcmp (t1->start, t2->start, t1->len) < 0);
+		}
+		else {
+			lua_pushboolean (L, t1->len < t2->len);
+		}
+	}
+
+	return 1;
+}
+
+static gint
+lua_text_concat (lua_State *L)
+{
+	LUA_TRACE_POINT;
+	struct rspamd_lua_text *t1 = lua_check_text_or_string (L, 1),
+			*t2 = lua_check_text_or_string (L, 2);
+
+	if (t1 && t2) {
+		struct rspamd_lua_text *final;
+
+		final = lua_new_text (L, NULL, t1->len + t2->len, TRUE);
+		memcpy ((void *)final->start, t1->start, t1->len);
+		memcpy ((void *)(final->start + t1->len), t2->start, t2->len);
+	}
+
+	return 1;
+}
+
 static gint
 lua_text_wipe (lua_State *L)
 {
@@ -945,6 +1245,145 @@ lua_text_wipe (lua_State *L)
 	return 0;
 }
 
+static gint
+lua_text_base32 (lua_State *L)
+{
+	LUA_TRACE_POINT;
+	struct rspamd_lua_text *t = lua_check_text (L, 1), *out;
+	enum rspamd_base32_type btype = RSPAMD_BASE32_DEFAULT;
+
+	if (t != NULL) {
+		if (lua_type (L, 2) == LUA_TSTRING) {
+			btype = rspamd_base32_decode_type_from_str (lua_tostring (L, 2));
+
+			if (btype == RSPAMD_BASE32_INVALID) {
+				return luaL_error (L, "invalid b32 type: %s", lua_tostring (L, 2));
+			}
+		}
+
+		out = lua_new_text (L, NULL, t->len * 8 / 5 + 2, TRUE);
+		out->len = rspamd_encode_base32_buf (t->start, t->len, (gchar *)out->start,
+				out->len, btype);
+	}
+	else {
+		return luaL_error (L, "invalid arguments");
+	}
+
+	return 1;
+}
+
+static gint
+lua_text_base64 (lua_State *L)
+{
+	LUA_TRACE_POINT;
+	struct rspamd_lua_text *t = lua_check_text (L, 1), *out;
+	gsize line_len = 0;
+	gboolean fold = FALSE;
+
+	if (t != NULL) {
+		if (lua_type (L, 2) == LUA_TNUMBER) {
+			line_len = lua_tointeger (L, 2);
+
+			if (line_len <= 8) {
+				return luaL_error (L, "too small line length (at least 8 is required)");
+			}
+		}
+
+		enum rspamd_newlines_type how = RSPAMD_TASK_NEWLINES_CRLF;
+
+		if (lua_type (L, 3) == LUA_TSTRING) {
+			const gchar *how_str = lua_tostring (L, 3);
+
+			if (g_ascii_strcasecmp (how_str, "cr") == 0) {
+				how = RSPAMD_TASK_NEWLINES_CR;
+			}
+			else if (g_ascii_strcasecmp (how_str, "lf") == 0) {
+				how = RSPAMD_TASK_NEWLINES_LF;
+			}
+			else if (g_ascii_strcasecmp (how_str, "crlf") != 0) {
+				return luaL_error (L, "invalid newline style: %s", how_str);
+			}
+		}
+
+		if (lua_type (L, 4) == LUA_TBOOLEAN) {
+			fold = lua_toboolean (L, 4);
+		}
+
+		gsize sz_len;
+
+		out = lua_newuserdata (L, sizeof (*t));
+		out->flags = RSPAMD_TEXT_FLAG_OWN;
+		out->start = rspamd_encode_base64_common (t->start, t->len,
+				line_len, &sz_len, fold, how);
+		out->len = sz_len;
+		rspamd_lua_setclass (L, "rspamd{text}", -1);
+	}
+	else {
+		return luaL_error (L, "invalid arguments");
+	}
+
+	return 1;
+}
+
+static gint
+lua_text_hex (lua_State *L)
+{
+	LUA_TRACE_POINT;
+	struct rspamd_lua_text *t = lua_check_text (L, 1), *out;
+
+	if (t != NULL) {
+
+		out = lua_new_text (L, NULL, t->len * 2, TRUE);
+		out->len = rspamd_encode_hex_buf (t->start, t->len, (gchar *)out->start,
+				out->len);
+	}
+	else {
+		return luaL_error (L, "invalid arguments");
+	}
+
+	return 1;
+}
+
+static gint
+lua_text_find (lua_State *L)
+{
+	LUA_TRACE_POINT;
+	struct rspamd_lua_text *t = lua_check_text (L, 1);
+	gsize patlen, init = 1;
+	const gchar *pat = luaL_checklstring (L, 2, &patlen);
+
+	if (t != NULL && pat != NULL) {
+
+		if (lua_isnumber (L, 3)) {
+			init = relative_pos_start (lua_tointeger (L, 3), t->len);
+		}
+
+		init --;
+
+		if (init > t->len) {
+			return luaL_error (L, "invalid arguments to find: init too large");
+		}
+
+		goffset pos = rspamd_substring_search (t->start + init,
+				t->len - init,
+				pat, patlen);
+
+		if (pos == -1) {
+			lua_pushnil (L);
+
+			return 1;
+		}
+
+		lua_pushinteger (L, pos + 1);
+		lua_pushinteger (L, pos + patlen);
+	}
+	else {
+		return luaL_error (L, "invalid arguments");
+	}
+
+	return 2;
+}
+
 #define BITOP(a,b,op) \
 		((a)[(gsize)(b)/(8*sizeof *(a))] op (gsize)1<<((gsize)(b)%(8*sizeof *(a))))
 
@@ -1200,10 +1639,53 @@ lua_text_oneline (lua_State *L)
 	return 1;
 }
 
+static gint
+lua_text_lower (lua_State *L)
+{
+	LUA_TRACE_POINT;
+	struct rspamd_lua_text *t = lua_check_text (L, 1), *nt;
+	gboolean is_utf8 = FALSE, is_inplace = FALSE;
+
+	if (t != NULL) {
+		if (lua_isboolean (L, 2)) {
+			is_utf8 = lua_toboolean (L, 2);
+		}
+		if (lua_isboolean (L, 3)) {
+			is_inplace = lua_toboolean (L, 3);
+		}
+
+		if (is_inplace) {
+			nt = t;
+			lua_pushvalue (L, 1);
+		}
+		else {
+			nt = lua_new_text (L, t->start, t->len, TRUE);
+		}
+
+		if (!is_utf8) {
+			rspamd_str_lc ((gchar *) nt->start, nt->len);
+		}
+		else {
+			rspamd_str_lc_utf8 ((gchar *) nt->start, nt->len);
+		}
+	}
+	else {
+		return luaL_error (L, "invalid arguments");
+	}
+
+	return 1;
+}
+
+/* Used to distinguish lua text metatable */
+static const guint rspamd_lua_text_cookie = 0x2b21ef6fU;
+
 static gint
 lua_load_text (lua_State * L)
 {
 	lua_newtable (L);
+	lua_pushstring (L, "cookie");
+	lua_pushnumber (L, rspamd_lua_text_cookie);
+	lua_settable (L, -3);
 	luaL_register (L, NULL, textlib_f);
 
 	return 1;
@@ -1213,6 +1695,9 @@ void
 luaopen_text (lua_State *L)
 {
 	rspamd_lua_new_class (L, "rspamd{text}", textlib_m);
+	lua_pushstring (L, "cookie");
+	lua_pushnumber (L, rspamd_lua_text_cookie);
+	lua_settable (L, -3);
 	lua_pop (L, 1);
 
 	rspamd_lua_add_preload (L, "rspamd_text", lua_load_text);
