@@ -14,7 +14,8 @@
  * limitations under the License.
  */
 #include "lua_common.h"
-#include "contrib/uthash/utlist.h"
+#include "lua_url.h"
+
 
 /***
  * @module rspamd_url
@@ -714,7 +715,7 @@ lua_url_to_table (lua_State *L)
 
 
 /***
- * @function url.create([mempool,] str)
+ * @function url.create([mempool,] str, [{flags_table}])
  * @param {rspamd_mempool} memory pool for URL, e.g. `task:get_mempool()`
  * @param {string} text that contains URL (can also contain other stuff)
  * @return {url} new url object that exists as long as the corresponding mempool exists
@@ -727,6 +728,7 @@ lua_url_create (lua_State *L)
 	const gchar *text;
 	size_t length;
 	gboolean own_pool = FALSE;
+	struct rspamd_lua_url *u;
 
 	if (lua_type (L, 1) == LUA_TUSERDATA) {
 		pool = rspamd_lua_check_mempool (L, 1);
@@ -752,6 +754,26 @@ lua_url_create (lua_State *L)
 		if (lua_type (L, -1) != LUA_TUSERDATA) {
 			/* URL is actually not found */
 			lua_pushnil (L);
+
+			return 1;
+		}
+
+		u = (struct rspamd_lua_url *)lua_touserdata (L, -1);
+
+		if (lua_type (L, 3) == LUA_TTABLE) {
+			/* Add flags */
+			for (lua_pushnil (L); lua_next (L, 3); lua_pop (L, 1)) {
+				int nmask = 0;
+				const gchar *fname = lua_tostring (L, -1);
+
+				if (rspamd_url_flag_from_string (fname, &nmask)) {
+					u->url->flags |= nmask;
+				}
+				else {
+					lua_pop (L, 1);
+					return luaL_error (L, "invalid flag: %s", fname);
+				}
+			}
 		}
 	}
 
@@ -853,9 +875,9 @@ lua_url_all (lua_State *L)
  * - `image`: URL is from src attribute of img HTML tag
  * @return {table} URL flags
  */
-#define PUSH_FLAG(fl, name) do { \
+#define PUSH_FLAG(fl) do { \
 	if (flags & (fl)) { \
-		lua_pushstring (L, (name)); \
+		lua_pushstring (L, rspamd_url_flag_to_string (fl)); \
 		lua_pushboolean (L, true); \
 		lua_settable (L, -3); \
 	} \
@@ -873,26 +895,27 @@ lua_url_get_flags (lua_State *L)
 
 		lua_createtable (L, 0, 4);
 
-		PUSH_FLAG (RSPAMD_URL_FLAG_PHISHED, "phished");
-		PUSH_FLAG (RSPAMD_URL_FLAG_NUMERIC, "numeric");
-		PUSH_FLAG (RSPAMD_URL_FLAG_OBSCURED, "obscured");
-		PUSH_FLAG (RSPAMD_URL_FLAG_REDIRECTED, "redirected");
-		PUSH_FLAG (RSPAMD_URL_FLAG_HTML_DISPLAYED, "html_displayed");
-		PUSH_FLAG (RSPAMD_URL_FLAG_FROM_TEXT, "text");
-		PUSH_FLAG (RSPAMD_URL_FLAG_SUBJECT, "subject");
-		PUSH_FLAG (RSPAMD_URL_FLAG_HOSTENCODED, "host_encoded");
-		PUSH_FLAG (RSPAMD_URL_FLAG_SCHEMAENCODED, "schema_encoded");
-		PUSH_FLAG (RSPAMD_URL_FLAG_PATHENCODED, "path_encoded");
-		PUSH_FLAG (RSPAMD_URL_FLAG_QUERYENCODED, "query_encoded");
-		PUSH_FLAG (RSPAMD_URL_FLAG_MISSINGSLASHES, "missing_slahes");
-		PUSH_FLAG (RSPAMD_URL_FLAG_IDN, "idn");
-		PUSH_FLAG (RSPAMD_URL_FLAG_HAS_PORT, "has_port");
-		PUSH_FLAG (RSPAMD_URL_FLAG_HAS_USER, "has_user");
-		PUSH_FLAG (RSPAMD_URL_FLAG_SCHEMALESS, "schemaless");
-		PUSH_FLAG (RSPAMD_URL_FLAG_UNNORMALISED, "unnormalised");
-		PUSH_FLAG (RSPAMD_URL_FLAG_ZW_SPACES, "zw_spaces");
-		PUSH_FLAG (RSPAMD_URL_FLAG_DISPLAY_URL, "url_displayed");
-		PUSH_FLAG (RSPAMD_URL_FLAG_IMAGE, "image");
+		PUSH_FLAG (RSPAMD_URL_FLAG_PHISHED);
+		PUSH_FLAG (RSPAMD_URL_FLAG_NUMERIC);
+		PUSH_FLAG (RSPAMD_URL_FLAG_OBSCURED);
+		PUSH_FLAG (RSPAMD_URL_FLAG_REDIRECTED);
+		PUSH_FLAG (RSPAMD_URL_FLAG_HTML_DISPLAYED);
+		PUSH_FLAG (RSPAMD_URL_FLAG_FROM_TEXT);
+		PUSH_FLAG (RSPAMD_URL_FLAG_SUBJECT);
+		PUSH_FLAG (RSPAMD_URL_FLAG_HOSTENCODED);
+		PUSH_FLAG (RSPAMD_URL_FLAG_SCHEMAENCODED);
+		PUSH_FLAG (RSPAMD_URL_FLAG_PATHENCODED);
+		PUSH_FLAG (RSPAMD_URL_FLAG_QUERYENCODED);
+		PUSH_FLAG (RSPAMD_URL_FLAG_MISSINGSLASHES);
+		PUSH_FLAG (RSPAMD_URL_FLAG_IDN);
+		PUSH_FLAG (RSPAMD_URL_FLAG_HAS_PORT);
+		PUSH_FLAG (RSPAMD_URL_FLAG_HAS_USER);
+		PUSH_FLAG (RSPAMD_URL_FLAG_SCHEMALESS);
+		PUSH_FLAG (RSPAMD_URL_FLAG_UNNORMALISED);
+		PUSH_FLAG (RSPAMD_URL_FLAG_ZW_SPACES);
+		PUSH_FLAG (RSPAMD_URL_FLAG_DISPLAY_URL);
+		PUSH_FLAG (RSPAMD_URL_FLAG_IMAGE);
+		PUSH_FLAG (RSPAMD_URL_FLAG_CONTENT);
 	}
 	else {
 		return luaL_error (L, "invalid arguments");
@@ -903,6 +926,289 @@ lua_url_get_flags (lua_State *L)
 
 #undef PUSH_FLAG
 
+void
+lua_tree_url_callback (gpointer key, gpointer value, gpointer ud)
+{
+	struct rspamd_lua_url *lua_url;
+	struct rspamd_url *url = (struct rspamd_url *)value;
+	struct lua_tree_cb_data *cb = ud;
+
+	if ((url->protocol & cb->protocols_mask) == url->protocol) {
+
+		if (cb->flags_mode == url_flags_mode_include_any) {
+			if (url->flags != (url->flags & cb->flags_mask)) {
+				return;
+			}
+		}
+		else {
+			if ((url->flags & cb->flags_mask) != cb->flags_mask) {
+				return;
+			}
+		}
+
+		if (cb->skip_prob > 0) {
+			gdouble coin = rspamd_random_double_fast_seed (cb->xoroshiro_state);
+
+			if (coin < cb->skip_prob) {
+				return;
+			}
+		}
+
+		lua_url = lua_newuserdata (cb->L, sizeof (struct rspamd_lua_url));
+		lua_pushvalue (cb->L, cb->metatable_pos);
+		lua_setmetatable (cb->L, -2);
+		lua_url->url = url;
+		lua_rawseti (cb->L, -2, cb->i++);
+	}
+}
+
+gboolean
+lua_url_cbdata_fill (lua_State *L,
+					 gint pos,
+					 struct lua_tree_cb_data *cbd,
+					 guint default_protocols,
+					 guint default_flags,
+					 gsize max_urls)
+{
+	gint protocols_mask = 0;
+
+	gint pos_arg_type = lua_type (L, pos);
+	guint flags_mask = default_flags;
+	gboolean seen_flags = FALSE, seen_protocols = FALSE;
+
+	memset (cbd, 0, sizeof (*cbd));
+	cbd->flags_mode = url_flags_mode_include_any;
+
+	if (pos_arg_type == LUA_TBOOLEAN) {
+		protocols_mask = default_protocols;
+		if (lua_toboolean (L, 2)) {
+			protocols_mask |= PROTOCOL_MAILTO;
+		}
+	}
+	else if (pos_arg_type == LUA_TTABLE) {
+		if (rspamd_lua_geti (L, 1, pos) == LUA_TNIL) {
+			/* New method: indexed table */
+
+			lua_getfield (L, pos, "flags");
+			if (lua_istable (L, -1)) {
+				gint top = lua_gettop (L);
+
+				lua_getfield (L, pos, "flags_mode");
+				if (lua_isstring (L, -1)) {
+					const gchar *mode_str = lua_tostring (L, -1);
+
+					if (strcmp (mode_str, "explicit") == 0) {
+						cbd->flags_mode = url_flags_mode_include_explicit;
+						/*
+						 * Ignore default flags in this mode and include
+						 * merely flags specified by a caller
+						 */
+						flags_mask = 0;
+					}
+				}
+				lua_pop (L, 1);
+
+				for (lua_pushnil (L); lua_next (L, top); lua_pop (L, 1)) {
+					int nmask = 0;
+					const gchar *fname = lua_tostring (L, -1);
+
+
+					if (rspamd_url_flag_from_string (fname, &nmask)) {
+						flags_mask |= nmask;
+					}
+					else {
+						msg_info ("bad url flag: %s", fname);
+						return FALSE;
+					}
+				}
+
+				seen_flags = TRUE;
+			}
+			else {
+				flags_mask |= default_flags;
+			}
+			lua_pop (L, 1);
+
+			lua_getfield (L, pos, "protocols");
+			if (lua_istable (L, -1)) {
+				gint top = lua_gettop (L);
+
+				for (lua_pushnil (L); lua_next (L, top); lua_pop (L, 1)) {
+					int nmask;
+					const gchar *pname = lua_tostring (L, -1);
+
+					nmask = rspamd_url_protocol_from_string (pname);
+
+					if (nmask != PROTOCOL_UNKNOWN) {
+						protocols_mask |= nmask;
+					}
+					else {
+						msg_info ("bad url protocol: %s", pname);
+						return FALSE;
+					}
+				}
+				seen_protocols = TRUE;
+			}
+			else {
+				protocols_mask = default_protocols;
+			}
+			lua_pop (L, 1);
+
+			if (!seen_protocols) {
+				lua_getfield (L, pos, "emails");
+				if (lua_isboolean (L, -1)) {
+					if (lua_toboolean (L, -1)) {
+						protocols_mask |= PROTOCOL_MAILTO;
+					}
+				}
+				lua_pop (L, 1);
+			}
+
+			if (!seen_flags) {
+				lua_getfield (L, pos, "images");
+				if (lua_isboolean (L, -1)) {
+					if (lua_toboolean (L, -1)) {
+						flags_mask |= RSPAMD_URL_FLAG_IMAGE;
+					}
+					else {
+						flags_mask &= ~RSPAMD_URL_FLAG_IMAGE;
+					}
+				}
+				else {
+					flags_mask &= ~RSPAMD_URL_FLAG_IMAGE;
+				}
+				lua_pop (L, 1);
+			}
+
+			if (!seen_flags) {
+				lua_getfield (L, pos, "content");
+				if (lua_isboolean (L, -1)) {
+					if (lua_toboolean (L, -1)) {
+						flags_mask |= RSPAMD_URL_FLAG_CONTENT;
+					}
+					else {
+						flags_mask &= ~RSPAMD_URL_FLAG_CONTENT;
+					}
+				}
+				else {
+					flags_mask &= ~RSPAMD_URL_FLAG_CONTENT;
+				}
+				lua_pop (L, 1);
+			}
+
+			lua_getfield (L, pos, "max_urls");
+			if (lua_isnumber (L, -1)) {
+				max_urls = lua_tonumber (L, -1);
+			}
+			lua_pop (L, 1);
+		}
+		else {
+			/* Plain table of the protocols */
+			for (lua_pushnil (L); lua_next (L, pos); lua_pop (L, 1)) {
+				int nmask;
+				const gchar *pname = lua_tostring (L, -1);
+
+				nmask = rspamd_url_protocol_from_string (pname);
+
+				if (nmask != PROTOCOL_UNKNOWN) {
+					protocols_mask |= nmask;
+				}
+				else {
+					msg_info ("bad url protocol: %s", pname);
+					return FALSE;
+				}
+			}
+		}
+
+		lua_pop (L, 1); /* After rspamd_lua_geti */
+	}
+	else if (pos_arg_type == LUA_TSTRING) {
+		const gchar *plist = lua_tostring (L, pos);
+		gchar **strvec;
+		gchar * const *cvec;
+
+		strvec = g_strsplit_set (plist, ",;", -1);
+		cvec = strvec;
+
+		while (*cvec) {
+			int nmask;
+
+			nmask = rspamd_url_protocol_from_string (*cvec);
+
+			if (nmask != PROTOCOL_UNKNOWN) {
+				protocols_mask |= nmask;
+			}
+			else {
+				msg_info ("bad url protocol: %s", *cvec);
+				return FALSE;
+			}
+
+			cvec ++;
+		}
+
+		g_strfreev (strvec);
+	}
+	else if (pos_arg_type == LUA_TNONE || pos_arg_type == LUA_TNIL) {
+		protocols_mask = default_protocols;
+		flags_mask = default_flags;
+	}
+	else {
+		return FALSE;
+	}
+
+	if (lua_type (L, pos + 1) == LUA_TBOOLEAN) {
+		if (lua_toboolean (L, pos + 1)) {
+			flags_mask |= RSPAMD_URL_FLAG_IMAGE;
+		}
+		else {
+			flags_mask &= ~RSPAMD_URL_FLAG_IMAGE;
+		}
+	}
+
+	cbd->i = 1;
+	cbd->L = L;
+	cbd->max_urls = max_urls;
+	cbd->protocols_mask = protocols_mask;
+	cbd->flags_mask = flags_mask;
+
+	/* This needs to be removed from the stack */
+	rspamd_lua_class_metatable (L, "rspamd{url}");
+	cbd->metatable_pos = lua_gettop (L);
+	(void)lua_checkstack (L, cbd->metatable_pos + 4);
+
+	return TRUE;
+}
+
+void
+lua_url_cbdata_dtor (struct lua_tree_cb_data *cbd)
+{
+	if (cbd->metatable_pos != -1) {
+		lua_remove (cbd->L, cbd->metatable_pos);
+	}
+}
+
+gsize
+lua_url_adjust_skip_prob (gdouble timestamp,
+						  guchar *digest,
+						  struct lua_tree_cb_data *cb,
+						  gsize sz)
+{
+	if (cb->max_urls > 0 && sz > cb->max_urls) {
+		cb->skip_prob = 1.0 - ((gdouble)cb->max_urls) / (gdouble)sz;
+		/*
+		 * Use task dependent probabilistic seed to ensure that
+		 * consequent task:get_urls return the same list of urls
+		 */
+		memcpy (&cb->xoroshiro_state[0], &timestamp,
+				MIN (sizeof (cb->xoroshiro_state[0]), sizeof (timestamp)));
+		memcpy (&cb->xoroshiro_state[1], digest,
+				sizeof (cb->xoroshiro_state[1]) * 3);
+		sz = cb->max_urls;
+	}
+
+	return sz;
+}
+
 static gint
 lua_load_url (lua_State * L)
 {
