@@ -14,22 +14,26 @@
  * limitations under the License.
  */
 #include "lua_common.h"
-#include "html.h"
-#include "tokenizers/tokenizers.h"
 #include "unix-std.h"
 #include "contrib/zstd/zstd.h"
-#include "contrib/uthash/utlist.h"
 #include "libmime/email_addr.h"
 #include "libmime/content_type.h"
 #include "libmime/mime_headers.h"
 #include "libutil/hash.h"
-#include "linenoise.h"
+
+#include "lua_parsers.h"
+
+#ifdef WITH_LUA_REPL
+#include "replxx.h"
+#endif
+
 #include <math.h>
 #include <glob.h>
 #include <zlib.h>
 
 #include "unicode/uspoof.h"
 #include "unicode/uscript.h"
+#include "contrib/fastutf8/fastutf8.h"
 
 /***
  * @module rspamd_util
@@ -71,6 +75,15 @@ LUA_FUNCTION_DEF (util, encode_base64);
  * @return {rspamd_text} encoded data chunk
  */
 LUA_FUNCTION_DEF (util, encode_qp);
+
+/***
+ * @function util.decode_qp(input)
+ * Decodes data from quouted printable
+ * @param {text or string} input input data
+ * @return {rspamd_text} decoded data chunk
+ */
+LUA_FUNCTION_DEF (util, decode_qp);
+
 /***
  * @function util.decode_base64(input)
  * Decodes data from base64 ignoring whitespace characters
@@ -80,16 +93,18 @@ LUA_FUNCTION_DEF (util, encode_qp);
 LUA_FUNCTION_DEF (util, decode_base64);
 
 /***
- * @function util.encode_base32(input)
+ * @function util.encode_base32(input, [b32type = 'default'])
  * Encodes data in base32 breaking lines if needed
  * @param {text or string} input input data
+ * @param {string} b32type base32 type (default, bleach, rfc)
  * @return {rspamd_text} encoded data chunk
  */
 LUA_FUNCTION_DEF (util, encode_base32);
 /***
- * @function util.decode_base32(input)
+ * @function util.decode_base32(input, [b32type = 'default'])
  * Decodes data from base32 ignoring whitespace characters
  * @param {text or string} input data to decode
+ * @param {string} b32type base32 type (default, bleach, rfc)
  * @return {rspamd_text} decoded data chunk
  */
 LUA_FUNCTION_DEF (util, decode_base32);
@@ -136,18 +151,6 @@ LUA_FUNCTION_DEF (util, parse_html);
  */
 LUA_FUNCTION_DEF (util, levenshtein_distance);
 
-/***
- * @function util.parse_addr(str)
- * Parse rfc822 address to components. Returns a table of components:
- *
- * - `name`: name of address (e.g. Some User)
- * - `addr`: address part (e.g. user@example.com)
- *
- * @param {string} str input string
- * @return {table} resulting table of components
- */
-LUA_FUNCTION_DEF (util, parse_addr);
-
 /***
  * @function util.fold_header(name, value, [how, [stop_chars]])
  * Fold rfc822 header according to the folding rules
@@ -197,13 +200,22 @@ LUA_FUNCTION_DEF (util, get_tld);
 LUA_FUNCTION_DEF (util, glob);
 
 /***
- * @function util.parse_mail_address(str, pool)
+ * @function util.parse_mail_address(str, [pool])
  * Parses email address and returns a table of tables in the following format:
  *
+ * - `raw` - the original value without any processing
  * - `name` - name of internet address in UTF8, e.g. for `Vsevolod Stakhov <blah@foo.com>` it returns `Vsevolod Stakhov`
  * - `addr` - address part of the address
  * - `user` - user part (if present) of the address, e.g. `blah`
  * - `domain` - domain part (if present), e.g. `foo.com`
+ * - `flags` - table with following keys set to true if given condition fulfilled:
+ *   - [valid] - valid SMTP address in conformity with https://tools.ietf.org/html/rfc5321#section-4.1.
+ *   - [ip] - domain is IPv4/IPv6 address
+ *   - [braced] - angled `<blah@foo.com>` address
+ *   - [quoted] - quoted user part
+ *   - [empty] - empty address
+ *   - [backslash] - user part contains backslash
+ *   - [8bit] - contains 8bit characters
  *
  * @param {string} str input string
  * @param {rspamd_mempool} pool memory pool to use
@@ -220,6 +232,14 @@ LUA_FUNCTION_DEF (util, parse_mail_address);
  */
 LUA_FUNCTION_DEF (util, strlen_utf8);
 
+/***
+ * @function util.lower_utf8(str)
+ * Converts utf8 string to lower case
+ * @param {string} str utf8 encoded string
+ * @return {string} lowercased utf8 string
+ */
+LUA_FUNCTION_DEF (util, lower_utf8);
+
 
 /***
  * @function util.strcasecmp(str1, str2)
@@ -343,7 +363,7 @@ LUA_FUNCTION_DEF (util, close_file);
 LUA_FUNCTION_DEF (util, random_hex);
 
 /***
- * @function util.zstd_compress(data)
+ * @function util.zstd_compress(data, [level=1])
  * Compresses input using zstd compression
  *
  * @param {string/rspamd_text} data input data
@@ -361,14 +381,25 @@ LUA_FUNCTION_DEF (util, zstd_compress);
 LUA_FUNCTION_DEF (util, zstd_decompress);
 
 /***
- * @function util.gzip_decompress(data)
+ * @function util.gzip_decompress(data, [size_limit])
  * Decompresses input using gzip algorithm
  *
  * @param {string/rspamd_text} data compressed data
- * @return {error,rspamd_text} pair of error + decompressed text
+ * @param {integer} size_limit optional size limit
+ * @return {rspamd_text} decompressed text
  */
 LUA_FUNCTION_DEF (util, gzip_decompress);
 
+/***
+ * @function util.inflate(data, [size_limit])
+ * Decompresses input using inflate algorithm
+ *
+ * @param {string/rspamd_text} data compressed data
+ * @param {integer} size_limit optional size limit
+ * @return {rspamd_text} decompressed text
+ */
+LUA_FUNCTION_DEF (util, inflate);
+
 /***
  * @function util.gzip_compress(data)
  * Compresses input using gzip compression
@@ -598,6 +629,23 @@ LUA_FUNCTION_DEF (util, parse_content_type);
  */
 LUA_FUNCTION_DEF (util, mime_header_encode);
 
+/***
+ *  @function util.btc_polymod(input_values)
+ * Performs bitcoin polymod function
+ * @param {table|numbers} input_values
+ * @return {boolean} true if polymod has been successful
+ */
+LUA_FUNCTION_DEF (util, btc_polymod);
+
+/***
+ * @function util.parse_smtp_date(str[, local_tz])
+ * Converts an SMTP date string to unix timestamp
+ * @param {string} str input string
+ * @param {boolean} local_tz convert to local tz if `true`
+ * @return {number} time as unix timestamp (converted to float)
+ */
+LUA_FUNCTION_DEF (util, parse_smtp_date);
+
 
 static const struct luaL_reg utillib_f[] = {
 	LUA_INTERFACE_DEF (util, create_event_base),
@@ -606,6 +654,7 @@ static const struct luaL_reg utillib_f[] = {
 	LUA_INTERFACE_DEF (util, process_message),
 	LUA_INTERFACE_DEF (util, encode_base64),
 	LUA_INTERFACE_DEF (util, encode_qp),
+	LUA_INTERFACE_DEF (util, decode_qp),
 	LUA_INTERFACE_DEF (util, decode_base64),
 	LUA_INTERFACE_DEF (util, encode_base32),
 	LUA_INTERFACE_DEF (util, decode_base32),
@@ -614,14 +663,15 @@ static const struct luaL_reg utillib_f[] = {
 	LUA_INTERFACE_DEF (util, tanh),
 	LUA_INTERFACE_DEF (util, parse_html),
 	LUA_INTERFACE_DEF (util, levenshtein_distance),
-	LUA_INTERFACE_DEF (util, parse_addr),
 	LUA_INTERFACE_DEF (util, fold_header),
 	LUA_INTERFACE_DEF (util, is_uppercase),
 	LUA_INTERFACE_DEF (util, humanize_number),
 	LUA_INTERFACE_DEF (util, get_tld),
 	LUA_INTERFACE_DEF (util, glob),
+	{"parse_addr", lua_util_parse_mail_address},
 	LUA_INTERFACE_DEF (util, parse_mail_address),
 	LUA_INTERFACE_DEF (util, strlen_utf8),
+	LUA_INTERFACE_DEF (util, lower_utf8),
 	LUA_INTERFACE_DEF (util, strcasecmp_ascii),
 	LUA_INTERFACE_DEF (util, strequal_caseless),
 	LUA_INTERFACE_DEF (util, get_ticks),
@@ -638,6 +688,7 @@ static const struct luaL_reg utillib_f[] = {
 	LUA_INTERFACE_DEF (util, zstd_decompress),
 	LUA_INTERFACE_DEF (util, gzip_compress),
 	LUA_INTERFACE_DEF (util, gzip_decompress),
+	LUA_INTERFACE_DEF (util, inflate),
 	LUA_INTERFACE_DEF (util, normalize_prob),
 	LUA_INTERFACE_DEF (util, caseless_hash),
 	LUA_INTERFACE_DEF (util, caseless_hash_fast),
@@ -659,6 +710,8 @@ static const struct luaL_reg utillib_f[] = {
 	LUA_INTERFACE_DEF (util, pack),
 	LUA_INTERFACE_DEF (util, unpack),
 	LUA_INTERFACE_DEF (util, packsize),
+	LUA_INTERFACE_DEF (util, btc_polymod),
+	LUA_INTERFACE_DEF (util, parse_smtp_date),
 	{NULL, NULL}
 };
 
@@ -695,11 +748,11 @@ static gint
 lua_util_create_event_base (lua_State *L)
 {
 	LUA_TRACE_POINT;
-	struct event_base **pev_base;
+	struct ev_loop **pev_base;
 
-	pev_base = lua_newuserdata (L, sizeof (struct event_base *));
+	pev_base = lua_newuserdata (L, sizeof (struct ev_loop *));
 	rspamd_lua_setclass (L, "rspamd{ev_base}", -1);
-	*pev_base = event_init ();
+	*pev_base = ev_loop_new (EVFLAG_SIGNALFD|EVBACKEND_ALL);
 
 	return 1;
 }
@@ -799,7 +852,6 @@ lua_util_config_from_ucl (lua_State *L)
 		cfg->lua_state = L;
 
 		cfg->rcl_obj = obj;
-		cfg->cache = rspamd_symcache_new (cfg);
 		top = rspamd_rcl_config_init (cfg, NULL);
 
 		if (!rspamd_rcl_parse (top, cfg, cfg, cfg->cfg_pool, cfg->rcl_obj, &err)) {
@@ -842,15 +894,15 @@ lua_util_process_message (lua_State *L)
 	const gchar *message;
 	gsize mlen;
 	struct rspamd_task *task;
-	struct event_base *base;
+	struct ev_loop *base;
 	ucl_object_t *res = NULL;
 
 	message = luaL_checklstring (L, 2, &mlen);
 
 	if (cfg != NULL && message != NULL) {
-		base = event_init ();
-		rspamd_init_filters (cfg, FALSE);
-		task = rspamd_task_new (NULL, cfg, NULL, NULL, base);
+		base = ev_loop_new (EVFLAG_SIGNALFD|EVBACKEND_ALL);
+		rspamd_init_filters (cfg, false, false);
+		task = rspamd_task_new (NULL, cfg, NULL, NULL, base, FALSE);
 		task->msg.begin = rspamd_mempool_alloc (task->task_pool, mlen);
 		rspamd_strlcpy ((gpointer)task->msg.begin, message, mlen);
 		task->msg.len = mlen;
@@ -865,7 +917,7 @@ lua_util_process_message (lua_State *L)
 		}
 		else {
 			if (rspamd_task_process (task, RSPAMD_TASK_PROCESS_ALL)) {
-				event_base_loop (base, 0);
+				ev_loop (base, 0);
 
 				if (res != NULL) {
 					ucl_object_push_lua (L, res, true);
@@ -885,7 +937,7 @@ lua_util_process_message (lua_State *L)
 			}
 		}
 
-		event_base_free (base);
+		ev_loop_destroy (base);
 	}
 	else {
 		lua_pushnil (L);
@@ -1032,70 +1084,13 @@ lua_util_encode_qp (lua_State *L)
 }
 
 static gint
-lua_util_decode_base64 (lua_State *L)
+lua_util_decode_qp (lua_State *L)
 {
 	LUA_TRACE_POINT;
-	struct rspamd_lua_text *t;
+	struct rspamd_lua_text *t, *out;
 	const gchar *s = NULL;
-	gsize inlen, outlen;
-	gboolean zero_copy = FALSE, grab_own = FALSE;
-	gint state = 0;
-	guint save = 0;
-
-	if (lua_type (L, 1) == LUA_TSTRING) {
-		s = luaL_checklstring (L, 1, &inlen);
-	}
-	else if (lua_type (L, 1) == LUA_TUSERDATA) {
-		t = lua_check_text (L, 1);
-
-		if (t != NULL) {
-			s = t->start;
-			inlen = t->len;
-		}
-	}
-
-	if (s != NULL) {
-		if (zero_copy) {
-			/* Decode in place */
-			outlen = g_base64_decode_step (s, inlen, (guchar *)s, &state, &save);
-			t = lua_newuserdata (L, sizeof (*t));
-			rspamd_lua_setclass (L, "rspamd{text}", -1);
-			t->start = s;
-			t->len = outlen;
-
-			if (grab_own) {
-				t->flags |= RSPAMD_TEXT_FLAG_OWN;
-			}
-			else {
-				t->flags = 0;
-			}
-		}
-		else {
-			t = lua_newuserdata (L, sizeof (*t));
-			rspamd_lua_setclass (L, "rspamd{text}", -1);
-			t->len = (inlen / 4) * 3 + 3;
-			t->start = g_malloc (t->len);
-			outlen = g_base64_decode_step (s, inlen, (guchar *)t->start,
-					&state, &save);
-			t->len = outlen;
-			t->flags = RSPAMD_TEXT_FLAG_OWN;
-		}
-	}
-	else {
-		lua_pushnil (L);
-	}
-
-	return 1;
-}
-
-static gint
-lua_util_encode_base32 (lua_State *L)
-{
-	LUA_TRACE_POINT;
-	struct rspamd_lua_text *t;
-	const gchar *s = NULL;
-	gchar *out;
-	gsize inlen, outlen;
+	gsize inlen = 0;
+	gssize outlen;
 
 	if (lua_type (L, 1) == LUA_TSTRING) {
 		s = luaL_checklstring (L, 1, &inlen);
@@ -1113,7 +1108,100 @@ lua_util_encode_base32 (lua_State *L)
 		lua_pushnil (L);
 	}
 	else {
-		out = rspamd_encode_base32 (s, inlen);
+		out = lua_newuserdata (L, sizeof (*t));
+		rspamd_lua_setclass (L, "rspamd{text}", -1);
+		out->start = g_malloc (inlen + 1);
+		out->flags = RSPAMD_TEXT_FLAG_OWN;
+		outlen = rspamd_decode_qp_buf (s, inlen, (gchar *)out->start, inlen + 1);
+
+		if (outlen > 0) {
+			out->len = outlen;
+		}
+		else {
+			/*
+			 * It removes out and frees memory on gc due to RSPAMD_TEXT_FLAG_OWN
+			 */
+			lua_pop (L, 1);
+			lua_pushnil (L);
+		}
+	}
+
+	return 1;
+}
+
+static gint
+lua_util_decode_base64 (lua_State *L)
+{
+	LUA_TRACE_POINT;
+	struct rspamd_lua_text *t;
+	const gchar *s = NULL;
+	gsize inlen = 0, outlen;
+
+	if (lua_type (L, 1) == LUA_TSTRING) {
+		s = luaL_checklstring (L, 1, &inlen);
+	}
+	else if (lua_type (L, 1) == LUA_TUSERDATA) {
+		t = lua_check_text (L, 1);
+
+		if (t != NULL) {
+			s = t->start;
+			inlen = t->len;
+		}
+	}
+
+	if (s != NULL) {
+		t = lua_newuserdata (L, sizeof (*t));
+		rspamd_lua_setclass (L, "rspamd{text}", -1);
+		t->len = (inlen / 4) * 3 + 3;
+		t->start = g_malloc (t->len);
+
+		rspamd_cryptobox_base64_decode (s, inlen, (guchar *)t->start,
+				&outlen);
+		t->len = outlen;
+		t->flags = RSPAMD_TEXT_FLAG_OWN;
+	}
+	else {
+		lua_pushnil (L);
+	}
+
+	return 1;
+}
+
+static gint
+lua_util_encode_base32 (lua_State *L)
+{
+	LUA_TRACE_POINT;
+	struct rspamd_lua_text *t;
+	const gchar *s = NULL;
+	gchar *out;
+	enum rspamd_base32_type btype = RSPAMD_BASE32_DEFAULT;
+	gsize inlen, outlen;
+
+	if (lua_type (L, 1) == LUA_TSTRING) {
+		s = luaL_checklstring (L, 1, &inlen);
+	}
+	else if (lua_type (L, 1) == LUA_TUSERDATA) {
+		t = lua_check_text (L, 1);
+
+		if (t != NULL) {
+			s = t->start;
+			inlen = t->len;
+		}
+	}
+
+	if (lua_type (L, 2) == LUA_TSTRING) {
+		btype = rspamd_base32_decode_type_from_str (lua_tostring (L, 2));
+
+		if (btype == RSPAMD_BASE32_INVALID) {
+			return luaL_error (L, "invalid b32 type: %s", lua_tostring (L, 2));
+		}
+	}
+
+	if (s == NULL) {
+		return luaL_error (L, "invalid arguments");
+	}
+	else {
+		out = rspamd_encode_base32 (s, inlen, btype);
 
 		if (out != NULL) {
 			t = lua_newuserdata (L, sizeof (*t));
@@ -1139,6 +1227,7 @@ lua_util_decode_base32 (lua_State *L)
 	struct rspamd_lua_text *t;
 	const gchar *s = NULL;
 	gsize inlen, outlen;
+	enum rspamd_base32_type btype = RSPAMD_BASE32_DEFAULT;
 
 	if (lua_type (L, 1) == LUA_TSTRING) {
 		s = luaL_checklstring (L, 1, &inlen);
@@ -1152,12 +1241,29 @@ lua_util_decode_base32 (lua_State *L)
 		}
 	}
 
+	if (lua_type (L, 2) == LUA_TSTRING) {
+		btype = rspamd_base32_decode_type_from_str (lua_tostring (L, 2));
+
+		if (btype == RSPAMD_BASE32_INVALID) {
+			return luaL_error (L, "invalid b32 type: %s", lua_tostring (L, 2));
+		}
+	}
+
 	if (s != NULL) {
-		t = lua_newuserdata (L, sizeof (*t));
-		rspamd_lua_setclass (L, "rspamd{text}", -1);
-		t->start = rspamd_decode_base32 (s, inlen, &outlen);
-		t->len = outlen;
-		t->flags = RSPAMD_TEXT_FLAG_OWN;
+		guchar *decoded;
+
+		decoded = rspamd_decode_base32 (s, inlen, &outlen, btype);
+
+		if (decoded) {
+			t = lua_newuserdata (L, sizeof (*t));
+			rspamd_lua_setclass (L, "rspamd{text}", -1);
+			t->start = (const gchar *)decoded;
+			t->len = outlen;
+			t->flags = RSPAMD_TEXT_FLAG_OWN;
+		}
+		else {
+			lua_pushnil (L);
+		}
 	}
 	else {
 		lua_pushnil (L);
@@ -1205,100 +1311,7 @@ lua_util_decode_url (lua_State *L)
 static gint
 lua_util_tokenize_text (lua_State *L)
 {
-	LUA_TRACE_POINT;
-	const gchar *in = NULL;
-	gsize len = 0, pos, ex_len, i;
-	GList *exceptions = NULL, *cur;
-	struct rspamd_lua_text *t;
-	struct rspamd_process_exception *ex;
-	UText utxt = UTEXT_INITIALIZER;
-	GArray *res;
-	rspamd_stat_token_t *w;
-
-	if (lua_type (L, 1) == LUA_TSTRING) {
-		in = luaL_checklstring (L, 1, &len);
-	}
-	else if (lua_type (L, 1) == LUA_TUSERDATA) {
-		t = lua_check_text (L, 1);
-
-		if (t) {
-			in = t->start;
-			len = t->len;
-		}
-	}
-
-	if (in == NULL) {
-		lua_pushnil (L);
-		return 1;
-	}
-
-	if (lua_gettop (L) > 1 && lua_type (L, 2) == LUA_TTABLE) {
-		lua_pushvalue (L, 2);
-		lua_pushnil (L);
-
-		while (lua_next (L, -2) != 0) {
-			if (lua_type (L, -1) == LUA_TTABLE) {
-				lua_rawgeti (L, -1, 1);
-				pos = luaL_checknumber (L, -1);
-				lua_pop (L, 1);
-				lua_rawgeti (L, -1, 2);
-				ex_len = luaL_checknumber (L, -1);
-				lua_pop (L, 1);
-
-				if (ex_len > 0) {
-					ex = g_malloc0 (sizeof (*ex));
-					ex->pos = pos;
-					ex->len = ex_len;
-					ex->type = RSPAMD_EXCEPTION_GENERIC;
-					exceptions = g_list_prepend (exceptions, ex);
-				}
-			}
-			lua_pop (L, 1);
-		}
-
-		lua_pop (L, 1);
-	}
-
-	if (exceptions) {
-		exceptions = g_list_reverse (exceptions);
-	}
-
-	UErrorCode uc_err = U_ZERO_ERROR;
-	utext_openUTF8 (&utxt,
-			in,
-			len,
-			&uc_err);
-
-	res = rspamd_tokenize_text ((gchar *)in, len,
-			&utxt,
-			RSPAMD_TOKENIZE_UTF, NULL,
-			exceptions,
-			NULL, NULL);
-
-	if (res == NULL) {
-		lua_pushnil (L);
-	}
-	else {
-		lua_createtable (L, res->len, 0);
-
-		for (i = 0; i < res->len; i ++) {
-			w = &g_array_index (res, rspamd_stat_token_t, i);
-			lua_pushlstring (L, w->original.begin, w->original.len);
-			lua_rawseti (L, -2, i + 1);
-		}
-	}
-
-	cur = exceptions;
-	while (cur) {
-		ex = cur->data;
-		g_free (ex);
-		cur = g_list_next (cur);
-	}
-
-	g_list_free (exceptions);
-	utext_close (&utxt);
-
-	return 1;
+	return lua_parsers_tokenize_text (L);
 }
 
 static gint
@@ -1315,49 +1328,7 @@ lua_util_tanh (lua_State *L)
 static gint
 lua_util_parse_html (lua_State *L)
 {
-	LUA_TRACE_POINT;
-	struct rspamd_lua_text *t;
-	const gchar *start = NULL;
-	gsize len;
-	GByteArray *res, *in;
-	rspamd_mempool_t *pool;
-	struct html_content *hc;
-
-	if (lua_type (L, 1) == LUA_TUSERDATA) {
-		t = lua_check_text (L, 1);
-
-		if (t != NULL) {
-			start = t->start;
-			len = t->len;
-		}
-	}
-	else if (lua_type (L, 1) == LUA_TSTRING) {
-		start = luaL_checklstring (L, 1, &len);
-	}
-
-	if (start != NULL) {
-		pool = rspamd_mempool_new (rspamd_mempool_suggest_size (), NULL);
-		hc = rspamd_mempool_alloc0 (pool, sizeof (*hc));
-		in = g_byte_array_sized_new (len);
-		g_byte_array_append (in, start, len);
-
-		res = rspamd_html_process_part (pool, hc, in);
-
-		t = lua_newuserdata (L, sizeof (*t));
-		rspamd_lua_setclass (L, "rspamd{text}", -1);
-		t->start = res->data;
-		t->len = res->len;
-		t->flags = RSPAMD_TEXT_FLAG_OWN;
-
-		g_byte_array_free (res, FALSE);
-		g_byte_array_free (in, TRUE);
-		rspamd_mempool_delete (pool);
-	}
-	else {
-		lua_pushnil (L);
-	}
-
-	return 1;
+	return lua_parsers_parse_html (L);
 }
 
 static gint
@@ -1386,50 +1357,6 @@ lua_util_levenshtein_distance (lua_State *L)
 	return 1;
 }
 
-static gint
-lua_util_parse_addr (lua_State *L)
-{
-	LUA_TRACE_POINT;
-	GPtrArray *addrs;
-	gsize len;
-	const gchar *str = luaL_checklstring (L, 1, &len);
-	rspamd_mempool_t *pool;
-	gboolean own_pool = FALSE;
-
-	if (str) {
-
-		if (lua_type (L, 2) == LUA_TUSERDATA) {
-			pool = rspamd_lua_check_mempool (L, 2);
-
-			if (pool == NULL) {
-				return luaL_error (L, "invalid arguments");
-			}
-		}
-		else {
-			pool = rspamd_mempool_new (rspamd_mempool_suggest_size (), "lua util");
-			own_pool = TRUE;
-		}
-
-		addrs = rspamd_email_address_from_mime (pool, str, len, NULL);
-
-		if (addrs == NULL) {
-			lua_pushnil (L);
-		}
-		else {
-			lua_push_emails_address_list (L, addrs, 0);
-		}
-
-		if (own_pool) {
-			rspamd_mempool_delete (pool);
-		}
-	}
-	else {
-		lua_pushnil (L);
-	}
-
-	return 1;
-}
-
 static gint
 lua_util_fold_header (lua_State *L)
 {
@@ -1565,11 +1492,10 @@ lua_util_glob (lua_State *L)
 	LUA_TRACE_POINT;
 	const gchar *pattern;
 	glob_t gl;
-	gint top, i, flags;
+	gint top, i, flags = 0;
 
 	top = lua_gettop (L);
 	memset (&gl, 0, sizeof (gl));
-	flags = GLOB_NOSORT;
 
 	for (i = 1; i <= top; i ++, flags |= GLOB_APPEND) {
 		pattern = luaL_checkstring (L, i);
@@ -1594,66 +1520,60 @@ lua_util_glob (lua_State *L)
 static gint
 lua_util_parse_mail_address (lua_State *L)
 {
-	LUA_TRACE_POINT;
-	GPtrArray *addrs;
-	gsize len;
-	const gchar *str = luaL_checklstring (L, 1, &len);
-	rspamd_mempool_t *pool;
-	gboolean own_pool = FALSE;
-
-	if (str) {
-
-		if (lua_type (L, 2) == LUA_TUSERDATA) {
-			pool = rspamd_lua_check_mempool (L, 2);
-
-			if (pool == NULL) {
-				return luaL_error (L, "invalid arguments");
-			}
-		}
-		else {
-			pool = rspamd_mempool_new (rspamd_mempool_suggest_size (), "lua util");
-			own_pool = TRUE;
-		}
-
-		addrs = rspamd_email_address_from_mime (pool, str, len, NULL);
-
-		if (addrs == NULL) {
-			lua_pushnil (L);
-		}
-		else {
-			lua_push_emails_address_list (L, addrs, 0);
-		}
-
-		if (own_pool) {
-			rspamd_mempool_delete (pool);
-		}
-	}
-	else {
-		lua_pushnil (L);
-	}
-
-	return 1;
+	return lua_parsers_parse_mail_address (L);
 }
 
 static gint
 lua_util_strlen_utf8 (lua_State *L)
 {
 	LUA_TRACE_POINT;
-	const gchar *str, *end;
+	const gchar *str;
 	gsize len;
 
 	str = lua_tolstring (L, 1, &len);
 
 	if (str) {
-		if (g_utf8_validate (str, len, &end)) {
-			len = g_utf8_strlen (str, len);
-		}
-		else if (end != NULL && end > str) {
-			len = (g_utf8_strlen (str, end - str)) /* UTF part */
-					+ (len - (end - str)) /* raw part */;
+		gint32 i = 0, nchars = 0;
+		UChar32 uc;
+
+		while (i < len) {
+			U8_NEXT ((guint8 *) str, i, len, uc);
+			nchars ++;
 		}
 
-		lua_pushinteger (L, len);
+		lua_pushinteger (L, nchars);
+	}
+	else {
+		return luaL_error (L, "invalid arguments");
+	}
+
+	return 1;
+}
+
+static gint
+lua_util_lower_utf8 (lua_State *L)
+{
+	LUA_TRACE_POINT;
+	const gchar *str;
+	gchar *dst;
+	gsize len;
+	UChar32 uc;
+	UBool err = 0;
+	gint32 i = 0, j = 0;
+
+	str = lua_tolstring (L, 1, &len);
+
+	if (str) {
+		dst = g_malloc (len);
+
+		while (i < len && err == 0) {
+			U8_NEXT ((guint8 *) str, i, len, uc);
+			uc = u_tolower (uc);
+			U8_APPEND (dst, j, len, uc, err);
+		}
+
+		lua_pushlstring (L, dst, j);
+		g_free (dst);
 	}
 	else {
 		return luaL_error (L, "invalid arguments");
@@ -1739,17 +1659,8 @@ static gint
 lua_util_get_time (lua_State *L)
 {
 	LUA_TRACE_POINT;
-	gdouble seconds;
-	struct timeval tv;
 
-	if (gettimeofday (&tv, NULL) == 0) {
-		seconds = tv_to_double (&tv);
-	}
-	else {
-		seconds = time (NULL);
-	}
-
-	lua_pushnumber (L, seconds);
+	lua_pushnumber (L, ev_time ());
 
 	return 1;
 }
@@ -1759,19 +1670,13 @@ lua_util_time_to_string (lua_State *L)
 {
 	LUA_TRACE_POINT;
 	gdouble seconds;
-	struct timeval tv;
 	char timebuf[128];
 
 	if (lua_isnumber (L, 1)) {
 		seconds = lua_tonumber (L, 1);
 	}
 	else {
-		if (gettimeofday (&tv, NULL) == 0) {
-			seconds = tv_to_double (&tv);
-		}
-		else {
-			seconds = time (NULL);
-		}
+		seconds = ev_time ();
 	}
 
 	rspamd_http_date_format (timebuf, sizeof (timebuf), seconds);
@@ -2048,6 +1953,7 @@ lua_util_zstd_compress (lua_State *L)
 	LUA_TRACE_POINT;
 	struct rspamd_lua_text *t = NULL, *res, tmp;
 	gsize sz, r;
+	gint comp_level = 1;
 
 	if (lua_type (L, 1) == LUA_TSTRING) {
 		t = &tmp;
@@ -2062,6 +1968,10 @@ lua_util_zstd_compress (lua_State *L)
 		return luaL_error (L, "invalid arguments");
 	}
 
+	if (lua_type (L, 2) == LUA_TNUMBER) {
+		comp_level = lua_tointeger (L, 2);
+	}
+
 	sz = ZSTD_compressBound (t->len);
 
 	if (ZSTD_isError (sz)) {
@@ -2075,7 +1985,7 @@ lua_util_zstd_compress (lua_State *L)
 	res->start = g_malloc (sz);
 	res->flags = RSPAMD_TEXT_FLAG_OWN;
 	rspamd_lua_setclass (L, "rspamd{text}", -1);
-	r = ZSTD_compress ((void *)res->start, sz, t->start, t->len, 1);
+	r = ZSTD_compress ((void *)res->start, sz, t->start, t->len, comp_level);
 
 	if (ZSTD_isError (r)) {
 		msg_err ("cannot compress data: %s", ZSTD_getErrorName (r));
@@ -2146,7 +2056,7 @@ lua_util_zstd_decompress (lua_State *L)
 
 		if (zin.pos < zin.size && zout.pos == zout.size) {
 			/* We need to extend output buffer */
-			zout.size = zout.size * 1.5 + 1.0;
+			zout.size = zout.size * 2;
 			out = g_realloc (zout.dst, zout.size);
 			zout.dst = out;
 		}
@@ -2188,7 +2098,6 @@ lua_util_gzip_compress (lua_State *L)
 	}
 
 
-
 	memset (&strm, 0, sizeof (strm));
 	rc = deflateInit2 (&strm, Z_DEFAULT_COMPRESSION, Z_DEFLATED,
 			MAX_WBITS + 16, MAX_MEM_LEVEL - 1, Z_DEFAULT_STRATEGY);
@@ -2200,14 +2109,14 @@ lua_util_gzip_compress (lua_State *L)
 	sz = deflateBound (&strm, t->len);
 
 	strm.avail_in = t->len;
-	strm.next_in = (guchar *)t->start;
+	strm.next_in = (guchar *) t->start;
 
 	res = lua_newuserdata (L, sizeof (*res));
 	res->start = g_malloc (sz);
 	res->flags = RSPAMD_TEXT_FLAG_OWN;
 	rspamd_lua_setclass (L, "rspamd{text}", -1);
 
-	p = (guchar *)res->start;
+	p = (guchar *) res->start;
 	remain = sz;
 
 	while (strm.avail_in != 0) {
@@ -2221,7 +2130,8 @@ lua_util_gzip_compress (lua_State *L)
 				break;
 			}
 			else {
-				msg_err ("cannot compress data: %s", zError (rc));
+				msg_err ("cannot compress data: %s (last error: %s)",
+						zError (rc), strm.msg);
 				lua_pop (L, 1); /* Text will be freed here */
 				lua_pushnil (L);
 				deflateEnd (&strm);
@@ -2235,9 +2145,9 @@ lua_util_gzip_compress (lua_State *L)
 		if (strm.avail_out == 0 && strm.avail_in != 0) {
 			/* Need to allocate more */
 			remain = res->len;
-			res->start = g_realloc ((gpointer)res->start, strm.avail_in + sz);
+			res->start = g_realloc ((gpointer) res->start, strm.avail_in + sz);
 			sz = strm.avail_in + sz;
-			p = (guchar *)res->start + remain;
+			p = (guchar *) res->start + remain;
 			remain = sz - remain;
 		}
 	}
@@ -2250,7 +2160,7 @@ lua_util_gzip_compress (lua_State *L)
 
 
 static gint
-lua_util_gzip_decompress (lua_State *L)
+lua_util_zlib_inflate (lua_State *L, int windowBits)
 {
 	LUA_TRACE_POINT;
 	struct rspamd_lua_text *t = NULL, *res, tmp;
@@ -2259,6 +2169,7 @@ lua_util_gzip_decompress (lua_State *L)
 	gint rc;
 	guchar *p;
 	gsize remain;
+	gssize size_limit = -1;
 
 	if (lua_type (L, 1) == LUA_TSTRING) {
 		t = &tmp;
@@ -2273,11 +2184,30 @@ lua_util_gzip_decompress (lua_State *L)
 		return luaL_error (L, "invalid arguments");
 	}
 
-	sz = t->len;
+	if (lua_type (L, 2) == LUA_TNUMBER) {
+		size_limit = lua_tointeger (L, 2);
+		if (size_limit <= 0) {
+			return luaL_error (L, "invalid arguments (size_limit)");
+		}
+
+		sz = MIN (t->len * 2, size_limit);
+	}
+	else {
+		sz = t->len * 2;
+	}
 
 	memset (&strm, 0, sizeof (strm));
 	/* windowBits +16 to decode gzip, zlib 1.2.0.4+ */
-	rc = inflateInit2 (&strm, MAX_WBITS + 16);
+
+	/* Here are dragons to distinguish between raw deflate and zlib */
+	if (windowBits == MAX_WBITS && t->len > 0) {
+		if ((int)(unsigned char)((t->start[0] << 4)) != 0x80) {
+			/* Assume raw deflate */
+			windowBits = -windowBits;
+		}
+	}
+
+	rc = inflateInit2 (&strm, windowBits);
 
 	if (rc != Z_OK) {
 		return luaL_error (L, "cannot init zlib");
@@ -2298,14 +2228,15 @@ lua_util_gzip_decompress (lua_State *L)
 		strm.avail_out = remain;
 		strm.next_out = p;
 
-		rc = inflate (&strm, Z_FINISH);
+		rc = inflate (&strm, Z_NO_FLUSH);
 
 		if (rc != Z_OK && rc != Z_BUF_ERROR) {
 			if (rc == Z_STREAM_END) {
 				break;
 			}
 			else {
-				msg_err ("cannot decompress data: %s", zError (rc));
+				msg_err ("cannot decompress data: %s (last error: %s)",
+						zError (rc), strm.msg);
 				lua_pop (L, 1); /* Text will be freed here */
 				lua_pushnil (L);
 				inflateEnd (&strm);
@@ -2317,10 +2248,21 @@ lua_util_gzip_decompress (lua_State *L)
 		res->len = strm.total_out;
 
 		if (strm.avail_out == 0 && strm.avail_in != 0) {
+
+			if (size_limit > 0 || res->len >= G_MAXUINT32 / 2) {
+				if (res->len > size_limit || res->len >= G_MAXUINT32 / 2) {
+					lua_pop (L, 1); /* Text will be freed here */
+					lua_pushnil (L);
+					inflateEnd (&strm);
+
+					return 1;
+				}
+			}
+
 			/* Need to allocate more */
 			remain = res->len;
-			res->start = g_realloc ((gpointer)res->start, strm.avail_in + sz);
-			sz = strm.avail_in + sz;
+			res->start = g_realloc ((gpointer)res->start, res->len * 2);
+			sz = res->len * 2;
 			p = (guchar *)res->start + remain;
 			remain = sz - remain;
 		}
@@ -2329,9 +2271,19 @@ lua_util_gzip_decompress (lua_State *L)
 	inflateEnd (&strm);
 	res->len = strm.total_out;
 
-	return 2;
+	return 1;
+}
+static gint
+lua_util_gzip_decompress (lua_State *L)
+{
+	return lua_util_zlib_inflate (L, MAX_WBITS + 16);
 }
 
+static gint
+lua_util_inflate (lua_State *L)
+{
+	return lua_util_zlib_inflate (L, MAX_WBITS);
+}
 
 static gint
 lua_util_normalize_prob (lua_State *L)
@@ -2509,7 +2461,7 @@ lua_util_is_utf_spoofed (lua_State *L)
 }
 
 static gint
-lua_util_is_utf_mixed_script(lua_State *L)
+lua_util_is_utf_mixed_script (lua_State *L)
 {
 	LUA_TRACE_POINT;
 	gsize len_of_string;
@@ -2575,7 +2527,8 @@ lua_util_get_string_stats (lua_State *L)
 		while (*string_to_check != '\0') {
 			if (g_ascii_isdigit(*string_to_check)) {
 				num_of_digits++;
-			} else if (g_ascii_isalpha(*string_to_check)) {
+			}
+			else if (g_ascii_isalpha(*string_to_check)) {
 				num_of_letters++;
 			}
 			string_to_check++;
@@ -2585,20 +2538,20 @@ lua_util_get_string_stats (lua_State *L)
 		return luaL_error (L, "invalid arguments");
 	}
 
-	lua_createtable(L, 0, 2);
-	lua_pushstring(L, "digits");
-	lua_pushinteger(L, num_of_digits);
-	lua_settable(L, -3);
-	lua_pushstring(L, "letters");
-	lua_pushinteger(L, num_of_letters);
-	lua_settable(L, -3);
+	lua_createtable (L, 0, 2);
+	lua_pushstring (L, "digits");
+	lua_pushinteger (L, num_of_digits);
+	lua_settable (L, -3);
+	lua_pushstring (L, "letters");
+	lua_pushinteger (L, num_of_letters);
+	lua_settable (L, -3);
 
 	return 1;
 }
 
 
 static gint
-lua_util_is_utf_outside_range(lua_State *L)
+lua_util_is_utf_outside_range (lua_State *L)
 {
 	LUA_TRACE_POINT;
 	gsize len_of_string;
@@ -2610,51 +2563,54 @@ lua_util_is_utf_outside_range(lua_State *L)
 	static rspamd_lru_hash_t *validators;
 
 	if (validators == NULL) {
-		validators = rspamd_lru_hash_new_full(16, g_free, (GDestroyNotify)uspoof_close, g_int64_hash, g_int64_equal);
+		validators = rspamd_lru_hash_new_full (16, g_free, (GDestroyNotify) uspoof_close, g_int64_hash, g_int64_equal);
 	}
 
 	if (string_to_check) {
-		guint64 hash_key = (guint64)range_end << 32 || range_start;
+		guint64 hash_key = (guint64) range_end << 32 || range_start;
 
-		USpoofChecker *validator = rspamd_lru_hash_lookup(validators, &hash_key, 0);
+		USpoofChecker *validator = rspamd_lru_hash_lookup (validators, &hash_key, 0);
 
 		UErrorCode uc_err = U_ZERO_ERROR;
 
 		if (validator == NULL) {
-			USet * allowed_chars;
-			guint64 * creation_hash_key = g_malloc(sizeof(guint64));
+			USet *allowed_chars;
+			guint64 *creation_hash_key = g_malloc (sizeof (guint64));
 			*creation_hash_key = hash_key;
 
 			validator = uspoof_open (&uc_err);
 			if (uc_err != U_ZERO_ERROR) {
 				msg_err ("cannot init spoof checker: %s", u_errorName (uc_err));
 				lua_pushboolean (L, false);
-				uspoof_close(validator);
-				g_free(creation_hash_key);
+				uspoof_close (validator);
+				g_free (creation_hash_key);
 				return 1;
 			}
 
-			allowed_chars = uset_openEmpty();
-			uset_addRange(allowed_chars, range_start, range_end);
-			uspoof_setAllowedChars(validator, allowed_chars, &uc_err);
+			allowed_chars = uset_openEmpty ();
+			uset_addRange (allowed_chars, range_start, range_end);
+			uspoof_setAllowedChars (validator, allowed_chars, &uc_err);
 
 			uspoof_setChecks (validator,
-				USPOOF_CHAR_LIMIT | USPOOF_ANY_CASE, &uc_err);
+					USPOOF_CHAR_LIMIT | USPOOF_ANY_CASE, &uc_err);
 
-			uset_close(allowed_chars);
+			uset_close (allowed_chars);
 
 			if (uc_err != U_ZERO_ERROR) {
 				msg_err ("Cannot configure uspoof: %s", u_errorName (uc_err));
 				lua_pushboolean (L, false);
-				uspoof_close(validator);
-				g_free(creation_hash_key);
+				uspoof_close (validator);
+				g_free (creation_hash_key);
 				return 1;
 			}
 
-			rspamd_lru_hash_insert(validators, creation_hash_key, validator, 0, 0);
+			rspamd_lru_hash_insert (validators, creation_hash_key, validator,
+					0, 0);
 		}
 
-		ret = uspoof_checkUTF8 (validator, string_to_check, len_of_string, NULL, &uc_err);
+		gint32 pos = 0;
+		ret = uspoof_checkUTF8 (validator, string_to_check, len_of_string, &pos,
+				&uc_err);
 	}
 	else {
 		return luaL_error (L, "invalid arguments");
@@ -2679,7 +2635,7 @@ lua_util_get_hostname (lua_State *L)
 		hostlen = 256;
 	}
 	else {
-		hostlen ++;
+		hostlen++;
 	}
 
 	hostbuf = g_alloca (hostlen);
@@ -2694,73 +2650,7 @@ lua_util_get_hostname (lua_State *L)
 static gint
 lua_util_parse_content_type (lua_State *L)
 {
-	LUA_TRACE_POINT;
-	gsize len;
-	const gchar *ct_str = luaL_checklstring (L, 1, &len);
-	rspamd_mempool_t *pool = rspamd_lua_check_mempool (L, 2);
-	struct rspamd_content_type *ct;
-
-	if (!ct_str || !pool) {
-		return luaL_error (L, "invalid arguments");
-	}
-
-	ct = rspamd_content_type_parse (ct_str, len, pool);
-
-	if (ct == NULL) {
-		lua_pushnil (L);
-	}
-	else {
-		GHashTableIter it;
-		gpointer k, v;
-
-		lua_createtable (L, 0, 4 + (ct->attrs ? g_hash_table_size (ct->attrs) : 0));
-
-		if (ct->type.len > 0) {
-			lua_pushstring (L, "type");
-			lua_pushlstring (L, ct->type.begin, ct->type.len);
-			lua_settable (L, -3);
-		}
-
-		if (ct->subtype.len > 0) {
-			lua_pushstring (L, "subtype");
-			lua_pushlstring (L, ct->subtype.begin, ct->subtype.len);
-			lua_settable (L, -3);
-		}
-
-		if (ct->charset.len > 0) {
-			lua_pushstring (L, "charset");
-			lua_pushlstring (L, ct->charset.begin, ct->charset.len);
-			lua_settable (L, -3);
-		}
-
-		if (ct->orig_boundary.len > 0) {
-			lua_pushstring (L, "boundary");
-			lua_pushlstring (L, ct->orig_boundary.begin, ct->orig_boundary.len);
-			lua_settable (L, -3);
-		}
-
-		if (ct->attrs) {
-			g_hash_table_iter_init (&it, ct->attrs);
-
-			while (g_hash_table_iter_next (&it, &k, &v)) {
-				struct rspamd_content_type_param *param =
-						(struct rspamd_content_type_param *)v, *cur;
-				guint i = 1;
-
-				lua_pushlstring (L, param->name.begin, param->name.len);
-				lua_createtable (L, 1, 0);
-
-				DL_FOREACH (param, cur) {
-					lua_pushlstring (L, cur->value.begin, cur->value.len);
-					lua_rawseti (L, -2, i++);
-				}
-
-				lua_settable (L, -3);
-			}
-		}
-	}
-
-	return 1;
+	return lua_parsers_parse_content_type (L);
 }
 
 
@@ -2790,10 +2680,33 @@ lua_util_is_valid_utf8 (lua_State *L)
 	const gchar *str;
 	gsize len;
 
-	str = lua_tolstring (L, 1, &len);
+	if (lua_isstring (L, 1)) {
+		str = lua_tolstring (L, 1, &len);
+	}
+	else {
+		struct rspamd_lua_text *t = lua_check_text (L, 1);
+
+		if (t) {
+			str = t->start;
+			len = t->len;
+		}
+		else {
+			return luaL_error (L, "invalid arguments (text expected)");
+		}
+	}
 
 	if (str) {
-		lua_pushboolean (L, g_utf8_validate (str, len, NULL));
+		goffset error_offset = rspamd_fast_utf8_validate (str, len);
+
+		if (error_offset == 0) {
+			lua_pushboolean (L, true);
+		}
+		else {
+			lua_pushboolean (L, false);
+			lua_pushnumber (L, error_offset);
+
+			return 2;
+		}
 	}
 	else {
 		return luaL_error (L, "invalid arguments");
@@ -2843,17 +2756,41 @@ lua_util_readline (lua_State *L)
 	if (lua_type (L, 1) == LUA_TSTRING) {
 		prompt = lua_tostring (L, 1);
 	}
+#ifdef WITH_LUA_REPL
+	static Replxx *rx_instance = NULL;
 
-	input = linenoise (prompt);
+	if (rx_instance == NULL) {
+		rx_instance = replxx_init ();
+	}
+
+	input = (gchar *)replxx_input (rx_instance, prompt);
 
 	if (input) {
 		lua_pushstring (L, input);
-		linenoiseHistoryAdd (input);
-		linenoiseFree (input);
 	}
 	else {
 		lua_pushnil (L);
 	}
+#else
+	size_t linecap = 0;
+	ssize_t linelen;
+
+	fprintf (stdout, "%s ", prompt);
+
+	linelen = getline (&input, &linecap, stdin);
+
+	if (linelen > 0) {
+		if (input[linelen - 1] == '\n') {
+			linelen --;
+		}
+
+		lua_pushlstring (L, input, linelen);
+		free (input);
+	}
+	else {
+		lua_pushnil (L);
+	}
+#endif
 
 	return 1;
 }
@@ -3665,11 +3602,28 @@ lua_util_unpack (lua_State *L)
 	Header h;
 	const char *fmt = luaL_checkstring(L, 1);
 	size_t ld;
-	const char *data = luaL_checklstring (L, 2, &ld);
-	size_t pos = (size_t) posrelat (luaL_optinteger (L, 3, 1), ld) - 1;
+	const char *data;
 	int n = 0;  /* number of results */
+
+	if (lua_type (L, 2) == LUA_TUSERDATA) {
+		struct rspamd_lua_text *t = lua_check_text (L, 2);
+
+		if (!t) {
+			return luaL_error (L, "invalid arguments");
+		}
+
+		data = t->start;
+		ld = t->len;
+	}
+	else {
+		data = luaL_checklstring (L, 2, &ld);
+	}
+
+	size_t pos = (size_t) posrelat (luaL_optinteger (L, 3, 1), ld) - 1;
 	luaL_argcheck(L, pos <= ld, 3, "initial position out of string");
+
 	initheader (L, &h);
+
 	while (*fmt != '\0') {
 		int size, ntoalign;
 		KOption opt = getdetails (&h, pos, &fmt, &size, &ntoalign);
@@ -3736,6 +3690,45 @@ lua_util_unpack (lua_State *L)
 	return n + 1;
 }
 
+static int
+lua_util_btc_polymod (lua_State *L)
+{
+	guint64 c = 1;
+
+	if (lua_type (L, 1) != LUA_TTABLE) {
+		return luaL_error (L, "invalid arguments");
+	}
+
+	for (lua_pushnil (L); lua_next (L, 1); lua_pop (L, 1)) {
+		guint8 c0 = c >> 35;
+		guint64 d = lua_tointeger (L, -1);
+
+		c = ((c & 0x07ffffffff) << 5) ^ d;
+
+		if (c0 & 0x01) c ^= 0x98f2bc8e61;
+		if (c0 & 0x02) c ^= 0x79b76d99e2;
+		if (c0 & 0x04) c ^= 0xf33e5fb3c4;
+		if (c0 & 0x08) c ^= 0xae2eabe2a8;
+		if (c0 & 0x10) c ^= 0x1e4f43e470;
+	}
+
+	if ((c ^ 1) == 0) {
+		lua_pushboolean (L, true);
+	}
+	else {
+		lua_pushboolean (L, false);
+	}
+
+	return 1;
+}
+
+static int
+lua_util_parse_smtp_date (lua_State *L)
+{
+	return lua_parsers_parse_smtp_date (L);
+}
+
+
 static gint
 lua_load_util (lua_State * L)
 {
@@ -3795,14 +3788,14 @@ static int
 lua_ev_base_loop (lua_State *L)
 {
 	int flags = 0;
-	struct event_base *ev_base;
+	struct ev_loop *ev_base;
 
 	ev_base = lua_check_ev_base (L, 1);
 	if (lua_isnumber (L, 2)) {
-		flags = lua_tonumber (L, 2);
+		flags = lua_tointeger (L, 2);
 	}
 
-	int ret = event_base_loop (ev_base, flags);
+	int ret = ev_run (ev_base, flags);
 	lua_pushinteger (L, ret);
 
 	return 1;
