@@ -22,11 +22,12 @@
  *   (e.g. if threshold is 0.1 than charset change should occure more often than in 10 symbols), default: 0.1
  */
 
+
 #include "config.h"
 #include "libmime/message.h"
 #include "rspamd.h"
 #include "libstat/stat_api.h"
-#include "libstat/tokenizers/tokenizers.h"
+#include "libmime/lang_detection.h"
 
 #include "unicode/utf8.h"
 #include "unicode/uchar.h"
@@ -57,7 +58,7 @@ INIT_LOG_MODULE(chartable)
 
 /* Initialization */
 gint chartable_module_init (struct rspamd_config *cfg, struct module_ctx **ctx);
-gint chartable_module_config (struct rspamd_config *cfg);
+gint chartable_module_config (struct rspamd_config *cfg, bool validate);
 gint chartable_module_reconfig (struct rspamd_config *cfg);
 
 module_t chartable_module = {
@@ -108,7 +109,7 @@ chartable_module_init (struct rspamd_config *cfg, struct module_ctx **ctx)
 
 
 gint
-chartable_module_config (struct rspamd_config *cfg)
+chartable_module_config (struct rspamd_config *cfg, bool validate)
 {
 	const ucl_object_t *value;
 	gint res = TRUE;
@@ -173,7 +174,7 @@ chartable_module_config (struct rspamd_config *cfg)
 gint
 chartable_module_reconfig (struct rspamd_config *cfg)
 {
-	return chartable_module_config (cfg);
+	return chartable_module_config (cfg, false);
 }
 
 static gint latin_confusable[] = {
@@ -356,14 +357,17 @@ rspamd_chartable_process_word_utf (struct rspamd_task *task,
 								   rspamd_stat_token_t *w,
 								   gboolean is_url,
 								   guint *ncap,
-								   struct chartable_ctx *chartable_module_ctx)
+								   struct chartable_ctx *chartable_module_ctx,
+								   const gchar *lang,
+								   gboolean ignore_diacritics)
 {
 	const UChar32 *p, *end;
 	gdouble badness = 0.0;
 	UChar32 uc;
 	UBlockCode sc;
+	guint cat;
 	gint last_is_latin = -1;
-	guint same_script_count = 0, nsym = 0;
+	guint same_script_count = 0, nsym = 0, nspecial = 0;
 	enum {
 		start_process = 0,
 		got_alpha,
@@ -383,8 +387,22 @@ rspamd_chartable_process_word_utf (struct rspamd_task *task,
 			break;
 		}
 
+		sc = ublock_getCode (uc);
+		cat = u_charType (uc);
+
+		if (!ignore_diacritics) {
+			if (cat == U_NON_SPACING_MARK ||
+				(sc == UBLOCK_LATIN_1_SUPPLEMENT) ||
+				(sc == UBLOCK_LATIN_EXTENDED_A) ||
+				(sc == UBLOCK_LATIN_EXTENDED_ADDITIONAL) ||
+				(sc == UBLOCK_LATIN_EXTENDED_B) ||
+				(sc == UBLOCK_COMBINING_DIACRITICAL_MARKS)) {
+				nspecial++;
+			}
+		}
+
 		if (u_isalpha (uc)) {
-			sc = ublock_getCode (uc);
+
 			if (sc <= UBLOCK_COMBINING_DIACRITICAL_MARKS ||
 					sc == UBLOCK_LATIN_EXTENDED_ADDITIONAL) {
 				/*
@@ -454,6 +472,16 @@ rspamd_chartable_process_word_utf (struct rspamd_task *task,
 		nsym ++;
 	}
 
+	if (nspecial > 0) {
+		if (!ignore_diacritics) {
+			/* Count diacritics  */
+			badness += nspecial;
+		}
+		else if (nspecial > 1) {
+			badness += (nspecial - 1.0) / 2.0;
+		}
+	}
+
 	/* Try to avoid FP for long words */
 	if (nsym > chartable_module_ctx->max_word_len) {
 		badness = 0;
@@ -557,10 +585,11 @@ rspamd_chartable_process_word_ascii (struct rspamd_task *task,
 	return badness;
 }
 
-static void
+static gboolean
 rspamd_chartable_process_part (struct rspamd_task *task,
 							   struct rspamd_mime_text_part *part,
-							   struct chartable_ctx *chartable_module_ctx)
+							   struct chartable_ctx *chartable_module_ctx,
+							   gboolean ignore_diacritics)
 {
 	rspamd_stat_token_t *w;
 	guint i, ncap = 0;
@@ -568,7 +597,7 @@ rspamd_chartable_process_part (struct rspamd_task *task,
 
 	if (part == NULL || part->utf_words == NULL ||
 			part->utf_words->len == 0) {
-		return;
+		return FALSE;
 	}
 
 	for (i = 0; i < part->utf_words->len; i++) {
@@ -578,7 +607,7 @@ rspamd_chartable_process_part (struct rspamd_task *task,
 
 			if (w->flags & RSPAMD_STAT_TOKEN_FLAG_UTF) {
 				cur_score += rspamd_chartable_process_word_utf (task, w, FALSE,
-						&ncap, chartable_module_ctx);
+						&ncap, chartable_module_ctx, part->language, ignore_diacritics);
 			}
 			else {
 				cur_score += rspamd_chartable_process_word_ascii (task, w,
@@ -596,15 +625,17 @@ rspamd_chartable_process_part (struct rspamd_task *task,
 
 	cur_score /= (gdouble)part->nwords;
 
-	if (cur_score > 2.0) {
-		cur_score = 2.0;
+	if (cur_score > 1.0) {
+		cur_score = 1.0;
 	}
 
 	if (cur_score > chartable_module_ctx->threshold) {
 		rspamd_task_insert_result (task, chartable_module_ctx->symbol,
 				cur_score, NULL);
-
+		return TRUE;
 	}
+
+	return FALSE;
 }
 
 static void
@@ -615,10 +646,32 @@ chartable_symbol_callback (struct rspamd_task *task,
 	guint i;
 	struct rspamd_mime_text_part *part;
 	struct chartable_ctx *chartable_module_ctx = chartable_get_context (task->cfg);
+	const gchar *language = NULL;
+	gboolean ignore_diacritics = FALSE, seen_violated_part = FALSE;
 
-	for (i = 0; i < task->text_parts->len; i ++) {
-		part = g_ptr_array_index (task->text_parts, i);
-		rspamd_chartable_process_part (task, part, chartable_module_ctx);
+	/* Check if we have parts with diacritic symbols language */
+	PTR_ARRAY_FOREACH (MESSAGE_FIELD (task, text_parts), i, part) {
+		if (part->languages && part->languages->len > 0) {
+			struct rspamd_lang_detector_res *lang =
+					(struct rspamd_lang_detector_res *)g_ptr_array_index (part->languages, 0);
+			gint flags;
+
+			flags = rspamd_language_detector_elt_flags (lang->elt);
+
+			if (flags & RS_LANGUAGE_DIACRITICS) {
+				ignore_diacritics = TRUE;
+			}
+		}
+
+		if (rspamd_chartable_process_part (task, part, chartable_module_ctx,
+				ignore_diacritics)) {
+			seen_violated_part = TRUE;
+		}
+	}
+
+	if (MESSAGE_FIELD (task, text_parts)->len == 0) {
+		/* No text parts, assume that we should ignore diacritics checks for metatokens */
+		ignore_diacritics = TRUE;
 	}
 
 	if (task->meta_words != NULL) {
@@ -629,16 +682,23 @@ chartable_symbol_callback (struct rspamd_task *task,
 		for (i = 0; i < arlen; i++) {
 			w = &g_array_index (task->meta_words, rspamd_stat_token_t, i);
 			cur_score += rspamd_chartable_process_word_utf (task, w, FALSE,
-					NULL, chartable_module_ctx);
+					NULL, chartable_module_ctx, language, ignore_diacritics);
 		}
 
 		cur_score /= (gdouble)arlen;
 
-		if (cur_score > 2.0) {
-			cur_score = 2.0;
+		if (cur_score > 1.0) {
+			cur_score = 1.0;
 		}
 
 		if (cur_score > chartable_module_ctx->threshold) {
+			if (!seen_violated_part) {
+				/* Further penalise */
+				if (cur_score > 0.25) {
+					cur_score = 0.25;
+				}
+			}
+
 			rspamd_task_insert_result (task, chartable_module_ctx->symbol,
 					cur_score, "subject");
 
