@@ -58,13 +58,14 @@ static const gchar default_sign_headers[] = ""
 		"(o)to:(o)cc:(x)mime-version:(x)content-type:(x)content-transfer-encoding:"
 		"resent-to:resent-cc:resent-from:resent-sender:resent-message-id:"
 		"(x)in-reply-to:(x)references:list-id:list-help:list-owner:list-unsubscribe:"
-		"list-subscribe:list-post:(x)openpgp:(x)autocrypt";
+		"list-unsubscribe-post:list-subscribe:list-post:(x)openpgp:(x)autocrypt";
 static const gchar default_arc_sign_headers[] = ""
 		"(o)from:(x)sender:(o)reply-to:(o)subject:(x)date:(x)message-id:"
 		"(o)to:(o)cc:(x)mime-version:(x)content-type:(x)content-transfer-encoding:"
 		"resent-to:resent-cc:resent-from:resent-sender:resent-message-id:"
 		"(x)in-reply-to:(x)references:list-id:list-help:list-owner:list-unsubscribe:"
-		"list-subscribe:list-post:dkim-signature:(x)openpgp:(x)autocrypt";
+		"list-unsubscribe-post:list-subscribe:list-post:dkim-signature:(x)openpgp:"
+		"(x)autocrypt";
 
 struct dkim_ctx {
 	struct module_ctx ctx;
@@ -112,7 +113,7 @@ static gint lua_dkim_canonicalize_handler (lua_State *L);
 
 /* Initialization */
 gint dkim_module_init (struct rspamd_config *cfg, struct module_ctx **ctx);
-gint dkim_module_config (struct rspamd_config *cfg);
+gint dkim_module_config (struct rspamd_config *cfg, bool validate);
 gint dkim_module_reconfig (struct rspamd_config *cfg);
 
 module_t dkim_module = {
@@ -308,7 +309,7 @@ dkim_module_init (struct rspamd_config *cfg, struct module_ctx **ctx)
 }
 
 gint
-dkim_module_config (struct rspamd_config *cfg)
+dkim_module_config (struct rspamd_config *cfg, bool validate)
 {
 	const ucl_object_t *value;
 	gint res = TRUE, cb_id = -1;
@@ -342,7 +343,7 @@ dkim_module_config (struct rspamd_config *cfg)
 	value = rspamd_config_get_module_opt (cfg, "dkim", "check_local");
 
 	if (value == NULL) {
-		rspamd_config_get_module_opt (cfg, "options", "check_local");
+		value = rspamd_config_get_module_opt (cfg, "options", "check_local");
 	}
 
 	if (value != NULL) {
@@ -352,10 +353,12 @@ dkim_module_config (struct rspamd_config *cfg)
 		dkim_module_ctx->check_local = FALSE;
 	}
 
-	value = rspamd_config_get_module_opt (cfg, "dkim", "check_authed");
+	value = rspamd_config_get_module_opt (cfg, "dkim",
+			"check_authed");
 
 	if (value == NULL) {
-		rspamd_config_get_module_opt (cfg, "options", "check_authed");
+		value = rspamd_config_get_module_opt (cfg, "options",
+				"check_authed");
 	}
 
 	if (value != NULL) {
@@ -435,7 +438,7 @@ dkim_module_config (struct rspamd_config *cfg)
 		rspamd_config_get_module_opt (cfg, "dkim", "whitelist")) != NULL) {
 
 		rspamd_config_radix_from_ucl (cfg, value, "DKIM whitelist",
-				&dkim_module_ctx->whitelist_ip, NULL, NULL);
+				&dkim_module_ctx->whitelist_ip, NULL, NULL, "dkim whitelist");
 	}
 
 	if ((value =
@@ -466,6 +469,10 @@ dkim_module_config (struct rspamd_config *cfg)
 				NULL, RSPAMD_MAP_DEFAULT)) {
 			msg_warn_config ("cannot load dkim domains list from %s",
 					ucl_object_tostring (value));
+
+			if (validate) {
+				return FALSE;
+			}
 		}
 		else {
 			got_trusted = TRUE;
@@ -520,8 +527,10 @@ dkim_module_config (struct rspamd_config *cfg)
 	}
 
 	if (dkim_module_ctx->trusted_only && !got_trusted) {
-		msg_err_config (
-			"trusted_only option is set and no trusted domains are defined; disabling dkim module completely as it is useless in this case");
+		msg_err_config ("trusted_only option is set and no trusted domains are defined");
+		if (validate) {
+			return FALSE;
+		}
 	}
 	else {
 		if (!rspamd_config_is_module_enabled (cfg, "dkim")) {
@@ -893,7 +902,7 @@ lua_dkim_sign_handler (lua_State *L)
 gint
 dkim_module_reconfig (struct rspamd_config *cfg)
 {
-	return dkim_module_config (cfg);
+	return dkim_module_config (cfg, false);
 }
 
 /*
@@ -1185,6 +1194,7 @@ dkim_symbol_callback (struct rspamd_task *task,
 
 			ctx = rspamd_create_dkim_context (rh_cur->decoded,
 					task->task_pool,
+					task->resolver,
 					dkim_module_ctx->time_jitter,
 					RSPAMD_DKIM_NORMAL,
 					&err);
@@ -1506,6 +1516,7 @@ lua_dkim_verify_handler (lua_State *L)
 
 		ctx = rspamd_create_dkim_context (sig,
 				task->task_pool,
+				task->resolver,
 				dkim_module_ctx->time_jitter,
 				type,
 				&err);
