@@ -19,13 +19,11 @@ limitations under the License.
 
 local rspamd_logger = require "rspamd_logger"
 local mempool = require "rspamd_mempool"
-local rspamd_tcp = require "rspamd_tcp"
 local rspamd_url = require "rspamd_url"
 local rspamd_util = require "rspamd_util"
 local rspamd_redis = require "lua_redis"
 local lua_util = require "lua_util"
-local check_local = false
-local check_authed = false
+local auth_and_local_conf
 
 if confighelp then
   return
@@ -46,15 +44,17 @@ local report_settings = {
   smtp_port = 25,
   retries = 2,
   from_name = 'Rspamd',
+  msgid_from = 'rspamd',
 }
-local report_template = [[From: "%s" <%s>
-To: %s
-Subject: Report Domain: %s
-	Submitter: %s
-	Report-ID: <%s>
-Date: %s
+local report_template = [[From: "{= from_name =}" <{= from_addr =}>
+To: {= rcpt =}
+{%+ if is_string(bcc) %}Bcc: {= bcc =}{%- endif %}
+Subject: Report Domain: {= reporting_domain =}
+	Submitter: {= submitter =}
+	Report-ID: {= report_id =}
+Date: {= report_date =}
 MIME-Version: 1.0
-Message-ID: <%s>
+Message-ID: <{= message_id =}>
 Content-Type: multipart/mixed;
 	boundary="----=_NextPart_000_024E_01CC9B0A.AFE54C00"
 
@@ -64,17 +64,17 @@ This is a multipart message in MIME format.
 Content-Type: text/plain; charset="us-ascii"
 Content-Transfer-Encoding: 7bit
 
-This is an aggregate report from %s.
+This is an aggregate report from {= submitter =}.
 
-Report domain: %s
-Submitter: %s
-Report ID: %s
+Report domain: {= reporting_domain =}
+Submitter: {= submitter =}
+Report ID: {= report_id =}
 
 ------=_NextPart_000_024E_01CC9B0A.AFE54C00
 Content-Type: application/gzip
 Content-Transfer-Encoding: base64
 Content-Disposition: attachment;
-	filename="%s!%s!%s!%s.xml.gz"
+	filename="{= submitter =}!{= reporting_domain =}!{= report_start =}!{= report_end =}.xml.gz"
 
 ]]
 local report_footer = [[
@@ -127,8 +127,18 @@ end
 
 local xml_grammar = gen_xml_grammar()
 
-local function escape_xml(goo)
-  return xml_grammar:match(goo)
+local function escape_xml(input)
+  if type(input) == 'string' or type(input) == 'userdata' then
+    return xml_grammar:match(input)
+  else
+    input = tostring(input)
+
+    if input then
+      return xml_grammar:match(input)
+    end
+  end
+
+  return ''
 end
 
 -- Default port for redis upstreams
@@ -179,10 +189,27 @@ end
 
 local dmarc_grammar = gen_dmarc_grammar()
 
+local function dmarc_key_value_case(elts)
+  if type(elts) ~= "table" then
+    return elts
+  end
+  local result = {}
+  for k, v in pairs(elts) do
+    k = k:lower()
+    if k ~= "v" then
+      v = v:lower()
+    end
+
+    result[k] = v
+  end
+
+  return result
+end
+
 local function dmarc_report(task, spf_ok, dkim_ok, disposition,
     sampled_out, hfromdom, spfdom, dres, spf_result)
   local ip = task:get_from_ip()
-  if not ip:is_valid() then
+  if ip and not ip:is_valid() then
     return nil
   end
   local rspamd_lua_utils = require "lua_util"
@@ -191,9 +218,10 @@ local function dmarc_report(task, spf_ok, dkim_ok, disposition,
   local dkim_fail = table.concat(dres.fail or E, '|')
   local dkim_temperror = table.concat(dres.temperror or E, '|')
   local dkim_permerror = table.concat(dres.permerror or E, '|')
+  local disposition_to_return = (disposition == "softfail") and "none" or disposition
   local res = table.concat({
     ip:to_string(), spf_ok, dkim_ok,
-    disposition, (sampled_out and 'sampled_out' or ''), hfromdom,
+    disposition_to_return, (sampled_out and 'sampled_out' or ''), hfromdom,
     dkim_pass, dkim_fail, dkim_temperror, dkim_permerror, spfdom, spf_result}, ',')
 
   return res
@@ -203,8 +231,7 @@ local function maybe_force_action(task, disposition)
   if disposition then
     local force_action = dmarc_actions[disposition]
     if force_action then
-      -- Don't do anything if pre-result has been already set
-      if task:has_pre_result() then return end
+      -- Set least action
       task:set_pre_result(force_action, 'Action set by DMARC', N, nil, nil, 'least')
     end
   end
@@ -229,6 +256,8 @@ local function dmarc_check_record(task, record, is_tld)
       record, is_tld, elts)
 
   if elts then
+    elts = dmarc_key_value_case(elts)
+
     local dkim_pol = elts['adkim']
     if dkim_pol then
       if dkim_pol == 's' then
@@ -560,7 +589,7 @@ local function dmarc_callback(task)
   local hfromdom = ((from or E)[1] or E).domain
   local dmarc_domain
   local ip_addr = task:get_ip()
-  local dmarc_checks = task:get_mempool():get_variable('dmarc_checks', 'int') or 0
+  local dmarc_checks = task:get_mempool():get_variable('dmarc_checks', 'double') or 0
   local seen_invalid = false
 
   if dmarc_checks ~= 2 then
@@ -568,8 +597,7 @@ local function dmarc_callback(task)
     return
   end
 
-  if ((not check_authed and task:get_user()) or
-      (not check_local and ip_addr and ip_addr:is_local())) then
+  if lua_util.is_skip_local_or_authed(task, auth_and_local_conf, ip_addr) then
     rspamd_logger.infox(task, "skip DMARC checks for local networks and authorized users")
     return
   end
@@ -664,6 +692,11 @@ local function dmarc_callback(task)
               end
             end
           end
+
+          if not has_valid_policy and not seen_invalid then
+            policy_target.err = lookup_domain .. ':' .. ' no valid DMARC record'
+            policy_target.symbol = dmarc_symbols['na']
+          end
         end
       end
 
@@ -705,29 +738,14 @@ local function dmarc_callback(task)
 end
 
 
-local function try_opts(where)
-  local ret = false
-  local opts = rspamd_config:get_all_opt(where)
-  if type(opts) == 'table' then
-    if type(opts['check_local']) == 'boolean' then
-      check_local = opts['check_local']
-      ret = true
-    end
-    if type(opts['check_authed']) == 'boolean' then
-      check_authed = opts['check_authed']
-      ret = true
-    end
-  end
-
-  return ret
-end
-
-if not try_opts(N) then try_opts('options') end
-
 local opts = rspamd_config:get_all_opt('dmarc')
 if not opts or type(opts) ~= 'table' then
   return
 end
+
+auth_and_local_conf = lua_util.config_check_local_or_authed(rspamd_config, N,
+    false, false)
+
 no_sampling_domains = rspamd_map_add(N, 'no_sampling_domains', 'map', 'Domains not to apply DMARC sampling to')
 no_reporting_domains = rspamd_map_add(N, 'no_reporting_domains', 'map', 'Domains not to apply DMARC reporting to')
 
@@ -763,7 +781,9 @@ if opts['reporting'] == true then
     take_report_id = rspamd_redis.add_redis_script(take_report_script, redis_params)
     rspamd_config:add_on_load(function(cfg, ev_base, worker)
       if not worker:is_primary_controller() then return end
+
       pool = mempool.create()
+
       rspamd_config:register_finish_script(function ()
         local stamp = pool:get_variable(VAR_NAME, 'double')
         if not stamp then
@@ -779,8 +799,11 @@ if opts['reporting'] == true then
         assert(f:close())
         pool:destroy()
       end)
-      local get_reporting_domain, reporting_domain, report_start, report_end, report_id, want_period, report_key
-      local reporting_addr = {}
+
+      local get_reporting_domain, reporting_domain, report_start,
+            report_end, report_id, want_period, report_key
+      local reporting_addrs = {}
+      local bcc_addrs = {}
       local domain_policy = {}
       local to_verify = {}
       local cursor = 0
@@ -794,7 +817,7 @@ if opts['reporting'] == true then
           }),
         }
         if data.override ~= '' then
-          table.insert(buf, string.format('<reason>%s</reason>', data.override))
+          table.insert(buf, string.format('<reason><type>%s</type></reason>', data.override))
         end
         table.insert(buf, table.concat({
           '</policy_evaluated></row><identifiers><header_from>', data.header_from,
@@ -817,7 +840,8 @@ if opts['reporting'] == true then
       end
       local function dmarc_report_xml()
         local entries = {}
-        report_id = string.format('%s.%d.%d', reporting_domain, report_start, report_end)
+        report_id = string.format('%s.%d.%d',
+            reporting_domain, report_start, report_end)
         lua_util.debugm(N, rspamd_config, 'new report: %s', report_id)
         local actions = {
           push = function(t)
@@ -854,13 +878,15 @@ if opts['reporting'] == true then
               end
             end
             if split[10] and split[10] ~= '' then
-              local tmp = rspamd_str_split(split[10], '|')
+              local tmp = lua_util.str_split(split[10], '|')
               for _, d in ipairs(tmp) do
-                table.insert(row.dkim_results, {domain = d, result = 'permerror'})
+                table.insert(row.dkim_results,
+                    {domain = d, result = 'permerror'})
               end
             end
             table.insert(entries, row)
           end,
+          -- TODO: please rework this shit
           header = function()
               return table.concat({
                 '<?xml version="1.0" encoding="utf-8"?><feedback><report_metadata><org_name>',
@@ -870,7 +896,8 @@ if opts['reporting'] == true then
                 '</begin><end>', report_end, '</end></date_range></report_metadata><policy_published><domain>',
                 reporting_domain, '</domain><adkim>', escape_xml(domain_policy.adkim), '</adkim><aspf>',
                 escape_xml(domain_policy.aspf), '</aspf><p>', escape_xml(domain_policy.p),
-                '</p><sp>', escape_xml(domain_policy.sp), '</sp><pct>', escape_xml(domain_policy.pct),
+                '</p><sp>', escape_xml(domain_policy.sp), '</sp><pct>',
+                escape_xml(domain_policy.pct),
                 '</pct></policy_published>'
               })
           end,
@@ -891,172 +918,93 @@ if opts['reporting'] == true then
           return f(p)
         end
       end
+
+
       local function send_report_via_email(xmlf, retry)
         if not retry then retry = 0 end
-        if retry > report_settings.retries then
-          rspamd_logger.errx(rspamd_config, "Couldn't send mail for %s: retries exceeded", reporting_domain)
-          return get_reporting_domain()
-        end
-        local tmp_addr = {}
-        for k in pairs(reporting_addr) do
-          table.insert(tmp_addr, k)
+
+        local function sendmail_cb(ret, err)
+          if not ret then
+            rspamd_logger.errx(rspamd_config, "Couldn't send mail for %s: %s", err)
+            if retry >= report_settings.retries then
+              rspamd_logger.errx(rspamd_config, "Couldn't send mail for %s: retries exceeded", reporting_domain)
+              return get_reporting_domain()
+            else
+              send_report_via_email(xmlf, retry + 1)
+            end
+          else
+            get_reporting_domain()
+          end
         end
+
+        -- Format message
+        local list_rcpt = lua_util.keys(reporting_addrs)
+
         local encoded = rspamd_util.encode_base64(rspamd_util.gzip_compress(
               table.concat(
                 {xmlf('header'),
                  xmlf('entries'),
                  xmlf('footer')})), 73)
-        local function mail_cb(err, data, conn)
-          local function no_error(merr, mdata, wantcode)
-            wantcode = wantcode or '2'
-            if merr then
-              rspamd_logger.errx(ev_base, 'got error in tcp callback: %s', merr)
-              if conn then
-                conn:close()
-              end
-              send_report_via_email(xmlf, retry+1)
-              return false
-            end
-            if mdata then
-              if type(mdata) ~= 'string' then
-                mdata = tostring(mdata)
-              end
-              if string.sub(mdata, 1, 1) ~= wantcode then
-                rspamd_logger.errx(ev_base, 'got bad smtp response: %s', mdata)
-                if conn then
-                  conn:close()
-                end
-                send_report_via_email(xmlf, retry+1)
-                return false
-              end
-            else
-              rspamd_logger.errx(ev_base, 'no data')
-              if conn then
-                conn:close()
-              end
-              send_report_via_email(xmlf, retry+1)
-              return false
-            end
-            return true
-          end
-          local function all_done_cb(merr, mdata)
-            if conn then
-              conn:close()
-            end
-            get_reporting_domain()
-            return true
-          end
-          local function quit_done_cb(merr, mdata)
-            conn:add_read(all_done_cb, '\r\n')
-          end
-          local function quit_cb(merr, mdata)
-            if no_error(merr, mdata) then
-              conn:add_write(quit_done_cb, 'QUIT\r\n')
-            end
-          end
-          local function pre_quit_cb(merr, mdata)
-            if no_error(merr, '2') then
-              conn:add_read(quit_cb, '\r\n')
-            end
-          end
-          local function data_done_cb(merr, mdata)
-            if no_error(merr, mdata, '3') then
-              local atmp = {}
-              for k in pairs(reporting_addr) do
-                table.insert(atmp, k)
-              end
-              local addr_string = table.concat(atmp, ', ')
-              -- TODO: migrate to templates and remove this shit
-              local rhead = string.format(report_template:gsub("\n", "\r\n"),
-                  report_settings.from_name,
-                  report_settings.email,
-                  addr_string,
-                  reporting_domain,
-                  report_settings.domain,
-                  report_id,
-                  rspamd_util.time_to_string(rspamd_util.get_time()),
-                  rspamd_util.random_hex(12) .. '@rspamd', -- Message-id
-                  report_settings.domain, -- Plain text part
-                  reporting_domain,
-                  addr_string,
-                  report_id,
-                  report_settings.domain,
-                  reporting_domain,
-                  report_start, report_end)
-              conn:add_write(pre_quit_cb, {rhead,
-                                           encoded,
-                                           report_footer:gsub("\n", "\r\n"),
-                                           '\r\n.\r\n'})
-            end
-          end
-          local function data_cb(merr, mdata)
-            if no_error(merr, '2') then
-              conn:add_read(data_done_cb, '\r\n')
-            end
-          end
-          local function rcpt_done_cb(merr, mdata)
-            if no_error(merr, mdata) then
-              conn:add_write(data_cb, 'DATA\r\n')
-            end
-          end
-          local from_done_cb
-          local function rcpt_cb(merr, mdata)
-            if no_error(merr, '2') then
-              if tmp_addr[1] then
-                conn:add_read(from_done_cb, '\r\n')
-              else
-                conn:add_read(rcpt_done_cb, '\r\n')
-              end
-            end
-          end
-          from_done_cb = function(merr, mdata)
-            if no_error(merr, mdata) then
-              conn:add_write(rcpt_cb, {'RCPT TO: <', table.remove(tmp_addr), '>\r\n'})
-            end
-          end
-          local function from_cb(merr, mdata)
-            if no_error(merr, '2') then
-              conn:add_read(from_done_cb, '\r\n')
-            end
-          end
-            local function hello_done_cb(merr, mdata)
-            if no_error(merr, mdata) then
-              conn:add_write(from_cb, {'MAIL FROM: <', report_settings.email, '>\r\n'})
-            end
-          end
-          local function hello_cb(merr)
-            if no_error(merr, '2') then
-              conn:add_read(hello_done_cb, '\r\n')
-            end
-          end
-          if no_error(err, data) then
-            conn:add_write(hello_cb, {'HELO ', report_settings.helo, '\r\n'})
-          end
+        local addr_string = table.concat(list_rcpt, ', ')
+
+        bcc_addrs = lua_util.keys(bcc_addrs)
+        local bcc_string
+        if #bcc_addrs > 0 then
+          bcc_string = table.concat(bcc_addrs, ', ')
         end
-        rspamd_tcp.request({
+
+        local rhead = lua_util.jinja_template(report_template,
+            {
+              from_name = report_settings.from_name,
+              from_addr = report_settings.email,
+              rcpt = addr_string,
+              bcc = bcc_string,
+              reporting_domain = reporting_domain,
+              submitter = report_settings.domain,
+              report_id = report_id,
+              report_date = rspamd_util.time_to_string(rspamd_util.get_time()),
+              message_id = rspamd_util.random_hex(16) .. '@' .. report_settings.msgid_from,
+              report_start = report_start,
+              report_end = report_end
+            }, true)
+        local message = {
+          (rhead:gsub("\n", "\r\n")),
+          encoded,
+          (report_footer:gsub("\n", "\r\n"))
+        }
+
+        local lua_smtp = require "lua_smtp"
+        lua_smtp.sendmail({
           ev_base = ev_base,
-          callback = mail_cb,
           config = rspamd_config,
-          stop_pattern = '\r\n',
           host = report_settings.smtp,
           port = report_settings.smtp_port,
           resolver = rspamd_config:get_resolver(),
-        })
+          from = report_settings.email,
+          recipients = list_rcpt,
+          helo =  report_settings.helo,
+        }, message, sendmail_cb)
       end
+
+
       local function make_report()
         if type(report_settings.override_address) == 'string' then
-          reporting_addr = {[report_settings.override_address] = true}
+          reporting_addrs = { [report_settings.override_address] = true}
         end
         if type(report_settings.additional_address) == 'string' then
-          reporting_addr[report_settings.additional_address] = true
+          if report_settings.additional_address_bcc then
+            bcc_addrs[report_settings.additional_address] = true
+          else
+            reporting_addrs[report_settings.additional_address] = true
+          end
         end
-        rspamd_logger.infox(ev_base, 'sending report for %s <%s>',
-            reporting_domain, table.concat(reporting_addr, ','))
+        rspamd_logger.infox(rspamd_config, 'sending report for %s <%s> (<%s> bcc)',
+            reporting_domain, reporting_addrs, bcc_addrs)
         local dmarc_xml = dmarc_report_xml()
         local dmarc_push_cb
         dmarc_push_cb = function(err, data)
           if err then
-            rspamd_logger.errx(ev_base, 'Redis request failed: %s', err)
+            rspamd_logger.errx(rspamd_config, 'redis request failed: %s', err)
             -- XXX: data is orphaned; replace key or delete data
             get_reporting_domain()
           elseif type(data) == 'table' then
@@ -1075,7 +1023,7 @@ if opts['reporting'] == true then
                 {report_key, cursor, 'COUNT', report_settings.hscan_count}
               )
               if not ret then
-                rspamd_logger.errx(ev_base, 'Failed to schedule redis request')
+                rspamd_logger.errx(rspamd_config, 'failed to schedule redis request')
                 get_reporting_domain()
               end
             else
@@ -1083,6 +1031,7 @@ if opts['reporting'] == true then
             end
           end
         end
+
         local ret = rspamd_redis.redis_make_request_taskless(ev_base,
           rspamd_config,
           redis_params,
@@ -1093,7 +1042,7 @@ if opts['reporting'] == true then
           {report_key, cursor, 'COUNT', report_settings.hscan_count}
         )
         if not ret then
-          rspamd_logger.errx(rspamd_config, 'Failed to schedule redis request')
+          rspamd_logger.errx(rspamd_config, 'failed to schedule redis request')
           -- XXX: data is orphaned; replace key or delete data
           get_reporting_domain()
         end
@@ -1101,9 +1050,9 @@ if opts['reporting'] == true then
       local function delete_reports()
         local function delete_reports_cb(err)
           if err then
-            rspamd_logger.errx(rspamd_config, 'Error deleting reports: %s', err)
+            rspamd_logger.errx(rspamd_config, 'error deleting reports: %s', err)
           end
-          rspamd_logger.infox(rspamd_config, 'Deleted reports for %s', reporting_domain)
+          rspamd_logger.infox(rspamd_config, 'deleted reports for %s', reporting_domain)
           get_reporting_domain()
         end
         local ret = rspamd_redis.redis_make_request_taskless(ev_base,
@@ -1116,7 +1065,7 @@ if opts['reporting'] == true then
           {report_key}
         )
         if not ret then
-          rspamd_logger.errx(rspamd_config, 'Failed to schedule redis request')
+          rspamd_logger.errx(rspamd_config, 'failed to schedule redis request')
           get_reporting_domain()
         end
       end
@@ -1126,10 +1075,10 @@ if opts['reporting'] == true then
           local function verify_cb(resolver, to_resolve, results, err, _, authenticated)
             if err then
               if err == 'no records with this name' or err == 'requested record is not found' then
-                rspamd_logger.infox(rspamd_config, 'Reports to %s for %s not authorised', test_addr, reporting_domain)
+                rspamd_logger.infox(rspamd_config, 'reports to %s for %s not authorised', test_addr, reporting_domain)
                 to_verify[test_addr] = nil
               else
-                rspamd_logger.errx(rspamd_config, 'Lookup error [%s]: %s', to_resolve, err)
+                rspamd_logger.errx(rspamd_config, 'lookup error [%s]: %s', to_resolve, err)
                 if retry < report_settings.retries then
                   retry = retry + 1
                   rspamd_config:get_resolver():resolve('txt', {
@@ -1156,12 +1105,12 @@ if opts['reporting'] == true then
                 rspamd_logger.infox(rspamd_config, 'Reports to %s for %s not authorised', test_addr, reporting_domain)
               else
                 to_verify[test_addr] = nil
-                reporting_addr[test_addr] = true
+                reporting_addrs[test_addr] = true
               end
             end
             local t, nvdom = next(to_verify)
             if not t then
-              if next(reporting_addr) then
+              if next(reporting_addrs) then
                 make_report()
               else
                 rspamd_logger.infox(rspamd_config, 'No valid reporting addresses for %s', reporting_domain)
@@ -1194,11 +1143,11 @@ if opts['reporting'] == true then
                   callback = check_addr_cb,
                 })
               else
-                rspamd_logger.errx(rspamd_config, 'No DMARC record found for %s', reporting_domain)
+                rspamd_logger.errx(rspamd_config, 'no DMARC record found for %s', reporting_domain)
                 delete_reports()
               end
             else
-              rspamd_logger.errx(rspamd_config, 'Lookup error [%s]: %s', to_resolve, err)
+              rspamd_logger.errx(rspamd_config, 'lookup error [%s]: %s', to_resolve, err)
               if retry < report_settings.retries then
                 retry = retry + 1
                 rspamd_config:get_resolver():resolve('txt', {
@@ -1207,7 +1156,7 @@ if opts['reporting'] == true then
                   callback = check_addr_cb,
                 })
               else
-                rspamd_logger.errx(rspamd_config, "Couldn't get reporting address for %s: retries exceeded",
+                rspamd_logger.errx(rspamd_config, "couldn't get reporting address for %s: retries exceeded",
                     reporting_domain)
                 delete_reports()
               end
@@ -1221,11 +1170,11 @@ if opts['reporting'] == true then
                 failed_policy = true
               elseif elts then
                 found_policy = true
-                policy = elts
+                policy = dmarc_key_value_case(elts)
               end
             end
             if not found_policy then
-              rspamd_logger.errx(rspamd_config, 'No policy: %s', to_resolve)
+              rspamd_logger.errx(rspamd_config, 'no policy: %s', to_resolve)
               if reporting_domain ~= esld then
                 rspamd_config:get_resolver():resolve('txt', {
                   ev_base = ev_base,
@@ -1236,25 +1185,25 @@ if opts['reporting'] == true then
                 delete_reports()
               end
             elseif failed_policy then
-              rspamd_logger.errx(rspamd_config, 'Duplicate policies: %s', to_resolve)
+              rspamd_logger.errx(rspamd_config, 'duplicate policies: %s', to_resolve)
               delete_reports()
             elseif not policy['rua'] then
-              rspamd_logger.errx(rspamd_config, 'No reporting address: %s', to_resolve)
+              rspamd_logger.errx(rspamd_config, 'no reporting address: %s', to_resolve)
               delete_reports()
             else
               local upool = mempool.create()
-              local split = rspamd_str_split(policy['rua'], ',')
+              local split = rspamd_str_split(policy['rua']:gsub('%s+', ''), ',')
               for _, m in ipairs(split) do
                 local url = rspamd_url.create(upool, m)
                 if not url then
-                  rspamd_logger.errx(rspamd_config, 'Couldnt extract reporting address: %s', policy['rua'])
+                  rspamd_logger.errx(rspamd_config, "couldn't extract reporting address: %s", policy['rua'])
                 else
                   local urlt = url:to_table()
                   if urlt['protocol'] ~= 'mailto' then
-                    rspamd_logger.errx(rspamd_config, 'Invalid URL: %s', url)
+                    rspamd_logger.errx(rspamd_config, 'invalid URL: %s', url)
                   else
                     if urlt['tld'] == rspamd_util.get_tld(reporting_domain) then
-                      reporting_addr[string.format('%s@%s', urlt['user'], urlt['host'])] = true
+                      reporting_addrs[string.format('%s@%s', urlt['user'], urlt['host'])] = true
                     else
                       to_verify[string.format('%s@%s', urlt['user'], urlt['host'])] = urlt['host']
                     end
@@ -1269,10 +1218,10 @@ if opts['reporting'] == true then
               domain_policy['sp'] = policy['sp'] or 'none'
               if next(to_verify) then
                 verify_reporting_address()
-              elseif next(reporting_addr) then
+              elseif next(reporting_addrs) then
                 make_report()
               else
-                rspamd_logger.errx(rspamd_config, 'No reporting address for %s', reporting_domain)
+                rspamd_logger.errx(rspamd_config, 'no reporting address for %s', reporting_domain)
                 delete_reports()
               end
             end
@@ -1287,12 +1236,12 @@ if opts['reporting'] == true then
       end
       get_reporting_domain = function()
         reporting_domain = nil
-        reporting_addr = {}
+        reporting_addrs = {}
         domain_policy = {}
         cursor = 0
         local function get_reporting_domain_cb(err, data)
           if err then
-            rspamd_logger.errx(cfg, 'Unable to get DMARC domain: %s', err)
+            rspamd_logger.errx(cfg, 'unable to get DMARC domain: %s', err)
           else
             if type(data) == 'userdata' then
               reporting_domain = nil
@@ -1302,7 +1251,7 @@ if opts['reporting'] == true then
               reporting_domain = tmp[2]
             end
             if not reporting_domain then
-              rspamd_logger.infox(cfg, 'No more domains to generate reports for')
+              rspamd_logger.infox(cfg, 'no more domains to generate reports for')
             else
               get_reporting_address()
             end
@@ -1319,11 +1268,11 @@ if opts['reporting'] == true then
           {idx_key}
         )
         if not ret then
-          rspamd_logger.errx(cfg, 'Unable to get DMARC domain')
+          rspamd_logger.errx(cfg, 'unable to get DMARC domain')
         end
       end
       local function send_reports(time)
-        rspamd_logger.infox(ev_base, 'sending reports ostensibly %1', time)
+        rspamd_logger.infox(rspamd_config, 'sending reports ostensibly %1', time)
         pool:set_variable(VAR_NAME, time)
         local yesterday = os.date('!*t', rspamd_util.get_time() - INTERVAL)
         local today = os.date('!*t', rspamd_util.get_time())
@@ -1363,7 +1312,7 @@ if opts['reporting'] == true then
       local stamp
       local f, err = io.open(statefile, 'r')
       if err then
-        rspamd_logger.errx('Failed to open statefile: %s', err)
+        rspamd_logger.errx(rspamd_config, 'failed to open statefile: %s', err)
       end
       if f then
         io.input(f)
@@ -1372,19 +1321,19 @@ if opts['reporting'] == true then
       end
       local time = rspamd_util.get_time()
       if not stamp then
-        lua_util.debugm(N, rspamd_config, 'No state found - sending reports immediately')
+        lua_util.debugm(N, rspamd_config, 'no state found - sending reports immediately')
         schedule_regular_send()
         send_reports(time)
         return
       end
       local delta = stamp - time + INTERVAL
       if delta <= 0 then
-        lua_util.debugm(N, rspamd_config, 'Last send is too old - sending reports immediately')
+        lua_util.debugm(N, rspamd_config, 'last send is too old - sending reports immediately')
         schedule_regular_send()
         send_reports(time)
         return
       end
-      lua_util.debugm(N, rspamd_config, 'Scheduling next send in %s seconds', delta)
+      lua_util.debugm(N, rspamd_config, 'scheduling next send in %s seconds', delta)
       schedule_intermediate_send(delta)
     end)
   end
@@ -1400,7 +1349,7 @@ end
 if opts['send_reports'] then
   for _, e in ipairs({'email', 'domain', 'org_name'}) do
     if not report_settings[e] then
-      rspamd_logger.errx(rspamd_config, 'Missing required setting: report_settings.%s', e)
+      rspamd_logger.errx(rspamd_config, 'missing required setting: report_settings.%s', e)
       return
     end
   end
@@ -1473,6 +1422,13 @@ rspamd_config:register_symbol({
   groups = {'dmarc'},
   type = 'virtual'
 })
+rspamd_config:register_symbol({
+  name = dmarc_symbols['badpolicy'],
+  parent = id,
+  group = 'policies',
+  groups = {'dmarc'},
+  type = 'virtual'
+})
 rspamd_config:register_symbol({
   name = dmarc_symbols['na'],
   parent = id,
@@ -1483,4 +1439,3 @@ rspamd_config:register_symbol({
 
 rspamd_config:register_dependency('DMARC_CALLBACK', symbols['spf_allow_symbol'])
 rspamd_config:register_dependency('DMARC_CALLBACK', symbols['dkim_allow_symbol'])
-
