@@ -22,7 +22,7 @@
 #include "libmime/message.h"
 #include "expression.h"
 #include "mime_expressions.h"
-#include "libutil/map.h"
+#include "libserver/maps/map.h"
 #include "lua/lua_common.h"
 
 static const guint64 rspamd_regexp_cb_magic = 0xca9d9649fc3e2659ULL;
@@ -46,7 +46,7 @@ static void process_regexp_item (struct rspamd_task *task,
 
 /* Initialization */
 gint regexp_module_init (struct rspamd_config *cfg, struct module_ctx **ctx);
-gint regexp_module_config (struct rspamd_config *cfg);
+gint regexp_module_config (struct rspamd_config *cfg, bool validate);
 gint regexp_module_reconfig (struct rspamd_config *cfg);
 
 module_t regexp_module = {
@@ -130,13 +130,13 @@ regexp_module_init (struct rspamd_config *cfg, struct module_ctx **ctx)
 }
 
 gint
-regexp_module_config (struct rspamd_config *cfg)
+regexp_module_config (struct rspamd_config *cfg, bool validate)
 {
 	struct regexp_ctx *regexp_module_ctx = regexp_get_context (cfg);
 	struct regexp_module_item *cur_item = NULL;
 	const ucl_object_t *sec, *value, *elt;
 	ucl_object_iter_t it = NULL;
-	gint res = TRUE, id, nre = 0, nlua = 0, nshots = cfg->default_max_shots;
+	gint res = TRUE, nre = 0, nlua = 0, nshots = cfg->default_max_shots;
 
 	if (!rspamd_config_is_module_enabled (cfg, "regexp")) {
 		return TRUE;
@@ -174,7 +174,9 @@ regexp_module_config (struct rspamd_config *cfg)
 			if (!read_regexp_expression (cfg->cfg_pool,
 					cur_item, ucl_object_key (value),
 					ucl_obj_tostring (value), &ud)) {
-				res = FALSE;
+				if (validate) {
+					return FALSE;
+				}
 			}
 			else {
 				rspamd_symcache_add_symbol (cfg->cache,
@@ -228,7 +230,9 @@ regexp_module_config (struct rspamd_config *cfg)
 					if (!read_regexp_expression (cfg->cfg_pool,
 							cur_item, ucl_object_key (value),
 							ucl_obj_tostring (elt), &ud)) {
-						res = FALSE;
+						if (validate) {
+							return FALSE;
+						}
 					}
 					else {
 						valid_expression = TRUE;
@@ -263,7 +267,9 @@ regexp_module_config (struct rspamd_config *cfg)
 								"mime_only attribute is not boolean for symbol: '%s'",
 								cur_item->symbol);
 
-						res = FALSE;
+						if (validate) {
+							return FALSE;
+						}
 					}
 					else {
 						if (ucl_object_toboolean (elt)) {
@@ -272,7 +278,7 @@ regexp_module_config (struct rspamd_config *cfg)
 					}
 				}
 
-				id = rspamd_symcache_add_symbol (cfg->cache,
+				rspamd_symcache_add_symbol (cfg->cache,
 						cur_item->symbol,
 						0,
 						process_regexp_item,
@@ -314,7 +320,9 @@ regexp_module_config (struct rspamd_config *cfg)
 								"score attribute is not numeric for symbol: '%s'",
 								cur_item->symbol);
 
-						res = FALSE;
+						if (validate) {
+							return FALSE;
+						}
 					}
 					else {
 						score = ucl_object_todouble (elt);
@@ -329,7 +337,9 @@ regexp_module_config (struct rspamd_config *cfg)
 								"one_shot attribute is not boolean for symbol: '%s'",
 								cur_item->symbol);
 
-						res = FALSE;
+						if (validate) {
+							return FALSE;
+						}
 					}
 					else {
 						if (ucl_object_toboolean (elt)) {
@@ -344,7 +354,9 @@ regexp_module_config (struct rspamd_config *cfg)
 								"any_shot attribute is not boolean for symbol: '%s'",
 								cur_item->symbol);
 
-						res = FALSE;
+						if (validate) {
+							return FALSE;
+						}
 					}
 					else {
 						if (ucl_object_toboolean (elt)) {
@@ -359,7 +371,9 @@ regexp_module_config (struct rspamd_config *cfg)
 								"nshots attribute is not numeric for symbol: '%s'",
 								cur_item->symbol);
 
-						res = FALSE;
+						if (validate) {
+							return FALSE;
+						}
 					}
 					else {
 						nshots = ucl_object_toint (elt);
@@ -374,7 +388,9 @@ regexp_module_config (struct rspamd_config *cfg)
 								"one_param attribute is not boolean for symbol: '%s'",
 								cur_item->symbol);
 
-						res = FALSE;
+						if (validate) {
+							return FALSE;
+						}
 					}
 					else {
 						if (ucl_object_toboolean (elt)) {
@@ -391,7 +407,9 @@ regexp_module_config (struct rspamd_config *cfg)
 								"priority attribute is not numeric for symbol: '%s'",
 								cur_item->symbol);
 
-						res = FALSE;
+						if (validate) {
+							return FALSE;
+						}
 					}
 					else {
 						priority = ucl_object_toint (elt);
@@ -427,8 +445,13 @@ regexp_module_config (struct rspamd_config *cfg)
 		}
 	}
 
-	msg_info_config ("init internal regexp module, %d regexp rules and %d "
-			"lua rules are loaded", nre, nlua);
+	if (res) {
+		msg_info_config ("init internal regexp module, %d regexp rules and %d "
+						 "lua rules are loaded", nre, nlua);
+	}
+	else {
+		msg_err_config ("fatal regexp module error");
+	}
 
 	return res;
 }
@@ -436,13 +459,13 @@ regexp_module_config (struct rspamd_config *cfg)
 gint
 regexp_module_reconfig (struct rspamd_config *cfg)
 {
-	return regexp_module_config (cfg);
+	return regexp_module_config (cfg, false);
 }
 
 static gboolean
 rspamd_lua_call_expression_func (struct ucl_lua_funcdata *lua_data,
 		struct rspamd_task *task,
-		GArray *args, gint *res,
+		GArray *args, gdouble *res,
 		const gchar *symbol)
 {
 	lua_State *L = lua_data->L;
@@ -511,7 +534,7 @@ process_regexp_item (struct rspamd_task *task,
 		void *user_data)
 {
 	struct regexp_module_item *item = user_data;
-	gint res = FALSE;
+	gdouble res = FALSE;
 
 	/* Non-threaded version */
 	if (item->lua_function) {
@@ -526,12 +549,7 @@ process_regexp_item (struct rspamd_task *task,
 	else {
 		/* Process expression */
 		if (item->expr) {
-			struct rspamd_expr_process_data process_data;
-			memset (&process_data, 0, sizeof process_data);
-
-			process_data.task = task;
-
-			res = rspamd_process_expression (item->expr, &process_data);
+			res = rspamd_process_expression (item->expr, 0, task);
 		}
 		else {
 			msg_warn_task ("FIXME: %s symbol is broken with new expressions",
@@ -539,7 +557,7 @@ process_regexp_item (struct rspamd_task *task,
 		}
 	}
 
-	if (res) {
+	if (res != 0) {
 		rspamd_task_insert_result (task, item->symbol, res, NULL);
 	}
 
