@@ -344,7 +344,13 @@ rspamadm_sign_file (const gchar *fname, struct rspamd_cryptobox_keypair *kp)
 	}
 
 	rspamd_snprintf (sigpath, sizeof (sigpath), "%s%s", fname, suffix);
-	g_assert (write (fd_sig, sig, rspamd_cryptobox_signature_bytes (mode)) != -1);
+
+	if (write (fd_sig, sig, rspamd_cryptobox_signature_bytes (mode)) == -1) {
+		rspamd_fprintf (stderr, "cannot write signature to %s: %s\n", sigpath,
+				strerror (errno));
+		exit (errno);
+	}
+
 	close (fd_sig);
 	munmap (map, st.st_size);
 
@@ -489,9 +495,12 @@ rspamadm_signtool (gint argc, gchar **argv, const struct rspamadm_command *cmd)
 	if (!g_option_context_parse (context, &argc, &argv, &error)) {
 		rspamd_fprintf (stderr, "option parsing failed: %s\n", error->message);
 		g_error_free (error);
+		g_option_context_free (context);
 		exit (1);
 	}
 
+	g_option_context_free (context);
+
 	if (openssl) {
 		mode = RSPAMD_CRYPTOBOX_MODE_NIST;
 	}
@@ -589,8 +598,13 @@ rspamadm_signtool (gint argc, gchar **argv, const struct rspamadm_command *cmd)
 
 		kp = rspamd_keypair_from_ucl (top);
 
-		if (kp == NULL || rspamd_keypair_type (kp) != RSPAMD_KEYPAIR_SIGN) {
-			rspamd_fprintf (stderr, "invalid or unsuitable for signing key\n");
+		if (kp == NULL) {
+			rspamd_fprintf (stderr, "invalid signing key\n");
+			exit (EXIT_FAILURE);
+		}
+
+		if (rspamd_keypair_type (kp) != RSPAMD_KEYPAIR_SIGN) {
+			rspamd_fprintf (stderr, "unsuitable for signing key\n");
 			exit (EXIT_FAILURE);
 		}
 
