@@ -3,7 +3,7 @@ import socket
 CONTROLLER_ERRORS = True
 KEY_PVT1 = 'ekd3x36tfa5gd76t6pa8hqif3ott7n1siuux68exbkk7ukscte9y'
 KEY_PUB1 = 'm8kneubpcjsb8sbsoj7jy7azj9fdd3xmj63txni86a8ye9ncomny'
-LOCAL_ADDR = u'127.0.0.1'
+LOCAL_ADDR = '127.0.0.1'
 MAP_WATCH_INTERVAL = '1min'
 PORT_CONTROLLER = 56790
 PORT_CONTROLLER_SLAVE = 56793
@@ -12,12 +12,14 @@ PORT_FUZZY_SLAVE = 56792
 PORT_NORMAL = 56789
 PORT_NORMAL_SLAVE = 56794
 PORT_PROXY = 56795
-PORT_CLAM = 56796
-PORT_FPROT = 56797
-PORT_FPROT2_DUPLICATE = 56798
-REDIS_ADDR = u'127.0.0.1'
+PORT_CLAM = 2100
+PORT_FPROT = 2101
+PORT_FPROT2_DUPLICATE = 2102
+PORT_AVAST = 2103
+P0F_SOCKET = '/tmp/p0f.sock'
+REDIS_ADDR = '127.0.0.1'
 REDIS_PORT = 56379
-NGINX_ADDR = u'127.0.0.1'
+NGINX_ADDR = '127.0.0.1'
 NGINX_PORT = 56380
 RSPAMD_GROUP = 'nogroup'
 RSPAMD_USER = 'nobody'
