@@ -1,14 +1,31 @@
 -- Run all unit tests in 'unit' directory
 
 local telescope = require "telescope"
+require "rspamd_assertions"
+local loaded, luacov = pcall(require, 'luacov.runner')
+if not loaded then
+  luacov = {
+    init = function() end,
+    shutdown = function() end,
+    run_report = function() end
+  }
+end
+luacov.init()
 
 local contexts = {}
 
 for _,t in ipairs(tests_list) do
   telescope.load_contexts(t, contexts)
 end
+local function test_filter(test)
+  return test.name:match(test_pattern)
+end
+if not test_pattern then
+  test_filter = function(_) return true end
+end
+
 local buffer = {}
-local results = telescope.run(contexts, callbacks, test_pattern)
+local results = telescope.run(contexts, callbacks, test_filter)
 local summary, data = telescope.summary_report(contexts, results)
 
 table.insert(buffer, telescope.test_report(contexts, results))
@@ -28,4 +45,7 @@ for _, v in pairs(results) do
     v.status_code == telescope.status_codes.fail then
     os.exit(1)
   end
-end
\ No newline at end of file
+end
+
+luacov:shutdown()
+luacov:run_report()
