@@ -1,38 +1,30 @@
-/*
- * Copyright (c) 2015, Vsevolod Stakhov
- * All rights reserved.
+/*-
+ * Copyright 2016 Vsevolod Stakhov
  *
- * Redistribution and use in source and binary forms, with or without
- * modification, are permitted provided that the following conditions are met:
- *	 * Redistributions of source code must retain the above copyright
- *	   notice, this list of conditions and the following disclaimer.
- *	 * Redistributions in binary form must reproduce the above copyright
- *	   notice, this list of conditions and the following disclaimer in the
- *	   documentation and/or other materials provided with the distribution.
+ * Licensed under the Apache License, Version 2.0 (the "License");
+ * you may not use this file except in compliance with the License.
+ * You may obtain a copy of the License at
  *
- * THIS SOFTWARE IS PROVIDED BY AUTHOR ''AS IS'' AND ANY
- * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
- * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
- * DISCLAIMED. IN NO EVENT SHALL AUTHOR BE LIABLE FOR ANY
- * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
- * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
- * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
- * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
- * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
- * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
+ *   http://www.apache.org/licenses/LICENSE-2.0
+ *
+ * Unless required by applicable law or agreed to in writing, software
+ * distributed under the License is distributed on an "AS IS" BASIS,
+ * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+ * See the License for the specific language governing permissions and
+ * limitations under the License.
  */
-
-
 #include "config.h"
-#include "main.h"
+#include "rspamd.h"
 #include "shingles.h"
 #include "fstring.h"
 #include "ottery.h"
 #include "cryptobox.h"
+#include "unix-std.h"
 
 static const int mapping_size = 64 * 8192 + 1;
 static const int max_seg = 32;
 static const int random_fuzz_cnt = 10000;
+enum rspamd_cryptobox_mode mode = RSPAMD_CRYPTOBOX_MODE_25519;
 
 static void *
 create_mapping (int mapping_len, guchar **beg, guchar **end)
@@ -55,12 +47,12 @@ create_mapping (int mapping_len, guchar **beg, guchar **end)
 
 static void
 check_result (const rspamd_nm_t key, const rspamd_nonce_t nonce,
-		const rspamd_sig_t mac, guchar *begin, guchar *end)
+		const rspamd_mac_t mac, guchar *begin, guchar *end)
 {
 	guint64 *t = (guint64 *)begin;
 
 	g_assert (rspamd_cryptobox_decrypt_nm_inplace (begin, end - begin, nonce, key,
-			mac));
+			mac, mode));
 
 	while (t < (guint64 *)end) {
 		g_assert (*t == 0);
@@ -167,10 +159,11 @@ rspamd_cryptobox_test_func (void)
 	guchar *begin, *end;
 	rspamd_nm_t key;
 	rspamd_nonce_t nonce;
-	rspamd_sig_t mac;
+	rspamd_mac_t mac;
 	struct rspamd_cryptobox_segment *seg;
 	double t1, t2;
 	gint i, cnt, ms;
+	gboolean checked_openssl = FALSE;
 
 	map = create_mapping (mapping_size, &begin, &end);
 
@@ -181,72 +174,89 @@ rspamd_cryptobox_test_func (void)
 	seg = g_slice_alloc0 (sizeof (*seg) * max_seg * 10);
 
 	/* Test baseline */
-	t1 = rspamd_get_ticks ();
-	rspamd_cryptobox_encrypt_nm_inplace (begin, end - begin, nonce, key, mac);
-	t2 = rspamd_get_ticks ();
+	t1 = rspamd_get_ticks (TRUE);
+	rspamd_cryptobox_encrypt_nm_inplace (begin, end - begin, nonce, key, mac,
+			mode);
+	t2 = rspamd_get_ticks (TRUE);
 	check_result (key, nonce, mac, begin, end);
 
-	msg_info ("baseline encryption: %.6f", t2 - t1);
+	msg_info ("baseline encryption: %.0f", t2 - t1);
+
+	mode = RSPAMD_CRYPTOBOX_MODE_NIST;
+	t1 = rspamd_get_ticks (TRUE);
+	rspamd_cryptobox_encrypt_nm_inplace (begin,
+			end - begin,
+			nonce,
+			key,
+			mac,
+			mode);
+	t2 = rspamd_get_ticks (TRUE);
+	check_result (key, nonce, mac, begin, end);
+
+	msg_info ("openssl baseline encryption: %.0f", t2 - t1);
+	mode = RSPAMD_CRYPTOBOX_MODE_25519;
+
+start:
 	/* A single chunk as vector */
 	seg[0].data = begin;
 	seg[0].len = end - begin;
-	t1 = rspamd_get_ticks ();
-	rspamd_cryptobox_encryptv_nm_inplace (seg, 1, nonce, key, mac);
-	t2 = rspamd_get_ticks ();
+	t1 = rspamd_get_ticks (TRUE);
+	rspamd_cryptobox_encryptv_nm_inplace (seg, 1, nonce, key, mac, mode);
+	t2 = rspamd_get_ticks (TRUE);
 
 	check_result (key, nonce, mac, begin, end);
 
-	msg_info ("bulk encryption: %.6f", t2 - t1);
+	msg_info ("bulk encryption: %.0f", t2 - t1);
 
 	/* Two chunks as vector */
 	seg[0].data = begin;
 	seg[0].len = (end - begin) / 2;
 	seg[1].data = begin + seg[0].len;
 	seg[1].len = (end - begin) - seg[0].len;
-	t1 = rspamd_get_ticks ();
-	rspamd_cryptobox_encryptv_nm_inplace (seg, 2, nonce, key, mac);
-	t2 = rspamd_get_ticks ();
+	t1 = rspamd_get_ticks (TRUE);
+	rspamd_cryptobox_encryptv_nm_inplace (seg, 2, nonce, key, mac, mode);
+	t2 = rspamd_get_ticks (TRUE);
 
 	check_result (key, nonce, mac, begin, end);
 
-	msg_info ("2 equal chunks encryption: %.6f", t2 - t1);
+	msg_info ("2 equal chunks encryption: %.0f", t2 - t1);
 
 	seg[0].data = begin;
 	seg[0].len = 1;
 	seg[1].data = begin + seg[0].len;
 	seg[1].len = (end - begin) - seg[0].len;
-	t1 = rspamd_get_ticks ();
-	rspamd_cryptobox_encryptv_nm_inplace (seg, 2, nonce, key, mac);
-	t2 = rspamd_get_ticks ();
+	t1 = rspamd_get_ticks (TRUE);
+	rspamd_cryptobox_encryptv_nm_inplace (seg, 2, nonce, key, mac, mode);
+	t2 = rspamd_get_ticks (TRUE);
 
 	check_result (key, nonce, mac, begin, end);
 
-	msg_info ("small and large chunks encryption: %.6f", t2 - t1);
+	msg_info ("small and large chunks encryption: %.0f", t2 - t1);
 
 	seg[0].data = begin;
 	seg[0].len = (end - begin) - 3;
 	seg[1].data = begin + seg[0].len;
 	seg[1].len = (end - begin) - seg[0].len;
-	t1 = rspamd_get_ticks ();
-	rspamd_cryptobox_encryptv_nm_inplace (seg, 2, nonce, key, mac);
-	t2 = rspamd_get_ticks ();
+	t1 = rspamd_get_ticks (TRUE);
+	rspamd_cryptobox_encryptv_nm_inplace (seg, 2, nonce, key, mac, mode);
+	t2 = rspamd_get_ticks (TRUE);
 
 	check_result (key, nonce, mac, begin, end);
 
-	msg_info ("large and small chunks encryption: %.6f", t2 - t1);
+	msg_info ("large and small chunks encryption: %.0f", t2 - t1);
 
 	/* Random two chunks as vector */
 	seg[0].data = begin;
 	seg[0].len = ottery_rand_range (end - begin - 1) + 1;
 	seg[1].data = begin + seg[0].len;
 	seg[1].len = (end - begin) - seg[0].len;
-	t1 = rspamd_get_ticks ();
-	rspamd_cryptobox_encryptv_nm_inplace (seg, 2, nonce, key, mac);
-	t2 = rspamd_get_ticks ();
+	t1 = rspamd_get_ticks (TRUE);
+	rspamd_cryptobox_encryptv_nm_inplace (seg, 2, nonce, key, mac, mode);
+	t2 = rspamd_get_ticks (TRUE);
 
 	check_result (key, nonce, mac, begin, end);
 
-	msg_info ("random 2 chunks encryption: %.6f", t2 - t1);
+	msg_info ("random 2 chunks encryption: %.0f", t2 - t1);
 
 	/* 3 specific chunks */
 	seg[0].data = begin;
@@ -255,47 +265,47 @@ rspamd_cryptobox_test_func (void)
 	seg[1].len = 2049;
 	seg[2].data = begin + seg[0].len + seg[1].len;
 	seg[2].len = (end - begin) - seg[0].len - seg[1].len;
-	t1 = rspamd_get_ticks ();
-	rspamd_cryptobox_encryptv_nm_inplace (seg, 3, nonce, key, mac);
-	t2 = rspamd_get_ticks ();
+	t1 = rspamd_get_ticks (TRUE);
+	rspamd_cryptobox_encryptv_nm_inplace (seg, 3, nonce, key, mac, mode);
+	t2 = rspamd_get_ticks (TRUE);
 
 	check_result (key, nonce, mac, begin, end);
 
-	msg_info ("small, medium and large chunks encryption: %.6f", t2 - t1);
+	msg_info ("small, medium and large chunks encryption: %.0f", t2 - t1);
 
 	cnt = create_random_split (seg, max_seg, begin, end);
-	t1 = rspamd_get_ticks ();
-	rspamd_cryptobox_encryptv_nm_inplace (seg, cnt, nonce, key, mac);
-	t2 = rspamd_get_ticks ();
+	t1 = rspamd_get_ticks (TRUE);
+	rspamd_cryptobox_encryptv_nm_inplace (seg, cnt, nonce, key, mac, mode);
+	t2 = rspamd_get_ticks (TRUE);
 
 	check_result (key, nonce, mac, begin, end);
 
-	msg_info ("random split of %d chunks encryption: %.6f", cnt, t2 - t1);
+	msg_info ("random split of %d chunks encryption: %.0f", cnt, t2 - t1);
 
 	cnt = create_realistic_split (seg, max_seg, begin, end);
-	t1 = rspamd_get_ticks ();
-	rspamd_cryptobox_encryptv_nm_inplace (seg, cnt, nonce, key, mac);
-	t2 = rspamd_get_ticks ();
+	t1 = rspamd_get_ticks (TRUE);
+	rspamd_cryptobox_encryptv_nm_inplace (seg, cnt, nonce, key, mac, mode);
+	t2 = rspamd_get_ticks (TRUE);
 
 	check_result (key, nonce, mac, begin, end);
 
-	msg_info ("realistic split of %d chunks encryption: %.6f", cnt, t2 - t1);
+	msg_info ("realistic split of %d chunks encryption: %.0f", cnt, t2 - t1);
 
 	cnt = create_constrainted_split (seg, max_seg + 1, 32, begin, end);
-	t1 = rspamd_get_ticks ();
-	rspamd_cryptobox_encryptv_nm_inplace (seg, cnt, nonce, key, mac);
-	t2 = rspamd_get_ticks ();
+	t1 = rspamd_get_ticks (TRUE);
+	rspamd_cryptobox_encryptv_nm_inplace (seg, cnt, nonce, key, mac, mode);
+	t2 = rspamd_get_ticks (TRUE);
 
 	check_result (key, nonce, mac, begin, end);
 
-	msg_info ("constrainted split of %d chunks encryption: %.6f", cnt, t2 - t1);
+	msg_info ("constrainted split of %d chunks encryption: %.0f", cnt, t2 - t1);
 
 	for (i = 0; i < random_fuzz_cnt; i ++) {
 		ms = ottery_rand_range (i % max_seg * 2) + 1;
 		cnt = create_random_split (seg, ms, begin, end);
-		t1 = rspamd_get_ticks ();
-		rspamd_cryptobox_encryptv_nm_inplace (seg, cnt, nonce, key, mac);
-		t2 = rspamd_get_ticks ();
+		t1 = rspamd_get_ticks (TRUE);
+		rspamd_cryptobox_encryptv_nm_inplace (seg, cnt, nonce, key, mac, mode);
+		t2 = rspamd_get_ticks (TRUE);
 
 		check_result (key, nonce, mac, begin, end);
 
@@ -306,9 +316,9 @@ rspamd_cryptobox_test_func (void)
 	for (i = 0; i < random_fuzz_cnt; i ++) {
 		ms = ottery_rand_range (i % max_seg * 2) + 1;
 		cnt = create_realistic_split (seg, ms, begin, end);
-		t1 = rspamd_get_ticks ();
-		rspamd_cryptobox_encryptv_nm_inplace (seg, cnt, nonce, key, mac);
-		t2 = rspamd_get_ticks ();
+		t1 = rspamd_get_ticks (TRUE);
+		rspamd_cryptobox_encryptv_nm_inplace (seg, cnt, nonce, key, mac, mode);
+		t2 = rspamd_get_ticks (TRUE);
 
 		check_result (key, nonce, mac, begin, end);
 
@@ -319,9 +329,9 @@ rspamd_cryptobox_test_func (void)
 	for (i = 0; i < random_fuzz_cnt; i ++) {
 		ms = ottery_rand_range (i % max_seg * 10) + 1;
 		cnt = create_constrainted_split (seg, ms, i, begin, end);
-		t1 = rspamd_get_ticks ();
-		rspamd_cryptobox_encryptv_nm_inplace (seg, cnt, nonce, key, mac);
-		t2 = rspamd_get_ticks ();
+		t1 = rspamd_get_ticks (TRUE);
+		rspamd_cryptobox_encryptv_nm_inplace (seg, cnt, nonce, key, mac, mode);
+		t2 = rspamd_get_ticks (TRUE);
 
 		check_result (key, nonce, mac, begin, end);
 
@@ -329,4 +339,10 @@ rspamd_cryptobox_test_func (void)
 			msg_info ("constrainted fuzz iterations: %d", i);
 		}
 	}
+
+	if (!checked_openssl) {
+		checked_openssl = TRUE;
+		mode = RSPAMD_CRYPTOBOX_MODE_NIST;
+		goto start;
+	}
 }
