@@ -1,5 +1,5 @@
 #include "config.h"
-#include "main.h"
+#include "rspamd.h"
 #include "cfg_file.h"
 #include "url.h"
 #include "tests.h"
